/****************************************
Sample V4L2 program for changing cropping
for Sentech cameras

Marius Silaghi msilaghi@fit.edu
 ***************************************/

#include <sys/ioctl.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <linux/videodev2.h>
#include <Sentech/sentech.h>

struct v4l2_control arg_ctl;
struct v4l2_format arg_fmt;
struct v4l2_crop arg_crop;
int fd, result;

const char* strformat(__u32 fmt) {
  switch(fmt){
  case V4L2_PIX_FMT_BGR24: return "V4L2_PIX_FMT_BGR24";
  case V4L2_PIX_FMT_RGB24: return "V4L2_PIX_FMT_RGB24";
  case V4L2_PIX_FMT_YUYV: return "V4L2_PIX_FMT_YUYV";
  case V4L2_PIX_FMT_YUV420: return "V4L2_PIX_FMT_YUV420";
  case V4L2_PIX_FMT_GREY: return "V4L2_PIX_FMT_GRAY";
  case V4L2_PIX_FMT_Y16: return "V4L2_PIX_FMT_Y16";
  case V4L2_PIX_FMT_SRGGB8: return "V4L2_PIX_FMT_SRGGB";
  default: return "Unknown";
  }
}

void getStatus(const char* status) {
  printf("%s\n", status);
  memset(&arg_fmt, 0, sizeof(arg_fmt));
  arg_fmt.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  arg_fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_YUYV;
  arg_fmt.fmt.pix.field = V4L2_FIELD_ANY;
  arg_fmt.fmt.pix.bytesperline = 0;
  arg_fmt.fmt.pix.priv = 0;
  result = ioctl(fd, VIDIOC_G_FMT, &arg_fmt);
  if(result != 0){
    perror("ioctl get format");
    exit(2);
  }
  printf("\tFMT: Type=%d, Fmt=%s, %dx%d, Bpl=%d, Size=%d\n",
	 arg_fmt.type,
	 strformat(arg_fmt.fmt.pix.pixelformat),
	 arg_fmt.fmt.pix.width,
	 arg_fmt.fmt.pix.height,
	 arg_fmt.fmt.pix.bytesperline,
	 arg_fmt.fmt.pix.sizeimage
	 );
  
  memset(&arg_crop, 0, sizeof(arg_crop));
  result = ioctl(fd, VIDIOC_G_CROP, &arg_crop);
  if(result != 0){
    perror("ioctl get crop");
    exit(2);
  }
  printf("\tCROP: Type=%d, left=%d, top=%d, size=%dx%d\n",
	 arg_crop.type, arg_crop.c.left, arg_crop.c.top,
	 arg_crop.c.width, arg_crop.c.height);
}

int main(int argc, char**argv) {
  int opt;
  int width=-1, height=-1, top=-1, left=-1;
  const char*filename = "/dev/video";
  int set_crop=0;

  if(argc==1){
      fprintf(stderr, "Usage: %s [-c] [-d device] [-l left] [-t top] [-h height] [-w width]\n",
	      argv[0]);
      exit(EXIT_FAILURE);
  }
  while ((opt = getopt(argc, argv, "d:t:l:h:w:c")) != -1) {
    switch (opt) {
    case 'd': filename =  optarg; break;
    case 't': top = atoi(optarg); set_crop=1; break;
    case 'l': left = atoi(optarg); set_crop=1; break; 
    case 'h': height = atoi(optarg); set_crop=1; break;
    case 'w': width = atoi(optarg); set_crop=1; break;
    case 'c': break;
    default: /* '?' */
      fprintf(stderr, "Usage: %s [-c] [-d device] [-l left] [-t top] [-h height] [-w width]\n",
	      argv[0]);
      exit(EXIT_FAILURE);
    }
  }

  fd = open(filename,O_RDONLY);
  if (fd < 0){
    perror("open");
    exit(2);
  }
  getStatus("Initial Status");
  
  if(set_crop){
    if(left>=0) arg_crop.c.left = left;
    if(top>=0) arg_crop.c.top = top;
    if(width>=0) arg_crop.c.width = width;
    if(height>=0) arg_crop.c.height = height;
    
    result = ioctl(fd, VIDIOC_S_CROP, &arg_crop);
    if(result != 0){
      perror("ioctl set crop");
      exit(2);
    }
    getStatus("Final Status");
  }
}
