/****************************************
Sample V4L2 program for changing controls
for Sentech cameras

Marius Silaghi msilaghi@fit.edu
 ***************************************/

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <linux/videodev2.h>
#include <Sentech/sentech.h>
const char* strformat(__u32 fmt) {
  switch(fmt){
  case V4L2_PIX_FMT_BGR24: return "V4L2_PIX_FMT_BGR24";
  case V4L2_PIX_FMT_RGB24: return "V4L2_PIX_FMT_RGB24";
  case V4L2_PIX_FMT_YUYV: return "V4L2_PIX_FMT_YUYV";
  case V4L2_PIX_FMT_YUV420: return "V4L2_PIX_FMT_YUV420";
  case V4L2_PIX_FMT_GREY: return "V4L2_PIX_FMT_GRAY";
  case V4L2_PIX_FMT_Y16: return "V4L2_PIX_FMT_Y16";
  case V4L2_PIX_FMT_SRGGB8: return "V4L2_PIX_FMT_SRGGB";
  default: return "Unknown";
  }
}
int main(int argc, char**argv) {
  int fd, result, opt;
  int width=-1, height=-1, Height=-1, fps=-1, scan=0;
  struct v4l2_control arg_ctl;
  struct v4l2_format arg_fmt;
  char*filename = "/dev/video";

  if(argc==1){
      fprintf(stderr, "Usage: %s [-s (for scan)] [-f pfs] [-d device] [-h height] [-w width]\n",
	      argv[0]);
      exit(EXIT_FAILURE);
  }

  while ((opt = getopt(argc, argv, "sd:f:H:h:w:")) != -1) {
    switch (opt) {
    case 's': scan = 1; break;
    case 'd': filename =  optarg; break;

    case 'f': fps = atoi(optarg); break;

    case 'H': Height = atoi(optarg); break;

    case 'h': height = atoi(optarg); break;
    case 'w': width = atoi(optarg); break;
 
    default: /* '?' */
      fprintf(stderr, "Usage: %s [-s] [-f fpsmode=0|1|2] [-d device] [-h height] [-w width] [-H heightmode=0|1|2]\n",
	      argv[0]);
      exit(EXIT_FAILURE);
    }
  }

  fd = open(filename,O_RDONLY);
  if (fd < 0){
    perror("open");
    exit(2);
  }

  if(Height>=0) {
    arg_ctl.id = V4L2_CID_STC_HEIGHT;
    arg_ctl.value = Height;
    result = ioctl(fd, VIDIOC_S_CTRL, &arg_ctl);
    if(result != 0){
      perror("ioctl Height");
      exit(2);
    }
  }

  if(fps>=0) {
    arg_ctl.id = V4L2_CID_STC_FPS;
    arg_ctl.value = fps;
    result = ioctl(fd, VIDIOC_S_CTRL, &arg_ctl);
    if(result != 0){
      perror("ioctl fps");
      exit(2);
    }
  }

  if(scan) {
    __s16 scan;
    result = ioctl(fd, SENTECH_IOC_GET_SCAN_HEIGHT, &scan);
    if(result != 0){
      perror("ioctl scan");
      exit(2);
    }
    printf("scan height=%d\n", scan);
  }

  if((width>=0) || (height>=0)) {
    memset(&arg_fmt, 0, sizeof(arg_fmt));
    arg_fmt.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    arg_fmt.fmt.pix.width = width;
    arg_fmt.fmt.pix.height = width;
    arg_fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_YUYV;
    arg_fmt.fmt.pix.field = V4L2_FIELD_ANY;
    arg_fmt.fmt.pix.bytesperline = 0;
    arg_fmt.fmt.pix.priv = 0;

    result = ioctl(fd, VIDIOC_G_FMT, &arg_fmt);
    if(result != 0){
      perror("ioctl get format");
      exit(2);
    }
    printf("Type=%d, Fmt=%s, %dx%d, Bpl=%d, Size=%d\n",
	   arg_fmt.type,
	   strformat(arg_fmt.fmt.pix.pixelformat),
	   arg_fmt.fmt.pix.width,
	   arg_fmt.fmt.pix.height,
	   arg_fmt.fmt.pix.bytesperline,
	   arg_fmt.fmt.pix.sizeimage
	   );

    if(width == -1) width = arg_fmt.fmt.pix.width;
    if(height == -1) height = arg_fmt.fmt.pix.height;

    arg_fmt.fmt.pix.width = width;
    arg_fmt.fmt.pix.height = height;
    result = ioctl(fd, VIDIOC_TRY_FMT, &arg_fmt);
    if(result != 0) {
      perror("ioctl try format");
      exit(2);
    }
    printf("Try w=%dx%d\n Type=%d, Fmt=%s, %dx%d, Bpl=%d, Size=%d\n",
	   width,height,
	   arg_fmt.type,
	   strformat(arg_fmt.fmt.pix.pixelformat),
	   arg_fmt.fmt.pix.width,
	   arg_fmt.fmt.pix.height,
	   arg_fmt.fmt.pix.bytesperline,
	   arg_fmt.fmt.pix.sizeimage
	   );

    if(arg_fmt.fmt.pix.width != width) {
      fprintf(stderr, 
	      "Can only set width=%d, not %d\n",
	      arg_fmt.fmt.pix.width,
	      width);
      exit(3);
    }

    arg_fmt.fmt.pix.width = width;
    arg_fmt.fmt.pix.height = height;
    result = ioctl(fd, VIDIOC_S_FMT, &arg_fmt);
    if(result != 0){
      perror("ioctl set format");
      exit(2);
    }
    if(arg_fmt.fmt.pix.width != width) {
      fprintf(stderr, 
	      "Have set width=%d, not %d\n",
	      arg_fmt.fmt.pix.width,
	      width);
      exit(3);
    }

    printf("Got Type=%d, Fmt=%s, %dx%d, Bpl=%d, Size=%d\n",
	   arg_fmt.type,
	   strformat(arg_fmt.fmt.pix.pixelformat),
	   arg_fmt.fmt.pix.width,
	   arg_fmt.fmt.pix.height,
	   arg_fmt.fmt.pix.bytesperline,
	   arg_fmt.fmt.pix.sizeimage
	   );
    
  }
  
  return 0;
}
