#include <sys/ioctl.h>
#include <stdio.h>
#include <getopt.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/types.h>
#include <asm/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/time.h>
#include <time.h>
#include <math.h>
#include <inttypes.h>
#include <sentech.h>
//#include <Sentech/sentech.h>
#define min(a,b) (((a)<(b))?(a):(b))




void ToMBS(struct camera_id *pCameraID)
{
	if(pCameraID->type & 0x80)
	{
		unsigned short org[STC_MAX_ID_NAME / 2];
		unsigned short nPos = 0;
		memcpy(org, pCameraID->name, STC_MAX_ID_NAME);
		memset(pCameraID->name, 0, STC_MAX_ID_NAME);

		while((org[nPos] != 0) && (nPos < STC_MAX_ID_NAME))
		{
			wchar_t src = (wchar_t)org[nPos];
			wcstombs(&pCameraID->name[nPos], &src, 1);
			nPos++;
		};

		pCameraID->length = strlen(pCameraID->name) + 1;
		pCameraID->type &= 0x7F;
	}
}


const char* dev_name = "/dev/video0"; 
static void usage (FILE * fp, int argc, char ** argv) {
  fprintf (fp,
         "Usage: %s [options]\n\n"
         "Options:\n"
           "-d name            video device name [/dev/video0]\n"
           "-v                 print version\n"
           "-i                 print ID\n"
           "-I name_ID         set name_ID\n"
           "", argv[0]);
}
static void errno_exit(const char * s) {
  perror(s);
  printf("exiting because of errors with: %s",s);
  exit (EXIT_FAILURE);
}
int main(int argc, char ** argv) {
  int getid=0;
  int getversion=0;
  char* setid=NULL;
  struct camera_id crtID;
  __s16 version=0;
  __s32 release=0;
  int fd;
  int opt;
  while ((opt = getopt(argc, argv, "d:iI:v")) != -1) {
    switch (opt) {
    case 'd':
      dev_name = optarg;
      break;
    case 'i':
      getid = 1;
      break;
    case 'v':
      getversion = 1;
      break;
    case 'I':
      setid = optarg;
      break;
    default: usage (stderr, argc, argv); exit (EXIT_FAILURE);
    }
  }
  if(!getid && !setid) getversion = 1;
  fd = open(dev_name, O_RDWR /* required */ | O_NONBLOCK, 0);
  if(getversion) {
#ifdef SENTECH_IOC_RELEASE
    if (-1 == ioctl (fd, SENTECH_IOC_RELEASE, &release)) {
      perror("SENTECH_IOC_RELEASE: driver too old");
    }else{
      printf("Sentech VERSION Major:%hhd Minor:%hhd Release:%hd\n", (release>>8)&0xFF, release&0xFF, (release>>16)&0xFFFF);
    }
#else
    if (-1 == ioctl (fd, SENTECH_IOC_VERSION, &version)) {
      perror("SENTECH_IOC_VERSION: driver too old, or not a Sentech device");
    }else{
      printf("Sentech VERSION Major:%hhd Minor:%hhd\n", (version>>8)&0xFF, version&0xFF);
    }
#endif
  }
  if(getid) {
    if (-1 == ioctl (fd, SENTECH_IOC_G_ID, &crtID)) {
      perror("SENTECH_IOC_G_ID: driver too old");
    }else{
      printf("Sentech crt ID:%uxh ID_type:%hhxh ID_len:%hhd  ID_name:\"%s\"\n",
	     crtID.camera_ID, crtID.type, crtID.length, crtID.name);
	if(0x80 & crtID.type)
	{
		ToMBS(&crtID);
		printf("Sentech crt ID:%uxh ID_type:%hhxh ID_len:%hhd  ID_name:\"%s\"\n",
			crtID.camera_ID, crtID.type, crtID.length, crtID.name);
	}
    }
  }
  if(setid != NULL) {
    crtID.type=0;
    crtID.length=min(strlen(setid)+1, STC_MAX_ID_NAME);
    strncpy(crtID.name, setid, STC_MAX_ID_NAME);
    if (-1 == ioctl (fd, SENTECH_IOC_S_ID, &crtID)) {
      perror("SENTECH_IOC_S_ID: driver too old");
    }else{
      //printf("Sentech ID:%uxh ID_type:%hhxh ID_len:%hhd  ID_name:\"%s\"\n",
      //     crtID.camera_ID, crtID.type, crtID.length, crtID.name);
      if (-1 == ioctl (fd, SENTECH_IOC_G_ID, &crtID)) {
	perror("SENTECH_IOC_G_ID: driver too old");
      }else{
	printf("Sentech new ID:%uxh ID_type:%hhxh ID_len:%hhd  ID_name:\"%s\"\n",
	       crtID.camera_ID, crtID.type, crtID.length, crtID.name);
      }
    }
  }
}
