/************************************/
/* Driver for Sentech USB Cameras   */
/* Release 3.2.32                   */
/************************************/
/* Author: Marius Calin Silaghi     */
/*         FLORIDA TECH             */
/*  <msilaghi@fit.edu>              */
/*    April           9, 2015       */
/************************************/
/* Copyright:                       */
/*  Marius C. Silaghi               */
/*  msilaghi@fit.edu                */
/************************************/

#ifndef __SENTECH_TUSB
#define __SENTECH_TUSB
#include <linux/types.h>
#include <linux/version.h>
#if  LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20)
#include "backport/bug.h"
#else
#include <linux/bug.h>
#endif
#define DRIVER_AUTHOR "Marius Calin Silaghi, msilaghi@fit.edu"
#define DRIVER_DESC "Sentech USB Cameras"
#define SENTECH_MAJOR_VERSION 3
#define SENTECH_MINOR_VERSION 2
#ifndef SENTECH_RELEASE
#define SENTECH_RELEASE 32
#endif 
#define SENTECH_VERSION KERNEL_VERSION(SENTECH_MAJOR_VERSION, SENTECH_MINOR_VERSION, SENTECH_RELEASE)
#define SENTECH_VENDOR_ID	0x1421
#define IMAGE_BULK_ENDPOINT 2
#define IMAGE_CTL_ENDPOINT 0
#define CTRL_TIMEOUT 2*HZ
#define IMAGE_BULK_PACKET 512
#define IMAGE_BULK_PACKET_MASK 511
#define STC_C33USB 0x0305 
#define STC_C83USB 0x0605 
#define STC_B33USB 0x0705 
#define STC_B83USB 0x0805 
#define STC_B512USB 0x0506 
#define STC_C512USB 0x0606 
#define STC_B152USB 0x0506 
#define STC_C152USB 0x0606 
#define STC_B202USB 0x0706 
#define STC_C202USB 0x0806 
#define STC_B33USB_T 0x0906 
#define STC_C33USB_T 0x1006 
#define STC_B83USB_T 0x1106
#define STC_C83USB_T 0x1206 
#define STC_B512USB_T 0x1306 
#define STC_C512USB_T 0x1406 
#define STC_B152USB_T 0x1306 
#define STC_C152USB_T 0x1406 
#define STC_B202USB_T 0x1506 
#define STC_C202USB_T 0x1606 
#define STC_B133USB_T 0x0109 
#define STC_C133USB_T 0x0209 
#define STC_MB33USB 0x0110
#define STC_MC33USB 0x0210
#define STC_MB83USB 0x0310
#define STC_MC83USB 0x0410
#define STC_MB133USB 0x0510
#define STC_MC133USB 0x0610
#define STC_MB152USB 0x0710
#define STC_MC152USB 0x0810
#define STC_MB202USB 0x0910
#define STC_MC202USB 0x1010
#define STC_MBA5MUSB3 0x0111
#define STC_MCA5MUSB3 0x0211
#define STC_MBE132U3V 0x0112
#define STC_MCE132U3V 0x0212
#define STC_MBCM401U3V 0x0113
#define STC_MCCM401U3V 0x0213
#define STC_MBCM200U3V 0x0313
#define STC_MCCM200U3V 0x0413
#define STC_BW     0x0001
#define STC_RGrGbB 0x0002
#define STC_GrRBGb 0x0003
#define STC_GbBRGr 0x0004
#define STC_BGbGrR 0x0005
#define STC_CUSTOM 0x00000001
#define STC_VGA    0x00000008
#define STC_XGA    0x00000020
#define STC_SXGA   0x00000100
#define STC_SXGA133   0x00000080
#define STC_UXGA   0x00000400
#define STC_QXGA   0x00000800
#define STC_QSXGA  0x80000000
#define STC_MBE132U3V_XGA  0x80000001
#define STC_MBCM401U3V_XGA  0x80000002
#define STC_MBCM200U3V_XGA  0x80000003
#pragma pack(push,1)
#define STC_ONE_BYTE_PER_PIXEL 1
#define STC_TWO_BYTES_PER_PIXEL_10 2
#define STC_TWO_BYTES_PER_PIXEL_12 4
#define STC_TWO_BYTES_PER_PIXEL_10_12 6
#define STC_BYTES_PER_PIXEL_8_10 3
#define STC_BYTES_PER_PIXEL_8_12 5
#define STC_TWO_BYTES_PER_PIXEL_8_10_12 7
#define STC_ONE_BYTE_PER_PIXEL_MONO 0x20
#define STC_TWO_BYTES_PER_PIXEL_10_MONO 0x40
#define STC_TWO_BYTES_PER_PIXEL_12_MONO 0x80
#define STC_THREE_BYTES_PER_PIXEL_RGB8 0x400
#define STC_SIX_BYTES_PER_PIXEL_RGB8 0x800
inline char* bytes_mode_name(uint32_t  code) {
  switch (code){
  case STC_ONE_BYTE_PER_PIXEL: return "STC_ONE_BYTE_PER_PIXEL";
  case STC_TWO_BYTES_PER_PIXEL_10: return "STC_TWO_BYTES_PER_PIXEL_10";
  case STC_TWO_BYTES_PER_PIXEL_12: return "STC_TWO_BYTES_PER_PIXEL_12";
  case STC_BYTES_PER_PIXEL_8_10: return "STC_BYTES_PER_PIXEL BITS 8 10";
  case STC_BYTES_PER_PIXEL_8_12: return "STC_BYTES_PER_PIXEL BITS 8 12";
  case STC_TWO_BYTES_PER_PIXEL_10_12: return "STC_TWO_BYTES_PER_PIXEL BITS 10 12";
  case STC_TWO_BYTES_PER_PIXEL_8_10_12: return "STC_BYTES_PER_PIXEL BITS 8 10 12";
  case STC_ONE_BYTE_PER_PIXEL_MONO: return "STC_ONE_BYTE_PER_PIXEL_MONO";
  case STC_TWO_BYTES_PER_PIXEL_10_MONO: return "STC_TWO_BYTES_PER_PIXEL_10_MONO";
  case STC_TWO_BYTES_PER_PIXEL_12_MONO: return "STC_TWO_BYTES_PER_PIXEL_12_MONO";
  case STC_THREE_BYTES_PER_PIXEL_RGB8: return "STC_THREE_BYTES_PER_PIXEL_RGB8";
  case STC_SIX_BYTES_PER_PIXEL_RGB8: return "STC_SIX_BYTES_PER_PIXEL_RGB8";
  }
  return "Unknown Bytes Mode";
}
#define STC_FULL_SCREEN 0
#define STC_HALF_SCREEN 1
#define STC_QUART_SCREEN 2
#define STC_WHOLE_SCREEN 4
#define STC_VARIABLE_SCREEN 8
#define STC_FULL_BIN_SCREEN 0x010
#define STC_WHOLE_BIN_SCREEN  0x020
#define STC_HALF_BIN_SCREEN  0x040
#define STC_QUART_BIN_SCREEN  0x080
#define STC_VARIABLE_BIN_SCREEN  0x0100
#define STC_AOI_SCREEN  0x08000
int getMaxBpp(uint32_t modes) {
  if(modes & STC_SIX_BYTES_PER_PIXEL_RGB8) return 6;
  if(modes & STC_THREE_BYTES_PER_PIXEL_RGB8) return 3;
  if(modes & STC_TWO_BYTES_PER_PIXEL_12_MONO) return 2;
  if(modes & STC_TWO_BYTES_PER_PIXEL_10_MONO) return 2;
  if(modes & STC_TWO_BYTES_PER_PIXEL_12) return 2;
  if(modes & STC_TWO_BYTES_PER_PIXEL_10) return 2;
  if(modes & STC_ONE_BYTE_PER_PIXEL) return 1;
  return 1;
}
int getMax_bpp(uint32_t modes) {
  if(modes & STC_SIX_BYTES_PER_PIXEL_RGB8) return 48;
  if(modes & STC_THREE_BYTES_PER_PIXEL_RGB8) return 24;
  if(modes & STC_TWO_BYTES_PER_PIXEL_12_MONO) return 12;
  if(modes & STC_TWO_BYTES_PER_PIXEL_12) return 12;
  if(modes & STC_TWO_BYTES_PER_PIXEL_10_MONO) return 10;
  if(modes & STC_TWO_BYTES_PER_PIXEL_10) return 10;
  if(modes & STC_ONE_BYTE_PER_PIXEL) return 8;
  return 8;
}
int getBppMode(int OR_Bpp_Modes, int Bpp) {
  switch(Bpp) {
  case 1:
    if(OR_Bpp_Modes & STC_ONE_BYTE_PER_PIXEL)
      return STC_ONE_BYTE_PER_PIXEL;
    break;
  case 2:
    if(OR_Bpp_Modes & STC_TWO_BYTES_PER_PIXEL_12)
      return STC_TWO_BYTES_PER_PIXEL_12;
    if(OR_Bpp_Modes & STC_TWO_BYTES_PER_PIXEL_10)
      return STC_TWO_BYTES_PER_PIXEL_10;
    break;
  }
  return 0;
}
struct Sentech_Header {
  __u8 Header_length;
  __u8 reservedAEF;
  __u8 data1[10];
  __u32 Frame_number_le;
  __u8 reserved_48[48];
} __attribute__ ((packed));
struct Sentech_Camera_Data_Payload {
  __u8 data1 [6];
  __u16 Camera_type_le; 
  __u16 Firmware_version_le;
  __u16 Color_layout_le; 
  __u16 data3;
  __u32 OR_Picture_size_modes_le;   
  __u32 OR_Bytes_per_pixel_le; 
  __u32 data2_1;
  __u16 OR_Scan_modes_le;     
  __u8 data2 [8];
  __u32 Standard_clock_speed_le;
  __u32 OR_Configurable_clock_le;
  __u8 data4 [12];
  __u32 Maximum_picture_width_le;
  __u32 Maximum_picture_height_le;
} __attribute__ ((packed));
#define Vertical_size_mode_le_MASK 0xFFFFFFFC
#define Vertical_offset_le_MASK 0xFFFFFFFE
struct Sentech_Picture_Size_Payload {
  __u8 data1[2];
  __u32 Picture_size_mode_le; 
  __u32 Horizontal_size_mode_le;
  __u32 Vertical_size_mode_le;
  __u32 Horizontal_offset_le;
  __u32 Vertical_offset_le;
  __u16 Scan_mode_le;
  __u16 reserved_1;
  __u16 line_numbers_le;
  __u16 one_line_clock_numbers_le;
  __u16 maximum_gain_le;      
  __u16 initial_gain_le;      
  __u16 initial_digital_gain_le;   
  __u32 Max_exposure_time_le;    
  __u32 Longest_FPS_exposure_le;  
  __u32 Longest_FPS_exposure_H_le; 
  __u32 Longest_ET_exposure_le; 
  __u32 Longest_ET_exposure_H_le; 
  __u8  reserved[8];         
} __attribute__ ((packed));
struct Sentech_Bytes_Per_Pixel_Data_Payload {
  __u8  Data_size;      
  __u8  Request;        
  __u32 Bytes_per_pixel_le; 
} __attribute__ ((packed));
struct Sentech_Shutter_Gain_Payload {
  __u8 data[2];
  __u16 Shutter_line_numbers_le;
  __u16 Shutter_clock_numbers_le;
  __u16 Gain_le;
  __u16 reserved1;
  __u16 Exposure_time_H_unit_le;
  __u8  reserved[52];
} __attribute__ ((packed));
struct Sentech_Clock_Payload {
  __u8 data1[2];
  __u32 Clock_mode_le;
  __u8 data2[4];
} __attribute__ ((packed));
struct Sentech_White_Balance_Payload {
  __u8  Data_size;
  __u8  Request;
  __u16 R_Gain_le;
  __u16 Gr_Gain_le;
  __u16 Gb_Gain_le;
  __u16 B_Gain_le;
  __u8 not_used[54];
} __attribute__ ((packed));
#define Sentech_EEPROM_100KHZ (0x00<<15)
#define Sentech_EEPROM_400KHZ (0x01<<15)
#define Sentech_EEPROM_FIX_SLAVE (0x00<<8)
#define Sentech_EEPROM_SET_SLAVE (0x01<<8)
#define Sentech_EEPROM_100KHZ_Aptima (0x00<<14)
#define Sentech_EEPROM_400KHZ_Aptima (0x02<<14)
#define Sentech_EEPROM_1MHZ_Aptima (0x03<<14)
#define Sentech_EEPROM_Request 0xF9
#define Sentech_EEPROM_ID_ADDRESS 0x6F00
#define Sentech_EEPROM_PRODUCT_INFO 0x6E00
#define Sentech_EEPROM_Request_Aptima 0xFA
#define Sentech_EEPROM_ID_ADDRESS_Aptima 0x0100
#define Sentech_EEPROM_PRODUCT_INFO_Aptima 0x0000
#define Sentech_SOFTWARE_TRIGGER_Request 0x30
#define Sentech_FPGA_Request 0x80
#define Sentech_FPGA2_Request 0x82
#define STC_FPGA_AUTOREFRESH_WVAL 0x8000
#define STC_FPGA_READ 0x00080
#define STC_FPGA_IO_DIRECTION_CMD 0x00
#define STC_FPGA_TRIGGER_MODE_CMD 0x30
#define STC_FPGA_IMAGE_FLIP_CMD 0x3F
#define O_TRIGGER O_SYNC
#define STC_FPGA_TRIGGER_MODE 0x01
#define STC_FPGA_TRIGGER_MODE_MASK 0x03
#define STC_FPGA_TRIGGER_MODE_SHIFT 0x0
#define STC_FPGA_HEADER_SPLIT 0x0100
#define STC_FPGA_HEADER_MASK 0x0100
#define STC_FPGA_HEADER_SHIFT 8
#define STC_FPGA_TRIGGER_SRC_SOFT 0x0400
#define STC_FPGA_TRIGGER_SRC_HARD 0x0800
#define STC_FPGA_TRIGGER_SRC_MASK 0x0c00
#define STC_FPGA_TRIGGER_SRC_SHIFT 10
#define STC_FPGA_TRIGGER_RO_SEL_MASK 0x040
#define STC_FPGA_TRIGGER_RO_SEL_SHIFT 6
#define STC_FPGA_TRIGGER_EXP_MODE_MASK 0x3000
#define STC_FPGA_TRIGGER_EXP_MODE_SHIFT 12
#define STC_FPGA_TRIGGER_EXP_D1_MASK 0x4000
#define STC_FPGA_TRIGGER_EXP_D1_SHIFT 13
#define STC_FPGA_TRIGGER_EXP_D2_MASK 0x8000
#define STC_FPGA_TRIGGER_EXP_D2_SHIFT 12
#define STC_FPGA_REFRESH_MASK 0x01
#define STC_FPGA_REFRESH_SHIFT 0
#define STC_FPGA_CNT_RST_MASK 0x02
#define STC_FPGA_CNT_RST_SHIFT 1
#define STC_FPGA_RO_MASK 0x04
#define STC_FPGA_RO_SHIFT 2
#define STC_FPGA_SUB_TRG_MASK 0x4000
#define STC_FPGA_SUB_TRG_SHIFT 14
#define STC_FPGA_TRIGGER_SRC_SOFT 0x0400
#define STC_FPGA_TRIGGER_SRC_HARD 0x0800
#define STC_FPGA_TRIGGER_SRC_MASK 0x0c00
#define STC_FPGA_TRIGGER_SRC_SHIFT 10
struct Sentech_FPGA_Payload {
  __u8  IO_Directions;      
  __u8 _00;
  __u8  IO_Reversing;        
  __u8 _01;
  __u8  IO_Status;           
  __u8 _02;
  __u16  reserved1;
  __u8  IO0_change; 
  __u8 _04;
  __u8  IO1_change; 
  __u8 _05;
  __u8  IO2_change;    
  __u8 _06;
  __u8  IO3_change;    
  __u8 _07;
  __u16  reserved2[7];
  __u8  LED;           
  __u8 _0F;
  __u16  reserved2b[4];
  __u8 Digital_gain;   
  __u8 _14;
  __u16 reserved3[0x1b];
  __u16 Trigger_Mode;   
  __u16 Trigger_Mask; 
  __u16 EXP_delay_TRG; 
  __u16 EXP_delay_TRG_high; 
  __u16 Strobe_start_delay_TRG; 
  __u16 Strobe_start_delay_TRG_high; 
  __u16 Strobe_end_delay_TRG; 
  __u16 Strobe_end_delay_TRG_high; 
  __u16 TRG_delay_TRG; 
  __u16 TRG_delay_TRG_high; 
  __u16 TRG_width_TRG; 
  __u16 TRG_width_TRG_high; 
  __u16 Readout_delay; 
  __u16 reserved3b[0x2];
  __u8 Image_Flip;
  __u8 _3F;
  __u16 White_Balance_R_gain;  
  __u16 White_Balance_GR_gain; 
  __u16 White_Balance_GB_gain; 
  __u16 White_Balance_B_gain;  
  __u8 White_Balance_mode;     
  __u8 _44;
  __u16 Max_exposure_Low_le; 
  __u8 Max_exposure_Upp;
  __u8 _46;
  __u16 Min_exposure_Low_le; 
  __u8 Min_exposure_Upp;
  __u8 _48;
  __u8 AGC_mode; 
  __u8 _49;
  __u8 ACL_Target_brightness; 
  __u8 _4A;
  __u8 Maximum_gain_AGC;
  __u8 _4B;
  __u8 Minimum_gain_AGC;
  __u8 _4C;
  __u8 Gamma;
  __u8 _4D;
  __u8 Pixel_blemish_correction;
  __u8 _4E;
  __u16 _4F;
  __u16 Refresh;       
  __u16 reserved5_1[0x04];
  __u16 output_frame_rate_decimal; 
  __u16 output_frame_rate; 
  __u16 reserved5_3[0x09];
  __u16 Altera_version; 
  __u16 reserved6[3];
  __u16 Variable_partial_ln_nb; 
  __u16 USB3Vision_mode; 
  __u8 SW_status; 
  __u8 _66;
  __u16 _67;
  __u8 capabilities;  
  __u8 _68;
  __u16 capabilities_3; 
  __u16 _6A;
  __u16 capabilities_4; 
  __u16 USB3Vision_support; 
} __attribute__ ((packed));
#pragma pack(pop)
#endif 
