#ifdef SENTECH_AWB
inline void interpolateRGBawb(__u8*R, __u8*G, __u8*B,
			      __u8* image, int _brightness,
			      int line_prev, int pix_prev,
			      int line_next, int pix_next,
			      int src_ptr, int line_start, int w,
			      int evenh, int evenw,
			      const __u8 * crt_gamma,
			      struct sentech_camera* cam_data) {
  int saturation = cam_data->saturation;
  if(evenh)
    if(evenw){ 
      *R = crt_gamma[image[src_ptr]];
      *G = crt_gamma[(((__u16)0)+image[line_prev+w]+image[line_next+w]+
		      image[line_start+pix_prev]+
		      image[line_start+pix_next])>>2];
      *B = crt_gamma[(((__u16)0)+image[line_prev+pix_prev]+
		      image[line_prev+pix_next]+
		      image[line_next+pix_prev]+
		      image[line_next+pix_next])>>2];
    }else{     
      *R = crt_gamma[(((__u16)0)+image[line_start+pix_prev]+image[line_start+pix_next])>>1];
      *G = crt_gamma[image[src_ptr]];
      *B = crt_gamma[(((__u16)0)+image[line_prev+w]+image[line_next+w])>>1];
    }
  else
    if(evenw){ 
      *B = crt_gamma[(((__u16)0)+image[line_start+pix_prev]+image[line_start+pix_next])>>1];
      *G = crt_gamma[image[src_ptr]];
      *R = crt_gamma[(((__u16)0)+image[line_prev+w]+image[line_next+w])>>1];
    }else{     
      *R = crt_gamma[(((__u16)0)+image[line_prev+pix_prev]+image[line_prev+pix_next]+
		      image[line_next+pix_prev]+image[line_next+pix_next])>>2];
      *G = crt_gamma[(((__u16)0)+image[line_prev+w]+image[line_next+w]+
		      image[line_start+pix_prev]+image[line_start+pix_next])>>2];
      *B = crt_gamma[image[src_ptr]];
    }
  if(unlikely((saturation != SATDEF) && cfg_saturation)){
    __u8 Rs,Gs,Bs;
#ifdef EXPENSIVE_SAT
    Rs = SaturateR(*R,*G,*B,saturation);
    Gs = SaturateG(*R,*G,*B,saturation);
    Bs = SaturateB(*R,*G,*B,saturation);
#else
    Rs = SaturateRo(*R,*G,*B, _KRR, _KRG, _KRB);
    Gs = SaturateGo(*R,*G,*B, _KGR, _KGG, _KRB);
    Bs = SaturateBo(*R,*G,*B, _KGR, _KRG, _KBB);
#endif
    *R=Rs;
    *G=Gs;
    *B=Bs;
  }
  if(unlikely(_brightness && brightness_correction)){
    *R = norm_rgb_bright(*R,_brightness);
    *G = norm_rgb_bright(*G,_brightness);
    *B = norm_rgb_bright(*B,_brightness);
  }
}
__attribute__((always_inline)) inline void
awb_gen_line_grey_sentech(struct sentech_camera * cam_data,
			  __u8* frame , char *data,
			  int  UNUSED(max_width) , int wmax ,
			  int wstart , int  UNUSED(h) ,
			  const __u8 *  crt_gamma,
			  int  UNUSED(max_height), int wratio, int color) {
  int w, sw=0;
  int line_start = 0;
  char *p=data;
  int brightness = cam_data->brightness;
  if(color) {
    for (w=line_start+wstart;sw<wmax;sw++,w+=wratio) {
      *p = frame[w];
      p++;
    }
  }else{
    if(unlikely(brightness && brightness_correction)){
      for (w=line_start+wstart;sw<wmax;sw++,w+=wratio) {
	*p = norm_rgb_bright(crt_gamma[frame[w]],brightness);
	p++;
      }    
    }else{
      for (w=line_start+wstart;sw<wmax;sw++,w+=wratio) {
	*p = crt_gamma[frame[w]];
	p++;
      }
    }
  }
}
__attribute__((always_inline)) inline void
awb_gen_line_y16_sentech(struct sentech_camera * UNUSED(cam_data),
			 __u8* frame , char *data,
			 int  UNUSED(max_width) , int wmax ,
			 int wstart , int  UNUSED(h) ,
			 const __u8 *  UNUSED(crt_gamma),
			 int UNUSED(max_height)) {
  int w, wmax_abs;
  int line_start = 0;
  char *p=data;
  wmax_abs = line_start+wmax+wstart;
  for (w=line_start+wstart;w<wmax_abs;w++) {
    *p = frame[w];
    p++;
    *p = frame[w];
    p++;
  }
}
__attribute__((always_inline)) inline void
awb_gen_line_bgr24_sentech(struct sentech_camera *cam_data,
	       __u8* frame , char *data,
	       int max_width , int wmax ,
	       int wstart , int h ,
	       const __u8 * crt_gamma,
	       int max_height, int UNUSED(rows), int wratio,
	       int hasprev, int evenh) {
  int brightness = cam_data->brightness;
  int w, w_idx;
  int line_start = 0, line_prev = -max_width, line_next = max_width;
  char *p=data;
  int src_ptr;
  int pix_prev, pix_next;
  int wmax_abs = (wmax>>1)-1;
;
  if(unlikely(!hasprev)) line_prev = line_next;
  else if(unlikely(h==(max_height-1))) line_next = line_prev;
  w = wstart&(~1);
  src_ptr = line_start+w;
  pix_prev = w-1; pix_next = w+1;
  if(likely(!w))
    pix_prev = pix_next;
  interpolateRGBawb(p+2, p+1, p,
		    frame, brightness,
		    line_prev, pix_prev,
		    line_next, pix_next,
		    src_ptr, line_start, w,
		    evenh, 1, crt_gamma, cam_data);    
  p+=3;
  w+=wratio;
  src_ptr+=wratio;
  pix_prev = w-1;
  pix_next = w+1;
  if(likely(wratio&1)){ 
    if(evenh) {
      for (w_idx=0;w_idx<wmax_abs;w_idx++) {
	interpolateRGBawb(p+2, p+1, p,
			  frame, brightness,
			  line_prev, pix_prev,
			  line_next, pix_next,
			  src_ptr, line_start, w,
			  1, 0, crt_gamma, cam_data);    
	p+=3;
	w+=wratio;
	src_ptr+=wratio;
	pix_prev = w-1;
	pix_next = w+1;
	interpolateRGBawb(p+2, p+1, p,
			frame, brightness,
			  line_prev, pix_prev,
			  line_next, pix_next,
			  src_ptr, line_start, w,
			  1, 1, crt_gamma, cam_data);    
      p+=3;
	w+=wratio;
	src_ptr+=wratio;
	pix_prev = w-1;
	pix_next = w+1;
      }
    }else{
      for (w_idx=0;w_idx<wmax_abs;w_idx++) {
	interpolateRGBawb(p+2, p+1, p,
			  frame, brightness,
			  line_prev, pix_prev,
			  line_next, pix_next,
			  src_ptr, line_start, w,
			  0, 0, crt_gamma, cam_data);    
	p+=3;
	w+=wratio;
	src_ptr+=wratio;
	pix_prev = w-1;
	pix_next = w+1;
	interpolateRGBawb(p+2, p+1, p,
			  frame, brightness,
			  line_prev, pix_prev,
			  line_next, pix_next,
			  src_ptr, line_start, w,
			  0, 1, crt_gamma, cam_data);    
      p+=3;
	w+=wratio;
	src_ptr+=wratio;
	pix_prev = w-1;
	pix_next = w+1;
      }
    }
  } else {
    if(evenh) {
      for (w_idx=0;w_idx<wmax_abs;w_idx++) {
	interpolateRGBawb(p+2, p+1, p,
			  frame, brightness,
			  line_prev, pix_prev,
			  line_next, pix_next,
			  src_ptr, line_start, w,
			  1, 1, crt_gamma, cam_data);    
	p+=3;
	w+=wratio;
	src_ptr+=wratio;
	pix_prev = w-1;
	pix_next = w+1;
	interpolateRGBawb(p+2, p+1, p,
			frame, brightness,
			  line_prev, pix_prev,
			  line_next, pix_next,
			  src_ptr, line_start, w,
			  1, 1, crt_gamma, cam_data);    
      p+=3;
	w+=wratio;
	src_ptr+=wratio;
	pix_prev = w-1;
	pix_next = w+1;
      }
    }else{
      for (w_idx=0;w_idx<wmax_abs;w_idx++) {
	interpolateRGBawb(p+2, p+1, p,
			  frame, brightness,
			  line_prev, pix_prev,
			  line_next, pix_next,
			  src_ptr, line_start, w,
			  0, 1, crt_gamma, cam_data);    
	p+=3;
	w+=wratio;
	src_ptr+=wratio;
	pix_prev = w-1;
	pix_next = w+1;
	interpolateRGBawb(p+2, p+1, p,
			  frame, brightness,
			  line_prev, pix_prev,
			  line_next, pix_next,
			  src_ptr, line_start, w,
			  0, 1, crt_gamma, cam_data);    
      p+=3;
	w+=wratio;
	src_ptr+=wratio;
	pix_prev = w-1;
	pix_next = w+1;
      }
    }
  }
  if(likely(w==(max_width-1)))
    pix_next = pix_prev;
  interpolateRGBawb(p+2, p+1, p,
		    frame, brightness,
		    line_prev, pix_prev,
		    line_next, pix_next,
		    src_ptr, line_start, w,
		    evenh, !(wratio&1), crt_gamma, cam_data);    
  p+=3;
;
}
__attribute__((always_inline)) inline void
awb_gen_line_rgb24_sentech(struct sentech_camera *cam_data,
	       __u8* frame , char *data,
	       int max_width , int wmax ,
	       int wstart , int h ,
	       const __u8 * crt_gamma,
	       int max_height, int UNUSED(rows), int wratio,
	       int hasprev, int evenh) {
  int brightness = cam_data->brightness;
  int w, w_idx;
  int line_start = 0, line_prev = -max_width, line_next = max_width;
  char *p=data;
  int src_ptr;
  int pix_prev, pix_next;
  int wmax_abs = (wmax>>1)-1;;
  if(unlikely(!hasprev)) line_prev = line_next;
  else if(unlikely(h==(max_height-1))) line_next = line_prev;
  w = wstart&(~1);
  src_ptr = line_start+w;
  pix_prev = w-1; pix_next = w+1;
  if(likely(!w))
    pix_prev = pix_next;
  interpolateRGBawb(p, p+1, p+2,
		    frame, brightness,
		    line_prev, pix_prev,
		    line_next, pix_next,
		    src_ptr, line_start, w,
		    evenh, 1, crt_gamma, cam_data);    
  p+=3;
  w+=wratio;
  src_ptr+=wratio;
  pix_prev = w-1;
  pix_next = w+1;
  if(likely(wratio&1)){ 
    if(evenh) {
      for (w_idx=0;w_idx<wmax_abs;w_idx++) {
	interpolateRGBawb(p, p+1, p+2,
			  frame, brightness,
			  line_prev, pix_prev,
			  line_next, pix_next,
			  src_ptr, line_start, w,
			  1, 0, crt_gamma, cam_data);    
	p+=3;
	w+=wratio;
	src_ptr+=wratio;
	pix_prev = w-1;
	pix_next = w+1;
	interpolateRGBawb(p, p+1, p+2,
			frame, brightness,
			  line_prev, pix_prev,
			  line_next, pix_next,
			  src_ptr, line_start, w,
			  1, 1, crt_gamma, cam_data);    
      p+=3;
	w+=wratio;
	src_ptr+=wratio;
	pix_prev = w-1;
	pix_next = w+1;
      }
    }else{
      for (w_idx=0;w_idx<wmax_abs;w_idx++) {
	interpolateRGBawb(p, p+1, p+2,
			  frame, brightness,
			  line_prev, pix_prev,
			  line_next, pix_next,
			  src_ptr, line_start, w,
			  0, 0, crt_gamma, cam_data);    
	p+=3;
	w+=wratio;
	src_ptr+=wratio;
	pix_prev = w-1;
	pix_next = w+1;
	interpolateRGBawb(p, p+1, p+2,
			  frame, brightness,
			  line_prev, pix_prev,
			  line_next, pix_next,
			  src_ptr, line_start, w,
			  0, 1, crt_gamma, cam_data);    
      p+=3;
	w+=wratio;
	src_ptr+=wratio;
	pix_prev = w-1;
	pix_next = w+1;
      }
    }
  } else {
    if(evenh) {
      for (w_idx=0;w_idx<wmax_abs;w_idx++) {
	interpolateRGBawb(p, p+1, p+2,
			  frame, brightness,
			  line_prev, pix_prev,
			  line_next, pix_next,
			  src_ptr, line_start, w,
			  1, 1, crt_gamma, cam_data);    
	p+=3;
	w+=wratio;
	src_ptr+=wratio;
	pix_prev = w-1;
	pix_next = w+1;
	interpolateRGBawb(p, p+1, p+2,
			frame, brightness,
			  line_prev, pix_prev,
			  line_next, pix_next,
			  src_ptr, line_start, w,
			  1, 1, crt_gamma, cam_data);    
      p+=3;
	w+=wratio;
	src_ptr+=wratio;
	pix_prev = w-1;
	pix_next = w+1;
      }
    }else{
      for (w_idx=0;w_idx<wmax_abs;w_idx++) {
	interpolateRGBawb(p, p+1, p+2,
			  frame, brightness,
			  line_prev, pix_prev,
			  line_next, pix_next,
			  src_ptr, line_start, w,
			  0, 1, crt_gamma, cam_data);    
	p+=3;
	w+=wratio;
	src_ptr+=wratio;
	pix_prev = w-1;
	pix_next = w+1;
	interpolateRGBawb(p, p+1, p+2,
			  frame, brightness,
			  line_prev, pix_prev,
			  line_next, pix_next,
			  src_ptr, line_start, w,
			  0, 1, crt_gamma, cam_data);    
      p+=3;
	w+=wratio;
	src_ptr+=wratio;
	pix_prev = w-1;
	pix_next = w+1;
      }
    }
  }
  if(likely(w==(max_width-1)))
    pix_next = pix_prev;
  interpolateRGBawb(p, p+1, p+2,
		    frame, brightness,
		    line_prev, pix_prev,
		    line_next, pix_next,
		    src_ptr, line_start, w,
		    evenh, !(wratio&1), crt_gamma, cam_data);    
  p+=3;
}
__attribute__((always_inline)) inline void
awb_gen_line_yuyv_sentech(struct sentech_camera *cam_data,
			  __u8* frame , char *data,
			  int max_width , int wmax ,
			  int wstart , int h ,
			  const __u8 * crt_gamma,
			  int max_height, int UNUSED(rows), int wratio,
			  int hasprev, int evenh) {
  int brightness = cam_data->brightness;
  int w, w_idx;
  int line_start = 0, line_prev = -max_width, line_next = max_width;
  __u8 R, G, B;
  __u8 Y, U, V;
  char *p=data;
  int src_ptr;
  int pix_prev, pix_next;
  int wmax_abs = (wmax>>1)-1;
  if(unlikely(!hasprev)) line_prev = line_next;
  else if(unlikely(h==(max_height-1))) line_next = line_prev;
  w = wstart&(~1);
  src_ptr = line_start+w;
  pix_prev = w-1; pix_next = w+1;
  if(likely(!w))
    pix_prev = pix_next;
  interpolateRGBawb(&R, &G, &B,
		    frame, brightness,
		    line_prev, pix_prev,
		    line_next, pix_next,
		    src_ptr, line_start, w,
		    evenh, 1, crt_gamma, cam_data);    
  Y = TO_Y(R,G,B);
  *p = Y;
  p++;
  U=TO_U(R,G,B);
  *p = U;
  p++;
  w+=wratio;
  src_ptr+=wratio;
  pix_prev = w-1;
  pix_next = w+1;
  if(likely(wratio&1)){ 
    if(evenh) {
      for (w_idx=0;w_idx<wmax_abs;w_idx++) {
	interpolateRGBawb(&R, &G, &B,
			  frame, brightness,
			  line_prev, pix_prev,
			  line_next, pix_next,
			  src_ptr, line_start, w,
			  1, 0, crt_gamma, cam_data);    
	Y = TO_Y(R,G,B);
	*p = Y;
	p++;
	V=TO_V(R,G,B);
	*p = V;
	p++;
	w+=wratio;
	src_ptr+=wratio;
	pix_prev = w-1;
	pix_next = w+1;
	interpolateRGBawb(&R, &G, &B,
			frame, brightness,
			  line_prev, pix_prev,
			  line_next, pix_next,
			  src_ptr, line_start, w,
			  1, 1, crt_gamma, cam_data);    
	Y = TO_Y(R,G,B);
	*p = Y;
	p++;
	U=TO_U(R,G,B);
	*p = U;
	p++;
	w+=wratio;
	src_ptr+=wratio;
	pix_prev = w-1;
	pix_next = w+1;
      }
    }else{
      for (w_idx=0;w_idx<wmax_abs;w_idx++) {
	interpolateRGBawb(&R, &G, &B,
			  frame, brightness,
			  line_prev, pix_prev,
			  line_next, pix_next,
			  src_ptr, line_start, w,
			  0, 0, crt_gamma, cam_data);    
	Y = TO_Y(R,G,B);
	*p = Y;
	p++;
	V=TO_V(R,G,B);
	*p = V;
	p++;
	w+=wratio;
	src_ptr+=wratio;
	pix_prev = w-1;
	pix_next = w+1;
	interpolateRGBawb(&R, &G, &B,
			  frame, brightness,
			  line_prev, pix_prev,
			  line_next, pix_next,
			  src_ptr, line_start, w,
			  0, 1, crt_gamma, cam_data);    
	Y = TO_Y(R,G,B);
	*p = Y;
	p++;
	U=TO_U(R,G,B);
	*p = U;
	p++;
	w+=wratio;
	src_ptr+=wratio;
	pix_prev = w-1;
	pix_next = w+1;
      }
    }
  } else {
    if(evenh) {
      for (w_idx=0;w_idx<wmax_abs;w_idx++) {
	interpolateRGBawb(&R, &G, &B,
			  frame, brightness,
			  line_prev, pix_prev,
			  line_next, pix_next,
			  src_ptr, line_start, w,
			  1, 1, crt_gamma, cam_data);    
	Y = TO_Y(R,G,B);
	*p = Y;
	p++;
	V=TO_V(R,G,B);
	*p = V;
	p++;
	w+=wratio;
	src_ptr+=wratio;
	pix_prev = w-1;
	pix_next = w+1;
	interpolateRGBawb(&R, &G, &B,
			frame, brightness,
			  line_prev, pix_prev,
			  line_next, pix_next,
			  src_ptr, line_start, w,
			  1, 1, crt_gamma, cam_data);    
	Y = TO_Y(R,G,B);
	*p = Y;
	p++;
	U=TO_U(R,G,B);
	*p = U;
	p++;
	w+=wratio;
	src_ptr+=wratio;
	pix_prev = w-1;
	pix_next = w+1;
      }
    }else{
      for (w_idx=0;w_idx<wmax_abs;w_idx++) {
	interpolateRGBawb(&R, &G, &B,
			  frame, brightness,
			  line_prev, pix_prev,
			  line_next, pix_next,
			  src_ptr, line_start, w,
			  0, 1, crt_gamma, cam_data);    
	Y = TO_Y(R,G,B);
	*p = Y;
	p++;
	V=TO_V(R,G,B);
	*p = V;
	p++;
	w+=wratio;
	src_ptr+=wratio;
	pix_prev = w-1;
	pix_next = w+1;
	interpolateRGBawb(&R, &G, &B,
			  frame, brightness,
			  line_prev, pix_prev,
			  line_next, pix_next,
			  src_ptr, line_start, w,
			  0, 1, crt_gamma, cam_data);    
	Y = TO_Y(R,G,B);
	*p = Y;
	p++;
	U=TO_U(R,G,B);
	*p = U;
	p++;
	w+=wratio;
	src_ptr+=wratio;
	pix_prev = w-1;
	pix_next = w+1;
      }
    }
  }
  if(likely(w==(max_width-1)))
    pix_next = pix_prev;
  interpolateRGBawb(&R, &G, &B,
		    frame, brightness,
		    line_prev, pix_prev,
		    line_next, pix_next,
		    src_ptr, line_start, w,
		    evenh, !(wratio&1), crt_gamma, cam_data);    
  Y = TO_Y(R,G,B);
  *p = Y;
  p++;
  V=TO_V(R,G,B);
  *p = V;
}
#define yuv420_even_rows(U, V, R, G, B, dataU, dataV) { \
U=TO_U(R,G,B); *dataU = U; dataU++; \
V=TO_V(R,G,B); *dataV = V; dataV++; }
#define yuv420_odd_rows(U16, V16, R, G, B, dataU, dataV) { \
    U16=TO_U(R,G,B); *dataU = (*dataU + U16)>>1; dataU++;	\
    V16=TO_V(R,G,B); *dataV = (*dataV + V16)>>1; dataV++; }
__attribute__((always_inline)) inline void
awb_gen_line_yuv420_sentech(struct sentech_camera *cam_data,
			    __u8* frame , __u8 *data ,
			    int max_width , int wmax ,
			    int wstart , int h ,
			    const __u8 * crt_gamma, __u8 *dataU, __u8 *dataV,
			    int max_height, int rows, int wratio,
			    int hasprev, int evenh) {
  int brightness = cam_data->brightness;
  int w, w_idx;
  int line_start = 0, line_prev = -max_width, line_next = max_width;
  __u8 R, G, B;
  register __u8 Y, U, V;
  __u16 U16, V16;
  int src_ptr;
  int pix_prev, pix_next;
  int wmax_abs = (wmax>>1)-1;
  int even_rows = !(1 & rows);
  int odd_rows = !even_rows;
  if(unlikely(!hasprev)) line_prev = line_next;
  else if(unlikely(h==(max_height-1))) line_next = line_prev;
  w = wstart&(~1);
  src_ptr = line_start+w;
  pix_prev = w-1; pix_next = w+1;
  if(likely(!w))
    pix_prev = pix_next;
  interpolateRGBawb(&R, &G, &B,
		    frame, brightness,
		    line_prev, pix_prev,
		    line_next, pix_next,
		    src_ptr, line_start, w,
		    evenh, 1, crt_gamma, cam_data);    
  Y = TO_Y(R,G,B);
  *data = Y; data++;
  if(even_rows) {
    yuv420_even_rows(U, V, R, G, B, dataU, dataV);
  }
  w+=wratio;
  src_ptr+=wratio;
  pix_prev = w-1;
  pix_next = w+1;
  if(likely(wratio&1)){ 
    if(evenh) {
      for (w_idx=0;w_idx<wmax_abs;w_idx++) {
	interpolateRGBawb(&R, &G, &B,
			  frame, brightness,
			  line_prev, pix_prev,
			  line_next, pix_next,
			  src_ptr, line_start, w,
			  1, 0, crt_gamma, cam_data);    
	Y = TO_Y(R,G,B);
	*data = Y; data++;
	if(odd_rows) {
	  yuv420_odd_rows(U16, V16, R, G, B, dataU, dataV);
	}
	w+=wratio;
	src_ptr+=wratio;
	pix_prev = w-1;
	pix_next = w+1;
	interpolateRGBawb(&R, &G, &B,
			frame, brightness,
			  line_prev, pix_prev,
			  line_next, pix_next,
			  src_ptr, line_start, w,
			  1, 1, crt_gamma, cam_data);    
	Y = TO_Y(R,G,B);
	*data = Y; data++;
	if(even_rows) {
	  yuv420_even_rows(U, V, R, G, B, dataU, dataV);
	}
	w+=wratio;
	src_ptr+=wratio;
	pix_prev = w-1;
	pix_next = w+1;
      }
    }else{
      for (w_idx=0;w_idx<wmax_abs;w_idx++) {
	interpolateRGBawb(&R, &G, &B,
			  frame, brightness,
			  line_prev, pix_prev,
			  line_next, pix_next,
			  src_ptr, line_start, w,
			  0, 0, crt_gamma, cam_data);    
	Y = TO_Y(R,G,B);
	*data = Y; data++;
	if(odd_rows) {
	  yuv420_odd_rows(U16, V16, R, G, B, dataU, dataV);
	}
	w+=wratio;
	src_ptr+=wratio;
	pix_prev = w-1;
	pix_next = w+1;
	interpolateRGBawb(&R, &G, &B,
			  frame, brightness,
			  line_prev, pix_prev,
			  line_next, pix_next,
			  src_ptr, line_start, w,
			  0, 1, crt_gamma, cam_data);    
	Y = TO_Y(R,G,B);
	*data = Y; data++;
	if(even_rows) {
	  yuv420_even_rows(U, V, R, G, B, dataU, dataV);
	}
	w+=wratio;
	src_ptr+=wratio;
	pix_prev = w-1;
	pix_next = w+1;
      }
    }
  } else {
    if(evenh) {
      for (w_idx=0;w_idx<wmax_abs;w_idx++) {
	interpolateRGBawb(&R, &G, &B,
			  frame, brightness,
			  line_prev, pix_prev,
			  line_next, pix_next,
			  src_ptr, line_start, w,
			  1, 1, crt_gamma, cam_data);    
	Y = TO_Y(R,G,B);
	*data = Y; data++;
	if(odd_rows) {
	  yuv420_odd_rows(U16, V16, R, G, B, dataU, dataV);
	}
	w+=wratio;
	src_ptr+=wratio;
	pix_prev = w-1;
	pix_next = w+1;
	interpolateRGBawb(&R, &G, &B,
			frame, brightness,
			  line_prev, pix_prev,
			  line_next, pix_next,
			  src_ptr, line_start, w,
			  1, 1, crt_gamma, cam_data);    
	Y = TO_Y(R,G,B);
	*data = Y; data++;
	if(even_rows) {
	  yuv420_even_rows(U, V, R, G, B, dataU, dataV);
	}
	w+=wratio;
	src_ptr+=wratio;
	pix_prev = w-1;
	pix_next = w+1;
      }
    }else{
      for (w_idx=0;w_idx<wmax_abs;w_idx++) {
	interpolateRGBawb(&R, &G, &B,
			  frame, brightness,
			  line_prev, pix_prev,
			  line_next, pix_next,
			  src_ptr, line_start, w,
			  0, 1, crt_gamma, cam_data);    
	Y = TO_Y(R,G,B);
	*data = Y; data++;
	if(odd_rows) {
	  yuv420_odd_rows(U16, V16, R, G, B, dataU, dataV);
	}
	w+=wratio;
	src_ptr+=wratio;
	pix_prev = w-1;
	pix_next = w+1;
	interpolateRGBawb(&R, &G, &B,
			  frame, brightness,
			  line_prev, pix_prev,
			  line_next, pix_next,
			  src_ptr, line_start, w,
			  0, 1, crt_gamma, cam_data);    
	Y = TO_Y(R,G,B);
	*data = Y; data++;
	if(even_rows) {
	  yuv420_even_rows(U, V, R, G, B, dataU, dataV);
	}
	w+=wratio;
	src_ptr+=wratio;
	pix_prev = w-1;
	pix_next = w+1;
      }
    }
  }
  if(likely(w==(max_width-1)))
    pix_next = pix_prev;
  interpolateRGBawb(&R, &G, &B,
		    frame, brightness,
		    line_prev, pix_prev,
		    line_next, pix_next,
		    src_ptr, line_start, w,
		    evenh, !(wratio&1), crt_gamma, cam_data);    
  Y = TO_Y(R,G,B);
  *data = Y; data++;
  if(odd_rows) {
    yuv420_odd_rows(U16, V16, R, G, B, dataU, dataV);
  }
}
#define yuv422_even_rows(U, V, R, G, B, dataU, dataV) { \
U=TO_U(R,G,B); *dataU = U; \
V=TO_V(R,G,B); *dataV = V; }
#define yuv422_odd_rows(U16, V16, R, G, B, dataU, dataV) { \
    U16=TO_U(R,G,B); *dataU = (*dataU + U16)>>1; dataU++;	\
    V16=TO_V(R,G,B); *dataV = (*dataV + V16)>>1; dataV++; }
__attribute__((always_inline)) inline void
awb_gen_line_yuv422P_sentech(struct sentech_camera *cam_data,
			    __u8* frame , __u8 *data ,
			    int max_width , int wmax ,
			    int wstart , int h ,
			    const __u8 * crt_gamma, __u8 *dataU, __u8 *dataV,
			    int max_height, int rows, int wratio,
			    int hasprev, int evenh) {
  int brightness = cam_data->brightness;
  int w, w_idx;
  int line_start = 0, line_prev = -max_width, line_next = max_width;
  __u8 R, G, B;
  register __u8 Y, U, V;
  register __u16 U16, V16;
  int src_ptr;
  int pix_prev, pix_next;
  int wmax_abs = (wmax>>1)-1;
  if(unlikely(!hasprev)) line_prev = line_next;
  else if(unlikely(h==(max_height-1))) line_next = line_prev;
  w = wstart&(~1);
  src_ptr = line_start+w;
  pix_prev = w-1; pix_next = w+1;
  if(likely(!w))
    pix_prev = pix_next;
  interpolateRGBawb(&R, &G, &B,
		    frame, brightness,
		    line_prev, pix_prev,
		    line_next, pix_next,
		    src_ptr, line_start, w,
		    evenh, 1, crt_gamma, cam_data);    
  Y = TO_Y(R,G,B);
  *data = Y; data++;
  U=TO_U(R,G,B);
  *dataU = U; 
  V=TO_V(R,G,B);
  *dataV = V; 
  w+=wratio;
  src_ptr+=wratio;
  pix_prev = w-1;
  pix_next = w+1;
  if(likely(wratio&1)){ 
    if(evenh) {
      for (w_idx=0;w_idx<wmax_abs;w_idx++) {
	interpolateRGBawb(&R, &G, &B,
			  frame, brightness,
			  line_prev, pix_prev,
			  line_next, pix_next,
			  src_ptr, line_start, w,
			  1, 0, crt_gamma, cam_data);    
	Y = TO_Y(R,G,B);
	*data = Y;
	data++;
	U16=TO_U(R,G,B);
	*dataU = (*dataU + U16)>>1;
	dataU++;
	V16=TO_V(R,G,B);
	*dataV = (*dataV + V16)>>1;
	dataV++;
	w+=wratio;
	src_ptr+=wratio;
	pix_prev = w-1;
	pix_next = w+1;
	interpolateRGBawb(&R, &G, &B,
			frame, brightness,
			  line_prev, pix_prev,
			  line_next, pix_next,
			  src_ptr, line_start, w,
			  1, 1, crt_gamma, cam_data);    
	Y = TO_Y(R,G,B);
	*data = Y; data++;
	U=TO_U(R,G,B);
	*dataU = U; 
	V=TO_V(R,G,B);
	*dataV = V; 
	w+=wratio;
	src_ptr+=wratio;
	pix_prev = w-1;
	pix_next = w+1;
      }
    }else{
      for (w_idx=0;w_idx<wmax_abs;w_idx++) {
	interpolateRGBawb(&R, &G, &B,
			  frame, brightness,
			  line_prev, pix_prev,
			  line_next, pix_next,
			  src_ptr, line_start, w,
			  0, 0, crt_gamma, cam_data);    
	Y = TO_Y(R,G,B);
	*data = Y; data++;
	U16=TO_U(R,G,B);
	*dataU = (*dataU + U16)>>1;
	dataU++;
	V16=TO_V(R,G,B);
	*dataV = (*dataV + V16)>>1;
	dataV++;
	w+=wratio;
	src_ptr+=wratio;
	pix_prev = w-1;
	pix_next = w+1;
	interpolateRGBawb(&R, &G, &B,
			  frame, brightness,
			  line_prev, pix_prev,
			  line_next, pix_next,
			  src_ptr, line_start, w,
			  0, 1, crt_gamma, cam_data);    
	Y = TO_Y(R,G,B);
	*data = Y; data++;
	U=TO_U(R,G,B);
	*dataU = U; 
	V=TO_V(R,G,B);
	*dataV = V; 
	w+=wratio;
	src_ptr+=wratio;
	pix_prev = w-1;
	pix_next = w+1;
      }
    }
  } else {
    if(evenh) {
      for (w_idx=0;w_idx<wmax_abs;w_idx++) {
	interpolateRGBawb(&R, &G, &B,
			  frame, brightness,
			  line_prev, pix_prev,
			  line_next, pix_next,
			  src_ptr, line_start, w,
			  1, 1, crt_gamma, cam_data);    
	Y = TO_Y(R,G,B);
	*data = Y; data++;
	U16=TO_U(R,G,B);
	*dataU = (*dataU + U16)>>1;
	dataU++;
	V16=TO_V(R,G,B);
	*dataV = (*dataV + V16)>>1;
	dataV++;
	w+=wratio;
	src_ptr+=wratio;
	pix_prev = w-1;
	pix_next = w+1;
	interpolateRGBawb(&R, &G, &B,
			frame, brightness,
			  line_prev, pix_prev,
			  line_next, pix_next,
			  src_ptr, line_start, w,
			  1, 1, crt_gamma, cam_data);    
	Y = TO_Y(R,G,B);
	*data = Y; data++;
	U=TO_U(R,G,B);
	*dataU = U; 
	V=TO_V(R,G,B);
	*dataV = V; 
	w+=wratio;
	src_ptr+=wratio;
	pix_prev = w-1;
	pix_next = w+1;
      }
    }else{
      for (w_idx=0;w_idx<wmax_abs;w_idx++) {
	interpolateRGBawb(&R, &G, &B,
			  frame, brightness,
			  line_prev, pix_prev,
			  line_next, pix_next,
			  src_ptr, line_start, w,
			  0, 1, crt_gamma, cam_data);    
	Y = TO_Y(R,G,B);
	*data = Y; data++;
	U16=TO_U(R,G,B);
	*dataU = (*dataU + U16)>>1;
	dataU++;
	V16=TO_V(R,G,B);
	*dataV = (*dataV + V16)>>1;
	dataV++;
	w+=wratio;
	src_ptr+=wratio;
	pix_prev = w-1;
	pix_next = w+1;
	interpolateRGBawb(&R, &G, &B,
			  frame, brightness,
			  line_prev, pix_prev,
			  line_next, pix_next,
			  src_ptr, line_start, w,
			  0, 1, crt_gamma, cam_data);    
	Y = TO_Y(R,G,B);
	*data = Y; data++;
	U=TO_U(R,G,B);
	*dataU = U; 
	V=TO_V(R,G,B);
	*dataV = V; 
	w+=wratio;
	src_ptr+=wratio;
	pix_prev = w-1;
	pix_next = w+1;
      }
    }
  }
  if(likely(w==(max_width-1)))
    pix_next = pix_prev;
  interpolateRGBawb(&R, &G, &B,
		    frame, brightness,
		    line_prev, pix_prev,
		    line_next, pix_next,
		    src_ptr, line_start, w,
		    evenh, !(wratio&1), crt_gamma, cam_data);    
  Y = TO_Y(R,G,B);
  *data = Y; data++;
  U16=TO_U(R,G,B);
  *dataU = (*dataU + U16)>>1;
  dataU++;
  V16=TO_V(R,G,B);
  *dataV = (*dataV + V16)>>1;
  dataV++;
}
static int
awb_move_sub_to_vmalloc(struct sentech_usb_buffer* sub) {
  struct sentech_camera* cam_data = sub->cam_data;
  char *data = sub->vbuffer;
  unsigned int count = sub->vbuffer_size;
  struct sentech_usb_fragment* suf;
  int row, rows=0;
  unsigned int retval = 0;
  int inc;
  __u8* lptr;
  int wratio=1;
  int hratio=1;
  int evenh = 0;
  int switchh = 1;
  int rows_output = 0; 
  int bytesperline = cam_data->pix_format.bytesperline;
  char * dataU;
  char * dataV;
  int color = 0; 
  const __u8 * crt_gamma = sentech_gamma[cam_data->gamma];
  if(cam_data->pix_format.sizeimage<sub->vbuffer_size)
    count = cam_data->pix_format.sizeimage;
  if(cam_data->pix_format.sizeimage > count) {
    return -EINVAL;
  }
  if(param_old_scale){
    if(scale) {
      if(cam_data->crop.width)
	wratio = cam_data->max_width / cam_data->crop.width;
      if(cam_data->crop.height) 
	hratio = cam_data->scan_height / cam_data->crop.height;
    }
  }else{
    wratio = cam_data->crop.width / cam_data->output.width;
    hratio = cam_data->crop.height / cam_data->output.height;
  }
  if(hratio<1) hratio = 1;
  switchh = (hratio & 1);
  suf = first_suf(sub);
  down(&cam_data->lock);
  if(!(cam_data->awb_buffer && cam_data->awb_object)) {
    up(&cam_data->lock);;
    return 0;
  }
  if (cam_data->height_changed) {
    int i;;
    Sentech_FreeAWBObject(cam_data->awb_object);
;
    Sentech_InitAWBObject(cam_data->awb_object,
			  cam_data->awb_buffer,
			  &cam_data->state12.line_numbers_le,
			  &cam_data->state12.one_line_clock_numbers_le,
			  &cam_data->payload21.Shutter_line_numbers_le,
			  &cam_data->payload21.Shutter_clock_numbers_le,
			  &cam_data->payload21.Gain_le,
			  &cam_data->max_width,
			  &cam_data->scan_height,
			  &cam_data->modif_flag21,
			  &cam_data->red,
			  &cam_data->green,
			  &cam_data->blue,
			  &cam_data->state23.R_Gain_le,
			  __le16_to_cpu(cam_data->state12.initial_gain_le),
			  __le16_to_cpu(cam_data->state12.initial_digital_gain_le),
			  cam_data->trigger_version 
#ifdef L_BOUNDS
			      ,&cam_data->gain_min
			      ,&cam_data->gain_max
			      ,&cam_data->exposure_min
			      ,&cam_data->exposure_max
#endif
			  );
    if(cam_data->auto_white_balance)
      Sentech_SetAutoAWB(cam_data->awb_object, STCAM_WB_FULLAUTO);
    if(cam_data->auto_gain) {
      Sentech_SetAutoGain(cam_data->awb_object,
			  (STCAM_AGC_ON|STCAM_AUTO_SHUTTER_ON));
      Sentech_SetAutoExposure(cam_data->awb_object,
			      (STCAM_AGC_ON|STCAM_AUTO_SHUTTER_ON));
    }
    for(i = 0; i < ARRAY_SIZE(sentech_qctrl); i++) {
      if (sentech_qctrl[i].id == V4L2_CID_STC_TARGET){
	BYTE target[]= {0,5,5};
	target[0] = cam_data->qctl_regs[i];
	Sentech_CLuminance_SetTarget(cam_data->awb_object,
				     target);
	break;
      }
    }
;
    cam_data->height_changed = 0;
  }else{;
  }
  if (cam_data->skiped_awb>=skip_awb
      ) {
    switch(cam_data->fmt->fourcc){
    case V4L2_PIX_FMT_SBGGR8:
    case V4L2_PIX_FMT_SRGGB8:
    case V4L2_PIX_FMT_GREY:
#ifdef ST_ImageCopyBalance
      Sentech_ImageCopyBalance(cam_data->awb_object, suf->framebuffer_image);
#else
      memcpy(cam_data->awb_buffer, suf->framebuffer_image, cam_data->scanimage_size);
      Sentech_ImageBalance(cam_data->awb_object);
#endif
      break;
    default:;
      Sentech_ImageProcessing(cam_data->awb_object, suf->framebuffer_image);;
      Sentech_ALCWBalance(cam_data->awb_object);
    }
    cam_data->recent_autobalance = 1;
    cam_data->skiped_awb = 0;
  } else{
    cam_data->skiped_awb++;
    switch(cam_data->fmt->fourcc){
    case V4L2_PIX_FMT_SBGGR8:
    case V4L2_PIX_FMT_SRGGB8:
    color = 1;
    case V4L2_PIX_FMT_GREY:
      if(param_old_scale){
	row = (scale?0:cam_data->crop.top);
	lptr = suf->framebuffer_image + row*cam_data->max_width;;
      }else{
	row = cam_data->crop.top;
	lptr = suf->framebuffer_image + row*cam_data->max_width;
      }
      goto grey_move;
    default:
;
      Sentech_ImageProcessing(cam_data->awb_object, suf->framebuffer_image);
    }
  }
  if(trace_buffers>1)
    memcpy(sub->vbuffer + cam_data->pix_format.sizeimage - (trace_buffers-1)*cam_data->max_frame_size,
	   suf->framebuffer_image, cam_data->scanimage_size);
  switch(cam_data->fmt->fourcc){
  case V4L2_PIX_FMT_SBGGR8:;
  case V4L2_PIX_FMT_SRGGB8:;
    color = 1;
  case V4L2_PIX_FMT_GREY:;
    if(param_old_scale)
      row = (scale?0:cam_data->crop.top);
    else
      row = cam_data->crop.top;
    lptr = cam_data->awb_buffer + row*cam_data->max_width;    
  grey_move:;
    if(param_old_scale)
      inc = cam_data->crop.width;
    else
      inc = cam_data->output.width;
    if(cam_data->hflip){;
      wratio = -wratio;
      lptr += cam_data->max_width-1;
    };
    if(
       cam_data->enable_native &&
       ((param_old_scale && !scale)||(wratio==1))) {
      if( ((param_old_scale && !scale)||hratio==1)&&
	  (((unsigned)abs(cam_data->crop.width)) == cam_data->max_width)){
	retval=cam_data->crop.height*cam_data->max_width;
	memcpy(data, lptr, retval);;
      }else{;
	if(param_old_scale){
	  for (rows=0; rows<cam_data->crop.height; rows++, row+=hratio){
	    memcpy(data, lptr+(scale?0:cam_data->crop.left), inc);
	    data += inc;
	    retval += inc;
	    lptr += cam_data->max_width*hratio;
	  }
	}else{
	  for (rows=0; rows<cam_data->output.height; rows++, row+=hratio){
	    memcpy(data, lptr+cam_data->crop.left, inc);
	    data += inc;
	    retval += inc;
	    lptr += cam_data->max_width*hratio;
	  }
	}
      }
      break;
    }
;
    if(cam_data->enable_native) color = 1; 
    if(param_old_scale){
      for (rows=0;rows<cam_data->crop.height;rows++, row+=hratio) {
	int crop_left = scale?0:cam_data->crop.left;
	if(cam_data->hflip) crop_left = -crop_left;
	awb_gen_line_grey_sentech(cam_data, 
		    lptr, data, 
		    cam_data->max_width, cam_data->crop.width,
		    crop_left, row, crt_gamma,
				  cam_data->max_height, wratio, color);
	data += inc;
	retval += inc;
	lptr += cam_data->max_width*hratio;
      }
    }else{
      for (rows=0;rows<cam_data->output.height;rows++, row+=hratio) {
	int crop_left = cam_data->crop.left;
	if(cam_data->hflip) crop_left = -crop_left;
	awb_gen_line_grey_sentech(cam_data, 
		    lptr, data, 
				  cam_data->max_width,
				  cam_data->output.width,
				  crop_left,
				  row, crt_gamma,
				  cam_data->max_height, wratio, color);
	data += inc;
	retval += inc;
	lptr += cam_data->max_width*hratio;
      }
    }
    break;
  case V4L2_PIX_FMT_Y16:;
    if(!param_old_scale) {retval=0;break;}
    inc = cam_data->crop.width << 1;
    suf = first_suf(sub);
    row = cam_data->crop.top;
    lptr = cam_data->awb_buffer + 
      row*cam_data->max_width;
    for (rows=0;rows<cam_data->crop.height;rows++, row++) {
      awb_gen_line_y16_sentech(cam_data, 
		    lptr, data, 
		    cam_data->max_width, cam_data->crop.width,
		    cam_data->crop.left, row, crt_gamma,
		    cam_data->max_height);
      data += inc;
      retval += inc;
      lptr += cam_data->max_width;
    }
    break;
  case V4L2_PIX_FMT_YUYV:;;
    if(param_old_scale){
      inc = cam_data->crop.width << 1;
      row = (scale?0:cam_data->crop.top);
      evenh = (scale)?(!gbrg):(!(cam_data->crop.top&1));
    }else{
      inc = cam_data->output.width << 1;
      row = cam_data->crop.top;
      evenh = !(cam_data->crop.top&1);
    }
    lptr = cam_data->awb_buffer + row*cam_data->max_width;
    rows_output = (param_old_scale)?
      cam_data->crop.height:
      cam_data->output.height;
;
    for (rows=0; rows < rows_output; rows++, row+=hratio) {
      if(lptr - cam_data->awb_buffer >= cam_data->scanimage_size) {;
	memset(data, 0, count - retval);
	retval = count;
	break;
      };
      awb_gen_line_yuyv_sentech (cam_data, 
				 lptr, data, 
				 cam_data->max_width,
				 ((param_old_scale)?
				  cam_data->crop.width:
				  cam_data->output.width),
				 ((param_old_scale&&scale)?
				  0:
				  cam_data->crop.left),
				 row, crt_gamma,
				 cam_data->max_height, rows, wratio, row, evenh);
      data += inc;
      retval += inc;
      lptr += cam_data->max_width*hratio;
      evenh ^= switchh;
;
    }
    if(count > retval);
    memset(data, 0, count - retval);
    retval = count;
    break;    
  case V4L2_PIX_FMT_BGR24:;;;
    if(param_old_scale){
      inc = cam_data->crop.width;
      row = (scale?0:cam_data->crop.top);
      evenh = (scale)?(!gbrg):(!(cam_data->crop.top&1));
    }else{
      inc = cam_data->output.width;
      row = cam_data->crop.top;
      evenh = !(cam_data->crop.top&1);
    }
    inc += (inc <<1);
    lptr = cam_data->awb_buffer + row*cam_data->max_width;
    rows_output = (param_old_scale)?
      cam_data->crop.height:
      cam_data->output.height;
    for (rows=0;rows < rows_output;rows++, row+=hratio) {
      if(lptr - cam_data->awb_buffer >= cam_data->scanimage_size) {;
	memset(data, 0, count - retval);
	retval = count;
	break;
      };
      awb_gen_line_bgr24_sentech(cam_data, 
				 lptr, data,
				 cam_data->max_width,
				 (param_old_scale)?
				   cam_data->crop.width:
				   cam_data->output.width,
				 ((param_old_scale&&scale)?
				  0:
				  cam_data->crop.left),
				 row, crt_gamma,
				 cam_data->max_height, rows,
				 wratio, row, evenh);
      data += inc;
      retval += inc;
      lptr += cam_data->max_width*hratio;
      evenh ^= switchh;;
    }
    if(count > retval);
    memset(data, 0, count - retval);
    retval = count;
    break;
  case V4L2_PIX_FMT_RGB24:;
;
    if(param_old_scale){
      inc = cam_data->crop.width;
      row = (scale?0:cam_data->crop.top);
      evenh = (scale)?(!gbrg):(!(cam_data->crop.top&1));
    }else{
      inc = cam_data->output.width;
      row = cam_data->crop.top;
      evenh = !(cam_data->crop.top&1);
    }
    inc += (inc <<1);
    lptr = cam_data->awb_buffer + row*cam_data->max_width;
    rows_output = (param_old_scale)?
      cam_data->crop.height:
      cam_data->output.height;
    for (rows=0;rows < rows_output ;rows++, row+=hratio) {
      if(lptr - cam_data->awb_buffer >= cam_data->scanimage_size) {;
	memset(data, 0, count - retval);
	retval = count;
	break;
      };
      awb_gen_line_rgb24_sentech(cam_data, 
				 lptr, data, 
				 cam_data->max_width,
				 ((param_old_scale)?
				  cam_data->crop.width:
				  cam_data->output.width),
				 ((param_old_scale&&scale)?
				  0:
				  cam_data->crop.left),
				 row, crt_gamma,
				 cam_data->max_height, rows, wratio,
				 row, evenh);
      data += inc;
      retval += inc;
      lptr += cam_data->max_width*hratio;
      evenh ^= switchh;;
    }
    if(count > retval);
    memset(data, 0, count - retval);
    retval = count;
    break;
  case V4L2_PIX_FMT_YUV420:;
;
    if(param_old_scale){
      inc = cam_data->crop.width;
      row = (scale?0:cam_data->crop.top);
      evenh = (scale)?(!gbrg):(!(cam_data->crop.top&1));
    }else{
      inc = cam_data->output.width;
      row = cam_data->crop.top;
      evenh = !(cam_data->crop.top&1);
    }
    lptr = cam_data->awb_buffer + row*cam_data->max_width;
    rows_output = (param_old_scale)?
      cam_data->crop.height:
      cam_data->output.height;
    dataU = data+cam_data->crop.height*bytesperline; 
    dataV = dataU + ((cam_data->crop.height*bytesperline)>>2);
    for (rows=0;rows<cam_data->crop.height;rows++, row+=hratio) {
      awb_gen_line_yuv420_sentech(cam_data, 
				  lptr, data, 
				  cam_data->max_width,
				  ((param_old_scale)?
				   cam_data->crop.width:
				   cam_data->output.width),
				  ((param_old_scale&&scale)?
				  0:
				  cam_data->crop.left), row, crt_gamma,
				  dataU, dataV, cam_data->max_height,
				  rows, wratio, row, evenh);
      data += inc;
      if(rows&1) {
	dataU += inc>>1;
	dataV += inc>>1;
	retval += inc;
      }
      retval += inc;
      lptr += cam_data->max_width*hratio;
      evenh ^= switchh;
    }
    break;    
  case V4L2_PIX_FMT_YUV422P:;
;
    if(param_old_scale){
      inc = cam_data->crop.width;
      row = (scale?0:cam_data->crop.top);
      evenh = (scale)?(!gbrg):(!(cam_data->crop.top&1));
    }else{
      inc = cam_data->output.width;
      row = cam_data->crop.top;
      evenh = !(cam_data->crop.top&1);
    }
    lptr = cam_data->awb_buffer + row*cam_data->max_width;
    rows_output = (param_old_scale)?
      cam_data->crop.height:
      cam_data->output.height;
    dataU = data+cam_data->crop.height*bytesperline; 
    dataV = dataU + ((cam_data->crop.height*bytesperline)>>1);
    for (rows=0;rows<cam_data->crop.height;rows++, row+=hratio) {
      awb_gen_line_yuv422P_sentech(cam_data, 
				  lptr, data, 
				  cam_data->max_width,
				  ((param_old_scale)?
				   cam_data->crop.width:
				   cam_data->output.width),
				  ((param_old_scale&&scale)?
				  0:
				  cam_data->crop.left), row, crt_gamma,
				  dataU, dataV, cam_data->max_height,
				  rows, wratio, row, evenh);
      data += inc;
      dataU += inc>>1;
      dataV += inc>>1;
      retval += inc;
      retval += inc;
      lptr += cam_data->max_width*hratio;
      evenh ^= switchh; 
    }
    break;    
    printk("Fourcc pixel format not supported: V4L2_PIX_FMT_YUV420\n");
    retval=0;
    break;
    printk("Fourcc pixel format not supported: V4L2_PIX_FMT_SBGGR8\n");
    retval=0;
    break;
  case V4L2_PIX_FMT_BGR32:
    printk("Fourcc pixel format not supported: V4L2_PIX_FMT_BGR32\n");
    retval=0;
    break;
  default:;
    retval=0;
  }
  up(&cam_data->lock);
  update_controls(cam_data);
  return retval;
}
#endif
