/************************************/
/* Driver for Sentech USB Cameras   */
/* Release 3.2.32                   */
/************************************/
/* Author: Marius Calin Silaghi     */
/*         FLORIDA TECH             */
/*  <msilaghi@fit.edu>              */
/*    April           9, 2015       */
/************************************/
/* Copyright:                       */
/*  Marius C. Silaghi               */
/*  msilaghi@fit.edu                */
/************************************/

#ifndef __KERNEL__
#define __KERNEL__
#endif
#ifndef MODULE
#define MODULE
#endif
#include "SentechTUSBC.h"
#ifdef UNUSED
#elif defined(__GNUC__)
# define UNUSED(x) UNUSED_ ## x __attribute__((unused))
#elif defined(__LCLINT__)
# define UNUSED(x) /*@unused@*/ x
#else
# define UNUSED(x) x
#endif
#ifdef SENTECH_AWB
#include <AWB/AWB.h>
#else
typedef uint8_t BYTE;
#endif
#define USB3_EXP_LOW_SHIFT 0
#define USB3_EXP_UP_SHIFT 16
#define USB3_EXP_LOW_MASK_16BITS 0x00FFFF
#define USB3_EXP_UP_MASK_4BITS 0x000F
#define USB3_MAX_EXP_DEFAULT 0x797
#define USB3_MIN_EXP_DEFAULT 0x1
#define USB3_UP_EXP_OR 1
#define USB3_SET_EXP_UP(variable,value,mask) { \
  if(USB3_UP_EXP_OR) {			       \
     variable &= ~mask;       \
     variable |= (value & mask);	\
}else{ \
  variable = value & mask; \
} \
}
#include "sentech_gamma.h"
#include "sentech_gamma16.h"
#define GAMMA_MAX 24
#define CAM_NAME_LENGTH 32
#include <asm/atomic.h>
#include <linux/spinlock.h>
#include <linux/vmalloc.h>
#include <linux/version.h>
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,26)
#include <linux/semaphore.h>
#else
#include <asm/semaphore.h>
#endif
#ifndef DECLARE_MUTEX
#ifdef __SEMAPHORE_INITIALIZER
#define DECLARE_MUTEX(name)     \
  struct semaphore name = __SEMAPHORE_INITIALIZER(name, 1)
#endif
#endif
#ifndef init_MUTEX
#define init_MUTEX(sem)         sema_init(sem, 1)
#endif
#ifndef DECLARE_MUTEX
#define DECLARE_MUTEX(name)     \
  DEFINE_SEMAPHORE(name)
#endif
#include <linux/module.h>
#include <linux/usb.h>
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,35)
#include <linux/usb/hcd.h>
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,39)
#include <linux/smp_lock.h>
#else
void lock_kernel(void){}
void unlock_kernel(void){}
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,7)
#include "backport/my_kthread_oldy.c"
#else
#include <linux/kthread.h>
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,27)
#include <linux/mm.h>
#include <media/v4l2-ioctl.h>
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,33)
#include <linux/slab.h>
#endif
#define MOD_INIT
#ifdef MOD_INIT
#include <linux/init.h>
#endif
#include <linux/version.h>
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18)
#include <media/v4l2-dev.h>
#endif
#include <linux/videodev2.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,38)
#include <linux/videodev.h>
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,18)
#include <linux/interrupt.h>
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,15)
#include <media/v4l2-common.h>
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,11,0)
#include <media/v4l2-device.h>
#endif
#include "sentech.h"
#ifndef USB_SPEED_WIRELESS
#define USB_SPEED_WIRELESS 4
#endif 
#ifndef USB_SPEED_SUPER
#define USB_SPEED_SUPER 5
#endif
#ifndef MAX_SUBS
#define MAX_SUBS 10
#endif
#ifndef NOPAGE_SIGBUS
#define NOPAGE_SIGBUS NULL
#endif
#ifdef EXPENSIVE_SAT
#define NB_SATURATIONS 0x0100
#define __sat_t __s32
#define SAT_SHIFT 0
#else
#define __sat_t __s16
#define SAT_SHIFT 8
#ifndef NB_SATURATIONS
#define NB_SATURATIONS 0x010
#endif
#endif
#define SATDEF (NB_SATURATIONS>>1)
#define SAT_MED 0x80
inline void setSaturations(__u8 K);
void initSaturation(void);
inline void setSaturations_16(__u8 K);
void initSaturation_16(void);
__sat_t* _KRR,* _KRG,* _KRB,* _KGR,* _KGG,* _KBB;
__sat_t* _KRR_16,* _KRG_16,* _KRB_16,* _KGR_16,* _KGG_16,* _KBB_16;
#define SBUF_IDX_SHIFT 24
#define COLOR_SHIFT 4
#define COLOR_NO_MONO 0
#define BRIGHTDEF -120
#define GAMMADEF 14
#define GAINDEF 0x80
#define MAXGAIN 0xff
#define CONTRASTDEF 106
#define REDDEF (16<<COLOR_SHIFT)-1
#define BLUEDEF (26<<COLOR_SHIFT)-1
#define MAX_EXP_CTRL 0xff
#undef BRIGHTDEF
#undef CONTRASTDEF
#undef REDDEF
#undef BLUEDEF
#undef GAINDEF
#undef MAXGAIN
#define BRIGHTDEF 0
#define CONTRASTDEF 0
#define REDDEF 255
#define BLUEDEF 255
#define GAINDEF 0x80
#define MAXGAIN 0xff
#ifndef STC_DEFAULT_BRIGHTNESS
#define STC_DEFAULT_BRIGHTNESS BRIGHTDEF
#endif
#ifndef STC_DEFAULT_CONTRAST
#define STC_DEFAULT_CONTRAST CONTRASTDEF
#endif
#ifndef STC_DEFAULT_SATURATION
#define STC_DEFAULT_SATURATION SATDEF
#endif
#define STC_DEFAULT_RED REDDEF
#define STC_DEFAULT_BLUE BLUEDEF
#define STC_DEFAULT_GREEN 0x10
#ifndef STC_DEFAULT_GAMMA
#define STC_DEFAULT_GAMMA GAMMADEF
#endif
#ifndef STC_DEFAULT_EXPOSURE
#define STC_DEFAULT_EXPOSURE STC_DEFAULT_CONTRAST
#endif
#ifndef STC_DEFAULT_EXPOSURE_TARGET
#define STC_DEFAULT_EXPOSURE_TARGET 90
#endif
#ifndef STC_DEFAULT_GAIN
#define STC_DEFAULT_GAIN GAINDEF
#endif
#define QUOTEME_(x) #x
#define QUOTEME(x) QUOTEME_(x)
DECLARE_MUTEX(sentech_lock);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,33)
DEFINE_RWLOCK(subs_list_lock);
#else
rwlock_t subs_list_lock = RW_LOCK_UNLOCKED;
#endif
#define sem_getcount(a) atomic_read(&((a)->count))
#include <linux/kref.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,9)
#define kref_put(a,b) kref_put(a)
#ifdef __CHECKER__
#define __bitwise__ __attribute__((bitwise))
#else
#define __bitwise__
#endif
#ifdef __CHECK_ENDIAN__
#define __bitwise __bitwise__
#else
#define __bitwise
#endif 
#endif 
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,9)
typedef __u16 __bitwise __le16;
typedef __u16 __bitwise __be16;
typedef __u32 __bitwise __le32;
typedef __u32 __bitwise __be32;
typedef __u64 __bitwise __le64;
typedef __u64 __bitwise __be64;
#endif
#include <linux/completion.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,7)
#define __sched     __attribute__((__section__(".sched.text")))
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11)
unsigned long 
fastcall __sched
wait_for_completion_timeout(struct completion *x, unsigned long timeout)
{
        might_sleep();
        spin_lock_irq(&x->wait.lock);
        if (!x->done) {
                DECLARE_WAITQUEUE(wait, current);
                wait.flags |= WQ_FLAG_EXCLUSIVE;
                __add_wait_queue_tail(&x->wait, &wait);
                do {
                        __set_current_state(TASK_UNINTERRUPTIBLE);
                        spin_unlock_irq(&x->wait.lock);
                        timeout = schedule_timeout(timeout);
                        spin_lock_irq(&x->wait.lock);
                        if (!timeout) {
                                __remove_wait_queue(&x->wait, &wait);
                                goto out;
                        }
                } while (!x->done);
                __remove_wait_queue(&x->wait, &wait);
        }
        x->done--;
out:
        spin_unlock_irq(&x->wait.lock);
        return timeout;
}
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)
#define list_first_entry(ptr, type, member) \
      list_entry((ptr)->next, type, member)
#endif
#include <linux/slab.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,14)
typedef unsigned __bitwise__ gfp_t;
/**
   * kzalloc - allocate memory. The memory is set to zero.
   * @size: how many bytes of memory are required.
   * @flags: the type of memory to allocate.
   */
void *kzalloc(size_t size, gfp_t flags)
{
        void *ret = kmalloc(size, flags);
        if (ret)
                memset(ret, 0, size);
        return ret;
}
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19)
#define bool int
#define false 0
#define true 1
struct v4l2_frmsize_discrete
 {
         __u32                   width;          
         __u32                   height;         
 };
#endif
#if  LINUX_VERSION_CODE > KERNEL_VERSION(2,6,15)
#include <media/v4l2-common.h>
#endif
#if  LINUX_VERSION_CODE < KERNEL_VERSION(2,6,18)
#define V4L2_CTRL_FLAG_NEXT_CTRL  0x80000000
#define V4L2_DEBUG_IOCTL     0x01
#define V4L2_DEBUG_IOCTL_ARG 0x02
#include "backport/ioctl2.c"
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,29)
#define V4L2_PIX_FMT_SGRBG8  v4l2_fourcc('G', 'R', 'B', 'G') 
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,8)
#define V4L2_PIX_FMT_SBGGR8  v4l2_fourcc('B','A','8','1') 
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,18)
#define V4L2_CTRL_FLAG_UPDATE           0x0008
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,52)
#include <linux/moduleparam.h>
#else
#define module_param(a,b,c) MODULE_PARM(a, "i")
#endif
static bool param_variable_width = false;
module_param(param_variable_width, bool, S_IRUGO);
MODULE_PARM_DESC(param_variable_width, "param_variable_width[0]:  1 for attempting it with USB3");
static bool param_luminance_bounds_us = true;
module_param(param_luminance_bounds_us, bool, S_IRUGO);
MODULE_PARM_DESC(param_luminance_bounds_us, "param_luminance_bounds_us[1]:  0 for absolute values");
#define MAX_VMALLOC_BUFFERS 4
#define VMALLOC_BUFFERS_URBS 5
static int VMALLOC_BUFFERS_URB_FRAMEBUFFER = 512<<10;
module_param(VMALLOC_BUFFERS_URB_FRAMEBUFFER, int, S_IRUGO);
MODULE_PARM_DESC(VMALLOC_BUFFERS_URB_FRAMEBUFFER, "[512] minimum");
static bool param_init_softtrigger = 0;
module_param(param_init_softtrigger, bool, S_IRUGO);
MODULE_PARM_DESC(param_init_softtrigger, "[0] 1 to init in trigger mode");
static int param_probe = 10000;
module_param(param_probe, int, S_IRUGO);
MODULE_PARM_DESC(param_probe, "[10000] for entire probe");
static int param_pret = ENODEV;
module_param(param_pret, int, S_IRUGO);
MODULE_PARM_DESC(param_pret, "ENODEV for exit");
static int param_have_buffers = 1;
module_param(param_have_buffers, int, S_IRUGO);
MODULE_PARM_DESC(param_have_buffers, "[0] if no buffer allocation is desired (due to too little memory)");
static int default_gamma = STC_DEFAULT_GAMMA;
module_param(default_gamma, int, S_IRUGO);
MODULE_PARM_DESC(default_gamma, "default_gamma[" QUOTEME(STC_DEFAULT_GAMMA) "]: out of 0.." QUOTEME(GAMMA_MAX));
static bool param_admit_unspecified_qbuf_index = true;
module_param(param_admit_unspecified_qbuf_index, bool, S_IRUGO);
MODULE_PARM_DESC(param_admit_unspecified_qbuf_index, "param_admit_unspecified_qbuf_index[true], false needed for multiple dequeued buffers, but with specified indexes");
static bool param_detect_multiple_prrobe_u3v = false;
module_param(param_detect_multiple_prrobe_u3v, bool, S_IRUGO);
MODULE_PARM_DESC(param_detect_multiple_prrobe_u3v, "param_detect_multiple_prrobe_u3v[true], false for only setting U3V Sentech mode on open/close");
static bool param_keep_vbuffer = true;
module_param(param_keep_vbuffer, bool, S_IRUGO);
MODULE_PARM_DESC(param_keep_vbuffer, "param_keep_vbuffer[true], true for not freeing vbuffer at each close");
static bool param_supports_bpp = true;
module_param(param_supports_bpp, bool, S_IRUGO);
MODULE_PARM_DESC(param_supports_bpp, "param_supports_bpp[true], true for no 2 byte support");
static int param_max_frame_size = 0;
module_param(param_max_frame_size, int, S_IRUGO);
MODULE_PARM_DESC(param_max_frame_size, "param_max_frame_size[0], nonzero sets max frame size");
static bool param_gamma16 = true;
module_param(param_gamma16, bool, S_IRUGO);
MODULE_PARM_DESC(param_gamma16, "param_gamma16[1], 0 to not enable gamma translation for 2 byte samples");
static bool param_read16 = 0;
module_param(param_read16, bool, S_IRUGO);
MODULE_PARM_DESC(param_read16, "param_read16[1], 0 for read 16 RGGB10/12 and YUYV on 8 bits, 1 to read them on 10/12 bits");
static int default_brightness = STC_DEFAULT_BRIGHTNESS;
module_param(default_brightness, int, S_IRUGO);
MODULE_PARM_DESC(default_brightness, "default_brightness[" QUOTEME(STC_DEFAULT_BRIGHTNESS) "]: out of -255..255");
static int default_exposure = STC_DEFAULT_EXPOSURE;
module_param(default_exposure, int, S_IRUGO);
MODULE_PARM_DESC(default_exposure, "default_exposure[" QUOTEME(STC_DEFAULT_EXPOSURE) "]: out of 0.." QUOTEME(MAX_EXP_CTRL));
static int default_exposure_target = STC_DEFAULT_EXPOSURE_TARGET; 
module_param(default_exposure_target, int, S_IRUGO);
MODULE_PARM_DESC(default_exposure_target, "default_exposure_target[" QUOTEME(STC_DEFAULT_EXPOSURE_TARGET) "]: out of 0..255");
static int default_gain = STC_DEFAULT_GAIN;
module_param(default_gain, int, S_IRUGO);
MODULE_PARM_DESC(default_gain, "default_gain[" QUOTEME(STC_DEFAULT_GAIN) "]: out of 0.." QUOTEME(MAXGAIN));
static int MARGINS = 0;
module_param(MARGINS, int, S_IRUGO);
MODULE_PARM_DESC(MARGINS, "MARGINS[0]:  10 for buffer margins");
#define MARGIN (MARGINS*PAGE_SIZE)
static int trace_buffers = 0;
module_param(trace_buffers, int, S_IRUGO);
MODULE_PARM_DESC(trace_buffers, "trace_buffers[0]:  2 for trace");
#define DEFAULT_RGB24 3
#define DEFAULT_YUYV 2
#define DEFAULT_RGGB 1
#define DEFAULT_BGR24 0
static int default_format = 0;
module_param(default_format, int, S_IRUGO);
MODULE_PARM_DESC(default_format, "rggb default_format[0=BGR24]: 1=RGGB 2=YUYV, 3=RGB24, ?=GBR24");
static int idle_buffers = 0;
module_param(idle_buffers, int, S_IRUGO);
MODULE_PARM_DESC(idle_buffers, "idle_buffers[0]:  0 for not reserving idle buffers");
static int debug = 0;
module_param(debug, int, S_IRUGO);
MODULE_PARM_DESC(debug, "debug[0]:  21 for debug");
static bool restart_on_bus_error = 0;
module_param(restart_on_bus_error, bool, S_IRUGO);
MODULE_PARM_DESC(restart_on_bus_error, "restart_on_bus_error[0]:  1 for automatic restart");
static bool dqueue_sleep = 1;
module_param(dqueue_sleep, bool, S_IRUGO);
MODULE_PARM_DESC(dqueue_sleep, "dqueue_sleep[0]:  1 for sleep in dqueue");
static bool scale = 1;
module_param(scale, bool, S_IRUGO);
MODULE_PARM_DESC(scale, "scale[1]:  0 for no scale");
static bool mono_no_color = 0;
module_param(mono_no_color, bool, S_IRUGO);
MODULE_PARM_DESC(mono_no_color, "mono_no_color[0]:  1 to only serve GREY for mono cameras");
static int param_integrated_rgb = 1;
module_param(param_integrated_rgb, int, S_IRUGO);
MODULE_PARM_DESC(param_integrated_rgb, "param_freq_auto[0]: integrate rgb in native buffer");
static int param_freq_auto = 3;
module_param(param_freq_auto, int, S_IRUGO);
MODULE_PARM_DESC(param_freq_auto, "param_freq_auto[3]: Hz");
#define param_max_exp_avg_shift 8
static int param_freq_auto_exp_avg = (1<<(param_max_exp_avg_shift))>>2;
module_param(param_freq_auto_exp_avg, int, S_IRUGO);
MODULE_PARM_DESC(param_freq_auto_exp_avg, "param_freq_auto_exp_avg[0-256]: 128 small is slower auto");
#ifdef SENTECH_AWB
static int param_old_scale = 0;
module_param(param_old_scale, int, S_IRUGO);
MODULE_PARM_DESC(param_old_scale, "param_old_scale[0]");
#else
static int param_old_scale = 1;
module_param(param_old_scale, int, S_IRUGO);
MODULE_PARM_DESC(param_old_scale, "param_old_scale[1]");
#endif
static int param_imagesize_claim_adjustment = 0;
module_param(param_imagesize_claim_adjustment, int, S_IRUGO);
MODULE_PARM_DESC(param_imagesize_claim_adjustment, "param_imagesize_claim_adjustment[0] gives an additional claimed # of bytesused in NATIVE dqbuf mode");
static int param_buffers = 3;
module_param(param_buffers, int, S_IRUGO);
MODULE_PARM_DESC(param_buffers, "param_buffers[3]");
static bool param_use_submitted = 0; 
module_param(param_use_submitted, bool, S_IRUGO|S_IWUSR);
MODULE_PARM_DESC(param_use_submitted, "param_use_submitted[false=0], can be modified later");
static int SUBS_SUBMITTED = 3; 
static int param_SUBS_SUBMITTED = 3;
module_param(param_SUBS_SUBMITTED, int, S_IRUGO);
MODULE_PARM_DESC(param_SUBS_SUBMITTED, "param_SUBS_SUBMITTED[2]");
static bool DYNAMIC_USE_SUBMITTED = 1;
module_param(DYNAMIC_USE_SUBMITTED, bool, S_IRUGO);
MODULE_PARM_DESC(DYNAMIC_USE_SUBMITTED, "DYNAMIC_USE_SUBMITTED[1] to use submissted if insufficient buffers are allocated");
static bool DYNAMIC_SUB_SUBMITTED = 0;
module_param(DYNAMIC_SUB_SUBMITTED, bool, S_IRUGO);
MODULE_PARM_DESC(DYNAMIC_SUB_SUBMITTED, "DYNAMIC_SUB_SUBMITTED[0] submit less if not enough buffers got");
static bool auto_SUBS_SUBMITTED = 1;
module_param(auto_SUBS_SUBMITTED, bool, S_IRUGO);
MODULE_PARM_DESC(auto_SUBS_SUBMITTED, "auto_SUBS_SUBMITTED[true=1]");
static bool param_bytesperline = 1;
module_param(param_bytesperline, bool, S_IRUGO);
MODULE_PARM_DESC(param_bytesperline, "param_bytesperline[0]: 1 to use byteperline");
static bool param_submit_user_buffers = 0;
module_param(param_submit_user_buffers, bool, S_IRUGO);
MODULE_PARM_DESC(param_submit_user_buffers, "param_submit_user_buffers[0], 1: to allow overwriting currently processed buffers");
static bool param_defrect_scan_height = 1;
module_param(param_defrect_scan_height, bool, S_IRUGO);
MODULE_PARM_DESC(param_defrect_scan_height, "param_defrect_scan_height[1]: 0 for static defrect.height=max_height rather than current scan_height");
static bool dynamic_scale = 1;
module_param(dynamic_scale, bool, S_IRUGO);
MODULE_PARM_DESC(dynamic_scale, "dynamic_scale[0]: 1 for dynamic scale for exposure");
static bool param_preallocate = 1;
module_param(param_preallocate, bool, S_IRUGO);
MODULE_PARM_DESC(param_preallocate, "param_preallocate[1], 0 for not");
static bool param_try_fragments = 0;
module_param(param_try_fragments, bool, S_IRUGO);
MODULE_PARM_DESC(param_try_fragments, "param_try_fragments[1], 0 for not");
static unsigned int param_max_fragment = 0x0400000;
module_param(param_max_fragment, uint, S_IRUGO);
MODULE_PARM_DESC(param_max_fragment, "param_max_fragment[0x0400000], 0 for none");
static bool param_continuous_capture = 1;
module_param(param_continuous_capture, bool, S_IRUGO);
MODULE_PARM_DESC(param_continuous_capture, "param_continuous_capture[1], 0 for not");
static int param_minor = -1;
module_param(param_minor, int, S_IRUGO);
MODULE_PARM_DESC(param_minor, "minor[-1]: -1 for autodetect");
#ifdef PARAM_FIRST_NB
static int first_nb = 0;
module_param(first_nb, int, S_IRUGO);
MODULE_PARM_DESC(first_nb, "first_nb[0]: the first number of an attached camera");
#else
static atomic_t first_nb = ATOMIC_INIT(0);
#endif
static bool param_streaming = 1;
module_param(param_streaming, bool, S_IRUGO);
MODULE_PARM_DESC(param_streaming, "param_streaming[1]: 0 for supporting streamming (mmap)");
static bool param_usrstreaming = 0;
module_param(param_usrstreaming, bool, S_IRUGO);
MODULE_PARM_DESC(param_usrstreaming, "param_usrstreaming[0]: 0 for supporting usr streamming (usrptr mmap)");
static bool param_read = 1;
module_param(param_read, bool, S_IRUGO);
MODULE_PARM_DESC(param_read, "param_read[1]: 1 for supporting read");
static bool param_overlay = 0;
module_param(param_overlay, bool, S_IRUGO);
MODULE_PARM_DESC(param_overlay, "param_overlay[1]: 0 for supporting read");
static bool param_native_fsizes = 1;
module_param(param_native_fsizes, bool, S_IRUGO);
MODULE_PARM_DESC(param_native_fsizes, "param_native_fsizes[1]: 0 for reporting only native frame sizes");
static bool y16 = 1;
module_param(y16, bool, S_IRUGO);
MODULE_PARM_DESC(y16, "y16[1]: 0 for no Y16 format");
static bool grey = 1;
module_param(grey, bool, S_IRUGO);
MODULE_PARM_DESC(grey, "grey[1]: 0 for no GREY format");
static bool rggb = 1;
module_param(rggb, bool, S_IRUGO);
MODULE_PARM_DESC(rggb, "rggb[1]: 0 for no RGGB format");
static bool bggr = 1;
module_param(bggr, bool, S_IRUGO);
MODULE_PARM_DESC(bggr, "bggr[0]: 0 for no BGGR format");
static bool rgb24 = 1;
module_param(rgb24, bool, S_IRUGO);
MODULE_PARM_DESC(rgb24, "rgb24[1]: 0 for no RGB24 format");
static bool bgr24 = 1;
module_param(bgr24, bool, S_IRUGO);
MODULE_PARM_DESC(bgr24, "bgr24[1]: 0 for no BGR24 format");
static bool yuyv = 1;
module_param(yuyv, bool, S_IRUGO);
MODULE_PARM_DESC(yuyv, "yuyv[1]: 0 for no YUYV format");
static bool yuv420 = 1;
module_param(yuv420, bool, S_IRUGO);
MODULE_PARM_DESC(yuv420, "yuv420[1]: 0 for no YUV420 format");
static bool yuv422P = 1;
module_param(yuv422P, bool, S_IRUGO);
MODULE_PARM_DESC(yuv422P, "yuv422P[1]: 0 for no YUV422P format");
static int skip_awb = 0;
module_param(skip_awb, int, S_IRUGO);
MODULE_PARM_DESC(skip_awb, "skip_awb[0]: n to skip n between any frame");
static int skip_awb_native = 3;
module_param(skip_awb_native, int, S_IRUGO);
MODULE_PARM_DESC(skip_awb_native, "skip_awb_native[3]: n to skip additional n frames between any frame for native capture");
static int cfg_saturation = 0;
module_param(cfg_saturation, int, S_IRUGO);
MODULE_PARM_DESC(cfg_saturation, "cfg_saturation[1]: 1 for supporting saturation control");
static bool color_correction = 1;
module_param(color_correction, bool, S_IRUGO);
MODULE_PARM_DESC(color_correction, "color_correction[1]: 1 for supporting color correction");
static bool brightness_correction = 1;
module_param(brightness_correction, bool, S_IRUGO);
MODULE_PARM_DESC(brightness_correction, "brightness_correction[0]: 1 for supporting brightness correction");
static bool trigger_on_fail = 1;
module_param(trigger_on_fail, bool, S_IRUGO);
MODULE_PARM_DESC(trigger_on_fail, "trigger_on_fail[1]: 0 for no trigger restart on bad data size");
static bool filter_short_frames = 1;
module_param(filter_short_frames, bool, S_IRUGO);
MODULE_PARM_DESC(filter_short_frames, "filter_short_frames[0]: 1 for no bad sized frame");
static bool gbrg = 0;
module_param(gbrg, bool, S_IRUGO);
MODULE_PARM_DESC(gbrg, "gbrg[0]: 1 for GBRG formats, rather than RGGB");
#define DEF_PSAT_AUTO 0
static bool param_auto_saturation = DEF_PSAT_AUTO;
module_param(param_auto_saturation, bool, S_IRUGO);
MODULE_PARM_DESC(param_auto_saturation, "param_auto_saturation[1]: 0");
#define DEF_PSAT_STEP 4
static int param_saturation_step = DEF_PSAT_STEP;
module_param(param_saturation_step, int, S_IRUGO);
MODULE_PARM_DESC(param_saturation_step, "param_sat_step[4]");
#define DEF_PSAT_MIN 220
static int param_saturation_min = DEF_PSAT_MIN;
module_param(param_saturation_min, int, S_IRUGO);
MODULE_PARM_DESC(param_saturation_min, "param_sat_min[220]");
#define DEF_PSAT_MAX 255
static int param_saturation_max = DEF_PSAT_MAX;
module_param(param_saturation_max, int, S_IRUGO);
MODULE_PARM_DESC(param_saturation_max, "param_sat_max[255]");
#define DEF_PSAT_AMORTIZ 50
static int param_sat_amortization = DEF_PSAT_AMORTIZ;
module_param(param_sat_amortization, int, S_IRUGO);
MODULE_PARM_DESC(param_sat_amortization, "param_sat_amortiz[50] divides first jump after a stability");
#define DEF_PSAT_AMORTIZ_LENGTH 5
static int param_sat_amortization_length = DEF_PSAT_AMORTIZ_LENGTH;
module_param(param_sat_amortization_length, int, S_IRUGO);
MODULE_PARM_DESC(param_sat_amortization_length, "param_sat_amortiz_length[5] divides first jump during this length after a stability");
#define DEF_PSAT_ACC 16
static int param_sat_acc = DEF_PSAT_ACC;
module_param(param_sat_acc, int, S_IRUGO);
MODULE_PARM_DESC(param_sat_acc, "param_sat_acc[8] jumps multiplication");
#define DEF_PSAT_UP_LIMIT 50
static int param_saturation_upper_limit = DEF_PSAT_UP_LIMIT; 
module_param(param_saturation_upper_limit, int, S_IRUGO);
MODULE_PARM_DESC(param_saturation_upper_limit, "param_sat_upper_limit[4]");
#define DEF_PSAT_LO_LIMIT 50
static int param_saturation_lower_limit = DEF_PSAT_LO_LIMIT; 
module_param(param_saturation_lower_limit, int, S_IRUGO);
MODULE_PARM_DESC(param_saturation_lower_limit, "param_sat_lower_limit[4]");
#define DEF_PSAT_EXP_INC 4
static int param_saturation_exp_inc = DEF_PSAT_EXP_INC;
module_param(param_saturation_exp_inc, int, S_IRUGO);
MODULE_PARM_DESC(param_saturation_exp_inc, "param_saturation_exp_inc[1]");
#define DEF_PSAT_GAIN_INC 4
static int param_saturation_gain_inc = DEF_PSAT_GAIN_INC;
module_param(param_saturation_gain_inc, int, S_IRUGO);
MODULE_PARM_DESC(param_saturation_gain_inc, "param_saturation_gain_inc[1]");
#define DEF_PSAT_SKIP 3
static int param_saturation_skip = DEF_PSAT_SKIP;
module_param(param_saturation_skip, int, S_IRUGO);
MODULE_PARM_DESC(param_saturation_skip, "param_saturation_skip[2]");
#define PROBE(l) if (l >= param_probe) return param_pret;
static int unloading_module = 0;
#define DEFAULT_NATIVE_MODES 1
static __u8 io_modes[] = {0x00,0x01,0x10,0x11,0x12,0x13,0x20,0x21,0x14,
			  0x00,0x01,0x10,0x30,0x40};
#define SENTECH_USB_C83 0
#define VGA_W 640
#define VGA_H 480
#define VGA_H2 224
#define VGA_H4 80
#define XGA_W 0x400
#define XGA_H 0x300
#define XGA_H2 344
#define XGA_H4 136
#define SXGA_W 1360
#define SXGA_H 1024
#define SXGA_H2 476
#define SXGA_H4 184
#define UXGA_W 1600
#define UXGA_H 1200
#define UXGA_H2 544
#define UXGA_H4 208
#define SXGA_133_W 1280
#define SXGA_133_H 960
#define SXGA_133_H2 440
#define SXGA_133_H4 168
#define QSXGA_W 2592
#define QSXGA_H 1944
#define QSXGA_H2 1944
#define QSXGA_H4 1944
#define ESXGA_W 1280
#define ESXGA_H 1024
#define CM4SXGA_W 2048
#define CM4SXGA_H 2048
#define CM2SXGA_W 2048
#define CM2SXGA_H 1088
static struct v4l2_frmsize_discrete fsizes_33[] = {
  {VGA_W, VGA_H}, {VGA_W, VGA_H2}, {VGA_W, VGA_H4}};
static struct v4l2_frmsize_discrete fsizes_83[] = {
  {XGA_W, XGA_H}, {XGA_W, XGA_H2}, {XGA_W, XGA_H4}};
static struct v4l2_frmsize_discrete fsizes_152[] = {
  {SXGA_W, SXGA_H}, {SXGA_W, SXGA_H2}, {SXGA_W, SXGA_H4}};
static struct v4l2_frmsize_discrete fsizes_202[] = {
  {UXGA_W, UXGA_H}, {UXGA_W, UXGA_H2}, {UXGA_W, UXGA_H4}};
static struct v4l2_frmsize_discrete fsizes_133[] = {
  {SXGA_133_W, SXGA_133_H}, {SXGA_133_W, SXGA_133_H2}, {SXGA_133_W, SXGA_133_H4}};
static struct v4l2_frmsize_discrete fsizes_A5USB3[] = {
  {QSXGA_W, QSXGA_H}, {QSXGA_W, QSXGA_H2}, {QSXGA_W, QSXGA_H4}};
static struct v4l2_frmsize_discrete fsizes_E132U3V[] = {
  {ESXGA_W, ESXGA_H}, {ESXGA_W, ESXGA_H}, {ESXGA_W, ESXGA_H}};
static struct v4l2_frmsize_discrete fsizes_CM401U3V[] = {
  {CM4SXGA_W, CM4SXGA_H}, {CM4SXGA_W, CM4SXGA_H}, {CM4SXGA_W, CM4SXGA_H}};
static struct v4l2_frmsize_discrete fsizes_CM200U3V[] = {
  {CM2SXGA_W, CM2SXGA_H}, {CM2SXGA_W, CM2SXGA_H}, {CM2SXGA_W, CM2SXGA_H}};
#define STC_CTRL_NAME_LEN 32
struct sentech_height_param {
  __u16 hardware_val;
  char name[32];
};
static struct sentech_height_param sentech_height_params[] = {
  { STC_QUART_SCREEN, "QUART SCREEN" },
  { STC_HALF_SCREEN, "HALF SCREEN" },
  { STC_FULL_SCREEN, "FULL SCREEN" },
  { STC_WHOLE_SCREEN, "WHOLE SCREEN" },
  { STC_VARIABLE_SCREEN, "VARIABLE SCREEN" },
  { STC_FULL_BIN_SCREEN, "FULL BIN SCREEN" },
  { STC_WHOLE_BIN_SCREEN, "WHOLE BIN SCREEN" },
  { STC_HALF_BIN_SCREEN, "HALF BIN SCREEN" },
  { STC_QUART_BIN_SCREEN, "QUART_BIN_SCREEN" },
  { STC_VARIABLE_BIN_SCREEN, "VARIABLE BIN SCREEN" }
};
struct sentech_params {
  struct v4l2_fract fps[3];
  u32   native_fourcc;
  u32   picture_size;
  __u16 model;
  __u16 fsizes;
  struct v4l2_frmsize_discrete* fsize;
  char name[STC_CTRL_NAME_LEN];
};
static
struct sentech_params sentech_fps[] = {
  {{{1,30},
    {1,15},
    {2,15}},
   V4L2_PIX_FMT_SRGGB8,
   STC_XGA,
   STC_C83USB, 3, fsizes_83,
   "STC_C83USB"
  },
  {{{1,60},
    {1,30},
    {1,15}},
   V4L2_PIX_FMT_SRGGB8,
   STC_VGA,
   STC_C33USB, 3, fsizes_33,
   "STC_C33USB"
  },
  {{{1,30},
    {1,15},
    {2,15}},
   V4L2_PIX_FMT_GREY,
   STC_XGA,
   STC_B83USB, 3, fsizes_83,
   "STC_B83USB"
  },
  {{{1,60},
    {1,30},
    {1,15}},
   V4L2_PIX_FMT_GREY,
   STC_VGA,
   STC_B33USB, 3, fsizes_33,
   "STC_B33USB"
  },
  {{{1,60},
    {1,30},
    {1,15}},
   V4L2_PIX_FMT_SRGGB8,
   STC_VGA,
   STC_C33USB_T, 3, fsizes_33,
   "STC_C33USB_T"
  },
  {{{1,60},
    {1,30},
    {1,15}},
   V4L2_PIX_FMT_GREY,
   STC_VGA,
   STC_MB33USB, 3, fsizes_33,
   "STC_MB33USB"
  },
  {{{1,60},
    {1,30},
    {1,15}},
   V4L2_PIX_FMT_SRGGB8,
   STC_VGA,
   STC_MC33USB, 3, fsizes_33,
   "STC_MC33USB"
  },
  {{{1,30},
    {1,15},
    {2,15}},
   V4L2_PIX_FMT_SRGGB8,
   STC_XGA,
   STC_C83USB_T, 3, fsizes_83,
   "STC_C83USB_T"
  },
  {{{1,30},
    {1,15},
    {2,15}},
   V4L2_PIX_FMT_GREY,
   STC_XGA,
   STC_B83USB_T, 3, fsizes_83,
   "STC_B83USB_T"
  },
  {{{1,30},
    {1,15},
    {2,15}},
   V4L2_PIX_FMT_SRGGB8,
   STC_XGA,
   STC_C83USB_T, 3, fsizes_83,
   "STC_C83USB_T"
  },
  {{{1,30},
    {1,15},
    {2,15}},
   V4L2_PIX_FMT_GREY,
   STC_XGA,
   STC_B83USB_T, 3, fsizes_83,
   "STC_B83USB_T"
  },
  {{{1,60},
    {1,30},
    {1,15}},
   V4L2_PIX_FMT_GREY,
   STC_VGA,
   STC_B33USB_T, 3, fsizes_33,
   "STC_B33USB_T"
  },
  {{{1,30},
    {1,15},
    {2,15}},
   V4L2_PIX_FMT_SRGGB8,
   STC_XGA,
   STC_MC83USB, 3, fsizes_83,
   "STC_MC83USB"
  },
  {{{1,30},
    {1,15},
    {2,15}},
   V4L2_PIX_FMT_GREY,
   STC_XGA,
   STC_MB83USB, 3, fsizes_83,
   "STC_MB83USB"
  },
  {{{1,19},
    {2,19},
    {4,19}},
   V4L2_PIX_FMT_SRGGB8,
   STC_SXGA,
   STC_C152USB_T, 3, fsizes_152,
   "STC_C152USB_T"
  },
  {{{1,19},
    {2,19},
    {4,19}},
   V4L2_PIX_FMT_GREY,
   STC_SXGA,
   STC_B152USB_T, 3, fsizes_152,
   "STC_B152USB_T"
  },
  {{{1,19},
    {2,19},
    {4,19}},
   V4L2_PIX_FMT_SRGGB8,
   STC_SXGA,
   STC_MC152USB, 3, fsizes_152,
   "STC_MC152USB"
  },
  {{{1,19},
    {2,19},
    {4,19}},
   V4L2_PIX_FMT_GREY,
   STC_SXGA,
   STC_MB152USB, 3, fsizes_152,
   "STC_MB152USB"
  },
  {{{4,15},
    {2,15},
    {1,15}},
   V4L2_PIX_FMT_GREY,
   STC_UXGA,
   STC_B202USB_T, 3, fsizes_202,
   "STC_B202USB_T"
  },
  {{{4,15},
    {2,15},
    {1,15}},
   V4L2_PIX_FMT_SRGGB8,
   STC_UXGA,
   STC_C202USB_T, 3, fsizes_202,
   "STC_C202USB_T"
  },
  {{{4,15},
    {2,15},
    {1,15}},
   V4L2_PIX_FMT_GREY,
   STC_UXGA,
   STC_MB202USB, 3, fsizes_202,
   "STC_MB202USB"
  },
  {{{4,15},
    {2,15},
    {1,15}},
   V4L2_PIX_FMT_SRGGB8,
   STC_UXGA,
   STC_MC202USB, 3, fsizes_202,
   "STC_MC202USB"
  },
  {{{1,22},
    {1,11},
    {2,11}},
   V4L2_PIX_FMT_GREY,
   STC_SXGA133,
   STC_B133USB_T, 3, fsizes_133,
   "STC_B133USB_T"
  },
  {{{1,22},
    {1,11},
    {2,11}},
   V4L2_PIX_FMT_SRGGB8,
   STC_SXGA133,
   STC_C133USB_T, 3, fsizes_133,
   "STC_C133USB_T"
  },
  {{{1,22},
    {1,11},
    {2,11}},
   V4L2_PIX_FMT_GREY,
   STC_SXGA133,
   STC_MB133USB, 3, fsizes_133,
   "STC_MB133USB"
  },
  {{{1,22},
    {1,11},
    {2,11}},
   V4L2_PIX_FMT_SRGGB8,
   STC_SXGA133,
   STC_MC133USB, 3, fsizes_133,
   "STC_MC133USB"
  },
  {{{1,14},
    {1,14},
    {1,14}},
   V4L2_PIX_FMT_GREY,
   STC_QSXGA,
   STC_MBA5MUSB3, 1, fsizes_A5USB3,
   "STC_MBA5MUSB3"
  },
  {{{1,14},
    {1,14},
    {1,14}},
   V4L2_PIX_FMT_SGRBG8,
   STC_QSXGA,
   STC_MCA5MUSB3, 1, fsizes_A5USB3,
   "STC_MCA5MUSB3"
  },
  {{{1,60},
    {1,60},
    {1,60}},
   V4L2_PIX_FMT_GREY,
   STC_SXGA,
   STC_MBE132U3V, 1, fsizes_E132U3V,
   "STC_MBE132U3V"
  },
  {{{1,60},
    {1,60},
    {1,60}},
   V4L2_PIX_FMT_SGRBG8,
   STC_SXGA,
   STC_MCE132U3V, 1, fsizes_E132U3V,
   "STC_MCE132U3V"
  },
  {{{1,86},
    {1,86},
    {1,86}},
   V4L2_PIX_FMT_GREY,
   STC_SXGA,
   STC_MBCM401U3V, 1, fsizes_CM401U3V,
   "STC_MBCM401U3V"
  },
  {{{1,86},
    {1,86},
    {1,86}},
   V4L2_PIX_FMT_SGRBG8,
   STC_SXGA,
   STC_MCCM401U3V, 1, fsizes_CM401U3V,
   "STC_MCCM401U3V"
  },
  {{{1,150},
    {1,150},
    {1,150}},
   V4L2_PIX_FMT_GREY,
   STC_SXGA,
   STC_MBCM200U3V, 1, fsizes_CM200U3V,
   "STC_MBCM200U3V"
  },
  {{{1,150},
    {1,150},
    {1,150}},
   V4L2_PIX_FMT_SGRBG8,
   STC_SXGA,
   STC_MCCM200U3V, 1, fsizes_CM200U3V,
   "STC_MCCM200U3V"
  },
  {{{1,30},
    {1,15},
    {2,15}},
   V4L2_PIX_FMT_GREY,
   STC_VGA,
   0 
   , 3, fsizes_33,
   "STC_UNKNOWN_VGA_MONO"
  },
  {{{1,30},
    {1,15},
    {2,15}},
   V4L2_PIX_FMT_SRGGB8,
   STC_VGA,
   0 
   , 3, fsizes_33,
   "STC_UNKNOWN_VGA_COLOR"
  }
};
#define ALL_CTRL
#define TRIGGER_CTRL
int max_exp_ctrl = MAX_EXP_CTRL;
static 
struct v4l2_queryctrl sentech_qctrl[] = {
  {
    .id            = V4L2_CID_BRIGHTNESS,
    .type          = V4L2_CTRL_TYPE_INTEGER,
    .name          = "Brightness",
    .minimum       = -255,
    .maximum       = 255,
    .step          = 1,
    .default_value = STC_DEFAULT_BRIGHTNESS,
    .flags         = 0,
  }, {
    .id            = V4L2_CID_CONTRAST,
    .type          = V4L2_CTRL_TYPE_INTEGER,
    .name          = "Contrast",
    .minimum       = 0,
    .maximum       = MAX_EXP_CTRL,
    .step          = 0x1,
    .default_value = STC_DEFAULT_CONTRAST,
    .flags         = V4L2_CTRL_FLAG_UPDATE, 
  }, {
    .id            = V4L2_CID_SATURATION,
    .type          = V4L2_CTRL_TYPE_INTEGER,
    .name          = "Saturation",
    .minimum       = 0,
    .maximum       = NB_SATURATIONS-1, 
    .default_value = STC_DEFAULT_SATURATION,
    .step          = 0x1,
    .flags         = 0,
  }, {
    .id            = V4L2_CID_HUE,
    .type          = V4L2_CTRL_TYPE_INTEGER,
    .name          = "Hue or Color Balance",
    .minimum       = 0,
    .maximum       = NB_SATURATIONS-1, 
    .step          = 0x1,
    .default_value = SATDEF,
    .flags         = V4L2_CTRL_FLAG_DISABLED,
  }, {
    .id            = V4L2_CID_HUE+1,
    .type          = V4L2_CTRL_TYPE_INTEGER,
    .name          = "Reserved Hue",
    .minimum       = 0,
    .maximum       = NB_SATURATIONS-1, 
    .step          = 0x1,
    .default_value = SATDEF,
    .flags         = V4L2_CTRL_FLAG_DISABLED,
  }
, {
    .id            = V4L2_CID_AUDIO_VOLUME,
    .type          = V4L2_CTRL_TYPE_INTEGER,
    .name          = "Audio Volume",
    .flags         = V4L2_CTRL_FLAG_DISABLED,
  }
, {
    .id            = V4L2_CID_AUDIO_BALANCE,
    .type          = V4L2_CTRL_TYPE_INTEGER,
    .name          = "Audio Balance",
    .flags         = V4L2_CTRL_FLAG_DISABLED,
  }
, {
    .id            = V4L2_CID_AUDIO_BASS,
    .type          = V4L2_CTRL_TYPE_INTEGER,
    .name          = "Audio Bass",
    .flags         = V4L2_CTRL_FLAG_DISABLED,
  }
, {
    .id            = V4L2_CID_AUDIO_TREBLE,
    .type          = V4L2_CTRL_TYPE_INTEGER,
    .name          = "Audio Treble",
    .flags         = V4L2_CTRL_FLAG_DISABLED,
  }
, {
    .id            = V4L2_CID_AUDIO_MUTE,
    .type          = V4L2_CTRL_TYPE_INTEGER,
    .name          = "Audio Mute",
    .flags         = V4L2_CTRL_FLAG_DISABLED,
  }
, {
    .id            = V4L2_CID_AUDIO_LOUDNESS,
    .type          = V4L2_CTRL_TYPE_INTEGER,
    .name          = "Audio Loudness",
    .flags         = V4L2_CTRL_FLAG_DISABLED,
  }
, {
    .id            = V4L2_CID_BLACK_LEVEL,
    .type          = V4L2_CTRL_TYPE_INTEGER,
    .name          = "Black Level",
    .flags         = V4L2_CTRL_FLAG_DISABLED,
  }
, {
    .id            = V4L2_CID_AUTO_WHITE_BALANCE,
    .type          = V4L2_CTRL_TYPE_BOOLEAN,
    .name          = "Auto White Balance",
    .minimum       = 0,
    .maximum       = 1,
    .step          = 0x1,
#ifdef SENTECH_AWB
    .default_value = 1,
#else
    .default_value = 0,
#endif
    .flags         = 0,
  }
, {
    .id            = V4L2_CID_DO_WHITE_BALANCE,
    .type          = V4L2_CTRL_TYPE_BUTTON,
    .name          = "Do White Balance",
    .minimum       = 0,
    .maximum       = 1,
    .step          = 0x1,
    .default_value = 0,
    .flags         = V4L2_CTRL_FLAG_UPDATE,
    }, {
    .id            = V4L2_CID_RED_BALANCE,
    .type          = V4L2_CTRL_TYPE_INTEGER,
    .name          = "Color Red Balance",
    .minimum       = 0,
    .maximum       = (256<<COLOR_SHIFT)-1,
    .step          = 0x1,
    .default_value = STC_DEFAULT_RED,
    .flags         = V4L2_CTRL_FLAG_UPDATE,
  }, {
    .id            = V4L2_CID_BLUE_BALANCE,
    .type          = V4L2_CTRL_TYPE_INTEGER,
    .name          = "Color Blue Balance",
    .minimum       = 0,
    .maximum       = (256<<COLOR_SHIFT)-1,
    .step          = 0x1,
    .default_value = STC_DEFAULT_BLUE,
    .flags         = V4L2_CTRL_FLAG_UPDATE,
    }, {
    .id            = V4L2_CID_GAMMA,
    .type          = V4L2_CTRL_TYPE_INTEGER,
    .name          = "Gamma (Whiteness)",
    .minimum       = 0,
    .maximum       = ARRAY_SIZE(sentech_gamma)-1,
    .step          = 1,
    .default_value = STC_DEFAULT_GAMMA, 
    .flags         = V4L2_CTRL_FLAG_UPDATE,
  }
, {
    .id            = V4L2_CID_EXPOSURE,
    .type          = V4L2_CTRL_TYPE_INTEGER,
    .name          = "Exposure",
    .minimum       = 0,
    .maximum       = MAX_EXP_CTRL,
    .step          = 0x1,
    .default_value = STC_DEFAULT_EXPOSURE,
    .flags         = V4L2_CTRL_FLAG_UPDATE, 
  }
, {
    .id            = V4L2_CID_AUTOGAIN,
    .type          = V4L2_CTRL_TYPE_BOOLEAN,
    .name          = "Automatic Gain",
    .minimum       = 0,
    .maximum       = 1,
    .step          = 0x1,
#ifdef SENTECH_AWB
    .default_value = 0x1,
#else
    .default_value = 0x0,
#endif
    .flags         = 0,
  }
, {
    .id            = V4L2_CID_GAIN,
    .type          = V4L2_CTRL_TYPE_INTEGER,
    .name          = "Gain",
    .minimum       = 0,
    .maximum       = MAXGAIN,
    .step          = 0x1,
    .default_value = STC_DEFAULT_GAIN, 
    .flags         = 0,
  }
, {
    .id            = V4L2_CID_HFLIP,
    .type          = V4L2_CTRL_TYPE_BOOLEAN,
    .name          = "Horizontal mirror (GREY/USB3)",
    .minimum       = 0,
    .maximum       = 1,
    .step          = 0x1,
    .default_value = 0,
    .flags         = 0, 
  }
, {
    .id            = V4L2_CID_VFLIP,
    .type          = V4L2_CTRL_TYPE_BOOLEAN,
    .name          = "Vertical mirror (USB3)",
    .minimum       = 0,
    .maximum       = 1,
    .step          = 0x1,
    .default_value = 0,
    .flags         = 0, 
  }
, {
    .id            = V4L2_CID_STC_FPS,
    .type          = V4L2_CTRL_TYPE_MENU, 
    .name          = "Frames per Second 15 30 60",
    .minimum       = 0,
    .maximum       = 2,
    .step          = 0x01,
    .default_value = 0x02,
    .flags         = 0,
  }
, {
    .id            = V4L2_CID_STC_HEIGHT,
    .type          = V4L2_CTRL_TYPE_MENU, 
    .name          = "Picture Height 1/4 1/2 1",
    .minimum       = 0,
    .maximum       = 9,
    .step          = 0x1,
    .default_value = STC_CTL_FULL_SCREEN,
#ifdef ALL_CTRL
    .flags         = 0,
#else
    .flags         = V4L2_CTRL_FLAG_DISABLED,
#endif
  }
, {
    .id            = V4L2_CID_STC_RED,
    .type          = V4L2_CTRL_TYPE_INTEGER,
    .name          = "Balance Red",
    .minimum       = 0,
    .maximum       = (256<<COLOR_SHIFT)-1,
    .step          = 0x1,
    .default_value = STC_DEFAULT_RED,
    .flags         = V4L2_CTRL_FLAG_UPDATE,
  }, {
    .id            = V4L2_CID_STC_GREEN,
    .type          = V4L2_CTRL_TYPE_INTEGER,  
    .name          = "Balance Green",
    .minimum       = 0,
    .maximum       = (256<<COLOR_SHIFT)-1,
    .step          = 0x1,
    .default_value = STC_DEFAULT_GREEN,
    .flags         = 0,
  }, {
    .id            = V4L2_CID_STC_BLUE,
    .type          = V4L2_CTRL_TYPE_INTEGER,
    .name          = "Balance Blue",
    .minimum       = 0,
    .maximum       = (256<<COLOR_SHIFT)-1,
    .step          = 0x1,
    .default_value = STC_DEFAULT_BLUE,
    .flags         = V4L2_CTRL_FLAG_UPDATE,
    }, {
    .id            = V4L2_CID_STC_GAMMA,
    .type          = V4L2_CTRL_TYPE_INTEGER,
    .name          = "STC Gamma (Whiteness)",
    .minimum       = 0,
    .maximum       = ARRAY_SIZE(sentech_gamma)-1,
    .step          = 1,
    .default_value = STC_DEFAULT_GAMMA, 
    .flags         = V4L2_CTRL_FLAG_UPDATE,
  }, {
    .id            = V4L2_CID_STC_TARGET,
    .type          = V4L2_CTRL_TYPE_INTEGER,
    .name          = "Auto Exposure Target",
    .minimum       = 0,
    .maximum       = 255,
    .step          = 1,
    .default_value = STC_DEFAULT_EXPOSURE_TARGET, 
    .flags         = 0,
  }, {
    .id            = V4L2_CID_STC_TRIGGER_MODE,
    .type          = V4L2_CTRL_TYPE_MENU, 
    .name          = ": Trigger: Norm Trigg Readout",
    .minimum       = 0,
    .maximum       = 2,
    .step          = 0x01,
    .default_value = 0x00,
#ifdef TRIGGER_CTRL
    .flags         = 0,
#else
    .flags         = V4L2_CTRL_FLAG_DISABLED,
#endif
  }, {
    .id            = V4L2_CID_STC_TRIGGER_SOURCE,
    .type          = V4L2_CTRL_TYPE_MENU, 
    .name          = ": Trigger Source: IGN Soft Hard",
    .minimum       = 0,
    .maximum       = 3,
    .step          = 0x01,
    .default_value = 0x00,
#ifdef TRIGGER_CTRL
    .flags         = 0,
#else
    .flags         = V4L2_CTRL_FLAG_DISABLED,
#endif
  }
, {
    .id            = V4L2_CID_STC_HEADER,
    .type          = V4L2_CTRL_TYPE_INTEGER,
    .name          = "Header: Normal Split",
    .minimum       = 0,
    .maximum       = 1,
    .step          = 0x01,
    .default_value = 0x00,
    .flags         = V4L2_CTRL_FLAG_DISABLED,
  }
, {
    .id            = V4L2_CID_STC_DIGITAL_GAIN,
    .type          = V4L2_CTRL_TYPE_INTEGER,
    .name          = "Digital Gain: 1-4 times",
    .minimum       = 0,
    .maximum       = 255,
    .step          = 0x01,
    .default_value = 0x80,
    .flags         = 0,
  }
, {
    .id            = V4L2_CID_STC_TRIGGER,
    .type          = V4L2_CTRL_TYPE_BUTTON,
    .name          = ": SOFTWARE TRIGGER",
    .minimum       = 0,
    .maximum       = 1,
    .step          = 0x01,
    .default_value = 0x00,
#ifdef TRIGGER_CTRL
    .flags         = 0,
#else
    .flags         = V4L2_CTRL_FLAG_DISABLED,
#endif
  }
, {
    .id            = V4L2_CID_STC_IO0_FUNCTION,
    .type          = V4L2_CTRL_TYPE_MENU, 
    .name          = "IO0 Function",
    .minimum       = 0,
    .maximum       = 13,
    .step          = 0x01,
    .default_value = IO_I_IGN,
#ifdef ALL_CTRL
    .flags         = 0,
#else
    .flags         = V4L2_CTRL_FLAG_DISABLED,
#endif
  }
, {
    .id            = V4L2_CID_STC_IO1_FUNCTION,
    .type          = V4L2_CTRL_TYPE_MENU, 
    .name          = "IO1 Function",
    .minimum       = 0,
    .maximum       = 13,
    .step          = 0x01,
    .default_value = IO_I_IGN,
#ifdef TRIGGER_CTRL
    .flags         = 0,
#else
    .flags         = V4L2_CTRL_FLAG_DISABLED,
#endif
  }, {
    .id            = V4L2_CID_STC_IO2_FUNCTION,
    .type          = V4L2_CTRL_TYPE_MENU, 
    .name          = "IO2 Function",
    .minimum       = 0,
    .maximum       = 13,
    .step          = 0x01,
    .default_value = IO_O_HIZ,
#ifdef TRIGGER_CTRL
    .flags         = 0,
#else
    .flags         = V4L2_CTRL_FLAG_DISABLED,
#endif
  }
, {
    .id            = V4L2_CID_STC_IO3_FUNCTION,
    .type          = V4L2_CTRL_TYPE_MENU, 
    .name          = "IO3 Function",
    .minimum       = 0,
    .maximum       = 13,
    .step          = 0x01,
    .default_value = IO_O_HIZ,
#ifdef ALL_CTRL
    .flags         = 0,
#else
    .flags         = V4L2_CTRL_FLAG_DISABLED,
#endif
  }
#ifdef  SENTECH_AUTO_AWB_CIDS
, {
    .id            = V4L2_CID_STC_AUTO_WHITE_BALANCE,
    .type          = V4L2_CTRL_TYPE_BOOLEAN,
    .name          = "Auto White Balance STC",
    .minimum       = 0,
    .maximum       = 1,
    .step          = 0x1,
#ifdef SENTECH_AWB
    .default_value = 1,
#else
    .default_value = 0,
#endif
    .flags         = 0,
  }
, {
    .id            = V4L2_CID_STC_AUTOGAIN,
    .type          = V4L2_CTRL_TYPE_BOOLEAN,
    .name          = "Automatic Gain/Exposure STC",
    .minimum       = 0,
    .maximum       = 1,
    .step          = 0x1,
#ifdef SENTECH_AWB
    .default_value = 0x1,
#else
    .default_value = 0x0,
#endif
    .flags         = 0,
  }
#endif
, {
    .id            = V4L2_CID_STC_LED,
    .type          = V4L2_CTRL_TYPE_MENU, 
    .name          = "LED",
    .minimum       = 0,
    .maximum       = 3,
    .step          = 0x01,
    .default_value = 0,
#ifdef ALL_CTRL
    .flags         = 0,
#else
    .flags         = V4L2_CTRL_FLAG_DISABLED,
#endif
  }
, {
    .id            = V4L2_CID_STC_RO_SEL,
    .type          = V4L2_CTRL_TYPE_MENU, 
    .name          = "ReadOut Select",
    .minimum       = 0,
    .maximum       = 1,
    .step          = 0x01,
    .default_value = 0,
#ifdef TRIGGER_CTRL
    .flags         = 0,
#else
    .flags         = V4L2_CTRL_FLAG_DISABLED,
#endif
  }
, {
    .id            = V4L2_CID_STC_W_TRG_OUT,
    .type          = V4L2_CTRL_TYPE_INTEGER,
    .name          = "Width Trigger Output",
    .minimum       = 0,
    .maximum       = 4094,
    .step          = 0x01,
    .default_value = 0x00,
#ifdef TRIGGER_CTRL
    .flags         = 0,
#else
    .flags         = V4L2_CTRL_FLAG_DISABLED,
#endif
  }, {
    .id            = V4L2_CID_STC_D_RO,
    .type          = V4L2_CTRL_TYPE_INTEGER,
    .name          = "Delay ReadOut",
    .minimum       = 0,
    .maximum       = 65535,
    .step          = 0x01,
    .default_value = 0x00,
#ifdef TRIGGER_CTRL
    .flags         = 0,
#else
    .flags         = V4L2_CTRL_FLAG_DISABLED,
#endif
  }, {
    .id            = V4L2_CID_STC_REFRESH,
    .type          = V4L2_CTRL_TYPE_BUTTON, 
    .name          = "Refresh",
    .minimum       = 0,
    .maximum       = 1,
    .step          = 0x01,
    .default_value = 1,
    .flags         = 0,
  }, {
    .id            = V4L2_CID_STC_CNT_RST,
    .type          = V4L2_CTRL_TYPE_BUTTON, 
    .name          = "Counter Reset",
    .minimum       = 0,
    .maximum       = 1,
    .step          = 0x01,
    .default_value = 1,
    .flags         = 0,
  }, {
    .id            = V4L2_CID_STC_RO,
    .type          = V4L2_CTRL_TYPE_BUTTON, 
    .name          = "SOFTWARE READ-OUT",
    .minimum       = 0,
    .maximum       = 1,
    .step          = 0x01,
    .default_value = 1,
#ifdef TRIGGER_CTRL
    .flags         = 0,
#else
    .flags         = V4L2_CTRL_FLAG_DISABLED,
#endif
  }, {
    .id            = V4L2_CID_STC_SUB_TRG,
    .type          = V4L2_CTRL_TYPE_BUTTON, 
    .name          = "SOFTWARE SUB-TRIGGER",
    .minimum       = 0,
    .maximum       = 1,
    .step          = 0x01,
    .default_value = 0x00,
#ifdef TRIGGER_CTRL
    .flags         = 0,
#else
    .flags         = V4L2_CTRL_FLAG_DISABLED,
#endif
  }
, {
    .id            = V4L2_CID_STC_EXP_MODE,
    .type          = V4L2_CTRL_TYPE_MENU, 
    .name          = "Exposure Mode",
    .minimum       = 0,
    .maximum       = 2,
    .step          = 0x01,
    .default_value = 0,
#ifdef ALL_CTRL
    .flags         = 0,
#else
    .flags         = V4L2_CTRL_FLAG_DISABLED,
#endif
  }, {
    .id            = V4L2_CID_STC_EXP_D1,
    .type          = V4L2_CTRL_TYPE_MENU, 
    .name          = "Exposure Start Delay 1",
    .minimum       = 0,
    .maximum       = 1,
    .step          = 0x01,
    .default_value = 0,
#ifdef ALL_CTRL
    .flags         = 0,
#else
    .flags         = V4L2_CTRL_FLAG_DISABLED,
#endif
  }, {
    .id            = V4L2_CID_STC_EXP_D2,
    .type          = V4L2_CTRL_TYPE_MENU, 
    .name          = "Exposure Start Delay 2",
    .minimum       = 0,
    .maximum       = 1,
    .step          = 0x01,
    .default_value = 0,
#ifdef ALL_CTRL
    .flags         = 0,
#else
    .flags         = V4L2_CTRL_FLAG_DISABLED,
#endif
  }
, {
    .id            = V4L2_CID_STC_D_TRG_EXP,
    .type          = V4L2_CTRL_TYPE_INTEGER,
    .name          = "Delay Trigger-Exposure",
    .minimum       = 0,
    .maximum       = 4094,
    .step          = 0x01,
    .default_value = 0x00,
#ifdef ALL_CTRL
    .flags         = 0,
#else
    .flags         = V4L2_CTRL_FLAG_DISABLED,
#endif
  }, {
    .id            = V4L2_CID_STC_D_TRG_ST,
    .type          = V4L2_CTRL_TYPE_INTEGER,
    .name          = "Delay Trigger-Start",
    .minimum       = 0,
    .maximum       = 4094,
    .step          = 0x01,
    .default_value = 0x00,
#ifdef ALL_CTRL
    .flags         = 0,
#else
    .flags         = V4L2_CTRL_FLAG_DISABLED,
#endif
  }, {
    .id            = V4L2_CID_STC_D_TRG_END,
    .type          = V4L2_CTRL_TYPE_INTEGER,
    .name          = "Delay Trigger-End",
    .minimum       = 0,
    .maximum       = 4094,
    .step          = 0x01,
    .default_value = 0x00,
#ifdef ALL_CTRL
    .flags         = 0,
#else
    .flags         = V4L2_CTRL_FLAG_DISABLED,
#endif
  }, {
    .id            = V4L2_CID_STC_D_TRG_I_O,
    .type          = V4L2_CTRL_TYPE_INTEGER,
    .name          = "Delay Trigger Input-Output",
    .minimum       = 0,
    .maximum       = 4094,
    .step          = 0x01,
    .default_value = 0x00,
#ifdef ALL_CTRL
    .flags         = 0,
#else
    .flags         = V4L2_CTRL_FLAG_DISABLED,
#endif
  }, {
    .id            = V4L2_CID_STC_CLEAR_STATE,
    .type          = V4L2_CTRL_TYPE_BUTTON, 
    .name          = "SOFTWARE CLEAR STATE",
    .minimum       = 0,
    .maximum       = 1,
    .step          = 0x01,
    .default_value = 1,
    .flags         = 0,
  }, {
    .id            = V4L2_CID_STC_TRIGGER_POLARITY_0,
    .type          = V4L2_CTRL_TYPE_MENU, 
    .name          = "TRIGGER POLARITY 0",
    .minimum       = 0,
    .maximum       = 1,
    .step          = 0x01,
    .default_value = 0,
    .flags         = 0,
  }, {
    .id            = V4L2_CID_STC_TRIGGER_POLARITY_1,
    .type          = V4L2_CTRL_TYPE_MENU, 
    .name          = "TRIGGER POLARITY 1",
    .minimum       = 0,
    .maximum       = 1,
    .step          = 0x01,
    .default_value = 0,
    .flags         = 0,
  }, {
    .id            = V4L2_CID_STC_TRIGGER_POLARITY_2,
    .type          = V4L2_CTRL_TYPE_MENU, 
    .name          = "TRIGGER POLARITY 2",
    .minimum       = 0,
    .maximum       = 1,
    .step          = 0x01,
    .default_value = 0,
    .flags         = 0,
  }, {
    .id            = V4L2_CID_STC_TRIGGER_POLARITY_3,
    .type          = V4L2_CTRL_TYPE_MENU, 
    .name          = "TRIGGER POLARITY 3",
    .minimum       = 0,
    .maximum       = 1,
    .step          = 0x01,
    .default_value = 0,
    .flags         = 0,
  }, {
    .id            = V4L2_CID_STC_EXPOSURE_US,
    .type          = V4L2_CTRL_TYPE_INTEGER,
    .name          = "Exposure (usecs)",
    .minimum       = 0,
    .maximum       = 1,
    .step          = 0x1,
    .default_value = 1,
    .flags         = V4L2_CTRL_FLAG_UPDATE, 
  }
 , {
    .id            = V4L2_CID_STC_GAIN_MIN,
    .type          = V4L2_CTRL_TYPE_INTEGER,
    .name          = "Gain Min",
    .minimum       = 0,
    .maximum       = GAINDEF,
    .step          = 0x1,
    .default_value = 0,
    .flags         = V4L2_CTRL_FLAG_UPDATE, 
  }, {
    .id            = V4L2_CID_STC_GAIN_MAX,
    .type          = V4L2_CTRL_TYPE_INTEGER,
    .name          = "Gain Max",
    .minimum       = GAINDEF,
    .maximum       = 255,
    .step          = 0x1,
    .default_value = 255,
    .flags         = V4L2_CTRL_FLAG_UPDATE, 
  }, {
    .id            = V4L2_CID_STC_EXPOSURE_MIN,
    .type          = V4L2_CTRL_TYPE_INTEGER,
    .name          = "Exposure Min in u_seconds",
    .minimum       = 0,
    .maximum       = 1,
    .step          = 0x1,
    .default_value = 0,
    .flags         = V4L2_CTRL_FLAG_UPDATE, 
  }, {
    .id            = V4L2_CID_STC_EXPOSURE_MAX,
    .type          = V4L2_CTRL_TYPE_INTEGER,
    .name          = "Exposure Max in u_seconds",
    .minimum       = 0,
    .maximum       = 1,
    .step          = 0x1,
    .default_value = 1,
    .flags         = V4L2_CTRL_FLAG_UPDATE, 
  }
#ifdef STC_CTRL_SPEED
, {
    .id            = V4L2_CID_STC_GAIN_CTRL_SPEED_LIMIT,
    .type          = V4L2_CTRL_TYPE_INTEGER,
    .name          = "AutoGain Ctrl Speed Limit",
    .minimum       = 0,
    .maximum       = 255,
    .step          = 0x1,
    .default_value = 0,
    .flags         = 0,
  }, {
    .id            = V4L2_CID_STC_SHUTTER_CTRL_SPEED_LIMIT,
    .type          = V4L2_CTRL_TYPE_INTEGER,
    .name          = "Auto-Exposure Ctrl Speed Limit",
    .minimum       = 0,
    .maximum       = 255,
    .step          = 0x1,
    .default_value = 0,
    .flags         = 0,
  }, {
    .id            = V4L2_CID_STC_AVERAGE_FRAME_COUNT,
    .type          = V4L2_CTRL_TYPE_INTEGER,
    .name          = "ALC Average Frame Count",
    .minimum       = 0,
    .maximum       = 255,
    .step          = 0x1,
    .default_value = 4, 
    .flags         = 0,
  }
#endif
, {
    .id            = V4L2_CID_STC_IO_STATUS_0,
    .type          = V4L2_CTRL_TYPE_MENU, 
    .name          = "IO STATUS 0",
    .minimum       = 0,
    .maximum       = 1,
    .step          = 0x01,
    .default_value = 0,
    .flags         = 0,
  }, {
    .id            = V4L2_CID_STC_IO_STATUS_1,
    .type          = V4L2_CTRL_TYPE_MENU, 
    .name          = "IO STATUS 1",
    .minimum       = 0,
    .maximum       = 1,
    .step          = 0x01,
    .default_value = 0,
    .flags         = 0,
  }, {
    .id            = V4L2_CID_STC_IO_STATUS_2,
    .type          = V4L2_CTRL_TYPE_MENU, 
    .name          = "IO STATUS 2",
    .minimum       = 0,
    .maximum       = 1,
    .step          = 0x01,
    .default_value = 0,
    .flags         = 0,
  }, {
    .id            = V4L2_CID_STC_IO_STATUS_3,
    .type          = V4L2_CTRL_TYPE_MENU, 
    .name          = "IO STATUS 3",
    .minimum       = 0,
    .maximum       = 1,
    .step          = 0x01,
    .default_value = 0,
    .flags         = 0,
  }
  , {
    .id            = V4L2_CID_STC_PRODUCT,
    .type          = V4L2_CTRL_TYPE_MENU, 
    .name          = "Product Name",
    .minimum       = 0,
    .maximum       = 7,
    .step          = 0x01,
    .default_value = 0,
    .flags         = 0,
  } 
#ifdef SENTECH_AWB
  , {
    .id            = V4L2_CID_STC_ALC_SPEED,
    .type          = V4L2_CTRL_TYPE_INTEGER,
    .name          = "ALC Skip Frame Count",
    .minimum       = 0,
    .maximum       = 10,
    .step          = 0x01,
    .default_value = 3, 
    .flags         = 0,
  }
#endif
  , {
    .id            = V4L2_CID_STC_FRAME_START,
    .type          = V4L2_CTRL_TYPE_INTEGER,
    .name          = "Variable Frame Start",
    .minimum       = 0,
    .maximum       = 10000,
    .step          = 0x01,
    .default_value = 0, 
    .flags         = 0,
  }
  , {
    .id            = V4L2_CID_STC_FRAME_HEIGHT,
    .type          = V4L2_CTRL_TYPE_INTEGER,
    .name          = "Variable Frame Height",
    .minimum       = 0,
    .maximum       = 10000,
    .step          = 0x01,
    .default_value = 0, 
    .flags         = 0,
  }
  , {
    .id            = V4L2_CID_STC_PIX_SATURATION_AUTO_LUMINANCE,
    .type          = V4L2_CTRL_TYPE_BOOLEAN,
    .name          = "AL limit pixel saturation",
    .minimum       = 0,
    .maximum       = 1,
    .step          = 0x01,
    .default_value = DEF_PSAT_AUTO,
    .flags         = 0,
  }
  , {
    .id            = V4L2_CID_STC_PIX_SATURATION_STEP,
    .type          = V4L2_CTRL_TYPE_INTEGER,
    .name          = "AL sampling step",
    .minimum       = 1,
    .maximum       = 50,
    .step          = 0x01,
    .default_value = DEF_PSAT_STEP,
    .flags         = 0,
  }
  , {
    .id            = V4L2_CID_STC_PIX_SATURATION_MIN,
    .type          = V4L2_CTRL_TYPE_INTEGER,
    .name          = "AL lighted threshold",
    .minimum       = 1,
    .maximum       = 255,
    .step          = 0x01,
    .default_value = DEF_PSAT_MIN,
    .flags         = 0,
  }
  , {
    .id            = V4L2_CID_STC_PIX_SATURATION_ACC,
    .type          = V4L2_CTRL_TYPE_INTEGER,
    .name          = "AL max step fraction acc",
    .minimum       = 1,
    .maximum       = 50,
    .step          = 0x01,
    .default_value = DEF_PSAT_ACC,
    .flags         = 0,
  }
  , {
    .id            = V4L2_CID_STC_PIX_SATURATION_UP_LIMIT,
    .type          = V4L2_CTRL_TYPE_INTEGER,
    .name          = "AL limit fraction saturated",
    .minimum       = 0,
    .maximum       = 100000,
    .step          = 0x01,
    .default_value = DEF_PSAT_UP_LIMIT,
    .flags         = 0,
  }
  , {
    .id            = V4L2_CID_STC_PIX_SATURATION_EXP_INC,
    .type          = V4L2_CTRL_TYPE_INTEGER,
    .name          = "AL exposure step fraction",
    .minimum       = 1,
    .maximum       = 100,
    .step          = 0x01,
    .default_value = DEF_PSAT_EXP_INC,
    .flags         = 0,
  }
  , {
    .id            = V4L2_CID_STC_PIX_SATURATION_GAIN_INC,
    .type          = V4L2_CTRL_TYPE_INTEGER,
    .name          = "AL gain step fraction",
    .minimum       = 1,
    .maximum       = 100,
    .step          = 0x01,
    .default_value = DEF_PSAT_GAIN_INC,
    .flags         = 0,
  }
  , {
    .id            = V4L2_CID_STC_PIX_SATURATION_LO_LIMIT,
    .type          = V4L2_CTRL_TYPE_INTEGER,
    .name          = "AL limit fraction lighted",
    .minimum       = 0, 
    .maximum       = 100000,
    .step          = 0x01,
    .default_value = DEF_PSAT_LO_LIMIT,
    .flags         = 0,
  }
  , {
    .id            = V4L2_CID_STC_PIX_SATURATION_AMORTIZ,
    .type          = V4L2_CTRL_TYPE_INTEGER,
    .name          = "AL amortization first adaptation",
    .minimum       = 1,
    .maximum       = 100,
    .step          = 0x01,
    .default_value = DEF_PSAT_AMORTIZ,
    .flags         = 0,
  }
  , {
    .id            = V4L2_CID_STC_PIX_SATURATION_AMORTIZ_LENGTH,
    .type          = V4L2_CTRL_TYPE_INTEGER,
    .name          = "AL amortization length",
    .minimum       = 1,
    .maximum       = 100,
    .step          = 0x01,
    .default_value = DEF_PSAT_AMORTIZ_LENGTH,
    .flags         = 0,
  }
  , {
    .id            = V4L2_CID_STC_PIX_SATURATION_MAX,
    .type          = V4L2_CTRL_TYPE_INTEGER,
    .name          = "AL saturation threshold",
    .minimum       = 1,
    .maximum       = 255,
    .step          = 0x01,
    .default_value = DEF_PSAT_MAX,
    .flags         = 0,
  }
  , {
    .id            = V4L2_CID_STC_PIX_SATURATION_SKIP,
    .type          = V4L2_CTRL_TYPE_INTEGER,
    .name          = "AL frames skip",
    .minimum       = 1, 
    .maximum       = 255,
    .step          = 0x01,
    .default_value = DEF_PSAT_SKIP,
    .flags         = 0,
  }
  , {
    .id            = V4L2_CID_STC_BYTES,
    .type          = V4L2_CTRL_TYPE_INTEGER,
    .name          = "Bytes per pixel",
    .minimum       = 1, 
    .maximum       = 1,
    .step          = 0x01,
    .default_value = 1,
    .flags         = 0,
  }
, {
    .id            = V4L2_CID_STC_NATIVE_MODES,
    .type          = V4L2_CTRL_TYPE_BOOLEAN,
    .name          = "Enable native modes, fast",
    .minimum       = 0,
    .maximum       = 1,
    .step          = 0x1,
    .default_value = DEFAULT_NATIVE_MODES,
    .flags         = 0,
  }
, {
    .id            = V4L2_CID_STC_LONG_EXPOSURE,
    .type          = V4L2_CTRL_TYPE_BOOLEAN,
    .name          = "Long Exposure (USB3)",
    .minimum       = 0,
    .maximum       = 1,
    .step          = 0x1,
    .default_value = 0,
    .flags         = V4L2_CTRL_FLAG_UPDATE, 
  }
, {
    .id            = V4L2_CID_STC_AUTOSHUTTERS,
    .type          = V4L2_CTRL_TYPE_BOOLEAN,
    .name          = "Automatic Shutters",
    .minimum       = 0,
    .maximum       = 1,
    .step          = 0x1,
    .default_value = 0x1,
    .flags         = 0,
  }
, {
    .id            = V4L2_CID_STC_OUTPUT_FRAME_RATE,
    .type          = V4L2_CTRL_TYPE_INTEGER,
    .name          = "Output Frame Rate",
    .minimum       = 0,
    .maximum       = 127,
    .step          = 0x1,
    .default_value = 0x0,
    .flags         = 0,
  }
, {
    .id            = V4L2_CID_STC_PIXEL_BLEMISH_CORRECTION,
    .type          = V4L2_CTRL_TYPE_BOOLEAN,
    .name          = "Pixel Blemish Correction",
    .minimum       = 0,
    .maximum       = 1,
    .step          = 0x1,
    .default_value = 0x1,
    .flags         = 0,
  }
, {
    .id            = V4L2_CID_STC_HORIZONTAL_SIZE_CAPTURE,
    .type          = V4L2_CTRL_TYPE_INTEGER,
    .name          = "Horizontal Size Capture",
    .minimum       = 0,
    .maximum       = 4096,
    .step          = 0x4,
    .default_value = 0x0,
    .flags         = V4L2_CTRL_FLAG_DISABLED,
  }
, {
    .id            = V4L2_CID_STC_VERTICAL_SIZE_CAPTURE,
    .type          = V4L2_CTRL_TYPE_INTEGER,
    .name          = "Vertical Size Capture",
    .minimum       = 0,
    .maximum       = 4096,
    .step          = 0x4,
    .default_value = 0x0,
    .flags         = V4L2_CTRL_FLAG_DISABLED,
  }
, {
    .id            = V4L2_CID_STC_HORIZONTAL_OFFSET_CAPTURE,
    .type          = V4L2_CTRL_TYPE_INTEGER,
    .name          = "Horizontal Offset Capture",
    .minimum       = 0,
    .maximum       = 4096,
    .step          = 0x2,
    .default_value = 0x0,
    .flags         = V4L2_CTRL_FLAG_DISABLED,
  }
, {
    .id            = V4L2_CID_STC_VERTICAL_OFFSET_CAPTURE,
    .type          = V4L2_CTRL_TYPE_INTEGER,
    .name          = "Vertical Offset Capture",
    .minimum       = 0,
    .maximum       = 4096,
    .step          = 0x2,
    .default_value = 0x0,
    .flags         = V4L2_CTRL_FLAG_DISABLED,
  }
, {
    .id            = V4L2_CID_STC_VERTICAL_SCAN,
    .type          = V4L2_CTRL_TYPE_INTEGER,
    .name          = "Vertical Scan",
    .minimum       = 0,
    .maximum       = 4096,
    .step          = 0x1,
    .default_value = 0x0,
    .flags         = 0,
  }
, {
    .id            = V4L2_CID_STC_SELECTED_FOURCC,
    .type          = V4L2_CTRL_TYPE_INTEGER,
    .name          = "SELECTED_FOURCC",
    .minimum       = 0,
    .maximum       = 0xffffffff,
    .step          = 0x1,
    .default_value = 0x0,
    .flags         = 0,
  }
};
struct sentech_fmt {
  char* name;
  u32   fourcc;          
  int   depth;
  u32   priv;
  enum v4l2_colorspace colorspace;
};
static struct sentech_fmt format_y16 = {
	.name     = "Y16",
	.fourcc   = V4L2_PIX_FMT_Y16,
	.depth    = 16,
	.priv     = 0,
	.colorspace = 0
};
static struct sentech_fmt format_grey = {
	.name     = "Gray, Native",
	.fourcc   = V4L2_PIX_FMT_GREY,
	.depth    = 8,
	.priv     = 0,
	.colorspace = 0
};
static struct sentech_fmt format_rggb = {
	.name     = "Bayern, RGGB, Native",
	.fourcc   = V4L2_PIX_FMT_SRGGB8,
	.depth    = 8,
	.priv     = 0,
	.colorspace = 0
};
static struct sentech_fmt format_grbg = {
	.name     = "Bayern, GRBG, Native",
	.fourcc   = V4L2_PIX_FMT_SGRBG8,
	.depth    = 8,
	.priv     = 0,
	.colorspace = 0
};
static struct sentech_fmt format_rggb10 = {
	.name     = "Bayern, RGGB10, Native",
	.fourcc   = V4L2_PIX_FMT_SRGGB10,
	.depth    = 16,
	.priv     = 0,
	.colorspace = 0
};
static struct sentech_fmt format_rggb12 = {
	.name     = "Bayern, RGGB12, Native",
	.fourcc   = V4L2_PIX_FMT_SRGGB12,
	.depth    = 16,
	.priv     = 0,
	.colorspace = 0
};
static struct sentech_fmt format_rggb16 = {
	.name     = "Bayern, RGGB16, Native",
	.fourcc   = V4L2_PIX_FMT_SRGGB16,
	.depth    = 16,
	.priv     = 0,
	.colorspace = 0
};
static struct sentech_fmt format_bggr = {
	.name     = "Bayern, BGGR",
	.fourcc   = V4L2_PIX_FMT_SBGGR8,
	.depth    = 8,
	.priv     = 0,
	.colorspace = 0,
};
static struct sentech_fmt format_bgr24 = {
	.name     = "BGR, 24 bits",
	.fourcc   = V4L2_PIX_FMT_BGR24,
	.depth    = 24,
	.priv     = 0,
	.colorspace = V4L2_COLORSPACE_SRGB
};
static struct sentech_fmt format_rgb24 = {
	.name     = "RGB, 24 bits",
	.fourcc   = V4L2_PIX_FMT_RGB24,
	.depth    = 24,
	.priv     = 0,
	.colorspace = V4L2_COLORSPACE_SRGB
};
static struct sentech_fmt format_yuyv = {
	.name     = "4:2:2, packed, YUYV",
	.fourcc   = V4L2_PIX_FMT_YUYV,
	.depth    = 16,
	.priv     = 0,
	.colorspace = 0
};
static struct sentech_fmt format_yuv420 = {
	.name     = "4:2:0, planar, YUYV",
	.fourcc   = V4L2_PIX_FMT_YUV420,
	.depth    = 12,
	.priv     = 0,
	.colorspace = 0
};
static struct sentech_fmt format_yuv422P = {
	.name     = "4:2:2, planar, YUV",
	.fourcc   = V4L2_PIX_FMT_YUV422P,
	.depth    = 16,
	.priv     = 0,
	.colorspace = 0
};
static LIST_HEAD(sentech_camlist);
static LIST_HEAD(sentech_sublist);
struct sentech_camera;
struct sentech_usb_fragment {
  struct list_head buffers; 
  __u8 * framebuffer; 
  __u8 * framebuffer_image; 
  struct Sentech_Header *framebuffer_header;
  __u32 fragment_size, raw_size;
  dma_addr_t dma, dma_header, dma_image;
  struct sentech_usb_buffer *sub;
  struct completion done;
  struct urb * urb;
  int margin_buffer_left, margin_buffer_right;
  int id;
  unsigned int copied_to_left:1; 
  unsigned int copied_to_right:1; 
  unsigned int submitted:1; 
  unsigned int has_data:1;  
  unsigned int shutdown:1;  
  unsigned int last_suf:1;  
  unsigned int first_suf:1; 
  unsigned int last_suf_submitted:1; 
  unsigned int recovery:1; 
  unsigned int displaced:1; 
  int error_code;  
  struct timeval timestamp;
};
struct sentech_usb_buffer {
  struct list_head sub_buffers; 
  struct list_head suf_buffers; 
  __u8* vbuffer;
  unsigned int vbuffer_size;
  struct sentech_camera * cam_data;
  unsigned int fragmented:1; 
  unsigned int submitted:1;  
  unsigned int has_data:1;   
  unsigned int user_owned:1; 
  int mmaped;
  int mapped_size;
  int error_code; 
  __u32 id;         
  __u32 size;       
  __u32 received;
  struct usb_bus *bus;  
  struct timeval timestamp;
};
struct sentech_indirect_buffers_frame {
  __u16 frame_nb; 
  struct timeval timestamp;  
  int data;  
  int done;  
};
struct sentech_indirect_buffers {
  int vmalloc_buffers_nb;
  int vmalloc_buffers_size; 
  struct timeval last_timestamp;
  __u16 last_frame;
  __u8* vmalloc_buffers[MAX_VMALLOC_BUFFERS]; 
  __u8* vmalloc_vbuffers[MAX_VMALLOC_BUFFERS]; 
  int vmalloc_vbuffers_size[MAX_VMALLOC_BUFFERS];
  int vmalloc_mapped_size[MAX_VMALLOC_BUFFERS];
  int vmalloc_mmaped[MAX_VMALLOC_BUFFERS];
  struct sentech_indirect_buffers_frame sib_frame[MAX_VMALLOC_BUFFERS];
  int vmalloc_buffers_idx[MAX_VMALLOC_BUFFERS]; 
  int vmalloc_buffers_user[MAX_VMALLOC_BUFFERS]; 
  int vmalloc_buffers_next_free; 
  int vmalloc_buffers_next_full; 
  int vmalloc_buffers_urbs_nb;
  int vmalloc_buffers_urbs_framebuffer_size;
  struct urb* vmalloc_buffers_urb[VMALLOC_BUFFERS_URBS];
  struct sentech_camera* vmalloc_buffers_context[VMALLOC_BUFFERS_URBS+MAX_VMALLOC_BUFFERS];
  struct completion  vmalloc_buffers_done[VMALLOC_BUFFERS_URBS]; 
  dma_addr_t vmalloc_buffers_dma[VMALLOC_BUFFERS_URBS];
  int vmalloc_buffers_submitted[VMALLOC_BUFFERS_URBS]; 
  __u8 * vmalloc_buffers_framebuffer[VMALLOC_BUFFERS_URBS]; 
};
struct sentech_camera {
  struct list_head sentech_devlist; 
  struct list_head buffers; 
  struct sentech_usb_buffer* idle_frame; 
  rwlock_t buffers_lock; 
  rwlock_t ctrl_lock; 
  wait_queue_head_t wq;    
  wait_queue_head_t autoq; 
  struct semaphore format_lock; 
  struct sentech_indirect_buffers sib;
  __u8* awb_buffer;
  void* awb_object;
  int recent_autobalance; 
  char cam_name[CAM_NAME_LENGTH];
  int cam_number;
  int open_desc; 
  int max_fps, fps_idx;
  int max_Bpp, Bpp; 
  int mask_Bpp;  
  int bpp; 
  u32 model, model_idx;
  u16 color_mode;
  enum v4l2_buf_type type;
  struct v4l2_queryctrl qctrl[ARRAY_SIZE(sentech_qctrl)];
  int qctl_regs[ARRAY_SIZE(sentech_qctrl)];
  unsigned int requested_buffers;
  unsigned int allocated_buffers;
  int skiped_awb;  
  atomic_t running_buffers; 
  unsigned int usb3:1;
  unsigned int supports_bpp:1;
  unsigned int height_changed:1;
  unsigned int trigger_version:1;
  unsigned int fragmented:1; 
  unsigned int shutdown:1; 
  unsigned int has_data:1; 
  unsigned int run:1; 
  unsigned int suspended_running:1;
  unsigned int preallocate:1;
  unsigned int try_fragments:1;
  unsigned int continuous_capture:1;
  unsigned int reading:1;  
  unsigned int streaming:1; 
  unsigned int connected:1; 
  unsigned int exclusive:1; 
  unsigned int hflip:1; 
  unsigned int h_flip_fpga:1; 
  unsigned int v_flip_fpga:1; 
  unsigned int h_flip_fpga_can:1; 
  unsigned int v_flip_fpga_can:1; 
  unsigned int auto_wb_fpga_can:1;
  unsigned int auto_gain_fpga_can:1;
  unsigned int auto_shutters_fpga_can:1;
  unsigned int pixel_blemish_correction_fpga_can:1;
  unsigned int saving_camera_settings_fpga_can:1;
  unsigned int exposure_priority_not_fps:1; 
  unsigned int native:1; 
  unsigned int enable_native:1; 
  unsigned int req_userptr_memap_overlay:2;  
  unsigned int do_white_balance:1;
  unsigned int auto_white_balance:1;
  unsigned int auto_gain:1;
  unsigned int auto_shutters:1;
  unsigned int kthread_running:1;
  unsigned int probe_done:1;
  unsigned int recovery_sufs; 
  unsigned int recovery; 
  struct sentech_usb_buffer* recovery_sub;
  void* streamer; 
  int shutters, gain;
  int min_sat_exp, min_sat_gain, max_sat_exp, max_sat_gain, sat_unchanged, psat_amortiz_length, last_sat_inc, psat_skip, last_sat_dir:1;
  int psat_auto:1, psat_step, psat_min, psat_max, psat_acc, psat_up, psat_lo, psat_exp, psat_gain, psat_amortiz, psat_calls;
  BYTE exposure_target;
  struct task_struct *kthread_ctrl; 
  int freq_auto; 
  int vmas;          
  enum v4l2_buf_type buffertype; 
  struct semaphore               lock;      
  unsigned int                   resources; 
  struct usb_device *udev;
  struct usb_interface *interface;
  struct device *dev;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,11,0)
  struct v4l2_device v4l2_dev;
#if defined(CONFIG_MEDIA_CONTROLLER)
  struct media_device media_dev;
#endif
#endif
  struct v4l2_rect defrect; 
  struct v4l2_rect bounds; 
  struct v4l2_rect crop; 
  struct v4l2_rect output; 
  struct v4l2_pix_format pix_format;
  struct video_device* v4l2dev;
  __u32 first_frame, last_frame, lost_frames;
#ifdef L_BOUNDS
  __s32 gain_min, gain_max, exposure_min, exposure_max;
#endif
  struct camera_id ID_buf;
  struct sentech_fpga2 fpga2_buf;
  rwlock_t fpga2_lock;
  __u32 fpga2_offset_analog_control;
  __u32 fpga2_offset_vendor_unique_control;
  __u32 fpga2_offset_USB_camera_original_function__save;
  __u32 fpga2_offset_USB_camera_status;
  __u32 fpga2_offset_USB_camera_feature;
  __u32 fpga2_offset_HDR_control;
  __u8 save_settings_cmd[2]; 
  int fpga2_offsets_known_flag;
  int u3v;
  struct Sentech_Camera_Data_Payload   payload11;
  struct Sentech_Picture_Size_Payload  payload12;
  struct Sentech_Bytes_Per_Pixel_Data_Payload payload13;
  struct Sentech_Shutter_Gain_Payload  payload21;
  struct Sentech_Clock_Payload         payload22;
  struct Sentech_White_Balance_Payload payload23;
  struct Sentech_FPGA_Payload          payload80;
  struct Sentech_Picture_Size_Payload  state12;
  struct Sentech_Bytes_Per_Pixel_Data_Payload state13;
  struct Sentech_Shutter_Gain_Payload  state21;
  struct Sentech_Clock_Payload         state22;
  struct Sentech_White_Balance_Payload state23;
  struct Sentech_FPGA_Payload          state80;
  struct list_head ctrl_queue; 
  int modif_flag12,modif_flag13,modif_flag21,modif_flag22,modif_flag_80_00,modif_flag_80_04,modif_flag_80_30,modif_flag_80_0F,modif_flag_80_14;
  int modif_flag_80_3F,modif_flag_80_49;
  int modif_flag_80_4A,modif_flag_80_4B,modif_flag_80_4C,modif_flag_80_4D,modif_flag_80_4E;
  int modif_flag_80_45_46; 
  int modif_flag_80_47_48; 
  struct sentech_fmt*                  fmt;
  int monochrome;
  __u32 crt_frame_size;
  __u32 max_frame_size;
  __u32 framebuffer_size; 
  int picture_size; 
  int width,height; 
  unsigned int max_width,max_height; 
  int scan_horizontal_offset; 
  int scan_vertical_offset; 
  int scan_width; 
  int scan_height; 
  int scan_size; 
  int scanimage_size;
  int status; 
  int pipe_image, pipe_ctl_rcv, pipe_ctl_snd;  
  int brightness, saturation, gamma;
  __u16 red, blue, green; 
  struct usb_bus *bus;  
  struct kref             kref;
};
static int sentech_trigger_mode(struct sentech_camera* cam_data, int memflags);
static struct sentech_usb_buffer*  alloc_sub(struct sentech_camera* cam_data);
static void free_first_sub (struct sentech_camera* cam_data);
static void sentech_submit_21(struct sentech_camera* cam_data, int memflags);
static 
int testAndClear(int* modif_flag) {
  int result = *modif_flag;
  *modif_flag = 0;
  return result;
}
static
void setFlag(int* modif_flag) {
  *modif_flag = 1;
}
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,13)
#define L2613
#else
#define L2613 struct device_attribute __attribute__((unused))  *attr,
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,17,0)                       
#define SENTECH_SHOW_FLAGS (S_IWUGO | S_IRUGO)
#else                                                                   
#define SENTECH_SHOW_FLAGS 0660            
#endif
#define show_set(value) \
static ssize_t show_##value(struct device *dev, L2613 char *buf)	\
{                                                                       \
        struct usb_interface *intf = to_usb_interface(dev);             \
        struct sentech_camera *cam_data = usb_get_intfdata(intf);       \
                                                                        \
        return sprintf(buf, "%.2d\n", cam_data->value);                 \
}                                                                       \
static ssize_t set_##value(struct device *dev, L2613 const char *buf, size_t count) \
{                                                                       \
        struct usb_interface *intf = to_usb_interface(dev);             \
        struct sentech_camera *cam_data = usb_get_intfdata(intf);       \
        int temp = simple_strtoul(buf, NULL, 10);                       \
        dev_info(&cam_data->udev->dev, "setting %s %d\n", buf, temp);   \
        cam_data->value = temp;                                         \
        return count;                                                   \
}                                                                       \
static DEVICE_ATTR(value, SENTECH_SHOW_FLAGS, show_##value, set_##value);
show_set(first_frame);
show_set(last_frame);
show_set(lost_frames);
#define show_string(value) \
static ssize_t show_##value(struct device *dev, L2613 char *buf)	\
{                                                                       \
        struct usb_interface *intf = to_usb_interface(dev);             \
        struct sentech_camera *cam_data = usb_get_intfdata(intf);       \
        if(cam_data->udev->value != NULL)		      		\
	  return sprintf(buf, "%s\n", cam_data->udev->value);		\
	return sprintf(buf, "No such information supplied!\n");         \
}                                                                       \
static ssize_t set_##value(struct device *dev, L2613 const char *buf, size_t count) \
{                                                                       \
        struct usb_interface *intf = to_usb_interface(dev);             \
        struct sentech_camera *cam_data = usb_get_intfdata(intf);       \
        int temp = simple_strtoul(buf, NULL, 10);                       \
        dev_info(&cam_data->udev->dev, "setting string %s %d\n", buf, temp);   \
        return count;                                                   \
}                                                                       \
static DEVICE_ATTR(value, S_IRUGO, show_##value, set_##value);
#if  LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,12)
show_string(serial);
show_string(product);
show_string(manufacturer);
#endif
#ifdef SYSFS_TEST
static ssize_t show_sysfs(struct device *dev, L2613 char *buf)
{
        struct kobject *kobj = &dev->kobj;                              
	char *s_name = kobject_name(kobj);
        if(s_name != NULL)		                		
	  return sprintf(buf, "%s\n", s_name);		        
	return sprintf(buf, "No such information supplied!\n");         
}                                                                       
static ssize_t set_sysfs(struct device *dev, L2613 const char *buf, size_t count) 
{                                                                       
        struct usb_interface *intf = to_usb_interface(dev);             
        struct sentech_camera *cam_data = usb_get_intfdata(intf);       
        int temp = simple_strtoul(buf, NULL, 10);                       
        dev_info(&cam_data->udev->dev, "setting string %s %d\n", buf, temp);   
        return count;                                                   
}                                                                       
static DEVICE_ATTR(sysfs_name, S_IRUGO, show_sysfs, set_sysfs);
#endif
#ifndef lprintk
#define lprintk(level,fmt, arg...) \
  do {                      \
  if (debug != 0)     \
  printk(KERN_DEBUG "sentech: " fmt , ## arg); \
  } while (0)
#endif
#ifndef dprintk
#define dprintk(level,fmt, arg...) \
  do {                      \
  if (debug >= (level))     \
  printk(KERN_DEBUG "sentech: " fmt , ## arg); \
  } while (0)
#endif
inline
void dump_data(int level, void* data, int size) {
  int k;
  unsigned char * _data = (unsigned char*)data;
  if(debug >=level){
    printk("<4> L:");
    for(k=0; k<size; k++) {
      printk("%hx.",_data[k]);
    }
    printk("\n");
  }
}
__s32 get_max_exposure(struct sentech_camera*cam_data) {
  __s32 max_exp;;
  if(cam_data->usb3) {
    if(cam_data->exposure_priority_not_fps){
      max_exp = (__le32_to_cpu(cam_data->state12.Longest_ET_exposure_H_le));
    }else{
      max_exp = (__le32_to_cpu(cam_data->state12.Longest_FPS_exposure_H_le));
    }
  }else{
    max_exp = (__le16_to_cpu(cam_data->state12.line_numbers_le)-1)*
      __le16_to_cpu(cam_data->state12.one_line_clock_numbers_le)+
      (__le16_to_cpu(cam_data->state12.one_line_clock_numbers_le)-1);
  };
  return max_exp;
}
__s32 get_crt_exposure(struct sentech_camera*cam_data) {
  __s32 crt_exp;
  if(cam_data->usb3) {
    crt_exp = __le16_to_cpu(cam_data->state21.Exposure_time_H_unit_le);
    crt_exp <<=16;
    crt_exp += __le16_to_cpu(cam_data->state21.Shutter_line_numbers_le);
    if(crt_exp==0) return get_max_exposure(cam_data);
  }else{
    __u16 lines, clocks;
    lines =  __le16_to_cpu(cam_data->state21.Shutter_line_numbers_le);
    clocks = __le16_to_cpu(cam_data->state21.Shutter_clock_numbers_le);;
    if((lines==0)&&(clocks==0)) return get_max_exposure(cam_data);
    crt_exp = (lines)*
      __le16_to_cpu(cam_data->state12.one_line_clock_numbers_le)+
      (clocks);
  };
  return crt_exp;
}
void set_crt_exposure(struct sentech_camera*cam_data, __s32 crt_exp) {
  __u16 line_numbers;
  __u16 clock_numbers;;
  if(cam_data->usb3) {
    if(crt_exp==0) crt_exp=1;
    cam_data->payload21.Shutter_line_numbers_le = __cpu_to_le16(crt_exp&0x00ffff);
    crt_exp >>= 16;
    cam_data->payload21.Exposure_time_H_unit_le = __cpu_to_le16(crt_exp&0x00ffff);
    setFlag(&cam_data->modif_flag21);;
  }else{
    if(crt_exp==0) crt_exp=1;
    line_numbers = crt_exp/__le16_to_cpu(cam_data->state12.one_line_clock_numbers_le);
    clock_numbers = crt_exp - (line_numbers) * __le16_to_cpu(cam_data->state12.one_line_clock_numbers_le);
    cam_data->payload21.Shutter_line_numbers_le = __cpu_to_le16(line_numbers);
    cam_data->payload21.Shutter_clock_numbers_le = __cpu_to_le16(clock_numbers);
    setFlag(&cam_data->modif_flag21);;
  }
}
void set_crt_exposure_async(struct sentech_camera*cam_data, __s32 crt_exp, int memflags) {
  __u16 line_numbers;
  __u16 clock_numbers;;
  if (cam_data->usb3) {
    if(crt_exp==0) crt_exp=1;
    cam_data->payload21.Shutter_line_numbers_le = __cpu_to_le16(crt_exp&0x00ffff);
    crt_exp >>= 16;
    cam_data->payload21.Exposure_time_H_unit_le = __cpu_to_le16(crt_exp&0x00ffff);
    sentech_submit_21(cam_data, memflags);;
  } else {
    if(crt_exp==0) crt_exp=1;
    line_numbers = crt_exp/__le16_to_cpu(cam_data->state12.one_line_clock_numbers_le);
    clock_numbers = crt_exp - (line_numbers) * __le16_to_cpu(cam_data->state12.one_line_clock_numbers_le);
    cam_data->payload21.Shutter_line_numbers_le = __cpu_to_le16(line_numbers);
    cam_data->payload21.Shutter_clock_numbers_le = __cpu_to_le16(clock_numbers);
    sentech_submit_21(cam_data, memflags);;
  }
}
__s16 get_initial_gain(struct sentech_camera*cam_data) {
  return __le16_to_cpu(cam_data->state12.initial_gain_le);
}
__s16 get_crt_gain(struct sentech_camera*cam_data) {
  __s16 crt_gain = __le16_to_cpu(cam_data->payload21.Gain_le);
  return crt_gain;
}
void set_crt_gain_async(struct sentech_camera*cam_data, __s16 crt_gain, int memflags) {;
  cam_data->payload21.Gain_le = __cpu_to_le16(crt_gain);
  sentech_submit_21(cam_data, memflags);
}
void set_crt_gain(struct sentech_camera*cam_data, __s16 crt_gain) {;
  cam_data->payload21.Gain_le = __cpu_to_le16(crt_gain);
  setFlag(&cam_data->modif_flag21);
}
__s32 useconds_2_exposure_usb3(struct sentech_camera*cam_data, __s32 usec) {
  return usec; 
}
int	GetAptina4MRowClockCount(__s32 dwImageWidth, __u8  byteHBinning, __u8  byteVBinning,
				 __s32* pdwRowClockCount, __s32* pdwShutterOverhead){
	int bReval = 1;
	__s32	dwHBMin;
	__s32	dwRowClock;
	do{
		__s32	pdwRowClockCandidate[2];
		__s32	pdwHBMinList[3][3] = 
		{
			{450, 430, 420},
			{796, 776, 766},
			{1488, 1468, 1458}
		};
		if(3 <= byteHBinning)	byteHBinning = 2;
		if(3 <= byteVBinning)	byteVBinning = 2;
		dwHBMin = pdwHBMinList[byteVBinning][byteHBinning];
		pdwRowClockCandidate[0] = dwImageWidth / 2 + dwHBMin;
		pdwRowClockCandidate[1] = 41 + 346 * (byteVBinning + 1) + 99;
		dwRowClock = pdwRowClockCandidate[0];
		if(dwRowClock < pdwRowClockCandidate[1])	dwRowClock = pdwRowClockCandidate[1];
		dwRowClock *= 2;
		*pdwRowClockCount = dwRowClock;
		*pdwShutterOverhead = 2 * (208 * (byteVBinning + 1) + 98 + 1 - 94);
	}while(0);
	return(bReval);
}
__s32 exposure_2_useconds_usb3(struct sentech_camera*cam_data, __s32 exposure) {
  return exposure;
}
__s32 exposure_2_useconds(struct sentech_camera*cam_data, __s32 exposure) {
  struct Sentech_Camera_Data_Payload* cd;
  struct Sentech_Clock_Payload* shg;
  __u32 clock_mode;
  __u32 clock;
  __s32 result;
  if(cam_data->usb3) return exposure_2_useconds_usb3(cam_data, exposure);
  cd = &cam_data->payload11;
  shg = &cam_data->state22;
  clock_mode = __le32_to_cpu(shg->Clock_mode_le);
  clock = __le32_to_cpu(cd->Standard_clock_speed_le);
  if(clock_mode<3) clock >>= clock_mode;
  result = exposure*1000 / (clock/1000);;
  return result;
}
__s32 useconds_2_exposure(struct sentech_camera*cam_data, __s32 usec) {
  struct Sentech_Camera_Data_Payload* cd = &cam_data->payload11;
  struct Sentech_Clock_Payload* shg = &cam_data->payload22;
  __u32 clock_mode = __le32_to_cpu(shg->Clock_mode_le);
  __u32 clock = __le32_to_cpu(cd->Standard_clock_speed_le);
  __s32 result;
  if(cam_data->usb3) return useconds_2_exposure_usb3(cam_data, usec);
  cd = &cam_data->payload11;
  shg = &cam_data->payload22;
  clock_mode = __le32_to_cpu(shg->Clock_mode_le);
  clock = __le32_to_cpu(cd->Standard_clock_speed_le);
  if(clock_mode<3) clock >>= clock_mode;
  result = (usec*(clock/1000))/1000;;
  return result;
}
static void set_controls(struct sentech_camera* cam_data, int memflags);
static void sentech_free_subs(struct sentech_camera* cam_data);
static void free_sub(struct sentech_usb_buffer* sub,
		     struct sentech_camera* cam_data);
static void sentech_free_non_dma_buffers (struct sentech_camera* cam_data);
#define to_cam_data_kref(d) container_of(d, struct sentech_camera, kref)
static void cam_data_delete(struct kref *kref){
  struct sentech_camera *cam_data;
  int running;
  DEFINE_WAIT(wait);;
  cam_data = to_cam_data_kref(kref);;
  if(!cam_data->has_data) {;
    return;
  }
  else
    cam_data->has_data = 0;
  running = atomic_read(&cam_data->running_buffers);;
  cam_data->run=0;
  while(atomic_read(&cam_data->running_buffers)) {;
    prepare_to_wait(&cam_data->wq,
		    &wait,TASK_INTERRUPTIBLE);;
    if(atomic_read(&cam_data->running_buffers)) {;
      schedule_timeout (HZ);
    }
    finish_wait(&cam_data->wq, &wait);
    running --;
    if(running <= 0){;
      break;
    }
  };
;
  if(cam_data->sentech_devlist.next &&
     cam_data->sentech_devlist.next != &cam_data->sentech_devlist) {
    list_del(&cam_data->sentech_devlist);
    cam_data->sentech_devlist.prev = cam_data->sentech_devlist.next = 0;
  }else{; 
   return;
  }
;
  sentech_free_subs(cam_data);
  if(!param_have_buffers)
    sentech_free_non_dma_buffers(cam_data);
  if(cam_data->idle_frame != NULL) {
    free_sub(cam_data->idle_frame, cam_data);
    cam_data->idle_frame = 0;
  }
#ifdef SENTECH_AWB
  if(cam_data->awb_object) {;
    kfree(cam_data->awb_object);
    cam_data->awb_object = NULL;
  }
  if(cam_data->awb_buffer) {;
    vfree(cam_data->awb_buffer);
    cam_data->awb_buffer = 0;
  }
#endif
  ;;
  memset(cam_data,0,sizeof(*cam_data));
  kfree (cam_data);;
}
#define SFH_MAGIC 0xA5A5
struct sentech_fh {
  struct sentech_camera  *cam_data;
  int priority;
  __u16 magic; 
};
inline struct sentech_fh*
alloc_sentech_fh(int flags) {
  struct sentech_fh* fh;
  fh = kzalloc(sizeof(*fh), flags);
  fh->magic = SFH_MAGIC;
  return fh;
}
static 
struct sentech_camera* get_cam_data(void* file_priv_data) {
  struct sentech_camera* result;
  if(!file_priv_data){;
    return NULL;
  }
  if(!(((struct sentech_fh *)file_priv_data)->magic == SFH_MAGIC))
    lprintk(1,"No magic fh data at %s\n", __FUNCTION__);
  result = ((struct sentech_fh *) file_priv_data) -> cam_data;
  if(!result)
    lprintk(1,"No private data at %s\n", __FUNCTION__);
  return result;
}
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,20)
#define ARG_COMPLETE
#else
#define ARG_COMPLETE  ,struct pt_regs *regs
#endif
static void sentech_usb_complete(struct urb * urb ARG_COMPLETE);
static void sentech_usb_indirect_complete(struct urb * urb ARG_COMPLETE);
static void sentech_complete_direction_change(struct urb * urb ARG_COMPLETE);
static void sentech_complete_fpga_set(struct urb * urb ARG_COMPLETE);
static int sentech_io_direction_change(struct sentech_camera* cam_data,int memflags);
static int sentech_fpga_set(struct sentech_camera* cam_data,int memflags, int cmd, int len);
static int sentech_fpga_set_lock(struct sentech_camera* cam_data, int memflags, int cmd, int len, int sync);
static int sentech_submit_indirect_kurbs(struct sentech_camera* cam_data, int mem_flags);
static struct sentech_usb_fragment*
first_suf(struct sentech_usb_buffer* sub) {
  struct sentech_usb_fragment* suf = NULL;
  suf = list_first_entry(&sub->suf_buffers,
			      struct sentech_usb_fragment, buffers);
  return suf;
}
static void free_sub_general(struct sentech_usb_buffer* sub,
		      struct sentech_camera* cam_data, int force);
static void free_sub(struct sentech_usb_buffer* sub,
		     struct sentech_camera* cam_data) {
  free_sub_general(sub, cam_data, 0);
}
static void free_sub_force(struct sentech_usb_buffer* sub,
		     struct sentech_camera* cam_data) {
  free_sub_general(sub, cam_data, 1);
}
static void free_sub_general(struct sentech_usb_buffer* sub,
			     struct sentech_camera* cam_data, int force){
  unsigned int k;
  struct page *page;
  unsigned long flags;;
  if(!sub){;
    return;
  }
  if(cam_data 
     && (!force)
     && !unloading_module) { 
    sub->cam_data = 0;
    write_lock_irqsave(&subs_list_lock, flags);
    list_add_tail(&sub->sub_buffers, &sentech_sublist);
    write_unlock_irqrestore(&subs_list_lock, flags);;
    return;
  }
;
  while(!list_empty(&sub->suf_buffers)) {
    struct sentech_usb_fragment* suf = first_suf(sub);
;
    if (!suf) {;
     break;
    }
    if(suf->submitted && cam_data && cam_data->connected) {;
      wait_for_completion_timeout(&suf->done,HZ);;
    }
    if (suf->urb) {
;
      usb_free_urb(suf->urb);
      suf->urb=NULL;
    }
    for(k=0; k<suf->fragment_size; k += PAGE_SIZE) {
      if(virt_addr_valid(suf->framebuffer_image + k)) {
	page = virt_to_page(suf->framebuffer_image + k);
      }else{
	page = vmalloc_to_page(suf->framebuffer_image + k);
      }
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,14,0)
      put_page_testzero (page); 
#else
    if (page_count(page) > 0) put_page_testzero (page);
#endif
    }
    {;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 35)
      usb_buffer_free(container_of(&(sub->bus), struct usb_device, bus),
		      suf->raw_size,suf->framebuffer,suf->dma);
#else
      usb_free_coherent(container_of(&(sub->bus),struct usb_device, bus),suf->raw_size, suf->framebuffer, suf->dma); 
#endif
      ;;
    }
    list_del(&suf->buffers);
;
    kfree(suf);;
  };
  if (param_keep_vbuffer) {
    if (sub->vbuffer) {
      vfree(sub->vbuffer);;
      sub->vbuffer = 0;
    }
  }
;
  kfree(sub);;
}
static void free_first_sub (struct sentech_camera* cam_data) {
  struct sentech_usb_buffer* sub=NULL;
  unsigned long flags;;
  read_lock_irqsave(&cam_data->buffers_lock, flags);
  if(!list_empty(&cam_data->buffers))
    sub =
      list_first_entry(&cam_data->buffers,
		       struct sentech_usb_buffer,
		       sub_buffers);
  read_unlock_irqrestore(&cam_data->buffers_lock, flags);
  if(!sub) goto end_ffs;
  write_lock_irqsave(&cam_data->buffers_lock, flags);
  list_del(&sub->sub_buffers);
  write_unlock_irqrestore(&cam_data->buffers_lock, flags);
;
  free_sub(sub, cam_data);
 end_ffs:;
}
static struct sentech_usb_buffer*
get_sub(struct sentech_camera* cam_data, __u32 index) {
  struct sentech_usb_buffer*result=NULL, * sub;
  unsigned long flags;
  read_lock_irqsave(&cam_data->buffers_lock, flags);  
  list_for_each_entry(sub, &cam_data->buffers, sub_buffers) {
    if(sub->id == index) {
      result = sub;
      break;
    }
  }
  read_unlock_irqrestore(&cam_data->buffers_lock, flags);
  return result;
}
#define round_bulk(a)				\
  ((a) + ((512-((a)&0x1FF))&0x1FF))
#define round_down_bulk(a)				\
  ((a)-((a)&0x1FF))
#define round_to_page_size(a)				\
  ((a) + ((PAGE_SIZE-((a)&(PAGE_SIZE-1)))&(PAGE_SIZE-1)))
#define round_down_to_page_size(a)				\
  ((a)-((a)&(PAGE_SIZE-1)))
static 
struct sentech_usb_buffer*  alloc_sub(struct sentech_camera* cam_data) {
  unsigned int size=cam_data->max_frame_size;
  int first_suf=1;
  unsigned int first_frag_size = size;
  unsigned int fragment_size = size;
  unsigned long flags;
  int reusing_sub = 0;
  struct sentech_usb_buffer* sub=0;
  unsigned int k;  
  struct page* page; 
  dma_addr_t dma;
  struct sentech_usb_fragment *suf, *next_suf;
  int raw_size; 
  int buffer_safety_padding;
  unsigned int urb_buf_size;
  u8* buffer_urb;
  int crt_suf_id = 0;;;
  write_lock_irqsave(&subs_list_lock, flags);
  if (!list_empty(&sentech_sublist)) {
    list_for_each_entry(sub, &sentech_sublist, sub_buffers) {
      if ((sub->size == size) &&
	  (sub->bus == cam_data->bus))
	reusing_sub = 1;
	break;
    }
    if(reusing_sub) {
      list_del(&sub->sub_buffers);
    }
  }
  write_unlock_irqrestore(&subs_list_lock, flags);
  buffer_safety_padding = sizeof(struct Sentech_Header);
  if(param_max_fragment>0) {
    if((param_max_fragment-buffer_safety_padding) < fragment_size)
      fragment_size = param_max_fragment-buffer_safety_padding;
    if(fragment_size>=PAGE_SIZE)  
      fragment_size &= ~(PAGE_SIZE-1); 
    else fragment_size = round_bulk(fragment_size); 
  }
  if(reusing_sub) {
    sub->cam_data = cam_data;
    sub->submitted = 0;
    sub->has_data = 0;
    sub->user_owned = 0;
    sub->error_code = 0;
    sub->mmaped = 0;
    if(sub->fragmented) cam_data->fragmented = 1;
    list_for_each_entry(suf, &sub->suf_buffers, buffers) {
      suf->error_code = 0;
      suf->submitted = suf->has_data = suf->shutdown = 0;
      if(suf->first_suf) first_frag_size = suf->fragment_size + sizeof(struct Sentech_Header);
      else first_frag_size = suf->fragment_size;
#ifdef SENTECH_ALL_BUFFER_FRAGMENTS_DATA_ALIGNED
      urb_buf_size = round_bulk(first_suf?first_frag_size:suf->fragment_size);
      buffer_urb = suf->first_suf?((__u8*)suf->framebuffer_header):
	suf->framebuffer_image;;;
#else
      if(first_suf && (size <= suf->fragment_size)) {
	urb_buf_size = 
	  (suf->fragment_size+sizeof(struct Sentech_Header));
      } else
	urb_buf_size = round_down_to_page_size(first_frag_size);
      buffer_urb = ((__u8*)suf->framebuffer_header);;;
#endif
      usb_fill_bulk_urb(suf->urb, 
			cam_data->udev, 
			cam_data->pipe_image, 
			buffer_urb, 
			urb_buf_size, 
			sentech_usb_complete,      
			suf 
			);
      suf->margin_buffer_left = (((__u8*)buffer_urb) - suf->framebuffer);
      suf->margin_buffer_right = suf->raw_size -
	(urb_buf_size + 
	 suf->margin_buffer_left);
      if ((suf->margin_buffer_left<0) || (suf->margin_buffer_right<0)) {
	printk(KERN_ALERT "Error in reused buffer allocation margins: %d  %d, sz= %x / %x %p\n"
	       ,suf->margin_buffer_left, suf->margin_buffer_right
	       ,suf->fragment_size, suf->raw_size, suf->framebuffer
	       );
	free_sub_force(sub, cam_data);
	goto noreuse;
      }
    };
    list_for_each_entry(next_suf, &sub->suf_buffers, buffers) {;
    }
    goto done;
  }
 noreuse:
  sub = kmalloc(sizeof(struct sentech_usb_buffer), GFP_KERNEL);
  if(!sub) return NULL;;
  memset (sub, 0x00, sizeof (struct sentech_usb_buffer));
  sub->size = size;
  sub->cam_data = cam_data;
  sub->bus = cam_data->bus; 
  INIT_LIST_HEAD(&sub->suf_buffers);;
  while (size>0) {
    __u8 * framebuffer = 0;
    raw_size = round_to_page_size(fragment_size+buffer_safety_padding); 
;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 35)
    framebuffer = usb_buffer_alloc (cam_data->udev,
				    raw_size,
				    GFP_KERNEL|__GFP_COLD|__GFP_REPEAT,
				    &dma);
#else
    framebuffer = usb_alloc_coherent (cam_data->udev,
				      raw_size,
				      GFP_KERNEL|__GFP_COLD|__GFP_REPEAT,
				      &dma);
#endif
;
    if(!framebuffer) {
      printk(KERN_ALERT "sentech: failed allocating suf buffer of size %x\n", raw_size);
      if(cam_data->try_fragments) {
	sub->fragmented = cam_data->fragmented = 1;
	fragment_size >>=1;
	if(fragment_size>PAGE_SIZE){ 
	  fragment_size &= ~(PAGE_SIZE-1);
	  fragment_size += PAGE_SIZE;  
	}
	fragment_size &= ~(IMAGE_BULK_PACKET-1);
	if(fragment_size>=IMAGE_BULK_PACKET) continue;
	else {free_sub_force(sub, cam_data); return NULL;}
      }else{
	kfree(sub);
	sub = NULL;
	goto done;
      }
    }
    suf = kmalloc(sizeof (*suf), GFP_KERNEL);
    if (!suf) {
      if(framebuffer){
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 35)
	usb_buffer_free(cam_data->udev,raw_size,framebuffer,dma);
#else
	usb_free_coherent(cam_data->udev,raw_size, framebuffer, dma);
#endif
      }else{;
      }
      free_sub_force(sub, cam_data);
      sub = NULL;
      goto done;
    }
    memset (framebuffer, 0x00, raw_size);
    memset (suf, 0x00, sizeof (*suf));
    suf->id = crt_suf_id++;
    suf->sub = sub;
    suf->fragment_size = fragment_size;;
    if (first_suf) {
      if (size > fragment_size) { 
	suf->fragment_size -= sizeof(struct Sentech_Header);;
      }
    }
    first_frag_size = fragment_size;
    suf->raw_size = raw_size;
    suf->framebuffer = framebuffer;
    suf->framebuffer_image = __va(((((__pa(framebuffer))+
				     sizeof(struct Sentech_Header)-
				     1)>>PAGE_SHIFT)+1)<<PAGE_SHIFT);
    for(k=0; k<suf->fragment_size; k += PAGE_SIZE) {
      if(virt_addr_valid(suf->framebuffer_image + k)) {
	page = virt_to_page(suf->framebuffer_image + k);
      }else{
	page = vmalloc_to_page(suf->framebuffer_image + k);
      }
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,14,0)
      get_page (page);
#else
      if (page_count(page) <= 0) init_page_count(page);
      else get_page(page);
#endif
    }
    suf->framebuffer_header = (struct Sentech_Header*)
      (suf->framebuffer_image -
       sizeof(struct Sentech_Header));
    suf->dma = dma;
    suf->dma_header = dma + (((__u8*)suf->framebuffer_header)
			     - suf->framebuffer);
    suf->dma_image  = dma + (suf->framebuffer_image - suf->framebuffer);
    suf->first_suf = first_suf;
;
    list_add_tail(&suf->buffers, &sub->suf_buffers);;
    if(size>=suf->fragment_size) size -= suf->fragment_size;
    else size = 0;
    if(0) if(size<fragment_size) fragment_size = size;
    suf->last_suf = (size<=0);
    if(! (suf->urb = usb_alloc_urb(0,GFP_KERNEL))){
      dev_err(cam_data->dev, "SentechUSB: Out of memory allocating urb\n");
      free_sub_force(sub, cam_data);
      sub= NULL;
      goto done;
    }
#ifdef SENTECH_ALL_BUFFER_FRAGMENTS_DATA_ALIGNED
    urb_buf_size = round_bulk(first_suf?first_frag_size:suf->fragment_size);
    buffer_urb = suf->first_suf?((__u8*)suf->framebuffer_header):
      suf->framebuffer_image;;;
#else
      if(first_suf && (size == 0)) {
	urb_buf_size = 
	  (suf->fragment_size+sizeof(struct Sentech_Header));
      } else
	urb_buf_size = round_down_to_page_size(first_frag_size);
    buffer_urb = ((__u8*)suf->framebuffer_header);
;;
#endif
    usb_fill_bulk_urb(suf->urb, 
		      cam_data->udev, 
		      cam_data->pipe_image, 
		      buffer_urb, 
		      urb_buf_size, 
		      sentech_usb_complete,      
		      suf 
		      );
    suf->margin_buffer_left = (((__u8*)buffer_urb) - suf->framebuffer);
    suf->margin_buffer_right = suf->raw_size -
      (urb_buf_size+ 
       suf->margin_buffer_left);
    if ((suf->margin_buffer_left<0) || (suf->margin_buffer_right<0)) {
      printk(KERN_ALERT "Error in buffer allocation margins: %d  %d, sz= %x / %x %p\n"
	     ,suf->margin_buffer_left, suf->margin_buffer_right
	     ,suf->fragment_size, suf->raw_size, suf->framebuffer
	     );
      free_sub_force(sub, cam_data);
      sub = NULL;
      goto done;
    }
;
    if (suf->first_suf) suf->fragment_size = urb_buf_size - sizeof(struct Sentech_Header);
    else suf->fragment_size = urb_buf_size;
    suf->urb->transfer_flags = URB_NO_TRANSFER_DMA_MAP;
#ifdef SENTECH_ALL_BUFFER_FRAGMENTS_DATA_ALIGNED
    suf->urb->transfer_dma = first_suf?suf->dma_header:suf->dma_image;
#else
    suf->urb->transfer_dma = suf->dma_header;
#endif
    first_suf=0;
  }
 done:
  if(sub!=NULL) {;
  };
  return sub;
}
#ifdef L_BOUNDS_ANY
void 
sentech_exposure_bounds(struct sentech_camera* cam_data, 
			unsigned int i, 
			__s32 max_exp) {
  if(param_luminance_bounds_us) {
    max_exp = exposure_2_useconds(cam_data, max_exp);
  };
  if (V4L2_CID_STC_EXPOSURE_MIN == sentech_qctrl[i].id) {
    cam_data->qctrl[i].maximum = max_exp;
#ifdef L_BOUNDS
    if(cam_data->exposure_min > max_exp)
      cam_data->qctl_regs[i] = cam_data->exposure_min = max_exp;
#endif
  }
  if (V4L2_CID_STC_EXPOSURE_MAX == sentech_qctrl[i].id) {
    if(cam_data->qctrl[i].maximum != max_exp) {
      cam_data->qctl_regs[i] = max_exp;
    }
    cam_data->qctrl[i].maximum = max_exp;
    cam_data->qctrl[i].default_value = max_exp;
#ifdef L_BOUNDS
    if(cam_data->qctrl[i].maximum != max_exp) {
      cam_data->exposure_max = max_exp;
    }
    if(cam_data->exposure_max > max_exp)
      cam_data->qctl_regs[i] = cam_data->exposure_max = max_exp;
#endif
  }
}
#endif
inline int f_min(int x, int y){
  if(x<y) return x;
  return y;
}
inline int f_max(int x, int y){
  if(x<y) return y;
  return x;
}
static void sentech_free_non_dma_buffers (struct sentech_camera* cam_data) {
  int i;
   for (i = 0; i < cam_data->sib.vmalloc_buffers_nb; i++) {
     if (cam_data->sib.vmalloc_buffers[i])
       vfree(cam_data->sib.vmalloc_buffers[i]);
     cam_data->sib.vmalloc_buffers[i] = 0;
   }
   cam_data->sib.vmalloc_buffers_nb = 0;
   for (i = 0; i < cam_data->sib.vmalloc_buffers_urbs_nb; i++) {
     if (cam_data->sib.vmalloc_buffers_framebuffer[i]) {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 35)
       usb_buffer_free(cam_data->udev,cam_data->sib.vmalloc_buffers_urbs_framebuffer_size, cam_data->sib.vmalloc_buffers_framebuffer[i], cam_data->sib.vmalloc_buffers_dma[i]);
#else
       usb_free_coherent(cam_data->udev,cam_data->sib.vmalloc_buffers_urbs_framebuffer_size, cam_data->sib.vmalloc_buffers_framebuffer[i], cam_data->sib.vmalloc_buffers_dma[i]);
#endif
       cam_data->sib.vmalloc_buffers_framebuffer[i] = 0;
     }
   }
   cam_data->sib.vmalloc_buffers_urbs_nb = 0;
}
static int
sentech_alloc_non_dma_buffers (struct sentech_camera* cam_data) { 
  int i, size;
  int nb_buffers = param_buffers;;
  if ((cam_data->allocated_buffers) || (cam_data->sib.vmalloc_buffers_nb)) {;
    return 0; 
  }
  if(!cam_data->max_frame_size) {;
    return 0; 
  }
  cam_data->framebuffer_size = cam_data->max_frame_size + 
      sizeof(struct Sentech_Header);
  size = cam_data->framebuffer_size;
  if (param_buffers > MAX_VMALLOC_BUFFERS)
    nb_buffers = MAX_VMALLOC_BUFFERS;
  for (i = 0; i < nb_buffers; i++) {
    cam_data->sib.vmalloc_buffers_context[i] = cam_data;
    cam_data->sib.vmalloc_buffers[i] = vmalloc(size);
    if (cam_data->sib.vmalloc_buffers[i] == 0)
      break;
    cam_data->sib.vmalloc_buffers_nb++;
    cam_data->sib.vmalloc_buffers_size = size;;
  }
  if (!cam_data->sib.vmalloc_buffers_nb) {;
    return ENOMEM;
  }
  for (i = 0; i < VMALLOC_BUFFERS_URBS; i ++) {
    cam_data->sib.vmalloc_buffers_urbs_framebuffer_size =
      VMALLOC_BUFFERS_URB_FRAMEBUFFER;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 35)
    cam_data->sib.vmalloc_buffers_framebuffer[i] =
      usb_buffer_alloc (cam_data->udev,
			cam_data->sib.vmalloc_buffers_urbs_framebuffer_size,
			GFP_KERNEL|__GFP_COLD|__GFP_REPEAT,
			&cam_data->sib.vmalloc_buffers_dma[i]);
#else
    cam_data->sib.vmalloc_buffers_framebuffer[i] =
      usb_alloc_coherent (cam_data->udev,
			  cam_data->sib.vmalloc_buffers_urbs_framebuffer_size,
			  GFP_KERNEL|__GFP_COLD|__GFP_REPEAT,
			  &cam_data->sib.vmalloc_buffers_dma[i]);
#endif
    if (cam_data->sib.vmalloc_buffers_framebuffer[i] == 0) {;
      break;
    };
    cam_data->sib.vmalloc_buffers_urb[i] = usb_alloc_urb(0,GFP_KERNEL);
    if (cam_data->sib.vmalloc_buffers_urb[i] == 0) {;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 35)
      usb_buffer_free(cam_data->udev,cam_data->sib.vmalloc_buffers_urbs_framebuffer_size, cam_data->sib.vmalloc_buffers_framebuffer[i], cam_data->sib.vmalloc_buffers_dma[i]);
#else
      usb_free_coherent(cam_data->udev,cam_data->sib.vmalloc_buffers_urbs_framebuffer_size, cam_data->sib.vmalloc_buffers_framebuffer[i], cam_data->sib.vmalloc_buffers_dma[i]);
#endif
      cam_data->sib.vmalloc_buffers_framebuffer[i] = 0;
      break;
    }
    cam_data->sib.vmalloc_buffers_context[i] = cam_data;
    init_completion(&cam_data->sib.vmalloc_buffers_done[i]);
    usb_fill_bulk_urb(cam_data->sib.vmalloc_buffers_urb[i], 
		      cam_data->udev, 
		      cam_data->pipe_image, 
		      cam_data->sib.vmalloc_buffers_framebuffer[i], 
		      cam_data->sib.vmalloc_buffers_urbs_framebuffer_size, 
		      sentech_usb_indirect_complete,      
		      &(cam_data->sib.vmalloc_buffers_context[i]) 
		      );
    cam_data->sib.vmalloc_buffers_urb[i]->transfer_dma = cam_data->sib.vmalloc_buffers_dma[i];
    cam_data->sib.vmalloc_buffers_urb[i]->transfer_flags = URB_NO_TRANSFER_DMA_MAP;
    cam_data->sib.vmalloc_buffers_urbs_nb++;
  }
;
  if (cam_data->sib.vmalloc_buffers_urbs_nb == 0) {
    for (i = 0; i < cam_data->sib.vmalloc_buffers_nb; i++) {;
      if (cam_data->sib.vmalloc_buffers[i] != 0)
	kfree (cam_data->sib.vmalloc_buffers[i]);
      cam_data->sib.vmalloc_buffers[i] = 0;
    }
    cam_data->sib.vmalloc_buffers_urbs_nb = 0;
    return ENOMEM;
  };
  return 0;
}
static int
sentech_alloc_subs(struct sentech_camera* cam_data) {
  unsigned long flags,k,retval = 0;
  int min_bufs = 1;
  if (cam_data->allocated_buffers) {;
    return 0; 
  }
  if (cam_data->max_frame_size) {
    cam_data->framebuffer_size = cam_data->max_frame_size + 
      sizeof(struct Sentech_Header);;
    if(!param_use_submitted)
      min_bufs = SUBS_SUBMITTED+1;
    else 
      min_bufs = 1;
    for (k = 0; k < cam_data->requested_buffers; k ++) {
      struct sentech_usb_buffer* sub;
;
      sub = alloc_sub(cam_data);
;
      if (!sub) {
	if ((k > 0) && (DYNAMIC_USE_SUBMITTED || DYNAMIC_SUB_SUBMITTED)) {
	  if (DYNAMIC_SUB_SUBMITTED) {
	    SUBS_SUBMITTED = f_min(f_max(1,k-1), SUBS_SUBMITTED);;
	  } else
	    SUBS_SUBMITTED = f_min(k, SUBS_SUBMITTED);
	  min_bufs = SUBS_SUBMITTED+1;;
	  if (DYNAMIC_USE_SUBMITTED && (k < min_bufs)) {
	    param_use_submitted = 1;
	    min_bufs = 1;;
	  }
	}
;
	if (k < min_bufs) {
	  printk(KERN_ALERT "SentechUSB: Error allocating buffer %lu. Try fragmentation!\n", k);
	  read_lock_irqsave(&cam_data->buffers_lock, flags);
	  while(!list_empty(&cam_data->buffers)){
	    read_unlock_irqrestore(&cam_data->buffers_lock, flags);
	    free_first_sub(cam_data);
	    read_lock_irqsave(&cam_data->buffers_lock, flags);
	  }
	  read_unlock_irqrestore(&cam_data->buffers_lock, flags);
	  retval = -ENOMEM;
	  goto done;
	}
	printk(KERN_ALERT "SentechUSB: Error allocating buffer %lu\n", k);
	cam_data->allocated_buffers = k;
	goto done;
      } else {;
	sub->id = k;
	write_lock_irqsave(&cam_data->buffers_lock, flags);
	list_add_tail(&sub->sub_buffers, &cam_data->buffers);
	write_unlock_irqrestore(&cam_data->buffers_lock, flags);;
      }
    }
  }
  cam_data->allocated_buffers = cam_data->requested_buffers;
  if (idle_buffers) 
    cam_data->idle_frame = alloc_sub(cam_data);
 done:
  return retval;
}
static void sentech_free_subs(struct sentech_camera* cam_data){;  
  while (!list_empty(&cam_data->buffers))
    free_first_sub(cam_data);;
}
static
void freeBufs_indirect(struct sentech_camera* cam_data){
  int index;
  for(index = 0; index < cam_data->sib.vmalloc_buffers_nb; index ++) {
    if (cam_data->sib.vmalloc_buffers_user[index]) {
        cam_data->sib.vmalloc_buffers_user[index] = 0;
	cam_data->sib.sib_frame[index].data = 0;
	cam_data->sib.sib_frame[index].done = 0;
    }
  }
}
static
void freeBufs(struct sentech_camera* cam_data){
  struct sentech_usb_buffer* sub;
  unsigned long flags;
  read_lock_irqsave(&cam_data->buffers_lock, flags);
  list_for_each_entry(sub, &cam_data->buffers, sub_buffers){
    if(sub->user_owned){
      sub->user_owned=0;
    }
  }
  read_unlock_irqrestore(&cam_data->buffers_lock, flags);
}
#define DPKLEV 1 
static void dprintksubs(struct sentech_camera* cam_data){
  struct sentech_usb_buffer* sub;
  struct sentech_usb_fragment* suf;
  unsigned long flags;;
  read_lock_irqsave(&cam_data->buffers_lock, flags);
  list_for_each_entry(sub, &cam_data->buffers, sub_buffers) {;
    list_for_each_entry(suf, &sub->suf_buffers, buffers) {;
    }
  }
  read_unlock_irqrestore(&cam_data->buffers_lock, flags);
;
}
#define TO_Y(R,G,B) ((9798*R+19234*G+3736*B)>>15)
inline int expavg (int new, int old, int alpha, int alpha_shift) {
  return (new*alpha+old*((1<<alpha_shift)-alpha)) >> alpha_shift;
}
inline
int get_min_max(struct sentech_camera* cam_data,
		__u8* src, int src_size,
		__u8 *minY, 
		__u8 *maxY, __u16*maxY_idx) {
  __u16 maxw = cam_data->max_width-1, 
    maxh = cam_data->scan_height-1;
  __u16 maxw1 = maxw = cam_data->max_width + 1;
  __u16 incw = 8, inch = 4; 
  int incptr = inch * cam_data->max_width;
  __u8* ptr = src + incptr;
  __u16 rows = inch;
  register __u16 w;
  register __u8 Y;;
  if(maxh*cam_data->max_width > src_size)
    maxh = src_size/cam_data->max_width;
  *minY = *maxY = TO_Y(ptr[2],ptr[3],ptr[2+maxw1]);
  *maxY_idx = incptr+2;
  for (;rows<maxh;rows+=inch) {
    for (w=incw;w<maxw; w+=incw) {
      Y = TO_Y(ptr[w],ptr[w+1],
	       ptr[w+maxw1]); 
      if(unlikely(Y < *minY)) *minY=Y;
      else
	if(unlikely(Y > *maxY)){
	  *maxY=Y;
	  *maxY_idx = w+ ptr - src;
	}
    }
    ptr += incptr;
  }
  return *maxY;
}
static int
update_controls(struct sentech_camera* cam_data){
  unsigned long flags;
  int max_exp, exposure;
  unsigned int i;
  __u16 gain =  __le16_to_cpu(cam_data->payload21.Gain_le);
  exposure = get_crt_exposure(cam_data);
  if(!dynamic_scale) {
    max_exp = get_max_exposure(cam_data);
    exposure = (MAX_EXP_CTRL*exposure) / max_exp;
  }
  for(i = 0; i < ARRAY_SIZE(sentech_qctrl); i++) {
    switch (sentech_qctrl[i].id) {
    case V4L2_CID_CONTRAST:
      {cam_data->qctl_regs[i] = exposure; break;}
    case V4L2_CID_EXPOSURE:
      {cam_data->qctl_regs[i] = exposure; break;}
    case V4L2_CID_STC_EXPOSURE_US:
      {cam_data->qctl_regs[i] = exposure_2_useconds(cam_data, exposure); break;}
    case V4L2_CID_GAIN:
      {cam_data->qctl_regs[i] = gain;;
	break;}
    case (V4L2_CID_STC_RED): 
    case V4L2_CID_RED_BALANCE:
      {cam_data->qctl_regs[i] = cam_data->red; break;}
    case (V4L2_CID_STC_GREEN): 
      {cam_data->qctl_regs[i] = cam_data->green; break;}
    case (V4L2_CID_STC_BLUE): 
    case V4L2_CID_BLUE_BALANCE:
      {cam_data->qctl_regs[i] = cam_data->blue; break;}
    }
    continue;
  }
  write_lock_irqsave(&cam_data->ctrl_lock, flags);
  set_controls(cam_data, GFP_ATOMIC);
  write_unlock_irqrestore(&cam_data->ctrl_lock, flags);
  cam_data->recent_autobalance = 1;
  return 0;
}
static 
int sentech_submit_kurbs(struct sentech_camera* cam_data, int mem_flags);
long long int 
sentech_compute_saturated_pixels16(struct sentech_camera* cam_data, __u16*_b,
				     int*_saturated, int*_lighted, int*_base,
				   __u16 lower_limit, __u16 upper_limit, int shift) {
  int start_x=cam_data->crop.left,start_y=cam_data->crop.top,
    max_x= cam_data->crop.left+cam_data->crop.width, 
    max_y=cam_data->crop.top+cam_data->crop.height, width_x=cam_data->max_width;
  int step_x = param_saturation_step;
  int step_y = param_saturation_step;
  __u16 * b=_b;
  __u16 * ptr=b+start_y*width_x;
  int saturated=0;
  int lighted=0;
  int x, y;
  long long int suma=0;
  for (y=start_y; y<max_y; y+=step_y) {
    for (x=start_x; x<max_x; x+=step_x) {
      register int pix = ptr[x]>>shift;
      if(pix >= upper_limit) saturated++;
      if(pix >= lower_limit) lighted++;
      suma += pix;  
    }
    ptr+=width_x*step_y;
  }
  *_saturated = saturated;
  *_lighted = lighted;
  *_base = ((max_y - start_y)/step_x) * ((max_x - start_x)/step_y);
  return suma;
}
long long int 
sentech_compute_saturated_pixels(struct sentech_camera* cam_data, __u8*_b,
				     int*_saturated, int*_lighted, int*_base,
				     __u8 lower_limit, __u8 upper_limit) {
  int start_x=cam_data->crop.left,start_y=cam_data->crop.top,
    max_x= cam_data->crop.left+cam_data->crop.width, 
    max_y=cam_data->crop.top+cam_data->crop.height, width_x=cam_data->max_width;
  int step_x = param_saturation_step;
  int step_y = param_saturation_step;
  __u8 * b=_b;
  __u8 * ptr=b+start_y*width_x;
  int saturated=0;
  int lighted=0;
  int x, y;
  long long int suma=0;
  for (y=start_y; y<max_y; y+=step_y) {
    for (x=start_x; x<max_x; x+=step_x) {
      register int pix = ptr[x];
      if(pix >= upper_limit) saturated++;
      if(pix >= lower_limit) lighted++;
      suma += pix;  
    }
    ptr+=width_x*step_y;
  }
  *_saturated = saturated;
  *_lighted = lighted;
  *_base = ((max_y - start_y)/step_x) * ((max_x - start_x)/step_y);
  return suma;
}
int adjust_inc(struct sentech_camera* cam_data, int range, int min_r, int max_r, int _m_inc, int amortiz, int dir){
  int limit_range = max((max_r-min_r+1)/2, 
			((cam_data->last_sat_dir == dir)?cam_data->last_sat_inc:0));
  int inc = min(range/_m_inc, limit_range);
  if(inc==0) inc = range/_m_inc;
  if(amortiz>1) inc=inc/amortiz;
  if(inc==0) inc = 1;
  inc=min(inc,(cam_data->last_sat_inc+1)*param_sat_acc/((amortiz>1)?2:1));
  cam_data->last_sat_inc = inc;
  cam_data->last_sat_dir = dir;
  return inc;
}
void sentech_reduce_exposure(struct sentech_camera* cam_data, int delta) {
  __s32 exposure = get_crt_exposure(cam_data);
  __s32 max_exposure = get_max_exposure(cam_data);
  __s16 gain = get_crt_gain(cam_data);
  __s16 ini_gain = get_initial_gain(cam_data);
  int amortization = 1;
  cam_data->max_sat_exp = exposure;
  cam_data->max_sat_gain = gain;
  if((exposure<=cam_data->min_sat_exp)&&(gain<=cam_data->min_sat_gain)){
    cam_data->sat_unchanged=cam_data->psat_amortiz_length;
    cam_data->min_sat_exp = 0;
    cam_data->min_sat_gain = 0;
  }
  if(cam_data->sat_unchanged>=1) amortization=cam_data->psat_amortiz;
;
  if (gain>ini_gain) {
    int inc = adjust_inc(cam_data, 255-ini_gain, cam_data->min_sat_gain, cam_data->max_sat_gain,
			 param_saturation_gain_inc, amortization, 0);
    gain = gain - inc;
    gain = max(gain, ini_gain);
    set_crt_gain(cam_data, gain);    
    return;
  }
  if (exposure>1){
    int inc = adjust_inc(cam_data, max_exposure,
			 cam_data->min_sat_exp, cam_data->max_sat_exp,
			 param_saturation_exp_inc, amortization, 0);
    if(gain<ini_gain) set_crt_gain(cam_data, ini_gain);
    exposure -= inc;
    exposure = max(exposure, 0);
    set_crt_exposure(cam_data, exposure);
    return;
  }
  if(gain>0) {
    int inc = adjust_inc(cam_data, ini_gain, cam_data->min_sat_gain, cam_data->max_sat_gain,
			 param_saturation_gain_inc, amortization, 0);
    gain = gain - inc;
    gain = max(gain, (__s16)0);
    set_crt_gain(cam_data, gain);
  }
}
void sentech_increase_exposure(struct sentech_camera* cam_data, int delta) {
  __s32 exposure = get_crt_exposure(cam_data);
  __s32 max_exposure = get_max_exposure(cam_data);
  __s16 gain = get_crt_gain(cam_data);
  __s16 ini_gain = get_initial_gain(cam_data);
  int amortization = 1;
  cam_data->min_sat_exp = exposure;
  cam_data->min_sat_gain = gain;
  if((exposure>=cam_data->max_sat_exp)&&(gain>=cam_data->max_sat_gain)){
    cam_data->sat_unchanged=cam_data->psat_amortiz_length;
    cam_data->max_sat_exp = max_exposure;
    cam_data->max_sat_gain = 255;
  }
  if(cam_data->sat_unchanged>=1) amortization=cam_data->psat_amortiz;
;
  if (gain<ini_gain) {
    int inc = adjust_inc(cam_data, ini_gain, cam_data->min_sat_gain, cam_data->max_sat_gain,
			 param_saturation_gain_inc, amortization, 1);
    gain = gain + inc;
    gain = min(gain, ini_gain);
    set_crt_gain(cam_data, gain);    
    return;
  }
  if (exposure<max_exposure){
    int inc = adjust_inc(cam_data, max_exposure,
			 cam_data->min_sat_exp, cam_data->max_sat_exp,
			 param_saturation_exp_inc, amortization, 1);
    if(gain>ini_gain) set_crt_gain(cam_data, ini_gain);
    exposure += inc;
    exposure = min(exposure, max_exposure);
    set_crt_exposure(cam_data, exposure);
    return;
  }
  if(gain<255) {
    int inc = adjust_inc(cam_data, 255-ini_gain, cam_data->min_sat_gain, cam_data->max_sat_gain,
			 param_saturation_gain_inc, amortization, 1);
    gain = gain + inc;
    gain = min(gain, (__s16)255);
    set_crt_gain(cam_data, gain);
  }
}
BYTE sentech_get_exp_target(struct sentech_camera* cam_data) {
  return cam_data->exposure_target;
}
void sentech_saturated_pixels(struct sentech_camera* cam_data, __u8*_b, char*_s) {
  int saturated=0;
  int lighted=0;
  int base=0;
  long long int suma, tgt, exp_tgt;
  int psat_skip = cam_data->psat_skip;
  __u16 psat_min_16=cam_data->psat_min, psat_max_16=cam_data->psat_max;
  int shift = cam_data->bpp-8;
  cam_data->psat_calls++; 
  if((psat_skip!=0)&&((cam_data->psat_calls % psat_skip) != 0)) return;;
  if(!cam_data->psat_auto) goto end;
  if(cam_data->Bpp==2) {
    suma = sentech_compute_saturated_pixels16(cam_data, (__u16*)_b, &saturated, &lighted, &base, psat_min_16, psat_max_16, shift);
  }else{
    suma = sentech_compute_saturated_pixels(cam_data, _b, &saturated, &lighted, &base, cam_data->psat_min, cam_data->psat_max);
  }
  tgt = base;
  exp_tgt = sentech_get_exp_target(cam_data);
  tgt *= exp_tgt;;
  if(((cam_data->psat_up==0) && (saturated > 0)) ||
     ((cam_data->psat_up!=0) && (saturated > base/cam_data->psat_up)) ||
      (suma > tgt)) {
    int var = saturated - base/((cam_data->psat_up)?cam_data->psat_up:-1);
    sentech_reduce_exposure(cam_data, var);
    if(cam_data->sat_unchanged > 0)cam_data->sat_unchanged--;
  }else
    if(((cam_data->psat_lo !=0 ) && (lighted < base/cam_data->psat_lo))||
       ((cam_data->psat_lo ==0 ) && (lighted < base))) {
      int var = base/(cam_data->psat_lo?cam_data->psat_lo:1) - lighted;
    sentech_increase_exposure(cam_data, var);
    if(cam_data->sat_unchanged > 0)cam_data->sat_unchanged--;
  }else{
    cam_data->sat_unchanged = cam_data->psat_amortiz_length;;
  }
 end:;
}
int sentech_thread_auto(void *data){
  int src_size;
  __u8* src;
  struct sentech_usb_buffer* sub;
  struct sentech_camera* cam_data =
    (struct sentech_camera*) data;
  struct sentech_usb_fragment *suf;
  int retval=0;
#ifndef SENTECH_AWB
  __u8 minY, maxY, maxY_R, maxY_G, maxY_B;
  __u16 maxY_idx;
  __u32 maxY_C, newR=1, newG=1, newB=1;
  int max_exp;
  int i;
#else
  int awb_balance=0;
#endif
  DEFINE_WAIT(wait);
;
  cam_data->kthread_running = 1;
  if (param_have_buffers || (cam_data->sib.vmalloc_buffers_nb <= 0)) {
    if ((param_buffers <= 1) || (cam_data->allocated_buffers <= 1)) {
      sub = get_sub(cam_data, 0);
      if (!sub) {;
	retval = -EINVAL;
	goto done;
      }
    } else {
      sub = get_sub(cam_data, 1);
      if (!sub) {;
	retval = -EINVAL;
	goto done;
      }
    }
    if (cam_data->fragmented) {;
      retval = -EINVAL;
      goto done;
    }
    suf = first_suf(sub);
    src_size = suf->fragment_size;
    src = suf->framebuffer_image;
  } else {
    src_size = cam_data->sib.vmalloc_buffers_size;
    src = cam_data->sib.vmalloc_buffers[0];
  }
  if (!cam_data->freq_auto) {;
    cam_data->freq_auto = 1;
  }
  for(;;) {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,7)
    if (my_kthread_should_stop()) break;
#else
    if (kthread_should_stop()) break;
#endif
    prepare_to_wait(&cam_data->autoq,
		    &wait,TASK_INTERRUPTIBLE);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,7)
    if (likely(!my_kthread_should_stop()))
#else
    if (likely(!kthread_should_stop()))
#endif
    {
      if((cam_data->run && !cam_data->status) || (cam_data->state80.Trigger_Mode & STC_FPGA_TRIGGER_MODE_MASK))
	schedule_timeout (HZ/cam_data->freq_auto);
      else
	schedule_timeout (HZ);
    }
    finish_wait(&cam_data->autoq, &wait);
    if (!cam_data->run
	){
      if(cam_data->shutdown) {
	retval = -ENODEV;
	goto done;
      }
      cam_data->status=0;
      continue;
    }
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,7)
    if (my_kthread_should_stop()) break;
#else
    if (kthread_should_stop()) break;
#endif
    if(cam_data->run && cam_data->status && !atomic_read(&cam_data->running_buffers)) {
      cam_data->status =
	usb_clear_halt(cam_data->udev, cam_data->pipe_image);      
      if(cam_data->shutdown) {retval = -ENODEV; goto done;}
      cam_data->status = 0;
      sentech_submit_kurbs(cam_data, GFP_KERNEL);
    }
    if(!cam_data->auto_gain) sentech_saturated_pixels(cam_data, src, "thread_auto");
    if (cam_data->auto_white_balance||
	cam_data->do_white_balance||
	cam_data->auto_gain||
	cam_data->psat_auto) {
#ifdef SENTECH_AWB
      lprintk(2,"Locking thread\n");
      down(&cam_data->lock);
      if (cam_data->height_changed) {;
	Sentech_FreeAWBObject(cam_data->awb_object);
;
	Sentech_InitAWBObject(cam_data->awb_object,
			      cam_data->awb_buffer,
			      &cam_data->state12.line_numbers_le,
			      &cam_data->state12.one_line_clock_numbers_le,
			      &cam_data->payload21.Shutter_line_numbers_le,
			      &cam_data->payload21.Shutter_clock_numbers_le,
			      &cam_data->payload21.Gain_le,
			      &cam_data->max_width,
			      &cam_data->scan_height,
			      &cam_data->modif_flag21,
			      &cam_data->red,
			      &cam_data->green,
			      &cam_data->blue,
			      &cam_data->state23.R_Gain_le,
			      __le16_to_cpu(cam_data->state12.initial_gain_le),
			      __le16_to_cpu(cam_data->state12.initial_digital_gain_le),
			      cam_data->trigger_version 
#ifdef L_BOUNDS
			      ,&cam_data->gain_min
			      ,&cam_data->gain_max
			      ,&cam_data->exposure_min
			      ,&cam_data->exposure_max
#endif
			      );
	if(cam_data->auto_white_balance)
	  Sentech_SetAutoAWB(cam_data->awb_object, STCAM_WB_FULLAUTO);
	if(cam_data->auto_gain) {
	  Sentech_SetAutoGain(cam_data->awb_object,
			      (STCAM_AGC_ON|STCAM_AUTO_SHUTTER_ON));
	  Sentech_SetAutoExposure(cam_data->awb_object,
				  (STCAM_AGC_ON|STCAM_AUTO_SHUTTER_ON));
	}
	{
	  BYTE target[]= {0,5,5};
	  target[0] = sentech_get_exp_target(cam_data);
	  Sentech_CLuminance_SetTarget(cam_data->awb_object,
				       target);
	};
	cam_data->height_changed = 0;
      }else{;
      }
      if(!cam_data->awb_buffer) {
	retval = -ENOMEM;
	up(&cam_data->lock);
	lprintk(2,"Unlocked thread\n");
	goto done;
      }
      if(!cam_data->awb_object || !cam_data->awb_buffer) {
	retval = -ENOMEM;
	up(&cam_data->lock);
	lprintk(2,"Unlocked thread\n");
	goto done;
      }
      if(!cam_data->recent_autobalance) { 
#ifdef ST_ImageCopyBalance
	awb_balance = Sentech_ImageCopyBalance(cam_data->awb_object, src);
#else
	int frag_size = cam_data->scanimage_size;;
	if (frag_size > src_size) frag_size = src_size;
	memcpy(cam_data->awb_buffer, src, frag_size);
	awb_balance = Sentech_ImageBalance(cam_data->awb_object);
#endif
	;;
      }
      up(&cam_data->lock);
      lprintk(2,"Unlocked thread\n");
      update_controls(cam_data);
      cam_data->recent_autobalance = 0; 
#else 
      get_min_max(cam_data, src, src_size,
		  &minY, &maxY, &maxY_idx);
      if (cam_data->auto_white_balance ||
	  cam_data->do_white_balance) {
	maxY_R = src[maxY_idx];
	maxY_G = src[maxY_idx+1];
	maxY_B = src[maxY_idx+1+
					cam_data->max_width];
	maxY_C = max(max(maxY_R, maxY_G),
		     max(maxY_B, maxY_G));
	if(maxY_R){
	  newR = (maxY_C<<COLOR_SHIFT) / maxY_R;
	  cam_data->red =  expavg(newR, cam_data->red, param_freq_auto_exp_avg,
				  param_max_exp_avg_shift);
	}
	if(maxY_G){
	  newG = (maxY_C<<COLOR_SHIFT) / maxY_G;
	  cam_data->green =  expavg(newG, cam_data->green, param_freq_auto_exp_avg,
				  param_max_exp_avg_shift);
	}
	if(maxY_B){
	  newB = (maxY_C<<COLOR_SHIFT) / maxY_B;
	  cam_data->blue =   expavg(newB, cam_data->blue, param_freq_auto_exp_avg,
				  param_max_exp_avg_shift);
	};
	if(cam_data->do_white_balance) {
	  cam_data->do_white_balance = 0;
	  for(i = 0; i < ARRAY_SIZE(sentech_qctrl); i++) {
	    if (sentech_qctrl[i].id == V4L2_CID_RED_BALANCE)
	      cam_data->qctl_regs[i] = cam_data->red;
	    if (sentech_qctrl[i].id == V4L2_CID_STC_RED)
	      cam_data->qctl_regs[i] = cam_data->red;
	    if (sentech_qctrl[i].id == V4L2_CID_BLUE_BALANCE)
	      cam_data->qctl_regs[i] = cam_data->blue;
	    if (sentech_qctrl[i].id == V4L2_CID_STC_BLUE)
	      cam_data->qctl_regs[i] = cam_data->blue;
	    if (sentech_qctrl[i].id == V4L2_CID_STC_GREEN)
	      cam_data->qctl_regs[i] = cam_data->green;
	  }
	}
      }
      if (cam_data->auto_gain) {
	if(minY>5) {
	  max_exp = get_max_exposure(cam_data);
	  if(cam_data->shutters < max_exp){
	    cam_data->shutters++;
	    set_crt_exposure(cam_data, cam_data->shutters);
	    setFlag(&cam_data->modif_flag21);
	  } else
	    if(cam_data->gain < MAXGAIN){
	      cam_data->gain--;
	      cam_data->payload21.Gain_le = __cpu_to_le16(cam_data->gain);
	      setFlag(&cam_data->modif_flag21);
	    }
	}
	if(maxY<250) {
	  max_exp = get_max_exposure(cam_data);
	  if(cam_data->shutters < max_exp){
	    cam_data->shutters--;
	    set_crt_exposure(cam_data, cam_data->shutters);
	    setFlag(&cam_data->modif_flag21);
	  } else
	    if(cam_data->gain < MAXGAIN){
	      cam_data->gain++;
	      cam_data->payload21.Gain_le = __cpu_to_le16(cam_data->gain);
	      setFlag(&cam_data->modif_flag21);
	    }
	}
      }
      update_controls(cam_data);
#endif  
    }else{
      update_controls(cam_data);
    }
  }
 done:
#ifdef SENTECH_AWB
  if(cam_data->awb_object) {;
    Sentech_FreeAWBObject(cam_data->awb_object);
    kfree(cam_data->awb_object);
    cam_data->awb_object = NULL;
  }
#endif
  cam_data->kthread_running = 0;
  mb();
  cam_data->kthread_ctrl = NULL;
  mb();;
  return retval;
}
void start_qthread(struct sentech_camera* cam_data) {;
  if(cam_data->kthread_ctrl){;
    return;
  }
  cam_data->kthread_running=1;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,7)
  cam_data->kthread_ctrl = my_kthread_run(sentech_thread_auto, cam_data,
		"sentech_auto_white_balance_gain");
#else
  cam_data->kthread_ctrl =
    kthread_run(sentech_thread_auto, cam_data,
		"sentech_auto_white_balance_gain");
#endif
  if (cam_data->kthread_ctrl == NULL) {;
    cam_data->kthread_running=0;
  };
}
void stop_qthread(struct sentech_camera* cam_data) {;
  if(cam_data->kthread_ctrl){;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,7)
   my_kthread_stop(cam_data->kthread_ctrl);
#else
    kthread_stop(cam_data->kthread_ctrl);
#endif
    wake_up_interruptible(&cam_data->autoq);
  }
;
}
static int usb_start_urb_async(struct urb *urb,
			      int mem_flag)
{ 
	int			status;
	urb->actual_length = 0;
	status = usb_submit_urb(urb, mem_flag);
	return status;
}
static int usb_internal_control_msg_async(struct usb_device *usb_dev,
				    unsigned int pipe, 
				    struct usb_ctrlrequest *cmd,
				    void *data, int len,
			 	    usb_complete_t complete, int mem_flag, void* context)
{
	struct urb *urb;
	urb = usb_alloc_urb(0, mem_flag);
	if (!urb)
		return -ENOMEM;
	usb_fill_control_urb(urb, usb_dev, pipe, (unsigned char *)cmd, data,
			     len, complete, context);
	return usb_start_urb_async(urb, mem_flag);
}
static int 
usb_control_msg_async(struct usb_device *dev, unsigned int pipe,
		      __u8 request,
		      __u8 requesttype,
		      __u16 value, __u16 index, void *data,
		      __u16 size, 
		      usb_complete_t complete, int mem_flag,
		      void*packet_buffer, void*context)
{
  struct usb_ctrlrequest *dr = packet_buffer ;
  int ret;
  if (!dr)
	return -ENOMEM;
  dr->bRequestType= requesttype;
  dr->bRequest = request;
  dr->wValue = cpu_to_le16p(&value);
  dr->wIndex = cpu_to_le16p(&index);
  dr->wLength = cpu_to_le16p(&size);
  ret = usb_internal_control_msg_async(dev, pipe, dr, data, size, complete, mem_flag, context);
  return ret;
}
struct ctrl_context{
  struct usb_ctrlrequest pb;
  struct sentech_camera* cam_data;
  struct completion done;
  int ID; 
  struct list_head queue;
  int end;
  int fpga_cmd;
  int fpga_len;
};
struct ctrl_fpga2_context{
  struct usb_ctrlrequest pb;
  struct sentech_camera* cam_data;
  struct completion done;
  int ID; 
  struct list_head queue;
  int end;
  __u32 fpga_cat;
  __u32 fpga_ins_cat;
  __u32 data;
  void* p_data;
};
static void
update_exposure_limits(struct sentech_camera* cam_data){
      unsigned int i;
      __s32 max_exp, max_exp_us;
      max_exp = get_max_exposure(cam_data);
      max_exp_us = exposure_2_useconds(cam_data, max_exp);
      for (i = 0; i < ARRAY_SIZE(sentech_qctrl); i++) {
	if ((V4L2_CID_EXPOSURE == sentech_qctrl[i].id) ||
	    ( V4L2_CID_CONTRAST == sentech_qctrl[i].id)) {
	  cam_data->qctrl[i].maximum = max_exp;
	  cam_data->qctrl[i].default_value = cam_data->qctrl[i].maximum;
	}
	if (V4L2_CID_STC_EXPOSURE_US == sentech_qctrl[i].id) {
	  cam_data->qctrl[i].maximum = max_exp_us;
	  cam_data->qctrl[i].default_value = cam_data->qctrl[i].maximum;
	}
#ifdef L_BOUNDS_ANY
	sentech_exposure_bounds(cam_data, i, max_exp);
#endif
      }
}
static void sentech_complete_get_clock(struct urb * urb ARG_COMPLETE){
  struct ctrl_context* ctx = (struct ctrl_context*)urb->context;
  struct sentech_camera* cam_data = ctx->cam_data;;
  if(urb->status<0);
  else{;
    if(dynamic_scale){
      update_exposure_limits(cam_data);
    }
    cam_data->fps_idx = __le32_to_cpu(cam_data->state22.Clock_mode_le);;
  }
  usb_free_urb(urb);
  complete(&ctx->done);
  ctx->end = 1;
}
static void sentech_complete_get_cam_data(struct urb * urb ARG_COMPLETE){
  struct ctrl_context* ctx = (struct ctrl_context*)urb->context;
  struct sentech_camera* cam_data = ctx->cam_data;;
  if(urb->status<0);
  else{;
    if(dynamic_scale){
      unsigned int i;
      __s32 max_exp, max_exp_us;
      max_exp = get_max_exposure(cam_data);
      max_exp_us = exposure_2_useconds(cam_data, max_exp);
      for (i = 0; i < ARRAY_SIZE(sentech_qctrl); i++) {
	if ((V4L2_CID_EXPOSURE == sentech_qctrl[i].id) ||
	    ( V4L2_CID_CONTRAST == sentech_qctrl[i].id)) {
	  cam_data->qctrl[i].maximum = max_exp;
	}
	if (V4L2_CID_STC_EXPOSURE_US == sentech_qctrl[i].id) {
	  cam_data->qctrl[i].maximum = max_exp_us;
	  cam_data->qctrl[i].default_value = cam_data->qctrl[i].maximum;
	}
#ifdef L_BOUNDS_ANY
	sentech_exposure_bounds(cam_data, i, max_exp);
#endif
      }
    }
  }
  usb_free_urb(urb);
  complete(&ctx->done);
  ctx->end = 1;
}
static void sentech_update_view(struct sentech_camera* cam_data){;
  cam_data->defrect.width = cam_data->scan_width;
  cam_data->output.width = cam_data->scan_width;
  cam_data->crop.width = cam_data->scan_width;
  cam_data->bounds.width = cam_data->scan_width;
  cam_data->output.left = 0;
  cam_data->crop.left = 0;
  cam_data->bounds.left = 0;
}
static void sentech_complete_get_pict_size(struct urb * urb ARG_COMPLETE){
  int retval;
  struct ctrl_context* ctx = (struct ctrl_context*)urb->context;
  struct sentech_camera* cam_data = ctx->cam_data;
  int old_scan_height = cam_data->scan_height;;
  if(urb->status<0);
  else{;
    if(cam_data->usb3) {
      __u32 old_scan_width = cam_data->scan_width;
      cam_data->scan_vertical_offset =
	__le32_to_cpu(cam_data->state12.Vertical_offset_le) & ~1;
      cam_data->scan_width = __le32_to_cpu(cam_data->state12.Horizontal_size_mode_le);
      cam_data->scan_horizontal_offset =
	__le32_to_cpu(cam_data->state12.Horizontal_offset_le) & ~1;;
      if (cam_data->scan_width == 0)
	cam_data->scan_width = cam_data->max_width; 
      if (old_scan_width != cam_data->scan_width) sentech_update_view(cam_data);
    }
    cam_data->scanimage_size = cam_data->Bpp*cam_data->scan_width*
      (cam_data->scan_height = 
       __le32_to_cpu(cam_data->state12.Vertical_size_mode_le));
    cam_data->scan_size = cam_data->scanimage_size + sizeof(struct Sentech_Header);;
    mb();
    if(cam_data->scan_height != old_scan_height)
      cam_data->height_changed = 1;
      cam_data->bounds.height = cam_data->scan_height;
      cam_data->bounds.width = cam_data->scan_width;
    if(dynamic_scale){
      unsigned int i;
      __s32 max_exp, max_exp_us;
      max_exp = get_max_exposure(cam_data);
      max_exp_us = exposure_2_useconds(cam_data, max_exp);
      for (i = 0; i < ARRAY_SIZE(sentech_qctrl); i++) {
	if ((V4L2_CID_EXPOSURE == sentech_qctrl[i].id) ||
	    ( V4L2_CID_CONTRAST == sentech_qctrl[i].id)) {
	  cam_data->qctrl[i].maximum = max_exp;
	}
	if (V4L2_CID_STC_EXPOSURE_US == sentech_qctrl[i].id) {
	  cam_data->qctrl[i].maximum = max_exp_us;
	  cam_data->qctrl[i].default_value = cam_data->qctrl[i].maximum;
	}
#ifdef L_BOUNDS_ANY
	sentech_exposure_bounds(cam_data, i, max_exp);
#endif	
      }
    }
  }
  usb_free_urb(urb);
  if((retval=usb_control_msg_async(cam_data->udev, cam_data->pipe_ctl_rcv,
				   0x11,
				   USB_DIR_IN |
				   USB_TYPE_VENDOR|
				   USB_RECIP_DEVICE,
				   0, 0, &cam_data->payload11,
				   sizeof(cam_data->payload11), 
				   sentech_complete_get_cam_data, GFP_ATOMIC,
				   &ctx->pb, ctx))){;
    complete(&ctx->done);
    ctx->end = 1;
  }
}
static void sentech_complete_pict_size(struct urb * urb ARG_COMPLETE){
  int retval;
  struct ctrl_context* ctx = (struct ctrl_context*)urb->context;
  struct sentech_camera* cam_data = ctx->cam_data;;
  if(urb->status<0);
  usb_free_urb(urb);
  if((retval=usb_control_msg_async(cam_data->udev, cam_data->pipe_ctl_rcv,
				   0x12,
				   USB_DIR_IN |
				   USB_TYPE_VENDOR|
				   USB_RECIP_DEVICE,
				   0, 0, &cam_data->state12,
				   sizeof(cam_data->state12), 
				   sentech_complete_get_pict_size, GFP_ATOMIC,
				   &ctx->pb, ctx))){;
    complete(&ctx->done);
    ctx->end = 1;
  }
}
static void sentech_complete_get_bpp(struct urb * urb ARG_COMPLETE){
  struct ctrl_context* ctx = (struct ctrl_context*)urb->context;
  struct sentech_camera* cam_data = ctx->cam_data;
  int i=0;;
  if(urb->status<0);
  usb_free_urb(urb);
  if(param_supports_bpp) {
    cam_data->Bpp = getMaxBpp(__le32_to_cpu(cam_data->state13.Bytes_per_pixel_le));
    cam_data->bpp = getMax_bpp(__le32_to_cpu(cam_data->payload11.OR_Bytes_per_pixel_le));
  }else{
    cam_data->Bpp = 1;
    cam_data->bpp = 8;
  }
  for(i = 0; i < ARRAY_SIZE(sentech_qctrl); i++) {
    switch (sentech_qctrl[i].id) {
    case (V4L2_CID_STC_BYTES):
      cam_data->qctl_regs[i] = cam_data->Bpp;
      break;
    }
  }
  cam_data->scanimage_size =  cam_data->Bpp*cam_data->scan_height*cam_data->scan_width;;
  cam_data->scan_size =  cam_data->scanimage_size+sizeof(struct Sentech_Header);;
  complete(&ctx->done);
  ctx->end = 1;
}
static void sentech_complete_bpp(struct urb * urb ARG_COMPLETE){
  int retval;
  struct ctrl_context* ctx = (struct ctrl_context*)urb->context;
  struct sentech_camera* cam_data = ctx->cam_data;;
  if(urb->status<0);
  usb_free_urb(urb);
  if((retval=usb_control_msg_async(cam_data->udev, cam_data->pipe_ctl_rcv,
				   0x13,
				   USB_DIR_IN |
				   USB_TYPE_VENDOR|
				   USB_RECIP_DEVICE,
				   0, 0, &cam_data->state13,
				   sizeof(cam_data->state13), 
				   sentech_complete_get_bpp, GFP_ATOMIC,
				   &ctx->pb, ctx))){;
    complete(&ctx->done);
    ctx->end = 1;
  }
}
static void sentech_complete_get_shutter_gain(struct urb * urb ARG_COMPLETE){
  struct ctrl_context* ctx = (struct ctrl_context*)urb->context;
  struct sentech_camera* cam_data = ctx->cam_data;
  int i;
  __u16 gain;
  __s32 exposure;;
  if(urb->status<0);
  usb_free_urb(urb);
  complete(&ctx->done);;;
  ctx->end = 1;
  gain =  __le16_to_cpu(cam_data->state21.Gain_le);
  exposure = get_crt_exposure(cam_data);;
  for(i = 0; i < ARRAY_SIZE(sentech_qctrl); i++) {
    switch (sentech_qctrl[i].id) {
    case V4L2_CID_CONTRAST:
      {cam_data->qctl_regs[i] = exposure; break;}
    case V4L2_CID_EXPOSURE:
      {cam_data->qctl_regs[i] = exposure; break;}
    case V4L2_CID_STC_EXPOSURE_US:
      {cam_data->qctl_regs[i] = exposure_2_useconds(cam_data, exposure);
	break;}
    case V4L2_CID_GAIN:
      {cam_data->qctl_regs[i] = gain;;
	break;}
    }
  }  
}
static void sentech_complete_shutter_gain(struct urb * urb ARG_COMPLETE){
  int retval;
  struct ctrl_context* ctx = (struct ctrl_context*)urb->context;
  struct sentech_camera* cam_data = ctx->cam_data;;
  if(urb->status<0);
  usb_free_urb(urb);
  if((retval=usb_control_msg_async(cam_data->udev, cam_data->pipe_ctl_rcv,
				   0x21,
				   USB_DIR_IN |
				   USB_TYPE_VENDOR|
				   USB_RECIP_DEVICE,
				   0, 0, &cam_data->state21,
				   sizeof(cam_data->state21), 
				   sentech_complete_get_shutter_gain, GFP_ATOMIC,
				   &ctx->pb, ctx))){;
    complete(&ctx->done);
    ctx->end = 1;
  }
}
static void sentech_complete_clock(struct urb * urb ARG_COMPLETE){
  int retval;
  struct ctrl_context* ctx = (struct ctrl_context*)urb->context;
  struct sentech_camera* cam_data = ctx->cam_data;;
  if(urb->status<0);
  usb_free_urb(urb);
  if((retval=usb_control_msg_async(cam_data->udev, cam_data->pipe_ctl_rcv,
				   0x22,
				   USB_DIR_IN |
				   USB_TYPE_VENDOR|
				   USB_RECIP_DEVICE,
				   0, 0, &cam_data->state22,
				   sizeof(cam_data->state22), 
				   sentech_complete_get_clock, GFP_ATOMIC,
				   &ctx->pb, ctx))){;
    complete(&ctx->done); 
    ctx->end = 1;
  }
}
static int cID = 0;
static void sentech_submit_21(struct sentech_camera* cam_data, int memflags) {
  struct ctrl_context* ctx;;;
  ctx = kmalloc(sizeof(struct ctrl_context), memflags);
  if(!ctx) return;
  ctx->ID = cID++;
  ctx->end = 0;
  ctx->cam_data = cam_data;
  init_completion(&ctx->done);
  list_add_tail(&ctx->queue, &cam_data->ctrl_queue);
  {
    int max_exp, exposure;
    unsigned int i;
    __u16 gain =  __le16_to_cpu(cam_data->payload21.Gain_le);
    exposure = get_crt_exposure(cam_data);
    if ( ! dynamic_scale ) {
      max_exp = get_max_exposure(cam_data);
      exposure = (MAX_EXP_CTRL*exposure) / max_exp;
    };
    for(i = 0; i < ARRAY_SIZE(sentech_qctrl); i++) {
	switch (sentech_qctrl[i].id) {
	case V4L2_CID_CONTRAST:
	  {cam_data->qctl_regs[i] = exposure; break;}
	case V4L2_CID_EXPOSURE:
	  {cam_data->qctl_regs[i] = exposure; break;}
	case V4L2_CID_STC_EXPOSURE_US:
	  {cam_data->qctl_regs[i] = exposure_2_useconds(cam_data, exposure);
	    break;}
	case V4L2_CID_GAIN:
	  {cam_data->qctl_regs[i] = gain;;
	    break;
	  }
	}
      };
  }
  usb_control_msg_async(cam_data->udev, cam_data->pipe_ctl_snd,
			  0x21,
			  USB_DIR_OUT |
			  USB_TYPE_VENDOR|
			  USB_RECIP_DEVICE,
			  0, 0, &cam_data->payload21,
			  sizeof(cam_data->payload21), 
			  sentech_complete_shutter_gain, memflags,
			  &ctx->pb, ctx);
}
static 
void set_controls(struct sentech_camera* cam_data, int memflags){
  struct ctrl_context* ctx;
  if(!list_empty(&cam_data->ctrl_queue));
  while(!list_empty(&cam_data->ctrl_queue)){
    ctx = list_entry(cam_data->ctrl_queue.next, struct ctrl_context, queue);
    if(ctx->end){
      list_del (&ctx->queue);;
      wait_for_completion_timeout(&ctx->done,HZ);;
      kfree(ctx);
    }else break;
  }
;
  if(testAndClear(&cam_data->modif_flag_80_00)) {
    sentech_io_direction_change(cam_data, memflags);
  }
  if(testAndClear(&cam_data->modif_flag_80_04)) {
    sentech_fpga_set(cam_data, memflags,0x04,4);
  }
  if(testAndClear(&cam_data->modif_flag_80_0F)) {
    sentech_fpga_set(cam_data, memflags, 0x0F, 1);
  }
  if(testAndClear(&cam_data->modif_flag_80_3F)) {
    sentech_fpga_set(cam_data, memflags, 0x3F, 1);
    cam_data->modif_flag_80_3F = 0;
  }
  if(testAndClear(&cam_data->modif_flag_80_45_46)) {
    sentech_fpga_set(cam_data, memflags, 0x45, 4);
    cam_data->modif_flag_80_45_46 = 0;
  }
  if(testAndClear(&cam_data->modif_flag_80_47_48)) {
    sentech_fpga_set(cam_data, memflags, 0x47, 3);
    cam_data->modif_flag_80_47_48 = 0;
  }
  if(testAndClear(&cam_data->modif_flag_80_49)) {
    sentech_fpga_set(cam_data, memflags, 0x49, 2);
    cam_data->modif_flag_80_49 = 0;
  }
  if(testAndClear(&cam_data->modif_flag_80_4A)) {
    sentech_fpga_set(cam_data, memflags, 0x4A, 1);
    cam_data->modif_flag_80_4A = 0;
  }
  if(testAndClear(&cam_data->modif_flag_80_4B)) {
    sentech_fpga_set(cam_data, memflags, 0x4B, 1);
    cam_data->modif_flag_80_4B = 0;
  }
  if(testAndClear(&cam_data->modif_flag_80_4C)) {
    sentech_fpga_set(cam_data, memflags, 0x4C, 1);
    cam_data->modif_flag_80_4C = 0;
  }
  if(testAndClear(&cam_data->modif_flag_80_4D)) {
    sentech_fpga_set(cam_data, memflags, 0x4D, 1);
    cam_data->modif_flag_80_4D = 0;
  }
  if(testAndClear(&cam_data->modif_flag_80_4E)) {
    sentech_fpga_set(cam_data, memflags, 0x4E, 1);
    cam_data->modif_flag_80_4E = 0;
  }
  if(testAndClear(&cam_data->modif_flag_80_14)) {;
    sentech_fpga_set(cam_data, memflags, 0x14, 1);;
  }
  if(testAndClear(&cam_data->modif_flag_80_30)){
    sentech_trigger_mode(cam_data, memflags);
  }
  if(testAndClear(&cam_data->modif_flag12)){;;
    ctx = kmalloc(sizeof(struct ctrl_context), memflags);
    if(!ctx) return;
    ctx->ID = cID++;
    ctx->end = 0;
    ctx->cam_data = cam_data;
    init_completion(&ctx->done);
    list_add_tail(&ctx->queue, &cam_data->ctrl_queue);
    usb_control_msg_async(cam_data->udev, cam_data->pipe_ctl_snd,
			  0x12,
			  USB_DIR_OUT |
			  USB_TYPE_VENDOR|
			  USB_RECIP_DEVICE,
			  0, 0, &cam_data->payload12,
			  sizeof(cam_data->payload12), 
			  sentech_complete_pict_size, memflags,
			  &ctx->pb, ctx);
  }
  if(testAndClear(&cam_data->modif_flag13)){;;
    ctx = kmalloc(sizeof(struct ctrl_context), memflags);
    if(!ctx) return;
    ctx->ID = cID++;
    ctx->end = 0;
    ctx->cam_data = cam_data;
    init_completion(&ctx->done);
    list_add_tail(&ctx->queue, &cam_data->ctrl_queue);
    usb_control_msg_async(cam_data->udev, cam_data->pipe_ctl_snd,
			  0x13,
			  USB_DIR_OUT |
			  USB_TYPE_VENDOR|
			  USB_RECIP_DEVICE,
			  0, 0, &cam_data->payload13,
			  sizeof(cam_data->payload13), 
			  sentech_complete_bpp, memflags,
			  &ctx->pb, ctx);
  }
  if(testAndClear(&cam_data->modif_flag21)){ 
    sentech_submit_21(cam_data, memflags);
  }
  if(testAndClear(&cam_data->modif_flag22)) {;;
    ctx = kmalloc(sizeof(struct ctrl_context), memflags);
    if(!ctx) return;
    ctx->end = 0;
    ctx->ID = cID++;
    init_completion(&ctx->done);
    ctx->cam_data = cam_data;
    list_add_tail(&ctx->queue, &cam_data->ctrl_queue);
    usb_control_msg_async(cam_data->udev, cam_data->pipe_ctl_snd,
			  0x22,
			  USB_DIR_OUT |
			  USB_TYPE_VENDOR|
			  USB_RECIP_DEVICE,
			  0, 0, &cam_data->payload22,
			  sizeof(cam_data->payload22), 
			  sentech_complete_clock, memflags,
			  &ctx->pb, ctx);
  };
}
static
void set_fps(struct sentech_camera* cam_data,
	     __s32 fps_idx){
  unsigned int i;
  unsigned long flags;
  cam_data->payload22.Clock_mode_le = __cpu_to_le32(fps_idx);
  for(i = 0; i < ARRAY_SIZE(sentech_qctrl); i++) {
    if (sentech_qctrl[i].id == V4L2_CID_STC_FPS) {
      cam_data->qctl_regs[i] = 2-fps_idx;
      break;
    }
  }
  setFlag(&cam_data->modif_flag22);;
  write_lock_irqsave(&cam_data->ctrl_lock, flags);
  set_controls(cam_data, GFP_ATOMIC);
  write_unlock_irqrestore(&cam_data->ctrl_lock, flags);
}
static void sentech_complete_trigger(struct urb * urb ARG_COMPLETE){
  struct ctrl_context* ctx = (struct ctrl_context*)urb->context;
  unsigned long flags;
;
  if(urb->status<0);
  usb_free_urb(urb);
  write_lock_irqsave(&ctx->cam_data->ctrl_lock, flags);
  complete(&ctx->done);
  ctx->end = 1;
  write_unlock_irqrestore(&ctx->cam_data->ctrl_lock, flags);;
}
static void unlink_subs(struct sentech_camera* cam_data){
  struct sentech_usb_fragment* suf = NULL;
  struct sentech_usb_buffer *sub = NULL;
  unsigned long flags;
  read_lock_irqsave(&cam_data->buffers_lock, flags);
  list_for_each_entry(sub, &cam_data->buffers, sub_buffers){
    if(!sub->submitted) continue;
    list_for_each_entry(suf, &sub->suf_buffers, buffers) {
      if(!suf->submitted) continue;
      usb_unlink_urb(suf->urb);;
    }
  }
  read_unlock_irqrestore(&cam_data->buffers_lock, flags);
}
static void
sentech_trigger_21(struct sentech_camera* cam_data, int memflags){
  struct ctrl_context* ctx;;;
  ctx = kmalloc(sizeof(struct ctrl_context), memflags);
  if(!ctx) return;
  ctx->ID = cID++;
  ctx->end = 0;
  ctx->cam_data = cam_data;
  init_completion(&ctx->done);
  list_add_tail(&ctx->queue, &cam_data->ctrl_queue);
  usb_control_msg_async(cam_data->udev, cam_data->pipe_ctl_snd,
			0x21,
			USB_DIR_OUT |
			USB_TYPE_VENDOR|
			USB_RECIP_DEVICE,
			0, 0, &cam_data->payload21,
			sizeof(cam_data->payload21), 
			sentech_complete_shutter_gain, memflags,
			&ctx->pb, ctx);
}
static void sentech_complete_trigger_read(struct urb * urb ARG_COMPLETE){
  struct ctrl_context* ctx = (struct ctrl_context*)urb->context;
  struct sentech_camera* cam_data = ctx->cam_data;;
  if(urb->status<0);
  usb_free_urb(urb);
  complete(&ctx->done);
  ctx->end = 1;;
  unlink_subs(cam_data);
  sentech_trigger_21(cam_data, GFP_ATOMIC);
;
}
static void sentech_complete_direction_read(struct urb * urb ARG_COMPLETE){
  struct ctrl_context* ctx = (struct ctrl_context*)urb->context;
;
  if(urb->status<0);
  usb_free_urb(urb);
  complete(&ctx->done);
  ctx->end = 1;;
;
}
static void sentech_complete_trigger_mode(struct urb * urb ARG_COMPLETE){
  int retval;
  struct ctrl_context* ctx = (struct ctrl_context*)urb->context;
  struct sentech_camera* cam_data = ctx->cam_data;;
  if(urb->status<0);;
  usb_free_urb(urb);
  if((retval=usb_control_msg_async(cam_data->udev, cam_data->pipe_ctl_rcv,
				   0x80,
				   USB_DIR_IN |
				   USB_TYPE_VENDOR|
				   USB_RECIP_DEVICE,
				   STC_FPGA_READ|STC_FPGA_TRIGGER_MODE_CMD,
				   0, &cam_data->state80.Trigger_Mode,
				   sizeof(cam_data->state80.Trigger_Mode), 
				   sentech_complete_trigger_read, GFP_ATOMIC,
				   &ctx->pb, ctx))){;
    complete(&ctx->done);
    ctx->end = 1;
  };
}
static void 
sentech_complete_direction_change(struct urb * urb ARG_COMPLETE){
  int retval;
  struct ctrl_context* ctx = (struct ctrl_context*)urb->context;
  struct sentech_camera* cam_data = ctx->cam_data;;
  if(urb->status<0);;
  usb_free_urb(urb);
  if((retval=usb_control_msg_async(cam_data->udev, cam_data->pipe_ctl_rcv,
				   0x80,
				   USB_DIR_IN |
				   USB_TYPE_VENDOR|
				   USB_RECIP_DEVICE,
				   STC_FPGA_READ|STC_FPGA_IO_DIRECTION_CMD,
				   0, &cam_data->state80.IO_Directions,
				   sizeof(cam_data->state80.IO_Directions), 
				   sentech_complete_direction_read, GFP_ATOMIC,
				   &ctx->pb, ctx))){;
    complete(&ctx->done);
    ctx->end = 1;
  };
}
static int 
sentech_trigger(struct sentech_camera* cam_data, int memflags) {
  int retval;
  struct ctrl_context* ctx;
  unsigned long flags;;;
  ctx = kmalloc(sizeof(struct ctrl_context), memflags);
  if(!ctx) return -ENOMEM;
  ctx->ID = cID++;
  ctx->end = 0;
  ctx->cam_data = cam_data;
  init_completion(&ctx->done);
  write_lock_irqsave(&ctx->cam_data->ctrl_lock, flags);
  list_add_tail(&ctx->queue, &cam_data->ctrl_queue);
  write_unlock_irqrestore(&ctx->cam_data->ctrl_lock, flags);
  retval =
    usb_control_msg_async(cam_data->udev, cam_data->pipe_ctl_snd,
			  0x30,
			  USB_DIR_OUT |
			  USB_TYPE_VENDOR|
			  USB_RECIP_DEVICE,
			  0, 0, NULL,
			  0, 
			  sentech_complete_trigger, memflags,
			  &ctx->pb, ctx);;
  return retval;
}
static int 
sentech_trigger_mode(struct sentech_camera* cam_data, int memflags) {
  int retval;
  struct ctrl_context* ctx;;
  ctx = kmalloc(sizeof(struct ctrl_context), memflags);
  if(!ctx) return -ENOMEM;
  ctx->ID = cID++;
  ctx->end = 0;
  ctx->cam_data = cam_data;
  init_completion(&ctx->done);
  list_add_tail(&ctx->queue, &cam_data->ctrl_queue);;
  retval =
    usb_control_msg_async(cam_data->udev, cam_data->pipe_ctl_snd,
			  0x80,
			  USB_DIR_OUT |
			  USB_TYPE_VENDOR|
			  USB_RECIP_DEVICE,
			  STC_FPGA_AUTOREFRESH_WVAL|STC_FPGA_TRIGGER_MODE_CMD, 
			  0, &cam_data->payload80.Trigger_Mode,
			  sizeof(cam_data->payload80.Trigger_Mode), 
			  sentech_complete_trigger_mode, memflags,
			  &ctx->pb, ctx);;
  return retval;
}
static int 
sentech_io_direction_change(struct sentech_camera* cam_data,int memflags){
  int retval;
  struct ctrl_context* ctx;;
  ctx = kmalloc(sizeof(struct ctrl_context), memflags);
  if(!ctx) return -ENOMEM;
  ctx->ID = cID++;
  ctx->end = 0;
  ctx->cam_data = cam_data;
  init_completion(&ctx->done);
  list_add_tail(&ctx->queue, &cam_data->ctrl_queue);;
  retval =
    usb_control_msg_async(cam_data->udev, cam_data->pipe_ctl_snd,
			  0x80,
			  USB_DIR_OUT |
			  USB_TYPE_VENDOR|
			  USB_RECIP_DEVICE,
			  STC_FPGA_AUTOREFRESH_WVAL|STC_FPGA_IO_DIRECTION_CMD, 
			  0, &cam_data->payload80.IO_Directions,
			  sizeof(cam_data->payload80.IO_Directions), 
			  sentech_complete_direction_change, memflags,
			  &ctx->pb, ctx);;
  return retval;
}
static void sentech_complete_fpga_read(struct urb * urb ARG_COMPLETE){
  int k;
  struct ctrl_context* ctx = (struct ctrl_context*)urb->context;;
  if(urb->status<0);
  usb_free_urb(urb);
  complete(&ctx->done);
  ctx->end = 1;;
  for(k=0; k<ctx->fpga_len; k++);
;
}
static void 
sentech_complete_fpga_refresh(struct urb * urb ARG_COMPLETE){
  int retval;
  struct ctrl_context* ctx = (struct ctrl_context*)urb->context;
  struct sentech_camera* cam_data = ctx->cam_data;;
  if(urb->status<0);;
  usb_free_urb(urb);
  if(((__u8*)(((__u16*)&cam_data->state80)+ctx->fpga_cmd)+ctx->fpga_len)-(__u8*)(((__u16*)&cam_data->state80))>
     sizeof(cam_data->state80));
  if((retval=usb_control_msg_async(cam_data->udev, cam_data->pipe_ctl_rcv,
				   0x80,
				   USB_DIR_IN |
				   USB_TYPE_VENDOR|
				   USB_RECIP_DEVICE,
				   STC_FPGA_READ|ctx->fpga_cmd,
				   0, (__u8*)(((__u16*)&cam_data->state80)+ctx->fpga_cmd),
				   ctx->fpga_len,
				   sentech_complete_fpga_read, GFP_ATOMIC,
				   &ctx->pb, ctx))){;
    complete(&ctx->done);
    ctx->end = 1;
  };
}
static void 
sentech_complete_fpga_set(struct urb * urb ARG_COMPLETE){
  int retval;
  struct ctrl_context* ctx = (struct ctrl_context*)urb->context;
  struct sentech_camera* cam_data = ctx->cam_data;;
  if(urb->status<0);;
  usb_free_urb(urb);
  if((retval=usb_control_msg_async(cam_data->udev, cam_data->pipe_ctl_snd,
			0x21,
			USB_DIR_OUT |
			USB_TYPE_VENDOR|
			USB_RECIP_DEVICE,
			0, 0, &cam_data->payload21,
			sizeof(cam_data->payload21), 
			sentech_complete_fpga_refresh, GFP_ATOMIC,
			&ctx->pb, ctx))){;
    complete(&ctx->done);
    ctx->end = 1;
  };
}
static int
sentech_fpga_set(struct sentech_camera* cam_data, int memflags,
		 int cmd, int len){
  return sentech_fpga_set_lock(cam_data, memflags, cmd, len, 0);
}
static int
sentech_fpga_set_lock(struct sentech_camera* cam_data, int memflags,
		      int cmd, int len, int _lockit){
  int retval,k;
  struct ctrl_context* ctx;
  unsigned long flags;;
  ctx = kmalloc(sizeof(struct ctrl_context), memflags);
  if (!ctx) return -ENOMEM;
  ctx->ID = cID++;
  ctx->end = 0;
  ctx->cam_data = cam_data;
  ctx->fpga_cmd=cmd;
  ctx->fpga_len=len;
  init_completion(&ctx->done);
;
  if (_lockit)
    write_lock_irqsave(&ctx->cam_data->ctrl_lock, flags);;
  list_add_tail(&ctx->queue, &cam_data->ctrl_queue);
  if (_lockit)
    write_unlock_irqrestore(&ctx->cam_data->ctrl_lock, flags);;;
  for(k=0; k<len; k++);
  retval =
    usb_control_msg_async(cam_data->udev, cam_data->pipe_ctl_snd,
			  0x80, 
			  USB_DIR_OUT |
			  USB_TYPE_VENDOR|
			  USB_RECIP_DEVICE,
			  STC_FPGA_AUTOREFRESH_WVAL|cmd, 
			  0, (((__u16*)&cam_data->payload80)+cmd),
			  len, 
			  sentech_complete_fpga_set, memflags,
			  &ctx->pb, ctx);;
  return retval;
}
static void sentech_complete_fpga2_set(struct urb * urb ARG_COMPLETE){
  int __attribute__((unused))  k;
  struct ctrl_fpga2_context* ctx = (struct ctrl_fpga2_context*)urb->context;;
  if (urb->status < 0);
  usb_free_urb(urb);
  complete(&ctx->done);
  ctx->end = 1;;
}
#define category_offset_expand(a)					\
  if (! ((a) & CATEGORY_OFFSET_PRESENCE_FLAG_MASK)) {			\
    lprintk (1, "%s: category not present: %x\n", __FUNCTION__, (a));	\
    return -ENODEV;							\
  }									\
  adr = ((a) & CATEGORY_OFFSET_PRESENCE_MASK) + category_inside;	\
  lprintk (1, "%s: category present: %x -> %x\n", __FUNCTION__, (a), adr);
static int
sentech_fpga2_set(struct sentech_camera* cam_data, int memflags,
		  __u32 category_offset, __u32 category_inside,
		  void * data, 
		  int __attribute__((unused))  len){
  int retval;
  struct ctrl_fpga2_context* ctx;
  __u32 adr = category_offset;
  __u32 adr_low = adr & 0x0000ff;
  __u32 adr_high = (adr >> 8) & 0x00ffff;
  __u8 _CMD = 0x01; 
  __u16 wVal = (adr_low << 8) | _CMD;
  __u16 wIndex = adr_high;
  __u16 wLen = 4;;
  if (cam_data->fpga2_buf.direct)
    adr = category_offset;
  else {
    switch(category_offset) {
    case CATEGORY_OFFSET_ANALOG_CTL:
      category_offset_expand(cam_data->fpga2_offset_analog_control);
      break;
    case CATEGORY_OFFSET_VENDOR_UNIQ_CTRL:
      category_offset_expand(cam_data->fpga2_offset_vendor_unique_control);
      break;
    case CATEGORY_OFFSET_USB_ORIG_SAVE:
      category_offset_expand(cam_data->fpga2_offset_USB_camera_original_function__save);
      break;
    case CATEGORY_OFFSET_CAMERA_STATUS:
      category_offset_expand(cam_data->fpga2_offset_USB_camera_status);
      break;
    case CATEGORY_OFFSET_CAMERA_FEATURE:
      category_offset_expand(cam_data->fpga2_offset_USB_camera_feature);
      break;
    case CATEGORY_OFFSET_HDR_CONTROL:
      category_offset_expand(cam_data->fpga2_offset_HDR_control);
      break;
    }
  }
  adr_low = adr & 0x0000ff;
  adr_high = (adr >> 8) & 0x00ffff;
  wVal = (adr_low << 8) | _CMD;
  wIndex = adr_high;
  ctx = kmalloc(sizeof(struct ctrl_fpga2_context), memflags);
  if (! ctx) return -ENOMEM;
  ctx->ID = cID++;
  ctx->end = 0;
  ctx->cam_data = cam_data;
  ctx->fpga_cat = category_offset; 
  ctx->fpga_ins_cat = category_inside; 
  ctx->data = *(__u32*)data;
  init_completion(&ctx->done);
  list_add_tail(&ctx->queue, &cam_data->ctrl_queue);
  retval =
    usb_control_msg_async(cam_data->udev, cam_data->pipe_ctl_snd,
			  0x82, 
			  USB_DIR_OUT |
			  USB_TYPE_VENDOR|
			  USB_RECIP_DEVICE, 
			  wVal, 
			  wIndex, 
			  &(ctx->data), 
			  wLen, 
			  sentech_complete_fpga2_set, 
			  memflags,
			  &ctx->pb, ctx);;
  return retval;
}
static void sentech_complete_fpga2_get(struct urb * urb ARG_COMPLETE){
  int __attribute__((unused))  k;
  struct ctrl_fpga2_context* ctx = (struct ctrl_fpga2_context*)urb->context;
  struct sentech_camera* cam_data = ctx->cam_data;;
  if (urb->status < 0)
    printk(KERN_ALERT "sentech_complete_fpga2_get: error %d completing fpga2 cmd 0x%x\n", urb->status, ctx->fpga_cat);
  cam_data->fpga2_buf.data = ctx->data;
  cam_data->fpga2_buf.filled = 1; 
  cam_data->fpga2_buf.status = urb->status; 
  usb_free_urb(urb);
  complete(&ctx->done);
  ctx->end = 1;;
}
static int
sentech_fpga2_get_sync(struct sentech_camera* cam_data, int memflags,
		  __u32 category_offset, __u32 category_inside,
		  void * data, 
		  int __attribute__((unused))  len){
  int retval;
  struct ctrl_fpga2_context* ctx;
  __u32 adr = category_offset; 
  __u32 adr_low = adr & 0x0000ff;
  __u32 adr_high = (adr >> 8) & 0x00ffff;
  __u8 _CMD = 0x01; 
  __u16 wVal = (adr_low << 8) | _CMD;
  __u16 wIndex = adr_high;
  __u16 wLen = 4;;
  if (cam_data->fpga2_buf.direct)
    adr = category_offset;
  else {
    switch(category_offset) {
    case CATEGORY_OFFSET_ANALOG_CTL:
      category_offset_expand(cam_data->fpga2_offset_analog_control);
      break;
    case CATEGORY_OFFSET_VENDOR_UNIQ_CTRL:
      category_offset_expand(cam_data->fpga2_offset_vendor_unique_control);
      break;
    case CATEGORY_OFFSET_USB_ORIG_SAVE:
      category_offset_expand(cam_data->fpga2_offset_USB_camera_original_function__save);
      break;
    case CATEGORY_OFFSET_CAMERA_STATUS:
      category_offset_expand(cam_data->fpga2_offset_USB_camera_status);
      break;
    case CATEGORY_OFFSET_CAMERA_FEATURE:
      category_offset_expand(cam_data->fpga2_offset_USB_camera_feature);
      break;
    case CATEGORY_OFFSET_HDR_CONTROL:
      category_offset_expand(cam_data->fpga2_offset_HDR_control);
      break;
    }
  }
  adr_low = adr & 0x0000ff;
  adr_high = (adr >> 8) & 0x00ffff;
  wVal = (adr_low << 8) | _CMD;
  wIndex = adr_high;
  ctx = kmalloc(sizeof(struct ctrl_fpga2_context), memflags);
  if (! ctx) return -ENOMEM;
  ctx->ID = cID++;
  ctx->end = 0;
  ctx->cam_data = cam_data;
  ctx->fpga_cat = category_offset; 
  ctx->fpga_ins_cat = category_inside; 
  ctx->data = *(__u32*)data;
  init_completion(&ctx->done);
  list_add_tail(&ctx->queue, &cam_data->ctrl_queue);
  retval =
    usb_control_msg_async(cam_data->udev, cam_data->pipe_ctl_rcv,
			  0x82, 
			  USB_DIR_IN |
			  USB_TYPE_VENDOR|
			  USB_RECIP_DEVICE, 
			  wVal, 
			  wIndex, 
			  &(ctx->data), 
			  wLen, 
			  sentech_complete_fpga2_get, 
			  memflags,
			  &ctx->pb, ctx);
  if (! retval) wait_for_completion_timeout(&ctx->done,HZ);;
  return retval;
}
static void sentech_complete_fpga2_read_offset_address(struct urb * urb ARG_COMPLETE) {
  int __attribute__((unused))  k;
  struct ctrl_fpga2_context* ctx = (struct ctrl_fpga2_context*)urb->context;
  struct sentech_camera* cam_data = ctx->cam_data;;
  if (urb->status < 0)
    printk(KERN_ALERT "sentech_complete_fpga2_read_offset_address: error %d completing fpga2 cmd 0x%x\n", urb->status, ctx->fpga_cat);
  else cam_data->fpga2_offsets_known_flag ++;
  usb_free_urb(urb);
  complete(&ctx->done);
  ctx->end = 1;;
}
static int
sentech_fpga2_read_offset_address(
		  struct sentech_camera* cam_data, int memflags,
		  __u32 category_offset,
		  void * data, int sync,
		  int __attribute__((unused))  len) {
  int retval;
  struct ctrl_fpga2_context* ctx;
  __u32 adr = category_offset;
  __u32 adr_low = adr & 0x0000ff;
  __u32 adr_high = (adr >> 8) & 0x00ffff;
  __u8 _CMD = 0x01; 
  __u16 wVal = (adr_low << 8) | _CMD;
  __u16 wIndex = adr_high;
  __u16 wLen = 4;;
  ctx = kmalloc(sizeof(struct ctrl_fpga2_context), memflags);
  if (! ctx) {
    retval = -ENOMEM;;
    return retval;
  }
  ctx->ID = cID++;
  ctx->end = 0;
  ctx->cam_data = cam_data;
  ctx->fpga_cat = category_offset; 
  ctx->fpga_ins_cat = 0; 
  ctx->data = *(__u32*)data; 
  ctx->p_data = data; 
  init_completion(&ctx->done);
  list_add_tail(&ctx->queue, &cam_data->ctrl_queue);
  retval =
    usb_control_msg_async(cam_data->udev, cam_data->pipe_ctl_rcv,
			  0x82, 
			  USB_DIR_IN |
			  USB_TYPE_VENDOR|
			  USB_RECIP_DEVICE, 
			  wVal, 
			  wIndex, 
			  data, 
			  wLen, 
			  sentech_complete_fpga2_read_offset_address, 
			  memflags,
			  &ctx->pb, ctx);;
  if (sync && ! retval) {
    wait_for_completion_timeout(&ctx->done,HZ);;
  };
  return retval;
}
#define MY_KERN_ERR KERN_ERR
static void sentech_usb_indirect_complete(struct urb * urb ARG_COMPLETE) {
  struct sentech_camera** _cam_data = (struct sentech_camera **) urb->context;
  struct sentech_camera* cam_data;
  int i, crt_urb = 0;
  int crt;
  __u16 selected_frame = 0;
  struct Sentech_Header* header = (struct Sentech_Header* ) urb->transfer_buffer;;
  switch (urb->status) {
  case -ESHUTDOWN:
  case -ENODEV:
  case -EPROTO:
    lprintk(0, "In %s status %d \n", __FUNCTION__, urb->status);
    cam_data = *_cam_data;
    cam_data->status = urb->status;
    cam_data->shutdown = 1;
    crt_urb = _cam_data - &cam_data->sib.vmalloc_buffers_context[0];
    goto saved;
  }
  cam_data = *_cam_data;
  cam_data->status |= urb->status;
  if ((! cam_data->run) || cam_data->status || !cam_data->connected || cam_data->shutdown) {;
    crt_urb = _cam_data - &cam_data->sib.vmalloc_buffers_context[0];
   goto saved;
  }
  crt_urb = _cam_data - &cam_data->sib.vmalloc_buffers_context[0];
  i = crt_urb;
  crt = cam_data->sib.vmalloc_buffers_next_free;;;
  if (!urb->status && urb->actual_length) {;
    if (crt < 0) {;
      goto search_available_buffer;
    }
    if (cam_data->sib.sib_frame[crt].data == 0) {;
      goto start_buffer;
    }
    if (cam_data->sib.vmalloc_buffers_idx[crt] + urb->actual_length
	> 
	cam_data->sib.vmalloc_buffers_size) {;
      goto save_to_new_buffer;
    }
    if (cam_data->sib.vmalloc_buffers_idx[crt] + urb->actual_length
	== 
	cam_data->sib.vmalloc_buffers_size) {;
      goto save_end_of_buffer;
    }
    if (urb->actual_length < urb->transfer_buffer_length) {;
      goto save_end_of_buffer;
    }
    goto save_to_crt;
  save_to_crt:;
    memcpy(cam_data->sib.vmalloc_buffers[crt]
	   + cam_data->sib.vmalloc_buffers_idx[crt],
	   urb->transfer_buffer,
	   urb->actual_length
	   );
    cam_data->sib.vmalloc_buffers_idx[crt] += urb->actual_length;;
    goto saved;
  save_end_of_buffer:;
    memcpy(cam_data->sib.vmalloc_buffers[crt]
	   + cam_data->sib.vmalloc_buffers_idx[crt],
	   urb->transfer_buffer,
	   urb->actual_length
	   );
    cam_data->sib.vmalloc_buffers_idx[crt] += urb->actual_length;;
    cam_data->sib.sib_frame[crt].done = 1;
    cam_data->sib.vmalloc_buffers_next_free = crt = -1;
    wake_up_interruptible(&cam_data->wq); 
    goto saved;
  save_to_new_buffer:
;
    cam_data->sib.sib_frame[crt].done = 1;
    wake_up_interruptible(&cam_data->wq); 
  search_available_buffer:;
    for (i = 0; i < cam_data->sib.vmalloc_buffers_nb; i++) {
      if (cam_data->sib.vmalloc_buffers_user[i] == 0) {
	mb();
	if (cam_data->sib.sib_frame[i].data == 0) {
	  crt = i;;
	  goto start_buffer;
	}
      }
    }
    for (i = 0; i < cam_data->sib.vmalloc_buffers_nb; i++) {
      if (cam_data->sib.vmalloc_buffers_user[i] == 1) {;
	continue;
      }
      if (crt == -1) {
	crt = i;
	selected_frame = cam_data->sib.sib_frame[i].frame_nb;;
	continue;
      }
      if ((cam_data->sib.sib_frame[i].frame_nb
	   > 
	   cam_data->sib.last_frame) &&
	  (selected_frame <= cam_data->sib.last_frame))
	{
	  selected_frame = cam_data->sib.sib_frame[i].frame_nb;
	  crt = i;;
	  continue;
	}
      if (cam_data->sib.sib_frame[i].frame_nb < selected_frame) {
	if ((selected_frame > cam_data->sib.last_frame) &&
	    (cam_data->sib.sib_frame[i].frame_nb < cam_data->sib.last_frame)) {;
	  continue;
	}
	selected_frame = cam_data->sib.sib_frame[i].frame_nb;
	crt = i;;
      }
    }
    if (crt == -1) {;
      goto stop;
    }
    goto start_buffer;
  start_buffer:;
    if (header->Header_length != 0x40) {;
      cam_data->sib.vmalloc_buffers_next_free = crt = -1;
      goto saved;
    }
    if ((header->reservedAEF | 1) != 0xAF) {;
      cam_data->sib.vmalloc_buffers_next_free = crt = -1;
      goto saved;
    }
    cam_data->sib.vmalloc_buffers_next_free = crt;
    cam_data->sib.sib_frame[crt].data = 1;
    cam_data->sib.sib_frame[crt].done = 0;
    cam_data->sib.vmalloc_buffers_idx[crt] = 0;
    do_gettimeofday(&(cam_data->sib.last_timestamp));
    cam_data->sib.sib_frame[crt].timestamp = cam_data->sib.last_timestamp;
    cam_data->sib.sib_frame[crt].frame_nb =
      __le32_to_cpu(header->Frame_number_le);
    if (urb->actual_length > sizeof(struct Sentech_Header)) {
      memcpy(cam_data->sib.vmalloc_buffers[crt]
	     + cam_data->sib.vmalloc_buffers_idx[crt],
	     urb->transfer_buffer+sizeof(struct Sentech_Header),
	     urb->actual_length-sizeof(struct Sentech_Header)
	     );
      cam_data->sib.vmalloc_buffers_idx[crt] += urb->actual_length-sizeof(struct Sentech_Header);
    }
;
    goto saved;
  }
 saved:;
  cam_data->sib.vmalloc_buffers_submitted[crt_urb] = 0;
  atomic_dec(&cam_data->running_buffers);
  if ((! cam_data->run) || cam_data->status) {;
    goto stop;
  }
;
  sentech_submit_indirect_kurbs(cam_data, GFP_ATOMIC);
  goto done;
 stop:;
 done:;
}
static void sentech_usb_complete(struct urb * urb ARG_COMPLETE){
  struct sentech_usb_fragment *prev_suf, *next_suf, *suf =
    (struct sentech_usb_fragment *)urb->context;
  struct sentech_usb_buffer   *next_sub, *sub, *tsub;
  struct sentech_camera *cam_data;
  __u8 *align_src_copy, *align_dst_copy;
  const int align_size_copy_to_right = sizeof(struct Sentech_Header);
  const int align_size_copy_to_left = PAGE_SIZE-sizeof(struct Sentech_Header);
  unsigned long flags;
  __u32 frame, lost;
  long unsigned int submitted_sub_size;
  int others_submitted = 0;
  int bytes_to_recover = 0;
;
  if(!suf) {
    printk(MY_KERN_ERR "complete: Unexpected USB bulk no suf context: %p\n", suf);
;
    return;
  }
;
  complete(&suf->done);
  suf->submitted = 0;
  suf->has_data = 1;
  do_gettimeofday(&suf->timestamp);
;
  sub = suf->sub;
  if(!sub) {
    printk(MY_KERN_ERR "complete: Unexpected USB bulk no sub context: %p\n", sub);
;
    return;
  };
  if(suf->first_suf) {;
    sub->timestamp = suf->timestamp;
    sub->error_code = 0;
    sub->has_data = 1;
  }
  cam_data = sub->cam_data;
  if(!cam_data) {
    printk(MY_KERN_ERR "complete: USB bulk no cam_data context: %p\n", cam_data);
;
    return;
  }
  if(suf->recovery) {;
    if(cam_data->recovery_sufs > 0) {
      cam_data->recovery_sufs --;
      cam_data->recovery -= urb->transfer_buffer_length;
    }
    if(cam_data->recovery_sufs == 0) {
      cam_data->recovery_sub = 0;
      cam_data->recovery = 0;
    }
    suf->recovery = 0;
    suf->displaced = 0;
    goto end_done;
  }
  if (sub == cam_data->idle_frame) {
    if(suf->last_suf_submitted) {;
      goto capture_submit_next;
    }else{;
      goto end_done;
    }
  }
;
  if(debug!=0) {
    printk("ST %d/%d %d %d %d|", suf->id, sub->id, urb->status, atomic_read(&cam_data->running_buffers), cam_data->recovery_sufs);
    list_for_each_entry(tsub, &cam_data->buffers, sub_buffers) {
      printk("%d:d%d s%d u%d !%d|",tsub->id,tsub->has_data, tsub->submitted,
	     tsub->user_owned, tsub->error_code);
    }
    printk("\n");
  }
  if (urb->status) {
    switch(urb->status) {
    case -EPIPE:  
      cam_data->status = urb->status;
      if(debug && printk_ratelimit()) printk(KERN_WARNING "USB PIPE Error!");
      break;      
    case -ENODEV: 
      cam_data->status = urb->status;
      if(debug && printk_ratelimit()) printk(KERN_WARNING "USB NODEV Error!");
      break;      
    case -EPROTO: 
      cam_data->status = urb->status;
      if(debug && printk_ratelimit()) printk(KERN_WARNING "USB PROTO Error!");
      break;      
    case -ESHUTDOWN: 
      cam_data->status = urb->status;
      cam_data->shutdown = 1;
      suf->shutdown = 1;;
      if(debug && printk_ratelimit()) printk(KERN_WARNING "USB SHUTDOWN Error!");
      break;      
    case -EILSEQ: 
      if(printk_ratelimit())
	printk(KERN_WARNING "CRC missmatch: change noisy USB cable!\n");
      break;
    case -EOVERFLOW: 
      cam_data->status = urb->status;
      if(printk_ratelimit())
	printk(KERN_WARNING "Babble on the bus:change noisy USB cable!\n");
      break;
    case -ECONNRESET: 
      if(printk_ratelimit()) if(debug>=1) printk(KERN_WARNING "Sentech URB reset!\n");      
      break;
    default:
      printk(MY_KERN_ERR "Unexpected USB bulk error: %d, %d?=%d\n",
	     urb->status, urb->actual_length, 
	     urb->transfer_buffer_length);
    }
    if(!sub->error_code)
      sub->error_code  = urb->status;
    suf->has_data = 0;
  }
  if(suf->displaced && !(suf->last_suf_submitted)) {;
    suf->displaced = 0;
    suf->recovery = 0;
    suf->has_data = 0;
    goto end_done;
  }
  if(suf->first_suf){
    sub->received = urb->actual_length;;
  }else {
    sub->received += urb->actual_length;;
    prev_suf = list_entry(suf->buffers.prev, struct sentech_usb_fragment, buffers);
    align_src_copy = urb->transfer_buffer;
    align_dst_copy = ((__u8*)prev_suf->urb->transfer_buffer)+
      prev_suf->urb->transfer_buffer_length;
    if
      (prev_suf->margin_buffer_right >= align_size_copy_to_right)
      {
	memcpy(align_dst_copy, align_src_copy, align_size_copy_to_right);
	suf->copied_to_right=1;;
      }
    else
      {;
	align_src_copy = (
			  ((__u8*)prev_suf->urb->transfer_buffer)+
			  prev_suf->urb->transfer_buffer_length - align_size_copy_to_left
			  );
	align_dst_copy = ((__u8*)urb->transfer_buffer) - align_size_copy_to_left;
	if(suf->margin_buffer_left >= align_size_copy_to_left) {
	  memcpy(align_dst_copy, align_src_copy, align_size_copy_to_left);
	  suf->copied_to_left=1;;
	}else{;
	}
      }
  }
  if((urb->actual_length!=urb->transfer_buffer_length)||(sub->received==cam_data->scan_size)) {
    int this_encountered = 0;
     list_for_each_entry(next_suf, &sub->suf_buffers, buffers) {
       if(!this_encountered) {
	 bytes_to_recover += next_suf->fragment_size; 
	 if(next_suf == suf) this_encountered = 1;
	 continue;
       }
       if (!next_suf->submitted) continue;
       next_suf->displaced = 1;
       others_submitted = 1;
;
    }   
  }else{
    int this_encountered = 0;
     list_for_each_entry(next_suf, &sub->suf_buffers, buffers) {
       if(!this_encountered) {
	 if(next_suf == suf) this_encountered = 1;
	 continue;
       }
       if (!next_suf->submitted) continue;
       others_submitted = 1;
       break;
     }
  }
  if(filter_short_frames
     &&(urb->actual_length!=urb->transfer_buffer_length)
     &&(sub->received!=cam_data->scan_size)
     ) {
    suf->has_data = 0;
    sub->has_data = 0;
    dprintk (0, "%d/%d Got %d(%x h)/%x instead of %d(%x h)\n",suf->id, sub->id,
	     urb->actual_length,urb->actual_length,sub->received,
	     cam_data->scan_size,cam_data->scan_size);
    if(urb->status != -ECONNRESET) {
      if((cam_data->payload80.Trigger_Mode & STC_FPGA_TRIGGER_MODE_MASK)
	 == STC_FPGA_TRIGGER_MODE)
	if((cam_data->payload80.Trigger_Mode & STC_FPGA_TRIGGER_SRC_MASK)
	   == STC_FPGA_TRIGGER_SRC_SOFT)
	  if(trigger_on_fail)
	    sentech_trigger(cam_data, GFP_ATOMIC);
    } else {;
    }
  }
  if(suf->first_suf && suf->has_data) {
    frame = __le32_to_cpu(suf->framebuffer_header->Frame_number_le);
    if(debug!=0) printk(MY_KERN_ERR "ST|%d|\n",frame);;
    if (!cam_data->first_frame)
      cam_data->first_frame = cam_data->last_frame = frame;
    else {
      lost = frame-cam_data->last_frame-1;
      cam_data->lost_frames += lost;
      cam_data->last_frame = frame;
    }
  }
  if(suf->first_suf) {
    sub->has_data = suf->has_data;
  }else{
    sub->has_data &= suf->has_data;
  }
 capture_submit_next:
  if(
     (suf->last_suf)
     || (suf->last_suf_submitted)
     || (!others_submitted) 
     || (urb->actual_length!=urb->transfer_buffer_length)
     || (sub->received==cam_data->scan_size)
     )
    { 
;
    if(!others_submitted) sub->submitted = 0;
    wake_up_interruptible(&cam_data->wq); 
    if(! cam_data->run || cam_data->status) {;
      if((suf->last_suf)
	 || (suf->last_suf_submitted)
	 || (!others_submitted)
	 || (sub->received==cam_data->scan_size))
	goto stop_buffer;
      else
	goto end_done;
    }
    if((others_submitted) && (urb->actual_length!=urb->transfer_buffer_length) && 
       (sub->received < cam_data->scan_size)) {;
      if((bytes_to_recover > cam_data->recovery)) {
	int rec_cnt = 0;
	int rec_sufs;
	if (cam_data->recovery_sub == NULL) {
	  cam_data->recovery_sub = sub;
	  cam_data->recovery_sufs = 0;
	  cam_data->recovery = 0;
	}
	rec_sufs = cam_data->recovery_sufs;
	list_for_each_entry(next_suf, &cam_data->recovery_sub->suf_buffers, buffers) { 
	  if(next_suf->recovery){
	    rec_cnt ++;
	    if(rec_cnt > rec_sufs)
	      dprintk (0, "Unexpected recovery sub %d/%d, %d vs %d\n", sub->id,cam_data->recovery_sub->id, rec_cnt, rec_sufs); 
	    bytes_to_recover -= next_suf->fragment_size;
	    if(bytes_to_recover <= 0) break;
	    continue;
	  }
	  if(rec_cnt < rec_sufs)
	    dprintk (0, "Unexpected missing recovery sub %d/%d, %d vs %d\n", sub->id,cam_data->recovery_sub->id, rec_cnt, rec_sufs); 
	  if(bytes_to_recover<=0) break;;
	  next_suf->recovery = 1;
	  next_suf->displaced = 0;
	  cam_data->recovery_sufs ++;
	  cam_data->recovery += next_suf->fragment_size;
	  usb_submit_urb(next_suf->urb, GFP_ATOMIC);
	  bytes_to_recover -= next_suf->fragment_size;
	}
      }
      goto end_done;
    }
    next_sub = NULL;
    if((!others_submitted)) {
      if(((sub->received != cam_data->scan_size) || (suf->displaced))
	 && filter_short_frames)
	sub->has_data = 0;
      read_lock_irqsave(&cam_data->buffers_lock, flags);
      mb();
      if (sub->has_data||(cam_data->recovery_sub==sub)) {
	next_sub =
	  list_first_entry(&cam_data->buffers,
			   struct sentech_usb_buffer, sub_buffers);
      } else {;
	next_sub = sub;
;	
      }
      mb();
      read_unlock_irqrestore(&cam_data->buffers_lock, flags);;
    } 
    if(suf->last_suf_submitted) suf->displaced = 0;
    if((!others_submitted)) {
      if(sub->has_data&&(next_sub->id == sub->id)&&(!param_use_submitted)){;
	goto stop_buffer; 
      } else {;	
      }
      if(!next_sub) {
	printk(MY_KERN_ERR "sentech: Unexpected USB bulk no next_sub: %p\n", next_sub);
	goto stop_buffer;
      }
      if(next_sub->user_owned){ 
	if(!param_submit_user_buffers) {
	  if(debug && printk_ratelimit());
	  read_lock_irqsave(&cam_data->buffers_lock, flags);
	  if (! list_is_last(&next_sub->sub_buffers, &cam_data->buffers)) {
	    next_sub = list_entry(next_sub->sub_buffers.next, 
				struct sentech_usb_buffer, sub_buffers);
	  }
	  read_unlock_irqrestore(&cam_data->buffers_lock, flags);;
	  if(next_sub->user_owned || next_sub->submitted){;
	    goto idle_buffer;
	  };
	}else{
	  if(debug && printk_ratelimit());
	}
      };
      if(next_sub->submitted){
	printk(MY_KERN_ERR "sentech: Unexpectedly buffer %d already submitted: %p\n", next_sub->id, next_sub);
	goto stop_buffer;
      };
      if((next_sub == cam_data->recovery_sub)){
	printk(MY_KERN_ERR "sentech: Stop recovery buffer: %p, %d\n", next_sub, next_sub->id);
	goto stop_buffer;
      };
      write_lock_irqsave(&cam_data->buffers_lock, flags);
      mb();
      list_del(&next_sub->sub_buffers);
      list_add_tail(&next_sub->sub_buffers, &cam_data->buffers);
      mb();
      write_unlock_irqrestore(&cam_data->buffers_lock, flags);;
      next_sub->submitted=1;
      submitted_sub_size = 0;
      if(!param_use_submitted){
	next_sub->has_data=0;
	next_sub->error_code=0;
      };
      list_for_each_entry(next_suf, &next_sub->suf_buffers, buffers) { 
	init_completion(&next_suf->done);
	next_suf->submitted = 1;
	next_suf->displaced = 0;
	next_suf->recovery = 0;
	if(!param_use_submitted){
	  next_suf->has_data = 0;
	}
;
	usb_submit_urb(next_suf->urb, GFP_ATOMIC);
	submitted_sub_size += next_suf->urb->transfer_buffer_length;
	if(submitted_sub_size >= cam_data->scan_size) {;
	  next_suf->last_suf_submitted = 1;
	  break;
	}else{
	  next_suf->last_suf_submitted = 0;;
	}
      }
    }
  }else{;;
  }
  goto end_done;
 idle_buffer:;
  if(cam_data->idle_frame == NULL) {
    goto stop_buffer;
  }
  if (cam_data->idle_frame->submitted) {
    printk(MY_KERN_ERR "sentech: IDLE frame already used\n");
    goto stop_buffer;
  }
  next_sub = cam_data->idle_frame;
  next_sub->submitted=1;
  if(!param_use_submitted){
    next_sub->has_data=0;
    next_sub->error_code=0;
  };
  submitted_sub_size = 0;
  list_for_each_entry(next_suf, &next_sub->suf_buffers, buffers) {
    init_completion(&next_suf->done);
    next_suf->submitted = 1;
    if(!param_use_submitted){
      next_suf->has_data = 0;
    };
    usb_submit_urb(next_suf->urb, GFP_ATOMIC);
    submitted_sub_size += next_suf->urb->transfer_buffer_length;
    if(submitted_sub_size >= cam_data->scan_size) {;
      next_suf->last_suf_submitted = 1;
      break;
    }else{
      next_suf->last_suf_submitted = 0;;
    }
  }
  goto end_done;
 stop_buffer:
  atomic_dec(&cam_data->running_buffers);;
  if(atomic_read(&cam_data->running_buffers)<0){
    atomic_inc(&cam_data->running_buffers);;
  }
  if(! cam_data->run || cam_data->status) {
    if(debug!=0)
      printk(MY_KERN_ERR "sentech: Camera stopped. run=%d:stat=%d. Stopping buffer %d. Left: %d\n", cam_data->run, cam_data->status, sub->id, atomic_read(&cam_data->running_buffers));
  }else{
    printk(MY_KERN_ERR "sentech: Slow USER. Stopping buffer %d. Left: %d\n", sub->id, atomic_read(&cam_data->running_buffers));
  }
 end_done:
  if(debug!=0){
    printk("SF     %d|", atomic_read(&cam_data->running_buffers));
    list_for_each_entry(tsub, &cam_data->buffers, sub_buffers) {
      printk("%d:d%d s%d u%d !%d|",tsub->id,tsub->has_data, tsub->submitted,
	     tsub->user_owned, tsub->error_code);
    }
    printk("\n\n");
  }
  dprintksubs(cam_data);;
  return;
}
static
int sentech_drop_images_direct(struct sentech_camera* cam_data){
  int retval=0;
  struct sentech_usb_buffer *sub = NULL;
  unsigned long flags;
  read_lock_irqsave(&cam_data->buffers_lock, flags);
  list_for_each_entry(sub, &cam_data->buffers, sub_buffers){
    if (sub->user_owned){;
      retval = -EBUSY;
    } 
    if (sub->has_data){;
      sub->has_data = 0;
    }
  }
  read_unlock_irqrestore(&cam_data->buffers_lock, flags);
  return retval;
}
static
int sentech_drop_images_indirect(struct sentech_camera* cam_data) {
  int retval=0;
  int i;
  for (i = 0; i < MAX_VMALLOC_BUFFERS ; i++) {
    cam_data->sib.sib_frame[i].data = 0;
    cam_data->sib.sib_frame[i].done = 0;
  }
  return retval;
}
static
int sentech_drop_images(struct sentech_camera* cam_data) {
  if (param_have_buffers) {
    return sentech_drop_images_direct(cam_data);
  } else {
    return sentech_drop_images_indirect(cam_data);
  }
}
static unsigned long submitting = 0;
static
int sentech_submit_indirect_kurbs(struct sentech_camera* cam_data, int mem_flags) {
  int i;
  int retval = 0;
  int max_loops = cam_data->sib.vmalloc_buffers_urbs_nb;
  int crt = 0;
;
  if (test_and_set_bit(0, &submitting)) {;
    return 0;
  }
  if (cam_data->shutdown) {retval = -ENODEV; goto done;}  
  while (
	 (
	  atomic_read(&cam_data->running_buffers)
	  <
	  cam_data->sib.vmalloc_buffers_urbs_nb)
	 &&
	 (
	  atomic_read(&cam_data->running_buffers)
	  <
	  param_SUBS_SUBMITTED)
	 ) {
;
    crt ++;
    if (crt > max_loops) {
      break;
    }
    for (i = 0; i < cam_data->sib.vmalloc_buffers_urbs_nb; i++) {
      if (cam_data->sib.vmalloc_buffers_submitted[i]) {;
	continue;
      }
      cam_data->sib.vmalloc_buffers_submitted[i] = 1;
      usb_submit_urb(cam_data->sib.vmalloc_buffers_urb[i], mem_flags);
      atomic_inc(&cam_data->running_buffers);;
      break; 
    }
  }
 done:
  mb();
  submitting = 0;
  mb();;
  return retval;
}
static
int sentech_submit_direct_kurbs(struct sentech_camera* cam_data, int mem_flags) {
  struct sentech_usb_fragment* suf = NULL;
  struct sentech_usb_buffer *sub = NULL;
  int retval = 0;
  unsigned long flags;
  unsigned long submitted_sub_size;;
  dprintksubs(cam_data);
  if(cam_data->shutdown) {retval = -ENODEV; goto done;}  
  for(;;){
    if(atomic_read(&cam_data->running_buffers) >= SUBS_SUBMITTED)
      return 0;;
    read_lock_irqsave(&cam_data->buffers_lock, flags);
    sub = list_first_entry(&cam_data->buffers,
			   struct sentech_usb_buffer, sub_buffers);
#ifdef SUBMITTED_DISOWNED
    sub->user_owned=1;
#endif
    read_unlock_irqrestore(&cam_data->buffers_lock, flags);
    if(unlikely(!sub || sub->submitted || (sub==cam_data->recovery_sub))){
      if(sub);
      break;
    }
    atomic_inc(&cam_data->running_buffers);
    sub->submitted = 1;;
#ifdef SUBMITTED_DISOWNED
    sub->has_data = 0;
    sub->error_code = 0;
#endif
    write_lock_irqsave(&cam_data->buffers_lock, flags);
#ifdef SUBMITTED_DISOWNED
    sub->user_owned=0;
#endif
    list_del(&sub->sub_buffers);
    list_add_tail(&sub->sub_buffers, &cam_data->buffers);
    write_unlock_irqrestore(&cam_data->buffers_lock, flags);
    submitted_sub_size = 0;
    list_for_each_entry(suf, &sub->suf_buffers, buffers) {
      if(suf->submitted) {
	suf->submitted = 0;
	dprintk (0, "Unexpected submitted buffer %d/%d \n", suf->id, sub->id);
	break;
      }
      init_completion(&suf->done);
      suf->submitted = 1;
      suf->displaced = 0;
      suf->recovery = 0;;
      usb_submit_urb(suf->urb, mem_flags);
      submitted_sub_size += suf->urb->transfer_buffer_length; 
      if(submitted_sub_size >= cam_data->scan_size) {
	suf->last_suf_submitted = 1;;
	break;
      }else{
	suf->last_suf_submitted = 0;;
      }
    };
  }
 done:
  dprintksubs(cam_data);;
  return retval;
}
static
int sentech_submit_kurbs(struct sentech_camera* cam_data, int mem_flags) {
  if (param_have_buffers) {
    return sentech_submit_direct_kurbs(cam_data, mem_flags);
  } else {
    if ( 
	! cam_data->probe_done) {
      printk(KERN_DEBUG "sentech:  Sentech USB Camera open during probe\n");
      return -EBUSY;
    }
    return sentech_submit_indirect_kurbs(cam_data, mem_flags);
  }
}
static int res_get(struct sentech_camera *cam_data){
  int result = 0;
  lprintk(1,"In %s\n", __FUNCTION__);
  down(&cam_data->lock);
  if (cam_data->resources) {;
    result = 0;
    goto done;
  }
  cam_data->resources =1;;
  result = 1;
 done:
  up(&cam_data->lock);
  lprintk(1,"Exiting %s\n", __FUNCTION__);
  return result;
}
static inline int res_locked(struct sentech_camera *cam_data){
  return (cam_data->resources);
}
static void res_free(struct sentech_camera *cam_data){
	down(&cam_data->lock);
	cam_data->resources = 0;;
	up(&cam_data->lock);
}
static 
int
sentech_vidioc_streamon (struct file *file, void * UNUSED(priv), enum v4l2_buf_type i) {
  struct sentech_camera* cam_data = get_cam_data(file->private_data);
  int retval = 0;
  int wratio = 1;
  int hratio = 1;
  char* format = "UNKNOWN";;
  lprintk(1,"In %s\n",__FUNCTION__);
  if (unlikely(!cam_data)) {;
    goto done_nowait;
  }
  if(!cam_data->connected) {
    printk("sentech streamon: Video device is disconnecting!\n");
    retval = -ENODEV;
    goto done_nowait;
  }
  if(!cam_data->run) {
    printk("sentech: Video device is stopped!\n");
    retval = -ENODEV;
    goto done_nowait;
  }
  if (!res_get(cam_data)){
    if (file==cam_data->streamer) goto on_done;
    dev_err(&cam_data->interface->dev, "Busy");
    retval = -EBUSY;
    return retval;
  }
  if(param_streaming && (i == cam_data->buffertype)){
    if(cam_data->fragmented){;
    }else{
;
      cam_data->streaming = 1;
      cam_data->streamer = file;
    }
  }else{;
    retval = -EINVAL;
  }
  if(!retval) cam_data->first_frame = 0;
 on_done:
  switch(cam_data->fmt->fourcc){
  case V4L2_PIX_FMT_SRGGB8: format = "SRGGB8"; break;
  case V4L2_PIX_FMT_SGRBG8: format = "SGRBG8"; break;
  case V4L2_PIX_FMT_SRGGB10: format = "SRGGB10"; break;
  case V4L2_PIX_FMT_SRGGB12: format = "SRGGB12"; break;
  case V4L2_PIX_FMT_SRGGB16: format = "SRGGB16"; break;
  case V4L2_PIX_FMT_GREY: format = "GREY"; break;
  case V4L2_PIX_FMT_Y16: format = "Y16"; break;
  case V4L2_PIX_FMT_YUYV: format = "YUYV"; break;
  case V4L2_PIX_FMT_BGR24: format = "BGR24"; break;
  case V4L2_PIX_FMT_RGB24: format = "RGB24"; break;
  case V4L2_PIX_FMT_YUV420: format = "YUV420"; break;
  case V4L2_PIX_FMT_YUV422P: format = "YUV422P"; break;
  default: format = "Unknown";
  }
  if(scale) {
    if(cam_data->crop.width)
      wratio = cam_data->max_width / cam_data->crop.width;
    if(cam_data->crop.height)
      hratio = cam_data->scan_height / cam_data->crop.height;
    if(hratio<1) hratio = 1;
  }
  if(debug != 0) {
#ifdef SENTECH_AWB
  printk(KERN_INFO "sentech: " "Streaming AWB "
#else
  printk(KERN_INFO "sentech: " "Streaming "
#endif
"%s (%i): width=%i x height=%i, top=%i, maxw=%i, scale=%i, gbrg=%i, wratio=%i hratio=%i\n",
	  format,cam_data->fmt->fourcc,
	  cam_data->crop.width,cam_data->crop.height,cam_data->crop.top,cam_data->max_width,
	  scale, gbrg, wratio, hratio);
	 }
  done_nowait:;
  lprintk(1,"Exiting %s\n",__FUNCTION__);
  return retval;
}
static 
int 
sentech_vidioc_streamoff(struct file *file,
			 void * UNUSED(priv), enum v4l2_buf_type type) {
  struct sentech_camera* cam_data = get_cam_data(file->private_data);
  int retval = 0;
;
  lprintk(1,"In %s\n",__FUNCTION__);
  if (type != V4L2_BUF_TYPE_VIDEO_CAPTURE){;
    return -EINVAL;
  }
  cam_data->streaming = 0;
  res_free(cam_data);
;
  lprintk(1,"Exiting %s\n",__FUNCTION__);
  return retval;
}
static
int sentech_vidioc_cropcap (struct file * UNUSED(file), void *fh,
			    struct v4l2_cropcap *a){
  struct sentech_camera* cam_data = get_cam_data(fh);
  int retval = 0;;
  lprintk(1,"In %s\n",__FUNCTION__);
  if(a->type != cam_data->type){
;
  }
  if(param_old_scale){
    a->bounds.left=0;
    a->bounds.top=0;
    a->bounds.width = cam_data->width;
    a->bounds.height = cam_data->height;
    cam_data->crop = a->defrect = a->bounds;;
  }else{
    a->bounds = cam_data->bounds;
    a->defrect = cam_data->defrect;;;
    if(param_defrect_scan_height) {
      a->defrect.height = cam_data->scan_height;;
    }
  }
  a->pixelaspect.numerator = a->pixelaspect.denominator = 1;
;
  lprintk(1,"Exiting %s\n",__FUNCTION__);
  return retval;
}
static
int sentech_vidioc_g_crop (struct file * UNUSED(file), void *fh,
			   struct v4l2_crop *a){
  struct sentech_camera* cam_data = get_cam_data(fh);
  int retval = 0;;
  lprintk(1,"In %s\n",__FUNCTION__);
  if(a->type != cam_data->type);
  a->c = cam_data->crop;
;
  lprintk(1,"Exiting %s\n",__FUNCTION__);
  return retval;
}
static
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,7,0)
int sentech_vidioc_s_crop (struct file *UNUSED(file), void *fh,
			   const struct v4l2_crop *request)
#else
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,18)
int sentech_vidioc_s_crop (struct file *UNUSED(file), void *fh,
			   const struct v4l2_crop *request)
#else
int sentech_vidioc_s_crop (struct file *UNUSED(file), void *fh,
			   struct v4l2_crop *request)
#endif
#endif
{
  struct sentech_camera* cam_data = get_cam_data(fh);
  struct v4l2_crop b;
  struct v4l2_crop* a = NULL;
  int retval = 0;
  int mask = ~1; 
  int minw = 2;;
  lprintk(1,"In %s\n",__FUNCTION__);
  if(request != NULL){
    b = *request;
    a = &b;
  }
  if(request->type != cam_data->type);
  if(a->c.width  < 0) a->c.width = minw;
  if(a->c.height < 0) a->c.height= minw;
  a->c.left = a->c.left & mask;
  a->c.top = a->c.top & mask;
  a->c.width = a->c.width & mask;
  a->c.height = a->c.height & mask;
  if(a->c.left >= cam_data->max_width) a->c.left=cam_data->max_width-minw;
  if(a->c.top  >= cam_data->max_height) a->c.top=cam_data->max_height-minw;
  if(a->c.left < 0) a->c.left=0;
  if(a->c.top  < 0) a->c.top=0;
  if(a->c.width + a->c.left > cam_data->max_width)
    a->c.width = cam_data->max_width - a->c.left;
  if(a->c.height + a->c.top > cam_data->max_height)
    a->c.height = cam_data->max_height - a->c.top;
  if ((retval = down_interruptible(& cam_data->format_lock))) {;
    retval = -ERESTARTSYS;
    goto scrop_exit;
  }
  if (cam_data->vmas) {;
    a->c.width = a->c.width - (a->c.width % cam_data->width);
    a->c.height = a->c.height - (a->c.height % cam_data->height);
    if (! a->c.width || ! a->c.height) {;
      retval = -EBUSY;
      goto scrop_unlock;
    }
    cam_data->crop = a->c;
    cam_data->output.left = a->c.width / cam_data->width;
    cam_data->output.top = a->c.height / cam_data->height;
  } else {;
    cam_data->crop = a->c;
    cam_data->width = cam_data->output.width = a->c.width;
    cam_data->height = cam_data->output.height = a->c.height;
    cam_data->output.left = 1;
    cam_data->output.top = 1;
  }
  scrop_unlock:
  up(&cam_data->format_lock);
 scrop_exit:;;
;
  lprintk(1,"Exiting %s\n",__FUNCTION__);
  return retval;
}
static
int sentech_vidioc_querycap (struct file *UNUSED(file), void *priv, 
		       struct v4l2_capability *cap){
  struct sentech_camera* dev = get_cam_data(priv);;
  lprintk(1,"In %s\n",__FUNCTION__);
  if(!cap){;
    return -EINVAL;
  }
  memset(cap, 0, sizeof *cap);
  strncpy((char*)cap->driver, "SentechTUSB", sizeof cap->driver); 
  strncpy((char*)cap->card, dev->cam_name, 32);
  strncpy((char*)cap->bus_info, dev->udev->bus->bus_name,
	  sizeof cap->bus_info);
  cap->version=SENTECH_VERSION; 
  cap->capabilities=
    V4L2_CAP_VIDEO_CAPTURE;
  if(param_streaming)
    cap->capabilities |=
      V4L2_CAP_STREAMING;
  if(param_read && param_have_buffers)
    cap->capabilities |=
      V4L2_CAP_READWRITE;
  if(param_overlay)
    cap->capabilities |=
      V4L2_CAP_VIDEO_OVERLAY;;
  lprintk(1,"Exiting %s\n",__FUNCTION__);
  return 0;
}
static
int sentech_vidioc_g_priority (struct file *file, void *fh,
			    enum v4l2_priority *p) {
  struct sentech_camera* cam_data = get_cam_data(file->private_data);
  struct sentech_fh *sfh= (struct sentech_fh *)fh;;
  lprintk(1,"In %s\n",__FUNCTION__);
  if(cam_data->exclusive)
    *p = V4L2_PRIORITY_RECORD;
  else
    *p = sfh -> priority;;
  lprintk(1,"Exiting %s\n",__FUNCTION__);
  return 0;
}
static int 
sentech_vidioc_s_priority (struct file *file, void *fh,
			   enum v4l2_priority p) {
  struct sentech_camera* cam_data = get_cam_data(file->private_data);
  struct sentech_fh *sfh= (struct sentech_fh *)fh;
  int retval = 0;;
  lprintk(1,"In %s\n",__FUNCTION__);
  switch(p) {
  case V4L2_PRIORITY_UNSET:
  case V4L2_PRIORITY_BACKGROUND:
  case V4L2_PRIORITY_INTERACTIVE:
    sfh -> priority = p;
    retval = 0;
    break;
  case V4L2_PRIORITY_RECORD:
    if(cam_data->open_desc >1){
      retval = -EBUSY;
      break;
    }
    sfh -> priority = p;
    cam_data->exclusive=1;
    retval = 0;
    break;
  default:
    retval = -EINVAL;
  };
  lprintk(1,"Exiting %s\n",__FUNCTION__);
  return retval;
}
static
int sentech_vidioc_enum_fmt_cap  (struct file *file, void *UNUSED(fh),
				  struct v4l2_fmtdesc *f) {
  struct sentech_camera* cam_data = get_cam_data(file->private_data);
  enum v4l2_buf_type type;
  unsigned int index;
  int retval = 0;
  if(f)
    lprintk(4,"In %s idx=%d (usb3=%d)\n",__FUNCTION__, f->index, cam_data->usb3);
  else
    lprintk(4,"In %s null param (usb3=%d)\n",__FUNCTION__, cam_data->usb3);
  if(!f){retval = -EINVAL; goto done;}
  index = f->index;
  type  = f->type;
  if (f->type != V4L2_BUF_TYPE_VIDEO_CAPTURE) {
    dev_err(&cam_data->interface->dev, "Type %d not video-cap", f->type);
    retval = -EINVAL;
    goto done;
  }
  memset (f, 0, sizeof(*f));
  f->index = index;
  f->type = type;
  retval = 0;
  if (cam_data->usb3) {
    switch (index) {
    case 0:
      strlcpy((char*)f->description, format_grbg.name, sizeof(f->description));
      f->pixelformat = format_grbg.fourcc;
      f->flags = 0;
      break;
    case 1:
      strlcpy((char*)f->description, format_grey.name, sizeof(f->description));
      f->pixelformat = format_grey.fourcc;
      f->flags = 0; 
      break;
    case 2:
      strlcpy((char*)f->description, format_bgr24.name, sizeof(f->description));
      f->pixelformat = format_bgr24.fourcc;
      f->flags = 0; 
      break;
    default:
      retval = -EINVAL;
    }
    goto done;
  }
  if (cam_data->monochrome && mono_no_color)
    index += 9;
  switch (index) {
  case 0:
    strlcpy((char*)f->description, format_rgb24.name, sizeof(f->description));
    f->pixelformat = format_rgb24.fourcc;
    f->flags = 0; 
    break;
  case 1:
    strlcpy((char*)f->description, format_yuyv.name, sizeof(f->description));
    f->pixelformat = format_yuyv.fourcc;
    f->flags = 0; 
    break;
  case 2:
    strlcpy((char*)f->description, format_yuv420.name, sizeof(f->description));
    f->pixelformat = format_yuv420.fourcc;
    f->flags = 0; 
    break;
  case 3:
    strlcpy((char*)f->description, format_bgr24.name, sizeof(f->description));
    f->pixelformat = format_bgr24.fourcc;
    f->flags = 0; 
    break;
  case 4:
    strlcpy((char*)f->description, format_rggb.name, sizeof(f->description));
    f->pixelformat = format_rggb.fourcc;
    f->flags = 0; 
    break;
  case 5:
    strlcpy((char*)f->description, format_bggr.name, sizeof(f->description));
    f->pixelformat = format_bggr.fourcc;
    f->flags = 0; 
    break;
  case 6:
    strlcpy((char*)f->description, format_rggb10.name, sizeof(f->description));
    f->pixelformat = format_rggb10.fourcc;
    f->flags = 0; 
    break;
  case 7:
    strlcpy((char*)f->description, format_rggb12.name, sizeof(f->description));
    f->pixelformat = format_rggb12.fourcc;
    f->flags = 0; 
    break;
  case 8:
    strlcpy((char*)f->description, format_rggb16.name, sizeof(f->description));
    f->pixelformat = format_rggb16.fourcc;
    f->flags = 0; 
    break;
  case 9:
    if(COLOR_NO_MONO && !cam_data->monochrome) {;
      retval = -EINVAL;
      break;
    }
    strlcpy((char*)f->description, format_y16.name, sizeof(f->description));
    f->pixelformat = format_y16.fourcc;
    f->flags = 0; 
    break;
  case 10:
    if(COLOR_NO_MONO && !cam_data->monochrome) {;
      retval = -EINVAL;
      break;
    }
    strlcpy((char*)f->description, format_grey.name, sizeof(f->description));
    f->pixelformat = format_grey.fourcc;
    f->flags = 0; 
    break;
  default:
    retval = -EINVAL;
  }
 done:
  lprintk(4,"Exiting %s\n",__FUNCTION__);
  return retval;
}
static
int sentech_vidioc_g_fmt_cap  (struct file *file, void *UNUSED(fh),
			       struct v4l2_format *f) {
  struct sentech_camera* cam_data = get_cam_data(file->private_data);
  int retval = 0;;
  lprintk(1,"In %s\n",__FUNCTION__);
  if(!f){retval = -EINVAL; goto done;};
  if (f->type != V4L2_BUF_TYPE_VIDEO_CAPTURE){
    dev_err(&cam_data->interface->dev, "Type %d not video-cap", f->type);
    retval = -EINVAL;
    goto done;
  }
  memset(&f->fmt.pix, 0, sizeof(struct v4l2_pix_format));
  f->fmt.pix.colorspace = V4L2_COLORSPACE_SRGB;
  f->fmt.pix = cam_data->pix_format;
  retval = 0;;;;
 done:;
  lprintk(1,"Exiting %s\n",__FUNCTION__);
  return retval;
}
static
  void sentech_setBytes(struct sentech_camera* cam_data, __u32 setBytes) {
  if(setBytes != __le32_to_cpu(cam_data->payload13.Bytes_per_pixel_le)) {
    cam_data->payload13.Bytes_per_pixel_le = __cpu_to_le32(setBytes);
    setFlag(&cam_data->modif_flag13);
  }
 }
static int get_depth(__u32 pixelformat){
  __u32 result=8;
  switch(pixelformat){
  case  V4L2_PIX_FMT_GREY: return format_grey.depth;
  case  V4L2_PIX_FMT_Y16: return format_y16.depth;
  case  V4L2_PIX_FMT_SRGGB8: return format_rggb.depth;
  case  V4L2_PIX_FMT_SGRBG8: return format_grbg.depth;
  case  V4L2_PIX_FMT_SRGGB10: return format_rggb10.depth;
  case  V4L2_PIX_FMT_SRGGB12: return format_rggb12.depth;
  case  V4L2_PIX_FMT_SRGGB16: return format_rggb16.depth;
  case  V4L2_PIX_FMT_SBGGR8: return format_bggr.depth;
  case  V4L2_PIX_FMT_BGR24: return format_bgr24.depth;
  case  V4L2_PIX_FMT_RGB24: return format_rgb24.depth;
  case  V4L2_PIX_FMT_YUYV:  return format_yuyv.depth;
  case  V4L2_PIX_FMT_YUV420:  return format_yuv420.depth;
  case  V4L2_PIX_FMT_YUV422P:  return format_yuv422P.depth;
  }
  return result;
}
 static int sentech_support_AOI(struct sentech_camera* cam_data) {
   return (__le16_to_cpu(cam_data->payload11.OR_Scan_modes_le) & STC_AOI_SCREEN)?1:0;
 }
static int sentech_vidioc_try_fmt_cap(struct file *file, void*priv, 
				      struct v4l2_format *f);
static
int sentech_vidioc_s_fmt_cap  (struct file *file, void *UNUSED(fh),
				struct v4l2_format *f) {
  struct sentech_camera* cam_data = get_cam_data(file->private_data);
  int wratio=1;
  int hratio=1;
  int scalable = 1;
  __u32 setBytes = STC_ONE_BYTE_PER_PIXEL;
  unsigned long flags;
  int retval = 0;;
  lprintk(1,"In %s\n",__FUNCTION__);
  if(!f){retval = -EINVAL; goto done;};
;
;
  if (f->type != V4L2_BUF_TYPE_VIDEO_CAPTURE){
    dev_err(&cam_data->interface->dev, "Type %d not video-cap", f->type);
    retval = -EINVAL;
    goto done;
  }
;;
  if (sentech_vidioc_try_fmt_cap(file, file->private_data, f) < 0) {
    retval = -EINVAL;
    goto done;
  };
  if ((retval = down_interruptible(&cam_data->format_lock))) {
    retval = -ERESTARTSYS;
    goto done;
  }
  if (cam_data->vmas) {;
    f->fmt.pix = cam_data->pix_format;
    retval = -EINVAL;
    goto format_exit;
  }
  if (cam_data->usb3) {
    switch ( f->fmt.pix.pixelformat) {
    case  V4L2_PIX_FMT_GREY:;
      break;
    case  V4L2_PIX_FMT_SGRBG8:;
      break;
    case  V4L2_PIX_FMT_RGB24:
    case  V4L2_PIX_FMT_BGR24:
      if (
	  (cam_data->mask_Bpp & STC_THREE_BYTES_PER_PIXEL_RGB8)
	  ) {;
	break;
      };
   default:;
      f->fmt.pix = cam_data->pix_format;
      retval = -EINVAL;
      goto format_exit; 
    }
    if (param_variable_width && cam_data->usb3 && sentech_support_AOI(cam_data)) {
      f->fmt.pix.width &= ~0x03;
      if (f->fmt.pix.width < 32) f->fmt.pix.width = 32;
      if (f->fmt.pix.height & 0x03) f->fmt.pix.height += 4;
      f->fmt.pix.height &= ~0x03;
      cam_data->payload12.Horizontal_size_mode_le = __cpu_to_le32(f->fmt.pix.width);
      cam_data->payload12.Vertical_size_mode_le = __cpu_to_le32(f->fmt.pix.height);
      cam_data->payload12.Horizontal_offset_le = __cpu_to_le32(0);
      cam_data->payload12.Vertical_offset_le = __cpu_to_le32(0);
      cam_data->payload12.Scan_mode_le = __cpu_to_le16(STC_AOI_SCREEN);
      setFlag( & cam_data->modif_flag12 );
      f->fmt.pix.bytesperline = f->fmt.pix.width*(get_depth(f->fmt.pix.pixelformat)>>3);;
    } else {;
      f->fmt.pix.bytesperline = cam_data->scan_width*(get_depth(f->fmt.pix.pixelformat)>>3);;
    }
    f->fmt.pix.sizeimage = 
      f->fmt.pix.height * f->fmt.pix.bytesperline;
  }
  if ( ! param_have_buffers ) {
    switch ( f->fmt.pix.pixelformat) {
    case  V4L2_PIX_FMT_GREY:;
      break;
    case  V4L2_PIX_FMT_SBGGR8:;
      break;
    case  V4L2_PIX_FMT_SRGGB8:;
      break;
    case  V4L2_PIX_FMT_SGRBG8:;
      break;
    case V4L2_PIX_FMT_Y16:;
      break;
    case V4L2_PIX_FMT_BGR24:;
      break;
    case V4L2_PIX_FMT_RGB24:;
      break;
    case V4L2_PIX_FMT_YUYV:;
      break;
    case V4L2_PIX_FMT_YUV422P:;
      break;
    case V4L2_PIX_FMT_YUV420:;
      break;
    default:;
      f->fmt.pix = cam_data->pix_format;
      retval = -EINVAL;
      goto format_exit; 
    }     
  }
  if (param_old_scale) {;
    cam_data->crop.left = cam_data->crop.top = 0;
    cam_data->width=cam_data->crop.width=f->fmt.pix.width;
    cam_data->height=cam_data->crop.height=f->fmt.pix.height;
    cam_data->output.width = f->fmt.pix.width;
    cam_data->output.height = f->fmt.pix.height;    
  } else {;
    cam_data->width=cam_data->output.width = f->fmt.pix.width;
    cam_data->height=cam_data->output.height = f->fmt.pix.height;    
  }
;;;;
;
  cam_data->crt_frame_size = cam_data->width * cam_data->height;
  if (cam_data->monochrome && mono_no_color)
    if ((f->fmt.pix.pixelformat != V4L2_PIX_FMT_GREY)&&(f->fmt.pix.pixelformat != V4L2_PIX_FMT_Y16)) {;
      retval = -EINVAL;
      goto format_exit;
  }
  switch ( f->fmt.pix.pixelformat ) {
  case  V4L2_PIX_FMT_GREY:;
    cam_data->fmt = &format_grey;
    break;
  case  V4L2_PIX_FMT_Y16:;
    if (param_read16 && ( cam_data->max_Bpp > 1 ) && (cam_data->mask_Bpp & STC_TWO_BYTES_PER_PIXEL_12)) {
      setBytes = STC_TWO_BYTES_PER_PIXEL_12;
      cam_data->fmt = &format_y16;
    } else
      if (param_read16 && ( cam_data->max_Bpp > 1 ) && (cam_data->mask_Bpp & STC_TWO_BYTES_PER_PIXEL_10)) {
	setBytes = STC_TWO_BYTES_PER_PIXEL_10;
	cam_data->fmt = &format_y16;
      } else {
	cam_data->fmt = &format_grey;
      }
    break;
  case  V4L2_PIX_FMT_SRGGB8:;
    cam_data->fmt = &format_rggb;
    scalable = 0;
    break;
  case  V4L2_PIX_FMT_SGRBG8:;
    cam_data->fmt = &format_grbg;
    scalable = 0;
    break;
  case  V4L2_PIX_FMT_SRGGB10:;
    if (param_read16 && ( cam_data->max_Bpp > 1 ) && (cam_data->mask_Bpp & STC_TWO_BYTES_PER_PIXEL_10)) {
      setBytes = STC_TWO_BYTES_PER_PIXEL_10;
      cam_data->fmt = &format_rggb10;
      scalable = 0;
    } else {;
      retval = -EINVAL;
      goto format_exit;
    }
    break;
  case  V4L2_PIX_FMT_SRGGB12:;
    if (param_read16 && ( cam_data->max_Bpp > 1 ) && (cam_data->mask_Bpp & STC_TWO_BYTES_PER_PIXEL_12)) {
      setBytes = STC_TWO_BYTES_PER_PIXEL_12;
      cam_data->fmt = &format_rggb12;
      scalable = 0;
    } else {;
      retval = -EINVAL;
      goto format_exit;
    }
    break;
  case  V4L2_PIX_FMT_SRGGB16:;
    setBytes = STC_TWO_BYTES_PER_PIXEL_12;
    cam_data->fmt = &format_rggb16;
    scalable = 0;
    break;
  case  V4L2_PIX_FMT_SBGGR8:;
    cam_data->fmt = &format_bggr;
    if (param_old_scale) {
      cam_data->crop.top = 1;
      cam_data->crop.left = 1;
    } else {
      cam_data->crop.top |= 1;
      cam_data->crop.left |= 1;
    }
    scalable = 0;
;
    break;
  case V4L2_PIX_FMT_BGR24:;
    cam_data->fmt = &format_bgr24;
    if (( cam_data->max_Bpp > 1 ) && (cam_data->mask_Bpp & STC_THREE_BYTES_PER_PIXEL_RGB8)) {
      setBytes =  STC_THREE_BYTES_PER_PIXEL_RGB8;
      scalable = 0;
    }
    break;
  case V4L2_PIX_FMT_RGB24:;
    cam_data->fmt = &format_rgb24;
    if (( cam_data->max_Bpp > 1 ) && (cam_data->mask_Bpp & STC_THREE_BYTES_PER_PIXEL_RGB8)) {
      setBytes =  STC_THREE_BYTES_PER_PIXEL_RGB8;
      scalable = 0;
    }
    break;
  case V4L2_PIX_FMT_YUYV:;
    if (param_read16 && ( cam_data->max_Bpp > 1 ) && (cam_data->mask_Bpp & STC_TWO_BYTES_PER_PIXEL_12)) {
      setBytes = STC_TWO_BYTES_PER_PIXEL_12;
    } else
      if (param_read16 && ( cam_data->max_Bpp > 1 ) && (cam_data->mask_Bpp & STC_TWO_BYTES_PER_PIXEL_10)) {
	setBytes = STC_TWO_BYTES_PER_PIXEL_10;
      }
    cam_data->fmt = &format_yuyv;
    break;
  case V4L2_PIX_FMT_YUV420:;
    cam_data->fmt = &format_yuv420;
    break;
  case V4L2_PIX_FMT_YUV422P:;
    cam_data->fmt = &format_yuv422P;
    break;
  default:
    retval = -EINVAL;
  }
  if (! param_old_scale) {;
    if (scale && scalable){;
      wratio = cam_data->crop.width / cam_data->output.width; 
      if (wratio < 1) wratio = 1;
      hratio = cam_data->crop.height / cam_data->output.height;
      if (hratio < 1) hratio = 1;
      if (cam_data->crop.width != wratio * cam_data->output.width) {
	wratio = cam_data->scan_width / cam_data->output.width;
	if(wratio < 1) wratio = 1; 
	cam_data->crop.width = wratio * cam_data->output.width;
      }
      if (cam_data->crop.height != hratio * cam_data->output.height) {
	hratio = cam_data->scan_height / cam_data->output.height;
	if (hratio < 1){
	  hratio = cam_data->max_height / cam_data->output.height;
	  if (hratio < 1) hratio = 1; 
	}
	cam_data->crop.height = hratio * cam_data->output.height;
      }
    } else {
      hratio = wratio = 1;
      cam_data->crop.width = wratio * cam_data->output.width;
      cam_data->crop.height = hratio * cam_data->output.height;
    }
    if(cam_data->crop.top+cam_data->crop.height>cam_data->max_height)
      cam_data->crop.top = cam_data->max_height - cam_data->crop.height;
    if(cam_data->crop.left+cam_data->crop.width>cam_data->scan_width)
      cam_data->crop.left = cam_data->scan_width - cam_data->crop.width;
    cam_data->output.top=hratio;
    cam_data->output.left=wratio;
  } else {;
    cam_data->output.top=1;
    cam_data->output.left=1;
  }
  cam_data->pix_format = f->fmt.pix;
;;
;
;;;
  sentech_setBytes(cam_data, setBytes);
 format_exit:
  up(&cam_data->format_lock);
 done:;
  write_lock_irqsave(&cam_data->ctrl_lock, flags);
  set_controls(cam_data, GFP_KERNEL);
  write_unlock_irqrestore(&cam_data->ctrl_lock, flags);;
  lprintk(1,"%s Exiting\n",__FUNCTION__);
  return retval;
}
static void
limit (__u32* val, __u32 min, __u32 max) {
  if (*val < min) *val = min;
  if (*val > max) *val = max;
}
static int 
sentech_vidioc_try_fmt_cap(struct file *UNUSED(file), void*priv, 
			   struct v4l2_format *f){
  struct sentech_camera* cam_data = get_cam_data(priv);
  __u32 maxw, maxh;
  int depth = get_depth(f->fmt.pix.pixelformat);
  lprintk(1,"In %s\n",__FUNCTION__);
;;;
  if (cam_data->vmas) {;
    f->fmt.pix = cam_data->pix_format;
  }
  f->fmt.pix.field = V4L2_FIELD_NONE;
  maxw  = cam_data->scan_width; 
  maxh  = cam_data->max_height;;
  f->fmt.pix.colorspace = 0;
  if (!f->fmt.pix.height && !f->fmt.pix.width) {
    f->fmt.pix.height = maxh;
    f->fmt.pix.width = maxw;
  }
  if (cam_data->usb3) {
    switch( f->fmt.pix.pixelformat){
    case  V4L2_PIX_FMT_GREY:;
      break;
    case  V4L2_PIX_FMT_SGRBG8:;
      break;
    case  V4L2_PIX_FMT_RGB24:
    case  V4L2_PIX_FMT_BGR24:
      if (
	  (cam_data->mask_Bpp & STC_THREE_BYTES_PER_PIXEL_RGB8)
  	  ) {;
	break;
      };
   default:;
      f->fmt.pix = cam_data->pix_format;
      return -EINVAL;
    }
  }
  limit(&f->fmt.pix.height, 0, maxh);
  limit(&f->fmt.pix.width, 0, maxw);;
  f->fmt.pix.width &= ~0x01;
  f->fmt.pix.bytesperline = f->fmt.pix.width*depth >> 3;
  f->fmt.pix.sizeimage = 
    f->fmt.pix.height * f->fmt.pix.bytesperline;;
  if (cam_data->monochrome && mono_no_color)
    if ((f->fmt.pix.pixelformat != V4L2_PIX_FMT_GREY) && (f->fmt.pix.pixelformat != V4L2_PIX_FMT_Y16)) {;
      return -EINVAL;
    }
  if ( f->fmt.pix.pixelformat == 0) {
    switch (sentech_fps[cam_data->model_idx].native_fourcc) {
    case V4L2_PIX_FMT_GREY:
      if (grey) {;
	f->fmt.pix.pixelformat =V4L2_PIX_FMT_GREY;
      } else if (y16) {;
	f->fmt.pix.pixelformat =V4L2_PIX_FMT_Y16;
      } else {;
	f->fmt.pix.pixelformat =V4L2_PIX_FMT_YUYV;
      }
      break;
    case V4L2_PIX_FMT_SRGGB8:
      switch (default_format) {
      case DEFAULT_YUYV:;
	f->fmt.pix.pixelformat =V4L2_PIX_FMT_YUYV;
	break;
      case DEFAULT_RGGB:
	if (rggb) {;
	  f->fmt.pix.pixelformat =V4L2_PIX_FMT_SRGGB8;
	  break;
	} 
      case DEFAULT_BGR24:
      default:
	if (bgr24) {;
	  f->fmt.pix.pixelformat =V4L2_PIX_FMT_BGR24;
	} else {;
	  f->fmt.pix.pixelformat =V4L2_PIX_FMT_SRGGB8;
	}
      }
      break;
    case V4L2_PIX_FMT_SGRBG8:
      switch (default_format) {
      case DEFAULT_RGB24:
 	if (rgb24) {;
	  f->fmt.pix.pixelformat =V4L2_PIX_FMT_RGB24;
	  break;
	} 
     default:;
	f->fmt.pix.pixelformat =V4L2_PIX_FMT_SGRBG8;
	break;	
      }
      break;
    default:;
      f->fmt.pix.pixelformat =V4L2_PIX_FMT_YUYV;
    }
    depth = get_depth(f->fmt.pix.pixelformat);
    f->fmt.pix.bytesperline = f->fmt.pix.width*depth >> 3;
    f->fmt.pix.sizeimage = 
      f->fmt.pix.height * f->fmt.pix.bytesperline;;
  }
  switch( f->fmt.pix.pixelformat){
  case  V4L2_PIX_FMT_GREY:;
    if(!grey || (COLOR_NO_MONO && !cam_data->monochrome)){;
      return -EINVAL;
    }
    f->fmt.pix.priv = 0;
    f->fmt.pix.bytesperline =
      (param_bytesperline?maxw:f->fmt.pix.width)*depth >> 3;
    f->fmt.pix.sizeimage = 
      f->fmt.pix.height * f->fmt.pix.bytesperline;
    break;
  case  V4L2_PIX_FMT_Y16:;
    if(!y16 || (COLOR_NO_MONO && !cam_data->monochrome)){;
      return -EINVAL;
    }
    break;
  case  V4L2_PIX_FMT_SRGGB8:;
    if(!rggb){;
      return -EINVAL;
    }
    f->fmt.pix.bytesperline =
      (param_bytesperline?maxw:f->fmt.pix.width)*depth >> 3;
    f->fmt.pix.sizeimage = 
      f->fmt.pix.height * f->fmt.pix.bytesperline;
    break;
  case  V4L2_PIX_FMT_SGRBG8:;
    if(!rggb){;
      return -EINVAL;
    }
    f->fmt.pix.bytesperline =
      (param_bytesperline?maxw:f->fmt.pix.width)*depth >> 3;
    f->fmt.pix.sizeimage = 
      f->fmt.pix.height * f->fmt.pix.bytesperline;
    break;
  case  V4L2_PIX_FMT_SRGGB10:;
  case  V4L2_PIX_FMT_SRGGB12:;
  case  V4L2_PIX_FMT_SRGGB16:;
    if(!rggb){;
      return -EINVAL;
    }
    f->fmt.pix.bytesperline =
      (param_bytesperline?maxw:f->fmt.pix.width)*depth >> 3;
    f->fmt.pix.sizeimage = 
      f->fmt.pix.height * f->fmt.pix.bytesperline;
    break;
  case  V4L2_PIX_FMT_SBGGR8:;
    if(!bggr){;
      return -EINVAL;
    }
    limit(&f->fmt.pix.height, 0, maxh-2);
    limit(&f->fmt.pix.width, 0, maxw-2);
    f->fmt.pix.width &= ~0x01;
    f->fmt.pix.bytesperline = f->fmt.pix.width*depth >> 3;
    f->fmt.pix.sizeimage = 
      f->fmt.pix.height * f->fmt.pix.bytesperline;
    break;
  case V4L2_PIX_FMT_BGR24:;
    if(!bgr24 ||
       cam_data->fragmented){;
      return -EINVAL;
    }
    f->fmt.pix.colorspace = V4L2_COLORSPACE_SRGB;
    break;
  case V4L2_PIX_FMT_RGB24:;
    if(!rgb24 ||
       cam_data->fragmented) {;
      return -EINVAL;
    }
    if (cam_data->usb3 && ( cam_data->max_Bpp > 1 ) && (cam_data->mask_Bpp & STC_THREE_BYTES_PER_PIXEL_RGB8)) {
      if (param_variable_width) {
	f->fmt.pix.width &= ~0x03;
	f->fmt.pix.height &= ~0x03;
	f->fmt.pix.bytesperline = f->fmt.pix.width*depth >> 3;
      } else {
	f->fmt.pix.bytesperline = cam_data->scan_width*depth >> 3;
      }
      f->fmt.pix.sizeimage = 
	f->fmt.pix.height * f->fmt.pix.bytesperline;
    }
    f->fmt.pix.colorspace = V4L2_COLORSPACE_SRGB;
    break;
  case  V4L2_PIX_FMT_YUYV:;
    if(!yuyv ||
       cam_data->fragmented){;
      return -EINVAL;
    }
    break;    
  case  V4L2_PIX_FMT_YUV420:;
    if(!yuv420 ||
       cam_data->fragmented){;
      return -EINVAL;
    }
    f->fmt.pix.width &= ~0x01;
    f->fmt.pix.height &= ~0x01;
    f->fmt.pix.bytesperline = f->fmt.pix.width;
    f->fmt.pix.sizeimage = 
      f->fmt.pix.height * f->fmt.pix.bytesperline;
    f->fmt.pix.sizeimage += (f->fmt.pix.sizeimage >> 1);
    break;    
  case  V4L2_PIX_FMT_YUV422P:;
    if(!yuv422P ||
       cam_data->fragmented){;
      return -EINVAL;
    }
    f->fmt.pix.width &= ~0x01;
    f->fmt.pix.bytesperline = f->fmt.pix.width;
    f->fmt.pix.sizeimage = 
      f->fmt.pix.height * f->fmt.pix.bytesperline;
    f->fmt.pix.sizeimage += f->fmt.pix.sizeimage;
    break;    
  case  V4L2_PIX_FMT_BGR32:;
  default:;
    f->fmt.pix.pixelformat=format_grey.fourcc;
    depth = format_grey.depth;
    f->fmt.pix.priv = format_grey.priv;
    limit(&f->fmt.pix.height, 0, maxh);
    limit(&f->fmt.pix.width, 0, maxw);
    f->fmt.pix.width &= ~0x03;
    f->fmt.pix.bytesperline =
      (maxw * depth) >> 3;
    f->fmt.pix.sizeimage =
      f->fmt.pix.height * f->fmt.pix.bytesperline;;
    return -EINVAL;
  }
  f->fmt.pix.sizeimage += trace_buffers*cam_data->max_frame_size;;;
;;;
  lprintk(1,"Exiting %s\n",__FUNCTION__);
  return 0;
}
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,15)
static
int sentech_vidioc_log_status (struct file *UNUSED(file), void *priv){
  struct sentech_camera* cam_data = get_cam_data(priv);;
  if(printk_ratelimit());
;
  return cam_data->status;
}
#endif
static
int sentech_vidioc_enum_input(struct file *UNUSED(file), void *private_data,
			 struct v4l2_input *input) {
  struct sentech_camera* cam_data = get_cam_data(private_data);
  int retval = 0;;
  if(input->index){;
    return -EINVAL;
  }
  memset(input, 0, sizeof(*input));
  strlcpy((char*)input->name,cam_data->cam_name ,31);
  input->name[31]=0;
  input->type = V4L2_INPUT_TYPE_CAMERA;
  input->audioset = 0; 
  input->std = 0; 
  input->status = cam_data->status; 
;
  return retval;  
}
static int sentech_vidioc_g_input (struct file *UNUSED(file), void *UNUSED(private_data), 
			    unsigned int *index){;
  *index = 0;;
  return 0;
}
static int sentech_vidioc_s_input (struct file *UNUSED(file), void *private_data, 
		       unsigned int index){
  struct sentech_camera* cam_data = get_cam_data(private_data);;
  if(index){;
    return -EINVAL;
  }
  if (cam_data->status){;
    return -EIO;
  }
  return 0; 
}
static 
int sentech_vidioc_querymenu (struct file *UNUSED(file), void *priv, 
			      struct v4l2_querymenu * qc){
  struct sentech_camera* cam_data = get_cam_data(priv);
  __u32 id = qc->id;
  int retval = 0;
  int i = cam_data->model_idx;
  char* fmt, f0, f1, f2, f3, g0='X', g1='X', g2='X', g3='X';
;
  id &= ~V4L2_CTRL_FLAG_NEXT_CTRL; 
  switch(id) {
  case V4L2_CID_STC_PRODUCT:
    switch(qc->index) {
    case 0: strncpy((char*)qc->name,sentech_fps[i].name,STC_CTRL_NAME_LEN); break;
    case 1: strncpy((char*)qc->name,kobject_name(&(cam_data->dev->kobj)),STC_CTRL_NAME_LEN); break;
    case 2:
#if  LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,12)
	if(cam_data->udev->serial != NULL)
	  strncpy((char*)qc->name,cam_data->udev->serial,STC_CTRL_NAME_LEN);
	else
#endif
	  strncpy((char*)qc->name,"No USB Serial Number",STC_CTRL_NAME_LEN);
	break;
    case 3:
#if  LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,12)
        if(cam_data->udev->product != NULL)
          strncpy((char*)qc->name,cam_data->udev->product,STC_CTRL_NAME_LEN);
	else
#endif
	  strncpy((char*)qc->name,"No Product String",STC_CTRL_NAME_LEN);
	break;
    case 4: 
#if  LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,12)
        if(cam_data->udev->manufacturer != NULL)
          strncpy((char*)qc->name,cam_data->udev->manufacturer,STC_CTRL_NAME_LEN);
        else
#endif
	  strncpy((char*)qc->name,"No Manufacturer String",STC_CTRL_NAME_LEN);
	break;
    case 5:
      snprintf((char*)qc->name, STC_CTRL_NAME_LEN, "Product ID: %d", cam_data->model);
        break;
    case 6:
      snprintf((char*)qc->name, STC_CTRL_NAME_LEN, "Firmware: %d / FPGA: %d", cam_data->payload11.Firmware_version_le,  cam_data->payload11.Camera_type_le);
        break;
    case 7:
      fmt = (char*)&(cam_data->pix_format.pixelformat);
      f0 = fmt[0];
      f1 = fmt[1];
      f2 = fmt[2];
      f3 = fmt[3];
      if(cam_data->fmt != 0){
	fmt = (char*)&(cam_data->fmt->fourcc);
	g0 = fmt[0];
	g1 = fmt[1];
	g2 = fmt[2];
	g3 = fmt[3];
      }
      snprintf((char*)qc->name, STC_CTRL_NAME_LEN, "Crt FOURCC: %c%c%c%c %c%c%c%c",f0,f1,f2,f3,g0,g1,g2,g3);
      break;
    default:
      dprintk (0, "V4L2_CID_STC_PRODUCT invalid %d\n", qc->index);
      retval = -EINVAL;
    }
    goto done;
    break;
  case V4L2_CID_STC_TRIGGER_MODE:
    switch(qc->index) {
    case STC_TRIGGER_MODE_NORMAL: strncpy((char*)qc->name,"NORMAL",32); break;
    case STC_TRIGGER_MODE_TRIGGER: strncpy((char*)qc->name,"TRIGGER",32); break;
    case STC_TRIGGER_MODE_READOUT: strncpy((char*)qc->name,"ReadOut",32); break;
    case 3: strncpy((char*)qc->name,"ReadOut",32); break;
    default:
      dprintk (0, "V4L2_CID_STC_TRIGGER_MODE invalid %d\n", qc->index);
      retval = -EINVAL;
    }
    goto done;
    break;
  case V4L2_CID_STC_TRIGGER_SOURCE:
    switch(qc->index) {
    case STC_TRIGGER_SOURCE_IGNORE: strncpy((char*)qc->name,"IGNORE",32); break;
    case STC_TRIGGER_SOURCE_SOFTWARE: strncpy((char*)qc->name,"SOFTWARE",32); break;
    case STC_TRIGGER_SOURCE_HARDWARE: strncpy((char*)qc->name,"HARDWARE",32); break;
    case 3: strncpy((char*)qc->name,"IGNORE",32); break;
    default:
      dprintk (0, "V4L2_CID_STC_TRIGGER_SOURCE invalid %d\n", qc->index);
      retval = -EINVAL;
    }
    goto done;
    break;
  case V4L2_CID_STC_TRIGGER_POLARITY_0:
  case V4L2_CID_STC_TRIGGER_POLARITY_1:
  case V4L2_CID_STC_TRIGGER_POLARITY_2:
  case V4L2_CID_STC_TRIGGER_POLARITY_3:
    switch(qc->index) {
    case 0: strncpy((char*)qc->name,"NORMAL",32); break;
    case 1: strncpy((char*)qc->name,"REVERSE",32); break;
    default:
      dprintk (0, "V4L2_CID_STC_TRIGGER_POLARITY invalid %d\n", qc->index);
      retval = -EINVAL;
    }
    goto done;
    break;
  case V4L2_CID_STC_IO0_FUNCTION:
  case V4L2_CID_STC_IO1_FUNCTION:
  case V4L2_CID_STC_IO2_FUNCTION:
  case V4L2_CID_STC_IO3_FUNCTION:
    switch(qc->index) {
    case IO_O_HIZ: strncpy((char*)qc->name,"OUTPUT Hi-Z",32); break;
    case IO_O_CT: strncpy((char*)qc->name,"OUTPUT common terminals",32); break;
    case IO_O_TRIG: strncpy((char*)qc->name,"OUTPUT Trigger Out",32); break;
    case IO_O_TRIG_T: strncpy((char*)qc->name,"OUTPUT Trigger through Out",32); break;
    case IO_O_EXP_END: strncpy((char*)qc->name,"OUTPUT Exposure End",32); break;
    case IO_O_TRAN_END: strncpy((char*)qc->name,"OUTPUT Transfer End",32); break;
    case IO_O_STROBE: strncpy((char*)qc->name,"OUTPUT Strobe out",32); break;
    case IO_O_EXP: strncpy((char*)qc->name,"OUTPUT Exposure time",32); break;
    case IO_O_TRIG_V: strncpy((char*)qc->name,"OUTPUT Triger Valid",32); break;
    case IO_I_IGN: strncpy((char*)qc->name,"INPUT Ignore",32); break;
    case IO_I_CT: strncpy((char*)qc->name,"INPUT common terminals",32); break;
    case IO_I_TRIG: strncpy((char*)qc->name,"INPUT Trigger In",32); break;
    case IO_I_READ_OUT: strncpy((char*)qc->name,"INPUT Read Out In",32); break;
    case IO_I_SUB_TRIG: strncpy((char*)qc->name,"INPUT Sub-Trigger In",32); break;
    default:
      dprintk (0, "V4L2_CID_STC_IO_FUNCTION invalid %d\n", qc->index);
      retval = -EINVAL;
    }
    goto done;
    break;
  case V4L2_CID_STC_IO_STATUS_0:
  case V4L2_CID_STC_IO_STATUS_1:
  case V4L2_CID_STC_IO_STATUS_2:
  case V4L2_CID_STC_IO_STATUS_3:
    switch(qc->index) {
    case 0: strncpy((char*)qc->name,"LOW",32); break;
    case 1: strncpy((char*)qc->name,"HIGH",32); break;
    default:
      dprintk (0, "V4L2_CID_STC_IO_STATUS invalid %d\n", qc->index);
      retval = -EINVAL;
    }
    goto done;
    break;
  case V4L2_CID_STC_FPS:
    switch(qc->index) {
    case 0: 
      sprintf((char*)qc->name,"%d/%d",sentech_fps[i].fps[2-qc->index].denominator,sentech_fps[i].fps[2-qc->index].numerator); break;
    case 1:
      sprintf((char*)qc->name,"%d/%d",sentech_fps[i].fps[2-qc->index].denominator,sentech_fps[i].fps[2-qc->index].numerator); break;
    case 2: 
      sprintf((char*)qc->name,"%d/%d",sentech_fps[i].fps[2-qc->index].denominator,sentech_fps[i].fps[2-qc->index].numerator); break;
    default:
      dprintk (0, "V4L2_CID_STC_FPS invalid %d\n", qc->index);
      retval = -EINVAL;
    }
    goto done;
    break;
  case V4L2_CID_STC_HEIGHT:
    if ((qc->index >= 0) && (qc->index < ARRAY_SIZE(sentech_height_params)))
      strncpy((char*)qc->name,sentech_height_params[qc->index].name,32);
    else {
      dprintk (0, "V4L2_CID_STC_HEIGHT invalid %d\n", qc->index);
      retval = -EINVAL;
    }
    goto done;
    break;
  case V4L2_CID_STC_LED:
    switch(qc->index) {
    case 0: strncpy((char*)qc->name,"NONE",32); break;
    case 1: strncpy((char*)qc->name,"GREEN",32); break;
    case 2: strncpy((char*)qc->name,"RED",32); break;
    case 3: strncpy((char*)qc->name,"GREEN + RED",32); break;
    default:
      dprintk (0, "V4L2_CID_STC_LED invalid %d\n", qc->index);
      retval = -EINVAL;
    }
    goto done;
    break;
  case V4L2_CID_STC_RO_SEL:
    switch(qc->index) {
    case 0: strncpy((char*)qc->name,"SOFTWARE",32); break;
    case 1: strncpy((char*)qc->name,"HARDWARE",32); break;
    default:
      dprintk (0, "V4L2_CID_STC_RO_SEL invalid %d\n", qc->index);
      retval = -EINVAL;
    }
    goto done;
    break;
  case V4L2_CID_STC_EXP_MODE:
    switch(qc->index) {
    case 0: strncpy((char*)qc->name,"EDGE PRESENCE",32); break;
    case 1: strncpy((char*)qc->name,"PULSE WIDTH",32); break;
    case 2: strncpy((char*)qc->name,"START / STOP",32); break;
    default:
      dprintk (0, "V4L2_CID_STC_EXP_MODE invalid %d\n", qc->index);
      retval = -EINVAL;
    }
    goto done;
    break;
  case V4L2_CID_STC_EXP_D1:
    switch(qc->index) {
    case 0: strncpy((char*)qc->name,"IMMEDIATE",32); break;
    case 1: strncpy((char*)qc->name,"AT NEXT HD",32); break;
    default:
      dprintk (0, "V4L2_CID_STC_EXP_D1 invalid %d\n", qc->index);
      retval = -EINVAL;
    }
    goto done;
    break;
  case V4L2_CID_STC_EXP_D2:
    switch(qc->index) {
    case 0: strncpy((char*)qc->name,"IMMEDIATE",32); break;
    case 1: strncpy((char*)qc->name,"BY EXPOSURE START AFTER RO",32); break;
    default:
      dprintk (0, "V4L2_CID_STC_EXP_D2 invalid %d\n", qc->index);
      retval = -EINVAL;
    }
    goto done;
    break;
  default:
    dprintk (0, "query_menu invalid ctl=%d(%xh %xh) %d\n", qc->id, qc->id, id, qc->index);
    retval = -EINVAL;
    goto done;
  }
  if(qc->index>2) {
    dprintk (0, "query_menu unknown ctl=%d(%xh %xh) invalid %d\n", qc->id, qc->id, id, qc->index);
    retval = -EINVAL;
    goto done;
  }
  sprintf((char*)qc->name,"%d",qc->index);
 done:
  return retval;
}
static 
int sentech_vidioc_queryctrl (struct file *UNUSED(file), void *priv, 
				      struct v4l2_queryctrl * qc){
  struct sentech_camera* cam_data = get_cam_data(priv);
  int retval = 0;
  unsigned int i;
  int next=0, found=0;;
  if(qc->id & V4L2_CTRL_FLAG_NEXT_CTRL) {
    next = 1;;
    qc->id &= ~V4L2_CTRL_FLAG_NEXT_CTRL;;
  }else{;
  }
  for (i = 0; i < ARRAY_SIZE(sentech_qctrl); i++) {
    if(next && found) {;
      memcpy(qc, &(cam_data->qctrl[i]), sizeof(*qc));
      goto done;
    }
    if ( (qc->id <= sentech_qctrl[i].id)) {
      if(next) {
	found = 1;;
	if(qc->id == sentech_qctrl[i].id) continue;
	else {
	  memcpy(qc, &(cam_data->qctrl[i]), sizeof(*qc));
	  goto done;
	}
      }else {
	if(qc->id == sentech_qctrl[i].id) {;
	  memcpy(qc, &(cam_data->qctrl[i]), sizeof(*qc));
	  goto done;
	}
      }
    }
  }
  retval = -EINVAL;
 done:;
  return retval;
}
static int sentech_vidioc_s_ctrl (struct file *UNUSED(file), void *priv, 
			   struct v4l2_control * ctrl){
  struct sentech_camera* cam_data = get_cam_data(priv);
  unsigned int i;
  unsigned long flags;
  int retval = 0;
  __u32 shutters;
  __u16 clock;
  int max_exp;
  int lockit = 1;;
  lprintk(1,"In %s for %d\n",__FUNCTION__, ctrl->id);
  for (i = 0; i < ARRAY_SIZE(sentech_qctrl); i++)
    if (ctrl->id == sentech_qctrl[i].id) {
      if (ctrl->value <
	  cam_data->qctrl[i].minimum
	  || ctrl->value >
	  cam_data->qctrl[i].maximum) {;
	retval = -ERANGE;
	goto done;
      }
      cam_data->qctl_regs[i]=ctrl->value;;
      goto vidioc_s_ctrl;
    }
;
  retval = -EINVAL;
  goto done;
 vidioc_s_ctrl:;
  switch (ctrl->id) {
  case (V4L2_CID_STC_VERTICAL_SCAN):
    retval = -EINVAL;
    goto done;
  case (V4L2_CID_STC_SELECTED_FOURCC):
    retval = -EINVAL;
    goto done;
  case (V4L2_CID_STC_HORIZONTAL_SIZE_CAPTURE):
    if (cam_data->usb3 && sentech_support_AOI(cam_data)) {
      int width = ctrl->value & ~3;
      if (width + cam_data->scan_horizontal_offset > cam_data->max_width) {
	retval = -EINVAL;
	goto done;
      }
      cam_data->payload12.Horizontal_size_mode_le = __cpu_to_le32(width);
      cam_data->payload12.Scan_mode_le = __cpu_to_le16(STC_AOI_SCREEN);
      setFlag(&cam_data->modif_flag12); 
    }
    goto done;
  case (V4L2_CID_STC_VERTICAL_SIZE_CAPTURE):
    if (cam_data->usb3 && sentech_support_AOI(cam_data)) {
      int height = ctrl->value & ~3;
      if (height + cam_data->scan_vertical_offset > cam_data->max_height) {
	retval = -EINVAL;
	goto done;
      }
      cam_data->payload12.Vertical_size_mode_le = __cpu_to_le32(height);
      cam_data->payload12.Scan_mode_le = __cpu_to_le16(STC_AOI_SCREEN);
      setFlag(&cam_data->modif_flag12); 
    }
    goto done;
  case (V4L2_CID_STC_HORIZONTAL_OFFSET_CAPTURE):
    if (cam_data->usb3 && sentech_support_AOI(cam_data)) {
      int h_off = ctrl->value & ~1;
      if (h_off + cam_data->scan_width > cam_data->max_width) {
	retval = -EINVAL;
	goto done;
      }
      cam_data->payload12.Horizontal_offset_le = __cpu_to_le32(h_off);
      cam_data->payload12.Scan_mode_le = __cpu_to_le16(STC_AOI_SCREEN);
      setFlag(&cam_data->modif_flag12); 
    }
    goto done;
  case (V4L2_CID_STC_VERTICAL_OFFSET_CAPTURE):
    if (cam_data->usb3 && sentech_support_AOI(cam_data)) {
      int v_off = ctrl->value & ~1;
      if (v_off + cam_data->scan_height > cam_data->max_height){
	retval = -EINVAL;
	goto done;
      }
      cam_data->payload12.Vertical_offset_le = __cpu_to_le32(v_off);
      cam_data->payload12.Scan_mode_le = __cpu_to_le16(STC_AOI_SCREEN);
      setFlag(&cam_data->modif_flag12); 
    }
    goto done;
  case (V4L2_CID_BRIGHTNESS):
    cam_data->brightness = ctrl->value;
    retval = 0;
    goto done;
  case (V4L2_CID_GAMMA):
  case (V4L2_CID_STC_GAMMA):
    if ((cam_data->usb3)&&
       ((cam_data->auto_gain_fpga_can)||(cam_data->auto_shutters_fpga_can))) {
	cam_data->payload80.Gamma = ctrl->value;
	setFlag(&cam_data->modif_flag_80_4D); 
	retval = 0;
	goto done;
    }else{
      cam_data->gamma = ctrl->value;
    }
    for(i = 0; i < ARRAY_SIZE(sentech_qctrl); i++) {
      if (sentech_qctrl[i].id == V4L2_CID_GAMMA)
	cam_data->qctl_regs[i] = ctrl->value;
      if (sentech_qctrl[i].id == V4L2_CID_STC_GAMMA)
	cam_data->qctl_regs[i] = ctrl->value;
    }
    retval = 0;
    goto done;
  case (V4L2_CID_SATURATION):
    setSaturations(cam_data->saturation = ctrl->value);
    setSaturations_16(cam_data->saturation = ctrl->value);
    retval = 0;
    goto done;
  case (V4L2_CID_HFLIP):
    if (cam_data->h_flip_fpga_can) {
      cam_data->h_flip_fpga = ctrl->value;
      if (cam_data->h_flip_fpga)
	cam_data->payload80.Image_Flip = cam_data->payload80.Image_Flip | 0x01;
      else
	cam_data->payload80.Image_Flip = cam_data->payload80.Image_Flip & ~0x01;
      setFlag(&cam_data->modif_flag_80_3F); 
      retval = 0;
    } else {
      if (cam_data->streaming && cam_data->native) {
	retval = -EBUSY;
      } else {
	if (! cam_data->usb3) cam_data->hflip = ctrl->value;
	retval = 0;
      }
    }
    goto done;
    break;
  case (V4L2_CID_VFLIP):
    if (cam_data->v_flip_fpga_can) {
      cam_data->v_flip_fpga = ctrl->value;
      if (cam_data->v_flip_fpga)
	cam_data->payload80.Image_Flip = cam_data->payload80.Image_Flip | 0x02;
      else
	cam_data->payload80.Image_Flip = cam_data->payload80.Image_Flip & ~0x02;
      setFlag(&cam_data->modif_flag_80_3F); 
      retval = 0;
    }
    goto done;
    break;
  case (V4L2_CID_STC_EXPOSURE_US): 
    ctrl->value = useconds_2_exposure(cam_data, ctrl->value);
  case (V4L2_CID_CONTRAST): 
  case (V4L2_CID_EXPOSURE):
    for(i = 0; i < ARRAY_SIZE(sentech_qctrl); i++) {
      if (sentech_qctrl[i].id == V4L2_CID_CONTRAST)
	cam_data->qctl_regs[i] = ctrl->value;
      if (sentech_qctrl[i].id == V4L2_CID_EXPOSURE)
	cam_data->qctl_regs[i] = ctrl->value;
      if (sentech_qctrl[i].id == V4L2_CID_STC_EXPOSURE_US)
	cam_data->qctl_regs[i] = exposure_2_useconds(cam_data, ctrl->value);
    }
    clock = __le16_to_cpu(cam_data->payload12.one_line_clock_numbers_le);
    shutters = ctrl->value;
    if ( ! dynamic_scale ) {
      max_exp = get_max_exposure(cam_data);
      shutters = shutters * max_exp / MAX_EXP_CTRL;
    }
    cam_data->shutters = shutters;
    if (0) {
      set_crt_exposure(cam_data, shutters);
      setFlag(&cam_data->modif_flag21);
    } else {
      testAndClear(&cam_data->modif_flag21);
      set_crt_exposure_async(cam_data, shutters, GFP_KERNEL);
    }
    retval = 0;
    goto done;
#ifdef  SENTECH_AUTO_AWB_CIDS
  case (V4L2_CID_STC_AUTO_WHITE_BALANCE):
#endif
  case (V4L2_CID_AUTO_WHITE_BALANCE):
    if (cam_data->monochrome) {
      if (ctrl->value) {;
	retval = -EINVAL;
	goto done;
      } else {
	retval = 0;
#ifdef SENTECH_AWB
	if (cam_data->awb_object)
	  Sentech_SetAutoAWB(cam_data->awb_object, STCAM_WB_OFF);
#endif
	goto done;
      }
    }
    cam_data->auto_white_balance = ctrl->value;
#ifdef SENTECH_AWB
    for (i = 0; i < ARRAY_SIZE(sentech_qctrl); i++) {
      if (sentech_qctrl[i].id == V4L2_CID_STC_GREEN) {;
	cam_data->green = cam_data->qctl_regs[i] = cam_data->qctrl[i].default_value;;
      }
    }
    Sentech_Set_Gain(cam_data->awb_object);
    if (cam_data->awb_object)
      Sentech_SetAutoAWB(cam_data->awb_object, (ctrl->value)?STCAM_WB_FULLAUTO:STCAM_WB_MANUAL);
#endif
    retval = 0;
    goto done;
  case (V4L2_CID_DO_WHITE_BALANCE):
    if (cam_data->monochrome) {;
      retval = -EINVAL;
      goto done;
    }
    cam_data->do_white_balance = 1;
#ifdef SENTECH_AWB
    if (cam_data->awb_object)
      Sentech_SetAutoAWB(cam_data->awb_object, STCAM_WB_ONESHOT);
#endif
    retval = 0;
    goto done;
  case (V4L2_CID_STC_AUTOSHUTTERS):;
    if (cam_data->auto_shutters_fpga_can) {
      cam_data->auto_shutters = ctrl->value;
      retval = 0;
      if (cam_data->auto_shutters){
	cam_data->payload80.AGC_mode = cam_data->payload80.AGC_mode | 0x01;
      } else {
	cam_data->payload80.AGC_mode = cam_data->payload80.AGC_mode & ~0x01;
      };
      setFlag(&cam_data->modif_flag_80_49); 
      goto done;
    }
    retval = -EINVAL;
    goto done;
#ifdef  SENTECH_AUTO_AWB_CIDS
  case (V4L2_CID_STC_AUTOGAIN):
#endif
  case (V4L2_CID_AUTOGAIN):;
    cam_data->auto_gain = ctrl->value;
    retval = 0;
    if (cam_data->usb3) {
      if (cam_data->auto_gain_fpga_can) {
	if (cam_data->auto_gain) {
	  cam_data->payload80.AGC_mode = cam_data->payload80.AGC_mode | 0x02;
	} else {
	  cam_data->payload80.AGC_mode = cam_data->payload80.AGC_mode & ~0x02;
	}
	setFlag(&cam_data->modif_flag_80_49); 
	goto done;
      }
    }
#ifdef SENTECH_AWB
    if (cam_data->awb_object) {;
      Sentech_SetAutoGain(cam_data->awb_object,
			 (ctrl->value)?
			 (STCAM_AGC_ON|STCAM_AUTO_SHUTTER_ON):
			 STCAM_AGC_OFF);
      Sentech_SetAutoExposure(cam_data->awb_object,
			 (ctrl->value)?
			 (STCAM_AGC_ON|STCAM_AUTO_SHUTTER_ON):
			 STCAM_AGC_OFF);
    }
#endif
    goto done;
  case (V4L2_CID_GAIN):
    cam_data->gain = ctrl->value;;
    if (0) {
      cam_data->payload21.Gain_le = __cpu_to_le16(ctrl->value);
      setFlag(&cam_data->modif_flag21);
    } else {
      testAndClear(&cam_data->modif_flag21);
      set_crt_gain_async(cam_data, ctrl->value, GFP_KERNEL);
    }
    retval = 0;
    goto done;
  case (V4L2_CID_STC_FPS): 
    cam_data->payload22.Clock_mode_le = __cpu_to_le32(2-ctrl->value);
    setFlag(&cam_data->modif_flag22);;
    retval = 0;
    goto done;
  case (V4L2_CID_STC_HEIGHT): 
    retval = 0;
    {
      __u16 val = 0xff;
      if ((ctrl->value >= 0) && (ctrl->value < ARRAY_SIZE(sentech_height_params))) {
	val = sentech_height_params[ctrl->value].hardware_val;
	cam_data->payload12.Scan_mode_le = __cpu_to_le16(val);
	setFlag(&cam_data->modif_flag12);
      } else retval = -EINVAL;;
    }
   goto done;
  case (V4L2_CID_STC_RED): 
  case (V4L2_CID_RED_BALANCE):
    cam_data->red = ctrl->value;
    for(i = 0; i < ARRAY_SIZE(sentech_qctrl); i++) {
      if (sentech_qctrl[i].id == V4L2_CID_RED_BALANCE)
	cam_data->qctl_regs[i] = ctrl->value;
      if (sentech_qctrl[i].id == V4L2_CID_STC_RED)
	cam_data->qctl_regs[i] = ctrl->value;
    }
#ifdef SENTECH_AWB
    if(!cam_data->auto_white_balance)Sentech_Set_Gain(cam_data->awb_object);
#endif
    retval = 0;
    goto done;
  case (V4L2_CID_STC_GREEN): 
    cam_data->green = ctrl->value;
#ifdef SENTECH_AWB
    if(!cam_data->auto_white_balance)Sentech_Set_Gain(cam_data->awb_object);
#endif
    retval = 0;
    goto done;
  case (V4L2_CID_STC_BLUE): 
  case (V4L2_CID_BLUE_BALANCE):
    cam_data->blue = ctrl->value;
    for(i = 0; i < ARRAY_SIZE(sentech_qctrl); i++) {
      if (sentech_qctrl[i].id == V4L2_CID_BLUE_BALANCE)
	cam_data->qctl_regs[i] = ctrl->value;
      if (sentech_qctrl[i].id == V4L2_CID_STC_BLUE)
	cam_data->qctl_regs[i] = ctrl->value;
    }
#ifdef SENTECH_AWB
    if(!cam_data->auto_white_balance)Sentech_Set_Gain(cam_data->awb_object);
#endif
    retval = 0;
    goto done;
  case (V4L2_CID_STC_TARGET):;
    cam_data->exposure_target = ctrl->value;
    if (cam_data->usb3) {
      if ((cam_data->auto_gain_fpga_can)||(cam_data->auto_shutters_fpga_can)) {
	cam_data->payload80.ACL_Target_brightness = ctrl->value;
	setFlag(&cam_data->modif_flag_80_4A); 
	retval = 0;
	goto done;
      }
    }
#ifdef SENTECH_AWB
    {
      BYTE target[]= {ctrl->value,5,5};
      if(cam_data->awb_object)
	Sentech_CLuminance_SetTarget(cam_data->awb_object,
				     target);
    }
#endif
    retval = 0;
    goto done;
  case (V4L2_CID_STC_TRIGGER):
    retval = sentech_trigger(cam_data, GFP_KERNEL); 
    goto done;
  case (V4L2_CID_STC_TRIGGER_MODE):
    if(((cam_data->payload80.Trigger_Mode & STC_FPGA_TRIGGER_MODE_MASK)>>STC_FPGA_TRIGGER_MODE_SHIFT) !=
       (ctrl->value))
      setFlag(&cam_data->modif_flag_80_30);
    cam_data->payload80.Trigger_Mode &= (~STC_FPGA_TRIGGER_MODE_MASK);
    cam_data->payload80.Trigger_Mode |= (ctrl->value<<STC_FPGA_TRIGGER_MODE_SHIFT);
    retval = 0;
    goto done;
  case (V4L2_CID_STC_TRIGGER_SOURCE):
    if(((cam_data->payload80.Trigger_Mode & STC_FPGA_TRIGGER_SRC_MASK)>>STC_FPGA_TRIGGER_SRC_SHIFT) !=
       (ctrl->value))
      setFlag(&cam_data->modif_flag_80_30);
    cam_data->payload80.Trigger_Mode &= (~STC_FPGA_TRIGGER_SRC_MASK);
    cam_data->payload80.Trigger_Mode |= (ctrl->value<<STC_FPGA_TRIGGER_SRC_SHIFT);
    retval = 0;
    goto done;
  case (V4L2_CID_STC_HEADER):
    setFlag(&cam_data->modif_flag_80_30);
    cam_data->payload80.Trigger_Mode &= (~STC_FPGA_HEADER_MASK);
    cam_data->payload80.Trigger_Mode |= (ctrl->value<<STC_FPGA_HEADER_SHIFT);
    retval = 0;
    goto done;
  case (V4L2_CID_STC_TRIGGER_POLARITY_0):
  case (V4L2_CID_STC_TRIGGER_POLARITY_1):
  case (V4L2_CID_STC_TRIGGER_POLARITY_2):
  case (V4L2_CID_STC_TRIGGER_POLARITY_3):
    {
      int id = ctrl->id - (V4L2_CID_STC_TRIGGER_POLARITY_0);
      __u8 reversing = cam_data->state80.IO_Reversing;;
      reversing &= ~(1<<id);
      if (ctrl->value) ctrl->value = 1;
      reversing |= (ctrl->value << id);;
      cam_data->payload80.IO_Reversing = reversing;
    }
    sentech_fpga_set_lock(cam_data, GFP_KERNEL, 0x01, 1, lockit); 
    retval = 0;
    goto done;
  case (V4L2_CID_STC_IO_STATUS_0):
  case (V4L2_CID_STC_IO_STATUS_1):
  case (V4L2_CID_STC_IO_STATUS_2):
  case (V4L2_CID_STC_IO_STATUS_3):
    {
      int id = ctrl->id - (V4L2_CID_STC_IO_STATUS_0);
      __u8 status = cam_data->payload80.IO_Status;;
      status &= ~(1<<id);
      status |= (ctrl->value << id);;
      cam_data->payload80.IO_Status = status;
    }
    sentech_fpga_set_lock(cam_data, GFP_KERNEL, 0x02, 1, lockit); 
    retval = 0;
    goto done;
  case (V4L2_CID_STC_IO0_FUNCTION):
  case (V4L2_CID_STC_IO1_FUNCTION):
  case (V4L2_CID_STC_IO2_FUNCTION):
  case (V4L2_CID_STC_IO3_FUNCTION):
    {
      int id = ctrl->id - (V4L2_CID_STC_IO0_FUNCTION);
      __u8 dir_out = (ctrl->value < IO_I_IGN)?1:0;
      __u8 old_dir = (cam_data->payload80.IO_Directions & (1<<id))?1:0;;
      if (1 || (old_dir ^ dir_out)) {
	cam_data->payload80.IO_Directions &= ~(1<<id);
	cam_data->payload80.IO_Directions |= (dir_out<<id);;
	sentech_fpga_set_lock(cam_data, GFP_KERNEL, 0x00, 1, lockit);
      }
      if ( 1 || ((&cam_data->payload80.IO0_change)[id<<1] != io_modes[ctrl->value])){
	(&cam_data->payload80.IO0_change)[id<<1] = io_modes[ctrl->value];
;
	sentech_fpga_set_lock(cam_data, GFP_KERNEL, 0x04+id, 1, lockit);
      } else {;
      }
    }
    retval = 0;
    goto done;
  case (V4L2_CID_STC_LED):
    cam_data->payload80.LED = ctrl->value;
    setFlag(&cam_data->modif_flag_80_0F);
    retval = 0;
    goto done;
  case (V4L2_CID_STC_DIGITAL_GAIN):;
    cam_data->payload80.Digital_gain = ctrl->value;
    setFlag(&cam_data->modif_flag_80_14);
    retval = 0;;
    goto done;
  case (V4L2_CID_STC_RO_SEL):
    cam_data->payload80.Trigger_Mode &= (~STC_FPGA_TRIGGER_RO_SEL_MASK);
    cam_data->payload80.Trigger_Mode |= (ctrl->value<<STC_FPGA_TRIGGER_RO_SEL_SHIFT);
    setFlag(&cam_data->modif_flag_80_30);
    retval = 0;;
    goto done;
  case (V4L2_CID_STC_EXP_MODE):
    cam_data->payload80.Trigger_Mode &= (~STC_FPGA_TRIGGER_EXP_MODE_MASK);
    cam_data->payload80.Trigger_Mode |= (ctrl->value<<STC_FPGA_TRIGGER_EXP_MODE_SHIFT);
    setFlag(&cam_data->modif_flag_80_30);
    retval = 0;;
    goto done;
  case (V4L2_CID_STC_EXP_D1):
    cam_data->payload80.Trigger_Mode &= (~STC_FPGA_TRIGGER_EXP_D1_MASK);
    cam_data->payload80.Trigger_Mode |= (ctrl->value<<STC_FPGA_TRIGGER_EXP_D1_SHIFT);
    setFlag(&cam_data->modif_flag_80_30);
    retval = 0;;
    goto done;
  case (V4L2_CID_STC_EXP_D2):
    cam_data->payload80.Trigger_Mode &= (~STC_FPGA_TRIGGER_EXP_D2_MASK);
    cam_data->payload80.Trigger_Mode |= (ctrl->value<<STC_FPGA_TRIGGER_EXP_D2_SHIFT);
    setFlag(&cam_data->modif_flag_80_30);
    retval = 0;;
    goto done;
  case (V4L2_CID_STC_REFRESH):
    cam_data->payload80.Refresh = 0;
    cam_data->payload80.Refresh |= (1<<STC_FPGA_REFRESH_SHIFT);
    sentech_fpga_set_lock(cam_data, GFP_ATOMIC, 0x50, 2, lockit);
    retval = 0;;
    goto done;
  case (V4L2_CID_STC_CNT_RST):
    cam_data->payload80.Refresh = 0;
    cam_data->payload80.Refresh |= (1<<STC_FPGA_CNT_RST_SHIFT);
    sentech_fpga_set_lock(cam_data, GFP_ATOMIC, 0x50, 2, lockit);
    retval = 0;;
    goto done;
  case (V4L2_CID_STC_RO):
    cam_data->payload80.Refresh = 0;
    cam_data->payload80.Refresh |= (1<<STC_FPGA_RO_SHIFT);;
    sentech_fpga_set_lock(cam_data, GFP_ATOMIC, 0x50, 2, lockit);
    retval = 0;
    goto done;
  case (V4L2_CID_STC_SUB_TRG):
    cam_data->payload80.Refresh = 0;
    cam_data->payload80.Refresh |= (1<<STC_FPGA_SUB_TRG_SHIFT);
    sentech_fpga_set_lock(cam_data, GFP_ATOMIC, 0x50, 2, lockit);;
    retval = 0;
    goto done;
  case (V4L2_CID_STC_D_TRG_EXP):
    cam_data->payload80.EXP_delay_TRG = ctrl->value;
    sentech_fpga_set_lock(cam_data, GFP_ATOMIC, 0x32, 2, lockit);
    retval = 0;;
    goto done;
  case (V4L2_CID_STC_D_TRG_ST):
    cam_data->payload80.Strobe_start_delay_TRG = ctrl->value;
    sentech_fpga_set_lock(cam_data, GFP_ATOMIC, 0x34, 2, lockit);
    retval = 0;
    goto done;
  case (V4L2_CID_STC_D_TRG_END):
    cam_data->payload80.Strobe_end_delay_TRG = ctrl->value;
    sentech_fpga_set_lock(cam_data, GFP_ATOMIC, 0x36, 2, lockit);
    retval = 0;
    goto done;
  case (V4L2_CID_STC_D_TRG_I_O):
    cam_data->payload80.TRG_delay_TRG = ctrl->value;
    sentech_fpga_set_lock(cam_data, GFP_ATOMIC, 0x38, 2, lockit);
    retval = 0;;
    goto done;
  case (V4L2_CID_STC_W_TRG_OUT):
    cam_data->payload80.TRG_width_TRG = ctrl->value;
    sentech_fpga_set_lock(cam_data, GFP_ATOMIC, 0x3A, 2, lockit);
    retval = 0;;
    goto done;
  case (V4L2_CID_STC_D_RO):;
    cam_data->payload80.Readout_delay = ctrl->value;
    sentech_fpga_set_lock(cam_data, GFP_ATOMIC, 0x3C, 2, lockit);
    retval = 0;;
    goto done;
  case (V4L2_CID_STC_CLEAR_STATE):;
    retval = sentech_drop_images(cam_data);
    goto done;
#ifndef L_BOUNDS_ANY
  case (V4L2_CID_STC_GAIN_MIN):;
    if (! cam_data->usb3 || ! cam_data->auto_gain_fpga_can) {
      retval = -EINVAL;
      goto done;
    }
    cam_data->payload80.Minimum_gain_AGC = ctrl->value & 0x00FF;
    setFlag(&cam_data->modif_flag_80_4C); 
    goto done;
  case (V4L2_CID_STC_GAIN_MAX):;
    if ( ! cam_data->usb3 || ! cam_data->auto_gain_fpga_can) {;
      retval = -EINVAL;
      goto done;
    }
    cam_data->payload80.Maximum_gain_AGC = ctrl->value & 0x00FF;
    setFlag(&cam_data->modif_flag_80_4B); 
    goto done;
  case (V4L2_CID_STC_EXPOSURE_MIN):
    {
      __s32 exposure = ctrl->value;;
      if (!cam_data->usb3 || !cam_data->auto_shutters_fpga_can) {;
	retval = -EINVAL;
	goto done;
      }
      if (param_luminance_bounds_us) {
	exposure = useconds_2_exposure(cam_data, exposure);
      }
      cam_data->payload80.Min_exposure_Low_le = __cpu_to_le16( (exposure>>USB3_EXP_LOW_SHIFT) & USB3_EXP_LOW_MASK_16BITS );
      USB3_SET_EXP_UP(cam_data->payload80.Min_exposure_Upp, (exposure>>USB3_EXP_UP_SHIFT), USB3_EXP_UP_MASK_4BITS);
      setFlag(&cam_data->modif_flag_80_47_48); 
      retval = 0;
    }
    goto done;
  case (V4L2_CID_STC_EXPOSURE_MAX):
    {
      __s32 exposure = ctrl->value;;
      if ( ! cam_data->usb3 || ! cam_data->auto_shutters_fpga_can) {;
	retval = -EINVAL;
	goto done;
      }
      if (param_luminance_bounds_us) {
	exposure = useconds_2_exposure(cam_data, exposure);
      }
      cam_data->payload80.Max_exposure_Low_le = __cpu_to_le16( (exposure>>USB3_EXP_LOW_SHIFT) & USB3_EXP_LOW_MASK_16BITS );
      USB3_SET_EXP_UP(cam_data->payload80.Max_exposure_Upp, (exposure>>USB3_EXP_UP_SHIFT), USB3_EXP_UP_MASK_4BITS);
      setFlag(&cam_data->modif_flag_80_45_46); 
      retval = 0;
    }
    goto done;
#endif
#ifdef L_BOUNDS_NATIVE
  case (V4L2_CID_STC_GAIN_MIN):
    if(!cam_data->usb3 || !cam_data->auto_gain_fpga_can) {;
      Sentech_Set_Auto_Min_Gain(cam_data->awb_object, ctrl->value);
      goto done;
    }
    cam_data->payload80.Minimum_gain_AGC = ctrl->value & 0x00FF;
    setFlag(&cam_data->modif_flag_80_4C); 
    goto done;
  case (V4L2_CID_STC_GAIN_MAX):;
    if(!cam_data->usb3 || !cam_data->auto_gain_fpga_can) {;
      Sentech_Set_Auto_Max_Gain(cam_data->awb_object, ctrl->value);
      goto done;
    }
    cam_data->payload80.Maximum_gain_AGC = ctrl->value & 0x00FF;
    setFlag(&cam_data->modif_flag_80_4B); 
    goto done;
  case (V4L2_CID_STC_EXPOSURE_MIN):
    {
      __u16 clock = __le16_to_cpu(cam_data->payload12.one_line_clock_numbers_le);
      __s32 lines, clocks, exposure = ctrl->value;;
      if(param_luminance_bounds_us) {
	exposure = useconds_2_exposure(cam_data, exposure);
      }
      if ( ! cam_data->usb3 || ! cam_data->auto_gain_fpga_can) {;
	lines = exposure/clock;
	clocks = exposure%clock;
	Sentech_Set_Auto_Min_Shutter(cam_data->awb_object, lines);
	Sentech_Set_Auto_Min_ShutterC(cam_data->awb_object, clocks);;
	goto done;
      } else {
	cam_data->payload80.Min_exposure_Low_le = __cpu_to_le16( (exposure>>USB3_EXP_LOW_SHIFT) & USB3_EXP_LOW_MASK_16BITS );
	USB3_SET_EXP_UP(cam_data->payload80.Min_exposure_Upp, (exposure>>USB3_EXP_UP_SHIFT), USB3_EXP_UP_MASK_4BITS);
	setFlag(&cam_data->modif_flag_80_47_48); 
      }
    }
    goto done;
  case (V4L2_CID_STC_EXPOSURE_MAX):
    {
      __u16 clock = __le16_to_cpu(cam_data->payload12.one_line_clock_numbers_le);
     __s32 lines, clocks, exposure = ctrl->value;;
     if (param_luminance_bounds_us) {
       exposure = useconds_2_exposure(cam_data, exposure);
     }
     if ( ! cam_data->usb3 || ! cam_data->auto_gain_fpga_can) {;
       lines = exposure/clock;
       clocks = exposure%clock;
       Sentech_Set_Auto_Max_Shutter(cam_data->awb_object, lines);
       Sentech_Set_Auto_Max_ShutterC(cam_data->awb_object, clocks);;
       goto done;
     }else{
       cam_data->payload80.Max_exposure_Low_le =
	 __cpu_to_le16( (exposure>>USB3_EXP_LOW_SHIFT) & USB3_EXP_LOW_MASK_16BITS);
       USB3_SET_EXP_UP(cam_data->payload80.Max_exposure_Upp, (exposure>>USB3_EXP_UP_SHIFT), USB3_EXP_UP_MASK_4BITS);
       setFlag(&cam_data->modif_flag_80_45_46); 
     }
    }
    goto done;
#endif
#ifdef L_BOUNDS
  case (V4L2_CID_STC_GAIN_MIN):;
    cam_data->gain_min = ctrl->value;
    if(!cam_data->usb3 || !cam_data->auto_gain_fpga_can) {
      if(cam_data->gain_min > cam_data->gain) {
	cam_data->gain = cam_data->gain_min;
	cam_data->payload21.Gain_le = cam_data->gain;
	setFlag(&cam_data->modif_flag21);      
      }
      goto done;
    }
    cam_data->payload80.Minimum_gain_AGC = ctrl->value & 0x00FF;
    setFlag(&cam_data->modif_flag_80_4C); 
    goto done;
  case (V4L2_CID_STC_GAIN_MAX):;
    cam_data->gain_max = ctrl->value;
    if(!cam_data->usb3 || !cam_data->auto_gain_fpga_can) {;
      if(cam_data->gain_max < cam_data->gain) {
	cam_data->gain = cam_data->gain_max;
	cam_data->payload21.Gain_le = cam_data->gain;
	setFlag(&cam_data->modif_flag21);      
      }
      goto done;
    }
    cam_data->payload80.Maximum_gain_AGC = ctrl->value & 0x00FF;
    setFlag(&cam_data->modif_flag_80_4B); 
    goto done;
  case (V4L2_CID_STC_EXPOSURE_MIN):;
    clock = __le16_to_cpu(cam_data->payload12.one_line_clock_numbers_le);
    max_exp = get_max_exposure(cam_data);
    if(max_exp < ctrl->value) {
      retval = -EINVAL;
      break;
    }
    cam_data->exposure_min = ctrl->value;
    if (cam_data->exposure_min > cam_data->shutters){
      shutters = cam_data->shutters = cam_data->exposure_min;
      set_crt_exposure(cam_data, shutters);
      setFlag(&cam_data->modif_flag21);
    }
    cam_data->payload80.Min_exposure_Low_le = __cpu_to_le16( (ctrl->value>>USB3_EXP_LOW_SHIFT) & USB3_EXP_LOW_MASK_16BITS );
    USB3_SET_EXP_UP(cam_data->payload80.Min_exposure_Upp, (ctrl->value>>USB3_EXP_UP_SHIFT), USB3_EXP_UP_MASK_4BITS);
    setFlag(&cam_data->modif_flag_80_47_48) = 1;
    retval = 0;
    goto done;
  case (V4L2_CID_STC_EXPOSURE_MAX):;
    clock = __le16_to_cpu(cam_data->payload12.one_line_clock_numbers_le);
    max_exp =  get_max_exposure(cam_data);
    if(max_exp < ctrl->value) {
      retval = -EINVAL;
      break;
    }
    cam_data->exposure_max = ctrl->value;
    if (cam_data->exposure_max < cam_data->shutters){
      shutters = cam_data->shutters = cam_data->exposure_max;
      set_crt_exposure(cam_data, shutters);
      setFlag(&cam_data->modif_flag21);
    }
    cam_data->payload80.Max_exposure_Low_le = __cpu_to_le16( (exposure>>USB3_EXP_LOW_SHIFT) & USB3_EXP_LOW_MASK_16BITS );
    USB3_SET_EXP_UP(cam_data->payload80.Max_exposure_Upp, (exposure>>USB3_EXP_UP_SHIFT), USB3_EXP_UP_MASK_4BITS);
    setFlag(&cam_data->modif_flag_80_45_46); 
    goto done;
#endif
#ifdef STC_CTRL_SPEED
  case (V4L2_CID_STC_GAIN_CTRL_SPEED_LIMIT):;
    {
      BYTE byteGainCtrlSpeedLimit = ctrl->value;
      Sentech_Set_GainCtrlSpeedLimit(cam_data->awb_object, 
				     byteGainCtrlSpeedLimit);
    }
    goto done;
  case (V4L2_CID_STC_SHUTTER_CTRL_SPEED_LIMIT):;
    {
      BYTE byteShutterCtrlSpeedLimit = ctrl->value;
      Sentech_Set_ShutterCtrlSpeedLimit(cam_data->awb_object, 
					byteShutterCtrlSpeedLimit);
    }
    goto done;
  case (V4L2_CID_STC_AVERAGE_FRAME_COUNT):;
    {
      BYTE byteAverageFrameCount = ctrl->value;
      Sentech_Set_AverageFrameCount(cam_data->awb_object, 
				    byteAverageFrameCount);
    }
    goto done;
#endif
#ifdef SENTECH_AWB
  case (V4L2_CID_STC_ALC_SPEED):;
    {
      BYTE byteShutterCtrlSpeedLimit = 0;
      BYTE byteGainCtrlSpeedLimit = 0;
      BYTE byteSkipFrameCount = ctrl->value;
      BYTE byteAverageFrameCount = 4;
      Sentech_CLuminance_SetCtrlSpeed(cam_data->awb_object,
				      byteShutterCtrlSpeedLimit,
				      byteGainCtrlSpeedLimit,
				      byteSkipFrameCount,
				      byteAverageFrameCount);
    }
    goto done;
#endif
  case (V4L2_CID_STC_PIXEL_BLEMISH_CORRECTION):
    {
      if (!cam_data->usb3 || !cam_data->pixel_blemish_correction_fpga_can) {
	retval = -EINVAL;
	goto done;
      }
      cam_data->payload80.Pixel_blemish_correction &= (~1); 
      cam_data->payload80.Pixel_blemish_correction |= (ctrl->value?1:0);
      setFlag(&cam_data->modif_flag_80_4E); 
      retval = 0;
    }
    goto done;
  case (V4L2_CID_STC_FRAME_START):;
    cam_data->payload12.Picture_size_mode_le = __cpu_to_le32(STC_CUSTOM);
    cam_data->payload12.Vertical_offset_le = __cpu_to_le32(ctrl->value & Vertical_offset_le_MASK);
    setFlag(&cam_data->modif_flag12);
    goto done;
  case (V4L2_CID_STC_FRAME_HEIGHT):;
    cam_data->payload12.Picture_size_mode_le = __cpu_to_le32(STC_CUSTOM);
    cam_data->payload12.Vertical_size_mode_le = __cpu_to_le32(ctrl->value & Vertical_size_mode_le_MASK);
    setFlag(&cam_data->modif_flag12);
    goto done;
  case (V4L2_CID_STC_PIX_SATURATION_AUTO_LUMINANCE):
    cam_data->psat_auto = ctrl->value;
    goto done;
  case (V4L2_CID_STC_PIX_SATURATION_STEP):
    param_saturation_step = cam_data->psat_step = ctrl->value;
    goto done;
  case (V4L2_CID_STC_PIX_SATURATION_MIN):
    cam_data->psat_min = ctrl->value;
    goto done;
  case (V4L2_CID_STC_PIX_SATURATION_MAX):
    cam_data->psat_max = ctrl->value;
    goto done;
  case (V4L2_CID_STC_PIX_SATURATION_ACC):
    param_sat_acc = cam_data->psat_acc = ctrl->value;
    goto done;
  case (V4L2_CID_STC_PIX_SATURATION_UP_LIMIT):
    cam_data->psat_up = ctrl->value;
    goto done;
  case (V4L2_CID_STC_PIX_SATURATION_LO_LIMIT):
    cam_data->psat_lo = ctrl->value;
    goto done;
  case (V4L2_CID_STC_PIX_SATURATION_EXP_INC):
    param_saturation_exp_inc = cam_data->psat_exp = ctrl->value;
    goto done;
  case (V4L2_CID_STC_PIX_SATURATION_GAIN_INC):
    param_saturation_gain_inc = cam_data->psat_gain = ctrl->value;
    goto done;
  case (V4L2_CID_STC_PIX_SATURATION_AMORTIZ):
    cam_data->psat_amortiz = ctrl->value;
    goto done;
  case (V4L2_CID_STC_PIX_SATURATION_AMORTIZ_LENGTH):
    cam_data->psat_amortiz_length = ctrl->value;
    goto done;
  case (V4L2_CID_STC_PIX_SATURATION_SKIP):
    cam_data->psat_skip = ctrl->value;
    goto done;
  case (V4L2_CID_STC_BYTES):
    if(cam_data->supports_bpp) {
      int new_mode = getBppMode(cam_data->mask_Bpp, ctrl->value);
      if (new_mode > 0) {
	cam_data->payload13.Bytes_per_pixel_le = __cpu_to_le32(new_mode);
	setFlag(&cam_data->modif_flag13);
      }
    }
    goto done;
  case (V4L2_CID_STC_NATIVE_MODES):
    cam_data->enable_native = ctrl->value;
    goto done;
  case (V4L2_CID_STC_LONG_EXPOSURE):
    cam_data->exposure_priority_not_fps = ctrl->value;
    update_exposure_limits(cam_data);
    if(get_crt_exposure(cam_data) > get_max_exposure(cam_data))
      set_crt_exposure(cam_data,get_max_exposure(cam_data));
    goto done;
  default:;
    retval = -EINVAL;
  }
 done:
;
  lprintk(1,"EXIT CTRL for %d, %d\n", ctrl->id, ctrl->value);
  write_lock_irqsave(&cam_data->ctrl_lock, flags);
  set_controls(cam_data, GFP_ATOMIC);
  write_unlock_irqrestore(&cam_data->ctrl_lock, flags);
  return retval;
}
 int sentech_update_shutter_gain(struct sentech_camera* cam_data);
#define CTRL_TIMEOUT 2*HZ
static int sentech_vidioc_g_ctrl (struct file *UNUSED(file), void *priv, 
			   struct v4l2_control * ctrl){
  struct sentech_camera* cam_data = get_cam_data(priv);
  unsigned int i;
  int retval=-EINVAL;;
  lprintk(1,"In %s\n",__FUNCTION__);
  switch (ctrl->id) {
  case (V4L2_CID_STC_HORIZONTAL_OFFSET_CAPTURE):
    ctrl->value = __le32_to_cpu(cam_data->state12.Horizontal_offset_le);
    return retval = 0;
  case (V4L2_CID_STC_VERTICAL_OFFSET_CAPTURE):
    ctrl->value = __le32_to_cpu(cam_data->state12.Vertical_offset_le); 
    return retval = 0;
  case (V4L2_CID_STC_HORIZONTAL_SIZE_CAPTURE):
    ctrl->value = __le32_to_cpu(cam_data->state12.Horizontal_size_mode_le);
    return retval = 0;
  case (V4L2_CID_STC_VERTICAL_SIZE_CAPTURE):
    ctrl->value =  __le32_to_cpu(cam_data->state12.Vertical_size_mode_le); 
    return retval = 0;
  case (V4L2_CID_STC_VERTICAL_SCAN):
    ctrl->value = cam_data->scan_height;
    return retval = 0;
  case (V4L2_CID_STC_SELECTED_FOURCC):
    ctrl->value = cam_data->pix_format.pixelformat;
    return retval = 0;
  case (V4L2_CID_STC_EXPOSURE_MAX):
    if(cam_data->usb3) {
      ctrl->value = (__le16_to_cpu(cam_data->state80.Max_exposure_Low_le)<<USB3_EXP_LOW_SHIFT)+
	((cam_data->state80.Max_exposure_Upp & USB3_EXP_UP_MASK_4BITS)<<USB3_EXP_UP_SHIFT);
      retval = 0;
      return retval;
    }
    break;
  case (V4L2_CID_STC_EXPOSURE_MIN):
    if(cam_data->usb3) {
      ctrl->value = (__le16_to_cpu(cam_data->state80.Min_exposure_Low_le)<<USB3_EXP_LOW_SHIFT)+
	((cam_data->state80.Min_exposure_Upp & USB3_EXP_UP_MASK_4BITS)<<USB3_EXP_UP_SHIFT);
      retval = 0;;
      return retval;
    }
    break;
  case (V4L2_CID_STC_GAIN_MAX):
    if(cam_data->usb3) {
      ctrl->value = cam_data->state80.Maximum_gain_AGC;
      retval = 0;;
      return retval;
    }
    break;
  case (V4L2_CID_STC_GAIN_MIN):
    if(cam_data->usb3) {
      ctrl->value = cam_data->state80.Minimum_gain_AGC;
      retval = 0;
      return retval;
    }
    break;
  case V4L2_CID_STC_IO_STATUS_0:
  case V4L2_CID_STC_IO_STATUS_1:
  case V4L2_CID_STC_IO_STATUS_2:
  case V4L2_CID_STC_IO_STATUS_3:
    { 
      int res; 
      int id = ctrl->id - (V4L2_CID_STC_IO_STATUS_0);
      res=usb_control_msg(cam_data->udev,
			  cam_data->pipe_ctl_rcv,
			  0x80,
			  USB_DIR_IN | USB_TYPE_VENDOR |USB_RECIP_DEVICE, 
			  STC_FPGA_READ|0x02,
			  0,
			  &cam_data->state80.IO_Status,
			  sizeof(cam_data->state80.IO_Status), CTRL_TIMEOUT);
      if(res<=0){ 
	dev_err(&cam_data->interface->dev,"SentechUSB: status result IOstatus = %d\n",res);
	retval = -ENODEV;
      }else{
	retval = 0;
      }
      ctrl->value = (cam_data->state80.IO_Status>>id)&1;;
      return retval;
    }
  case V4L2_CID_STC_PIXEL_BLEMISH_CORRECTION:
    {
      if(!cam_data->usb3){
	ctrl->value = 0;
	return 0;
      }
      ctrl->value = cam_data->state80.Pixel_blemish_correction & 1;
      return retval = 0;
    }
  case V4L2_CID_GAIN:
    {
      sentech_update_shutter_gain(cam_data);
      ctrl->value = __le16_to_cpu(cam_data->state21.Gain_le);
      return  retval = 0;
    }
  case V4L2_CID_STC_EXPOSURE_US:
    {
      __s32 crt_exp;
      sentech_update_shutter_gain(cam_data);;
      crt_exp = get_crt_exposure(cam_data);;
      ctrl->value = exposure_2_useconds(cam_data, crt_exp);;
      return  retval = 0;
    }
  case V4L2_CID_STC_OUTPUT_FRAME_RATE:
    {
      int res;
      unsigned int dec, val;
      res=usb_control_msg(cam_data->udev,
			  cam_data->pipe_ctl_rcv,
			  0x80,
			  USB_DIR_IN | USB_TYPE_VENDOR |USB_RECIP_DEVICE, 
			  STC_FPGA_READ|0x02,
			  0,
			  &cam_data->state80.output_frame_rate_decimal,
			  4, CTRL_TIMEOUT);
      if(res<=0){ 
	dev_err(&cam_data->interface->dev,"SentechUSB: status result output frame rate = %d\n",res);
	retval = -ENODEV;
      }else{
	retval = 0;
      }
      ctrl->value = (cam_data->state80.output_frame_rate)&0x003F;
      dec = cam_data->state80.output_frame_rate_decimal;
      val = cam_data->state80.output_frame_rate;;
      return retval;
    }
  }
  for (i = 0; i < ARRAY_SIZE(sentech_qctrl); i++)
    if (ctrl->id == sentech_qctrl[i].id) {
      ctrl->value=cam_data->qctl_regs[i];
      retval = 0;
      break;
    }
;
  lprintk(1,"Exiting %s\n",__FUNCTION__);
  return retval;
}
static
int sentech_vidioc_reqbufs (struct file *UNUSED(file), void *fh, struct v4l2_requestbuffers *b){
  struct sentech_camera* cam_data = get_cam_data(fh);
  int retval = 0;;
  lprintk(1,"In %s\n",__FUNCTION__);
  if (cam_data->shutdown || !cam_data->connected) {;
    retval = -EINVAL; goto vidioc_reqbufs_exit;
  }
  if(b->count==0){;
    retval = 0; 
    goto  vidioc_reqbufs_exit;
  }
  if(param_streaming) {
    if(cam_data->reading||
       cam_data->streaming||
       cam_data->vmas){;
      retval = -EBUSY;
      goto vidioc_reqbufs_exit;
    } 
    switch(b->type) {
    case V4L2_BUF_TYPE_VIDEO_OVERLAY:;
      retval = -EINVAL;
      goto vidioc_reqbufs_exit;
    case V4L2_BUF_TYPE_VIDEO_CAPTURE:
       cam_data->buffertype = b->type;
       break;
    case V4L2_BUF_TYPE_PRIVATE:
    default:;
      retval = -EINVAL;
      goto vidioc_reqbufs_exit;
    }
    switch(b->memory) {
    case V4L2_MEMORY_MMAP:
      cam_data->req_userptr_memap_overlay = 1;
      if(b->type != V4L2_BUF_TYPE_VIDEO_CAPTURE){
	retval = -EINVAL;;
	goto vidioc_reqbufs_exit;
      }
      if(b->count==0){ 
	retval = 0;    
;
	break;
      }
      if(b->count<1){
	retval = -EINVAL;
	printk("Invalid buffer count %d for MMAP request\n", b->count);
	goto vidioc_reqbufs_exit;
      }
      if(param_have_buffers) {
	if(!cam_data->allocated_buffers){
	  cam_data->requested_buffers = b->count;;
	  if((retval = sentech_alloc_subs(cam_data))){;
	    goto vidioc_reqbufs_exit;
	  };
	}else{;
	}
	b->count = cam_data->allocated_buffers;
      } else {
	b->count = cam_data->sib.vmalloc_buffers_nb;;
      }
      break;
    case V4L2_MEMORY_USERPTR:;
      cam_data->req_userptr_memap_overlay = 2;
      if(b->type != V4L2_BUF_TYPE_VIDEO_CAPTURE){
	retval = -EINVAL;;
	goto vidioc_reqbufs_exit;
      }
      if(!param_usrstreaming) {
	retval = -EINVAL;;
      }
      break;
    case V4L2_MEMORY_OVERLAY:;
	retval = -EINVAL;;
	goto vidioc_reqbufs_exit;
      cam_data->req_userptr_memap_overlay = 3;
      if(b->type != V4L2_BUF_TYPE_VIDEO_OVERLAY){
	retval = -EINVAL;;
	goto vidioc_reqbufs_exit;
      }
      if(!param_overlay) {
	retval = -EINVAL;;
      }
      break;
    default:
      retval = -EINVAL;;
      goto vidioc_reqbufs_exit;
    }
  }else{;
    retval = -EINVAL;
  }
 vidioc_reqbufs_exit:;
  lprintk(1,"Exiting %s\n",__FUNCTION__);
  return retval;
}
 inline int sentech_pagerounded(int size){
   if (size & ((1<<PAGE_SHIFT)-1)) {;
     size += 1<<PAGE_SHIFT;
     size &= ~((1<<PAGE_SHIFT)-1);;
   }
   return size;
 }
int sentech_vidioc_querybuf(struct file *UNUSED(file), void *fh, struct v4l2_buffer *b){
  struct sentech_camera* cam_data = get_cam_data(fh);
  int retval = 0;
  int size = cam_data->pix_format.sizeimage;
  __u32 input = 0; 
  lprintk(1,"In %s\n",__FUNCTION__);;
  if(cam_data->shutdown) {;
    retval = -ENODEV; goto vidioc_querybuf_exit;
  }
  if(param_streaming) {
    if (param_have_buffers) {
      if(unlikely( (b->index>=cam_data->allocated_buffers))) {
	retval = -EINVAL;;
	goto vidioc_querybuf_exit;
      }
    } else {
      if(unlikely( (b->index>=cam_data->sib.vmalloc_buffers_nb))) {
	retval = -EINVAL;;
	goto vidioc_querybuf_exit;
      }
    }
    if(unlikely((b->type != cam_data->buffertype) ||
		(input) ||
		(b->type != V4L2_BUF_TYPE_VIDEO_CAPTURE))) {
      retval = -EINVAL;;
      goto vidioc_querybuf_exit;
    }else{;
    }
    size = sentech_pagerounded(size);
    b->bytesused = cam_data->pix_format.sizeimage; 
    b->field = V4L2_FIELD_NONE;
    b->flags = 0;
    b->m.offset = b->index<<SBUF_IDX_SHIFT; 
    b->length = size+MARGIN;
    b->memory = V4L2_MEMORY_MMAP;
  }else{;
    retval = -EINVAL;
  }
 vidioc_querybuf_exit:;
  lprintk(1,"Exiting %s\n",__FUNCTION__);
  return retval;
}
 int sentech_vidioc_qbuf_indirect (struct file *file,  struct v4l2_buffer *b) {
  struct sentech_camera* cam_data;
  __u32 index = b->index;
  int retval = 0;;
  lprintk(2,"In %s\n",__FUNCTION__);
  lprintk(2,"Queue Index %d\n",b->index);
  cam_data = get_cam_data(file->private_data);
  if (!cam_data) goto done;
  if (!cam_data->connected) {
    printk("sentech qbuf: Video device is disconnecting!\n");
    retval = -ENODEV;
    goto done;
  }
  if (!cam_data->run) {
    printk("sentech: Video device is stopped!\n");
    retval = -ENODEV;
    goto done;
  }
  if (b->type != cam_data->buffertype) {
    retval = -EINVAL;;
    goto done;
  }
  if ((index < 0) || (index >= cam_data->sib.vmalloc_buffers_nb)) {
    retval = -EINVAL;;
    goto done;
  }
  cam_data->sib.vmalloc_buffers_user[index] = 0;
  if(param_admit_unspecified_qbuf_index)
    freeBufs_indirect(cam_data);
  if(likely((restart_on_bus_error || !cam_data->status) && cam_data->run)) {
    if(unlikely( atomic_read(&cam_data->running_buffers) < SUBS_SUBMITTED)) {;
      sentech_submit_kurbs(cam_data, GFP_KERNEL);
    }
  }
  b->flags |= V4L2_BUF_FLAG_QUEUED|V4L2_BUF_FLAG_MAPPED;
  b->flags &= ~V4L2_BUF_FLAG_DONE;;
 done:
  lprintk(2,"Exiting %s\n",__FUNCTION__);
  return retval;
 }
 int sentech_vidioc_qbuf_direct (struct file *file,  struct v4l2_buffer *b){
  struct sentech_camera* cam_data;
  __u32 index = b->index;
  struct sentech_usb_buffer* sub;
  int retval = 0;
  lprintk(2,"In %s\n",__FUNCTION__);
  lprintk(2,"Queue Index %d\n",b->index);
  cam_data = get_cam_data(file->private_data);
  if(!cam_data) goto done;
  if(!cam_data->connected) {
    printk("sentech qbuf: Video device is disconnecting!\n");
    retval = -ENODEV;
    goto done;
  }
  if(!cam_data->run) {
    printk("sentech: Video device is stopped!\n");
    retval = -ENODEV;
    goto done;
  }
  if (b->type != cam_data->buffertype) {
    retval = -EINVAL;;
    goto done;
  }
  sub = get_sub(cam_data, index);
  if(!sub){
    retval = -EINVAL;;
    goto done;
  }
  sub->user_owned = 0;
  if(param_admit_unspecified_qbuf_index)
    freeBufs(cam_data);
  if(likely((restart_on_bus_error || !cam_data->status) && cam_data->run)) {
    if(unlikely( atomic_read(&cam_data->running_buffers) < SUBS_SUBMITTED)) {;
      sentech_submit_kurbs(cam_data, GFP_KERNEL);
    }
  }
  b->flags |= V4L2_BUF_FLAG_QUEUED|V4L2_BUF_FLAG_MAPPED;
  b->flags &= ~V4L2_BUF_FLAG_DONE;;
 done:
  lprintk(2,"Exiting %s\n",__FUNCTION__);
  return retval;
}
int sentech_vidioc_qbuf (struct file *file, void *UNUSED(fh), struct v4l2_buffer *b){
  if (param_have_buffers) return sentech_vidioc_qbuf_direct (file,  b);
  return sentech_vidioc_qbuf_indirect (file,  b);
}
#include "sentech_color_read16.c"
#include "sentech_color_read.c"
#include "sentech_awb_vmalloc.c"
#include "sentech_awb_indirect_vmalloc.c"
#include "sentech_color_vmalloc.c"
#include "sentech_color_indirect_vmalloc.c"
static struct sentech_usb_buffer*
get_has_data(struct sentech_camera *cam_data, int _lock){
  struct sentech_usb_buffer* sub, *retval=NULL;
  unsigned long flags;
#ifdef LOCK_HAS_RAW_LOCK
  ;;
#endif
  read_lock_irqsave(&cam_data->buffers_lock, flags);
  list_for_each_entry(sub, &cam_data->buffers, sub_buffers) {;
    if (sub->has_data && 
	(
	 !sub->user_owned) && 
	(param_use_submitted || !sub->submitted)) {
      if(_lock){
	sub->user_owned = 1;
      }
      retval = sub;;
      break;
    }
  }
  read_unlock_irqrestore(&cam_data->buffers_lock, flags);
#ifdef LOCK_HAS_RAW_LOCK
  ;;
#endif
  return retval;
}
static struct sentech_usb_buffer*
get_first_submitted(struct sentech_camera *cam_data, int _lock){
  struct sentech_usb_buffer* sub, *retval=NULL;
  unsigned long flags;
#ifdef LOCK_HAS_RAW_LOCK
  ;;
#endif
  read_lock_irqsave(&cam_data->buffers_lock, flags);
  list_for_each_entry(sub, &cam_data->buffers, sub_buffers) {;
    if (sub->has_data || sub->submitted) {
      if(_lock && !sub->submitted){
	sub->user_owned = 1;
      }
      retval = sub;;
      break;
    }
  }
  read_unlock_irqrestore(&cam_data->buffers_lock, flags);
#ifdef LOCK_HAS_RAW_LOCK
  ;;
#endif
  return retval;
}
#define __wait_event_interruptible_exclusive_timeout(wq, condition, ret)          \
do {                                                                    \
        DEFINE_WAIT(__wait);                                            \
                                                                        \
        for (;;) {                                                      \
                prepare_to_wait_exclusive(&wq, &__wait, TASK_INTERRUPTIBLE);      \
                if (condition)                                          \
                        break;                                          \
                if (!signal_pending(current)) {                         \
                        ret = schedule_timeout(ret);                    \
                        if (!ret)                                       \
                                break;                                  \
                        continue;                                       \
                }                                                       \
                ret = -ERESTARTSYS;                                     \
                break;                                                  \
        }                                                               \
        if (!ret && (condition))                                        \
                ret = 1;                                                \
        finish_wait(&wq, &__wait);                                      \
} while (0)
#define wait_event_interruptible_exclusive_timeout(wq, condition, timeout)        \
({                                                                      \
        long __ret = timeout;                                           \
        if (!(condition))                                               \
                __wait_event_interruptible_exclusive_timeout(wq, condition, __ret); \
        __ret;                                                          \
})
 int find_oldest_buffer(struct sentech_camera* cam_data, int just_test){
   int found = 0;
   int found_buf = -1;
   int i;
;
   for (i = 0; i < cam_data->sib.vmalloc_buffers_nb; i++) {
     if (cam_data->sib.vmalloc_buffers_user[i]) {;
       continue;
     }
     if (!cam_data->sib.sib_frame[i].done) {;
       continue;
     }
     if (cam_data->sib.vmalloc_buffers_idx[i] + sizeof(struct Sentech_Header) < cam_data->scan_size) {;
       continue;
     }
     if (just_test){;
       return i;
     };
     if (found) {
       if (
	   ((cam_data->sib.sib_frame[i].frame_nb < cam_data->sib.sib_frame[found_buf].frame_nb) &&
	    ((cam_data->sib.sib_frame[found_buf].frame_nb <= cam_data->sib.last_frame) ||
	     (cam_data->sib.last_frame < cam_data->sib.sib_frame[i].frame_nb)))
	   ||
	   ((cam_data->sib.last_frame < cam_data->sib.sib_frame[i].frame_nb)
	    && (cam_data->sib.sib_frame[found_buf].frame_nb <= cam_data->sib.last_frame))) {
;
	 cam_data->sib.vmalloc_buffers_user[i] = 1;
	 found_buf = i;
       } else {;
       }
     } else {
       found = 1;
       cam_data->sib.vmalloc_buffers_user[i] = 1;
       found_buf = i;;
     }
   }
   if (found_buf < 0) {
     if ((!cam_data->run) || (!cam_data->streaming) || (cam_data->shutdown) || (cam_data->status))
       return -2;
   }
   return found_buf;
 }
 int sentech_vidioc_dqbuf_indirect (struct file *file,  struct v4l2_buffer *b) {
  struct sentech_camera* cam_data;
  __u32 index = 0;
  unsigned long flags;
  int retval = -EINVAL;
  int found_buf;
  lprintk(1,"In %s\n",__FUNCTION__);
  cam_data = get_cam_data(file->private_data);
  if (unlikely(!cam_data)) {;
    goto done_nowait;
  }
  if (!cam_data->connected) {
    printk("sentech dqbuf: Video device is disconnecting!\n");
    retval = -ENODEV;
    goto done_nowait;
  }
  if (unlikely(b->type != cam_data->buffertype)) {;
    retval = -EINVAL;
    goto done_nowait;
  }
  if (likely((restart_on_bus_error || !cam_data->status) && cam_data->run)) {
    if (unlikely( atomic_read(&cam_data->running_buffers) < SUBS_SUBMITTED)) {
;
      sentech_submit_kurbs(cam_data, GFP_KERNEL); 
    }
  } else {
    if (! cam_data->run);
    else {;;
    }
    retval = -EINVAL;
    retval = -ENODEV;
    goto done_nowait;
  }
  for (;;) {
    read_lock_irqsave(&cam_data->buffers_lock, flags);
    mb();
    found_buf = find_oldest_buffer(cam_data, 0);
    mb();
    read_unlock_irqrestore(&cam_data->buffers_lock, flags);
    if (found_buf >= 0){;
      break;
    }
    if (file->f_flags & O_NONBLOCK){
      printk(MY_KERN_ERR "sentech: TRY AGAIN LATER. USB STATE CONDITION. NONBLOCKING request!\n");;
      retval = -EAGAIN;
      goto done;
    }
    if ((cam_data->shutdown) || (!cam_data->run) || (!cam_data->streaming) || (cam_data->status) || (!dqueue_sleep)) {;
      retval = -ENODEV;
      goto done;
    }
    if(wait_event_interruptible_exclusive_timeout(cam_data->wq, (find_oldest_buffer(cam_data, 1) != -1), HZ) == -ERESTARTSYS) {
      retval = -ERESTARTSYS;;
      goto done;
    }
  }
  goto success;
 success:
  if ((retval = down_interruptible(&cam_data->format_lock))) {  
    retval = -ERESTARTSYS;
    goto format_exit;
  }
  index = found_buf;
  if (debug != 0) printk("dqbuf:transfer %d\n",found_buf);;
  cam_data->sib.vmalloc_buffers_user[found_buf] = 1; 
  mb();
  cam_data->sib.sib_frame[found_buf].data = 0;
  cam_data->sib.sib_frame[found_buf].done = 0;
  b->index = index;
  b->timestamp = cam_data->sib.sib_frame[found_buf].timestamp;
  b->sequence = cam_data->sib.sib_frame[found_buf].frame_nb; 
  b->field = V4L2_FIELD_NONE;
  b->flags = V4L2_BUF_FLAG_DONE;
  b->m.offset = index<<SBUF_IDX_SHIFT;;
  if (cam_data->sib.vmalloc_vbuffers[found_buf]) {;
    if (!cam_data->auto_gain && !cam_data->usb3)
      sentech_saturated_pixels(cam_data, cam_data->sib.vmalloc_buffers[found_buf], "vbuffer");
    if (trace_buffers > 0) {
      int frag_size = cam_data->scanimage_size;
      if (frag_size < cam_data->sib.vmalloc_buffers_size)
	frag_size = cam_data->sib.vmalloc_buffers_size;
      memcpy(cam_data->sib.vmalloc_vbuffers[found_buf]
	     + cam_data->pix_format.sizeimage 
	     - trace_buffers*cam_data->max_frame_size,
	     cam_data->sib.vmalloc_buffers[found_buf],
	     frag_size);
    }
#ifdef SENTECH_AWB
    if (cam_data->usb3) {
    } else {
      retval = awb_indirect_move_to_vmalloc (cam_data, found_buf);;
    }
#else
    if (cam_data->usb3) {
    } else {
      retval = indirect_move_to_vmalloc (cam_data, found_buf);;
    }
#endif
    b->bytesused = retval;
    if(retval > 
       param_imagesize_claim_adjustment+cam_data->pix_format.sizeimage)
      b->bytesused =  param_imagesize_claim_adjustment+cam_data->pix_format.sizeimage;
    b->length = cam_data->sib.vmalloc_vbuffers_size[found_buf]; 
    if (retval > 0) retval = 0;;
  } else { 
;
    b->bytesused = cam_data->pix_format.sizeimage+param_imagesize_claim_adjustment;
    b->length = cam_data->sib.vmalloc_mapped_size[found_buf];;
    if ( ! cam_data->auto_gain )
      sentech_saturated_pixels(cam_data, cam_data->sib.vmalloc_buffers[found_buf], "native");
#ifdef SENTECH_AWB
    if ( ! cam_data->usb3 ) {
      if (cam_data->skiped_awb>=skip_awb){;
#ifdef ST_ImageCopyBalance
	;;
	Sentech_ImageCopyBalance(cam_data->awb_object, cam_data->sib.vmalloc_buffers[found_buf]);
#else
	Sentech_ImageProcessing(cam_data->awb_object, cam_data->sib.vmalloc_buffers[found_buf]);
	Sentech_ALCWBalance(cam_data->awb_object); 
#endif
	cam_data->skiped_awb = 0;
	cam_data->recent_autobalance = 1;
      } else {
	cam_data->skiped_awb++;
      }
    }
#endif
    if (param_integrated_rgb && cam_data->auto_white_balance) {
      *((__u16*)cam_data->sib.vmalloc_buffers[found_buf])=cam_data->red;
      *(((__u16*)cam_data->sib.vmalloc_buffers[found_buf])+1)=cam_data->green;
      *(((__u16*)cam_data->sib.vmalloc_buffers[found_buf])+2)=cam_data->blue;
    }
  }
  up(&cam_data->format_lock);
  goto done;
 format_exit: 
  cam_data->sib.vmalloc_buffers_user[found_buf] = 0; 
  cam_data->sib.sib_frame[found_buf].done = 0; 
  cam_data->sib.sib_frame[found_buf].data = 0; 
 done:;
 done_nowait:;
  lprintk(1,"Exiting %s\n",__FUNCTION__);
  return retval;
}
 int sentech_vidioc_dqbuf_direct (struct file *file,  struct v4l2_buffer *b){
  struct sentech_camera* cam_data;
  __u32 index = 0;
  struct sentech_usb_buffer* sub, *tsub;
  struct sentech_usb_fragment* suf;
  int retval = -EINVAL;
  int found = 0;
  lprintk(0,"In %s\n",__FUNCTION__);
  cam_data = get_cam_data(file->private_data);
  if (unlikely(!cam_data)) {;
    goto done_nowait;
  }
  if(!cam_data->connected) {
    printk("sentech dqbuf: Video device is disconnecting!\n");
    retval = -ENODEV;
    goto done_nowait;
  }
  if (unlikely(b->type != cam_data->buffertype)) {;
    retval = -EINVAL;
    goto done_nowait;
  }
  if(likely((restart_on_bus_error || !cam_data->status) && cam_data->run)) {
    if(unlikely( atomic_read(&cam_data->running_buffers) < SUBS_SUBMITTED)) {
;
      sentech_submit_kurbs(cam_data, GFP_KERNEL); 
    }
  }else{
    if(!cam_data->run);
    else {;;
    }
    retval = -EINVAL;
    retval = -ENODEV;
    goto done_nowait;
  }
  for(;;) {
    mb();
    found = 0;
    if(debug==-2) printk("In dqbuf\n");
    sub=get_has_data(cam_data, 1); 
    if(sub==NULL) sub= get_first_submitted(cam_data, 1); 
    if(sub!=NULL){
      index = sub->id;
      found = 1;
    }
    if(unlikely(!found)){
      if(unlikely(atomic_read(&cam_data->running_buffers))){
;
	printk(MY_KERN_ERR "sentech: ASSERTION 1' FAILS. Please Report to msilaghi@fit.edu!\n"); 
	retval = -EIO;
	goto done;
      }else{
;
	printk(MY_KERN_ERR "sentech: dqbuf dir ASSERTION 1'' FAILS. Please Report to msilaghi@fit.edu!\n"); 
	cam_data->status =
	  usb_clear_halt(cam_data->udev, cam_data->pipe_image);
	if(!cam_data->status)
	  sentech_submit_kurbs(cam_data, GFP_KERNEL);
	else{
	  if(!cam_data->shutdown) {
	    printk(MY_KERN_ERR "sentech: dqbuf dir Brute status cleaning\n");
	    cam_data->status = 0;
	    sentech_submit_kurbs(cam_data, GFP_KERNEL);
	  }else{
	    retval = -ENODEV;
	    goto done;
	  }
	}
	if(file->f_flags & O_NONBLOCK){
	  printk(MY_KERN_ERR "sentech: TRY AGAIN LATER. USB STATE CONDITION. NONBLOCKING request!\n");;
	  retval = -EAGAIN;
	  goto done;
	}else{
	  printk(MY_KERN_ERR "sentech: TRY AGAIN LATER. USB STATE CONDITION!\n");;
	  retval = -EIO;
	  goto done;
	}
	printk(MY_KERN_ERR "sentech: ASSERTION 1''' FAILS. Please Report to msilaghi@fit.edu!\n");
	retval = -EIO;
	goto done;
	continue;
      } 
    } 
    if(unlikely(!sub)) {
      printk(MY_KERN_ERR "sentech: ASSERTION 2 FAILS. Please Report to msilaghi@fit.edu!\n");;
      retval = -EIO; goto done;
    }
    suf = first_suf(sub);
    if(unlikely(!suf)) {
      printk(MY_KERN_ERR "sentech: ASSERTION 3 FAILS. Please Report to msilaghi@fit.edu!\n");;
      retval = -EIO;
      goto done;
    }
    if (sub->user_owned){;
      retval = 0;
      goto success; 
    }
    if(debug!=0){
      printk("No data ready!\n");
      printk("ST %d %d|", sub->id, atomic_read(&cam_data->running_buffers));
      list_for_each_entry(tsub, &cam_data->buffers, sub_buffers) {
	printk("%d:d%d s%d u%d !%d|",tsub->id,tsub->has_data, tsub->submitted,
	       tsub->user_owned, tsub->error_code);
      }
    }
    if(file->f_flags & O_NONBLOCK){;
      retval = -EAGAIN;
      goto done;
    }
    if(!cam_data->streaming){;
      if(debug!=0) printk("non streaming\n");
      retval = -EAGAIN;
      goto done;
    };
    if(!dqueue_sleep) { 
      printk("Should never sleep!\n");
      retval = -EIO;
      goto done;
    }
    if(wait_event_interruptible_exclusive_timeout(cam_data->wq, ((sub=get_has_data(cam_data, 0)) != NULL), HZ) == -ERESTARTSYS){
      retval = -ERESTARTSYS;;
      goto done;
    }
  }
 success:
  if((retval = down_interruptible(&cam_data->format_lock))){  
    retval = -ERESTARTSYS;
    goto format_exit;
  }
  if(debug!=0) printk("dqbuf:transfer %d\n",sub->id);;
  sub->has_data = 0;
  sub->user_owned = 1; 
  b->index = index;
  b->timestamp = sub->timestamp;
  b->sequence = le32_to_cpu(suf->framebuffer_header->Frame_number_le);
  b->field = V4L2_FIELD_NONE;
  b->flags = V4L2_BUF_FLAG_DONE;
  b->m.offset = index<<SBUF_IDX_SHIFT;;
  if (sub->vbuffer) {;
    if (! cam_data->auto_gain && ! cam_data->usb3)
      sentech_saturated_pixels(cam_data, suf->framebuffer_image, "vbuffer");
    if (trace_buffers > 0) {
      int frag_size = cam_data->scanimage_size;
      if (frag_size < suf->fragment_size)
	frag_size = suf->fragment_size;
      memcpy(sub->vbuffer + cam_data->pix_format.sizeimage - trace_buffers*cam_data->max_frame_size,
	     suf->framebuffer_image, frag_size);
    }
#ifdef SENTECH_AWB
    if (cam_data->usb3) {
      retval = move_sub_to_vmalloc(sub);;
    } else {
      retval = awb_move_sub_to_vmalloc(sub);;
    }
#else
    retval = move_sub_to_vmalloc(sub);;
#endif
    b->bytesused = retval;
    if (retval > 
       param_imagesize_claim_adjustment+cam_data->pix_format.sizeimage)
      b->bytesused =  param_imagesize_claim_adjustment+cam_data->pix_format.sizeimage;
    b->length = sub->vbuffer_size; 
    if (retval > 0) {;
     retval = 0;
    };
  } else { 
;
    b->bytesused = cam_data->pix_format.sizeimage+param_imagesize_claim_adjustment;
    b->length = sub->mapped_size;;
    if ( ! cam_data->auto_gain )
      sentech_saturated_pixels(cam_data, suf->framebuffer_image, "native");
#ifdef SENTECH_AWB
    if ( ! cam_data->usb3 ) {
      if (cam_data->skiped_awb >= skip_awb){;
	if (cam_data->awb_object != NULL) {
#ifdef ST_ImageCopyBalance
	  Sentech_ImageCopyBalance(cam_data->awb_object, suf->framebuffer_image);
#else
	  Sentech_ImageProcessing(cam_data->awb_object, suf->framebuffer_image);
	  Sentech_ALCWBalance(cam_data->awb_object); 
#endif
	} else {
	  dprintk (0,"%s: balance with NULL object %p\n",__FUNCTION__, cam_data->awb_object);
	}
	cam_data->skiped_awb = 0;
	cam_data->recent_autobalance = 1;
      } else{
	cam_data->skiped_awb++;
      }
    }
#endif
    if(param_integrated_rgb && cam_data->auto_white_balance) {
      *((__u16*)suf->framebuffer_image)=cam_data->red;
      *(((__u16*)suf->framebuffer_image)+1)=cam_data->green;
      *(((__u16*)suf->framebuffer_image)+2)=cam_data->blue;
    }
  }
  up(&cam_data->format_lock);
  goto done;
 format_exit: 
  sub->user_owned = 0; 
 done:;
 done_nowait:
;
  lprintk(0,"Exiting %s with r=%d\n",__FUNCTION__, retval);
  return retval;
}
int sentech_vidioc_dqbuf (struct file *file, void *UNUSED(fh), struct v4l2_buffer *b){
  if (param_have_buffers) return sentech_vidioc_dqbuf_direct(file,  b);
  return sentech_vidioc_dqbuf_indirect(file,  b);
}
#ifdef HAVE_V4L1
static int 
sentech_vidiocgmbuf  (struct file *file, void *fh, struct video_mbuf *p) {
  struct sentech_camera* cam_data = get_cam_data(file->private_data);
  int retval = 0;
  struct video_mbuf *mbuf = arg;
  struct v4l2_requestbuffers req;
  unsigned int i;
;
  memset(&req,0,sizeof(req));
  req.type   = q->type;
  req.count  = 8;
  req.memory = V4L2_MEMORY_MMAP;
  if (retval < 0){
    break;
  }
  memset(mbuf,0,sizeof(*mbuf));
  mbuf->frames = req.count;
  mbuf->size   = 0;
  for (i = 0; i < mbuf->frames; i++) {
    mbuf->offsets[i]  = q->bufs[i]->boff;
    mbuf->size       += q->bufs[i]->bsize;
  };
  return retval;
}
#endif
int sentech_vidioc_g_parm(struct file *UNUSED(file), void *fh,
			  struct v4l2_streamparm *a) {
  struct sentech_camera* cam_data = get_cam_data(fh);
  int retval = 0;;
  if(a->type != V4L2_BUF_TYPE_VIDEO_CAPTURE){
    retval=-EINVAL;;
    goto done;
  }
  a->parm.capture.capability = V4L2_CAP_TIMEPERFRAME;
  a->parm.capture.capturemode = V4L2_MODE_HIGHQUALITY;
  a->parm.capture.timeperframe.numerator =
    sentech_fps[cam_data->model_idx].fps[cam_data->fps_idx].numerator;
  a->parm.capture.timeperframe.denominator =
    sentech_fps[cam_data->model_idx].fps[cam_data->fps_idx].denominator;
  a->parm.capture.extendedmode = 0; 
  a->parm.capture.readbuffers = cam_data->allocated_buffers;;
 done:
  return retval;
}
int sentech_vidioc_s_parm(struct file *UNUSED(file), void *fh,
			  struct v4l2_streamparm *a) {
  struct sentech_camera* cam_data = get_cam_data(fh);
  int retval = -EINVAL;
  unsigned int model;
  int k;;
  if(a->type != V4L2_BUF_TYPE_VIDEO_CAPTURE){
;
    goto done;
  }
  if (a->parm.capture.capability == V4L2_CAP_TIMEPERFRAME) {
    for (model = 0; model < ARRAY_SIZE(sentech_fps); model ++) {
      if (cam_data->model != sentech_fps[model].model) continue;
      for ( k = 0; k < 3; k ++) {;
	if ((a->parm.capture.timeperframe.numerator ==
	    sentech_fps[model].fps[k].numerator)
	   &&
	   (a->parm.capture.timeperframe.denominator ==
	    sentech_fps[model].fps[k].denominator)) {
	  set_fps(cam_data, k);
	  cam_data->fps_idx = k;;
	  retval = 0;
	  break;
	}
      }
    }
  }
  if (a->parm.capture.readbuffers !=
      cam_data->allocated_buffers) {;
    cam_data->requested_buffers = a->parm.capture.readbuffers;
  }
 done:;
  return retval;
}
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,29)
int sentech_vidioc_enum_framesizes (struct file *file, void *fh,
				   struct v4l2_frmsizeenum *fsize) {
  struct sentech_camera* cam_data = get_cam_data(fh);
  int retval = 0;
  __u32 index = (*fsize).index;;
;
  if(param_native_fsizes > 0) {
    (*fsize).type = V4L2_FRMSIZE_TYPE_DISCRETE;
    if ((index < 0) ||
	(index >= sentech_fps[cam_data->model_idx].fsizes)) {retval=-EINVAL; goto done;}
    (*fsize).discrete = sentech_fps[cam_data->model_idx].fsize[index];
  } else { 
    if(index != 0) { retval=-EINVAL; goto done;}
    (*fsize).type = V4L2_FRMSIZE_TYPE_STEPWISE;
    (*fsize).stepwise.min_width = 0; 
    (*fsize).stepwise.min_height = 0; 
    (*fsize).stepwise.max_width = cam_data->max_width;
    (*fsize).stepwise.max_height = cam_data->max_height;
    (*fsize).stepwise.step_width = 2;  
    switch((*fsize).pixel_format) {
    case  V4L2_PIX_FMT_GREY:
    case  V4L2_PIX_FMT_Y16:
    case  V4L2_PIX_FMT_SRGGB8:
    case  V4L2_PIX_FMT_SGRBG8:
    case  V4L2_PIX_FMT_SRGGB10:
    case  V4L2_PIX_FMT_SRGGB12:
    case  V4L2_PIX_FMT_SRGGB16:
    case V4L2_PIX_FMT_BGR24:
    case  V4L2_PIX_FMT_YUYV:
      break;
    case  V4L2_PIX_FMT_SBGGR8:
      (*fsize).stepwise.step_width = 2; 
      (*fsize).stepwise.step_height = 1; 
      (*fsize).stepwise.max_width = cam_data->max_width-2;
      (*fsize).stepwise.max_height = cam_data->max_height-2;
      break;
    case  V4L2_PIX_FMT_YUV420:
      (*fsize).stepwise.step_width = 2;  
      (*fsize).stepwise.step_height = 2; 
      break;
    case  V4L2_PIX_FMT_YUV422P:
      (*fsize).stepwise.step_width = 2;  
      (*fsize).stepwise.step_height = 1; 
      break;
    default:
      (*fsize).stepwise.step_width = 4;  
      (*fsize).stepwise.step_height = 1; 
    }
  }
 done:;
  return retval;
}
int sentech_vidioc_enum_frameintervals (struct file *file, void *fh,
					struct v4l2_frmivalenum *fival) {
  struct sentech_camera* cam_data = get_cam_data(fh);
  int retval = 0;
  __u32 index = (*fival).index;;
  (*fival).type = V4L2_FRMIVAL_TYPE_DISCRETE;
  if ((index < 0) ||
      (index >= 3)) {retval=-EINVAL; goto done;}
  (*fival).discrete = sentech_fps[cam_data->model_idx].fps[index];
 done:;
  return retval;
}
#endif
int set_SentechMode(struct sentech_camera* cam_data, int on);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,29)
static int sentech_open(struct inode *inode, struct file *file) {
#else
static int sentech_open(struct file *file) {
#endif
  enum v4l2_buf_type type = 0;
  struct sentech_fh *fh;
  struct video_device *dev = video_devdata(file);
  unsigned int retval = 0;
  struct sentech_camera* cam_data = video_get_drvdata(dev); 
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,29)
  int minor = iminor(inode);
#else
#endif
;
  lprintk(2,"sentech: In %s %d\n",__FUNCTION__, cam_data->open_desc);;
  if((! cam_data)||(! cam_data->connected)){
    printk(KERN_DEBUG "sentech:  Sentech USB Camera open has no data\n");
    retval = -ENODEV;
    goto open_exit_unlock;
  }
  if(cam_data->shutdown) {retval = -ENODEV; goto open_exit_unlock;}
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,29)
  if(minor != cam_data->v4l2dev->minor){
    printk(KERN_WARNING " Sentech USB Camera open UFO: try minor %d, got %d\n",
	   minor, cam_data->v4l2dev->minor);
    retval =  -ENODEV;
    goto open_exit_unlock;
  }
#endif
  if((cam_data->open_desc>0) && ((cam_data->exclusive))){
    printk(KERN_DEBUG " Sentech USB Camera BUSY\n");
    retval = -EBUSY;
    goto open_exit_unlock;
  }
  cam_data->type = type =  V4L2_BUF_TYPE_VIDEO_CAPTURE; 
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,29)
  dprintk (1,"open type=%s users=%d\n", v4l2_type_names[type], cam_data->open_desc);
#endif
  fh = alloc_sentech_fh(GFP_KERNEL);
  if (!fh) {
    retval = -ENOMEM;
    printk(KERN_DEBUG " Sentech USB Camera memory allocation failed\n");
    goto open_exit_unlock;
  }
  if(!cam_data->open_desc){
    if(!param_detect_multiple_prrobe_u3v)
      set_SentechMode(cam_data, 1);
    cam_data->run = 1;
    wake_up_interruptible(&cam_data->autoq); 
    if(cam_data->shutdown) {retval = -ENODEV; kfree (fh); goto open_exit_unlock;}
    cam_data->status = 0;
    cam_data->first_frame = 0;
    cam_data->last_frame = 0;
    cam_data->lost_frames = 0;;
    sentech_drop_images(cam_data);
    sentech_submit_kurbs(cam_data, GFP_KERNEL);
  }else;
  cam_data->open_desc++;  
  fh -> cam_data = cam_data;
  file->private_data = fh;
;
  kref_get(&cam_data->kref);
 open_exit_unlock:;
  lprintk(2,"sentech: Exiting %s od=%d\n",__FUNCTION__,cam_data->open_desc);
  return retval;
}
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,29)
static int sentech_release(struct inode * inode, struct file *file){
#else
static int sentech_release(struct file *file){
#endif
  struct sentech_camera* cam_data = get_cam_data(file->private_data);
  struct sentech_fh* fh = file->private_data;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,29)
#else
#endif
  int retval = 0;
  if(!cam_data || !fh){
    printk("sentech:  Sentech USB Camera RELEASE null\n");
    return retval;
  };
  lprintk(1,"In (l) %s od=%d\n",__FUNCTION__,cam_data->open_desc);
  if(cam_data->open_desc==1) {
    cam_data->reading=0;
    cam_data->streaming=0;
    cam_data->buffertype = 0; 
    cam_data->run=0;
    if(!param_detect_multiple_prrobe_u3v)
      set_SentechMode(cam_data, 0);
    lprintk(1,"%s closing device\n",__FUNCTION__);
    if(res_locked(cam_data)) {
      lprintk(1,"will free open %s\n",__FUNCTION__);
      res_free(cam_data);
      lprintk(1,"freed open %s\n",__FUNCTION__);
    }
  }
  lprintk(1,"Middle %s\n",__FUNCTION__);
  if(cam_data->exclusive) cam_data->exclusive=0;;
  kfree (fh);
  cam_data->open_desc--;;;
  kref_put(&cam_data->kref, cam_data_delete);;
  lprintk(1,"Exiting (l) %s od=%d\n",__FUNCTION__,cam_data->open_desc);
  return retval;
}
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,19)
 void v4l_printk_ioctl_arg(char *UNUSED(s),unsigned int UNUSED(cmd), void *UNUSED(arg)){}
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,29)
 static int sentech_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long ul_arg);
#else
 static long sentech_ioctl(struct file *file, unsigned int cmd, unsigned long ul_arg);
#endif
static ssize_t 
sentech_read(struct file *file, char __user *data, size_t count, loff_t *ppos){
  struct sentech_camera *cam_data = NULL;
  int retval = 0;
  int data_available = 0;
  struct sentech_usb_buffer* sub = NULL;
  struct sentech_usb_fragment* suf = NULL;
  struct video_device *vfd = video_devdata(file);
  DEFINE_WAIT(wait);;
  if(!vfd) {
    printk("sentech: No video device!\n");
    retval = -ENODEV;
    goto done;
  }
  cam_data = get_cam_data(file->private_data);
  if(cam_data == NULL) {
    printk("sentech: Video device is absent!\n");
    retval = -ENODEV;
    goto done;
  }
  if(!cam_data->connected) {
    printk("sentech read: Video device is disconnecting!\n");
    retval = -ENODEV;
    goto done;
  }
  if(!cam_data->run) {
    printk("sentech: Video device is stopped!\n");
    retval = -ENODEV;
    goto done;
  }
  if ((file->f_flags & O_TRIGGER)) {;
    retval = sentech_trigger(cam_data, GFP_KERNEL);
    if (retval != 0) goto done;
  }  
  if (!param_have_buffers) {;
    retval = -EINVAL;
    goto done;
  }
  do {
    if ((sub = get_has_data(cam_data,1)))
	data_available = 1;
    if (data_available)  break;
    if (cam_data->status) {
      retval = -EIO;;
      break;
    }
    if ((file->f_flags & O_NONBLOCK)) {;
      retval = -EAGAIN;
      break;
    }
    if(atomic_read(&cam_data->running_buffers) > 0) {
      prepare_to_wait(&cam_data->wq,
		      &wait,TASK_INTERRUPTIBLE);;
      dprintksubs(cam_data);
      if ((sub=get_has_data(cam_data, 1))) {
	data_available=1;
#ifdef LOCK_HAS_RAW_LOCK
	;;
#endif
      }else
	schedule_timeout (HZ);
      finish_wait(&cam_data->wq, &wait);
      if(signal_pending(current)) {;
	return -ERESTARTSYS;
      }
    } else {;
      if(cam_data->shutdown) {retval = -ENODEV; break;}
      cam_data->status = 0;
      sentech_submit_kurbs(cam_data, GFP_KERNEL);
      retval = -EAGAIN;
      break;
    }
  } while (!data_available);
;
  if(data_available) {;
    if((retval = down_interruptible(&cam_data->format_lock))){  
      retval = -ERESTARTSYS;
      goto format_exit;
    }
    if(cam_data->Bpp<=1){;
      retval = move_sub_to_user(sub,data,count);
    }else{;
      retval = move_sub_to_user16(sub,data,count);
    }
    up(&cam_data->format_lock);
  format_exit:;
    mb();;
    sub->has_data = 0;
    list_for_each_entry(suf, &sub->suf_buffers, buffers)
      suf->has_data=0;
    mb();
    sub->user_owned = 0;
  }
 done:;
  return retval;
}
static unsigned int
sentech_poll_indirect(struct file *file, struct poll_table_struct *wait){
  struct sentech_fh *fh = file->private_data;
  struct sentech_camera *cam_data = get_cam_data(fh);
  struct video_device *vfd = video_devdata(file);
  int found_buf;
  int retval = 0;
;
  lprintk(1,"In %s\n",__FUNCTION__);
  if(!vfd) {
    printk("sentech: No video device!\n");
    retval |= POLLERR;
    goto done;
  }
  if(cam_data == NULL) {
    printk("sentech: Video device is absent!\n");
    retval |= POLLERR;
    goto done;
  }
  if(!cam_data->connected) {
    printk("sentech poll: Video device is disconnecting!\n");
    retval |= POLLERR | POLLHUP; 
    goto done;
  }
  if(!cam_data->run) {
    printk("sentech: Video device is stopped!\n");
    retval |= POLLERR;
    goto done;
  }
  if(cam_data->shutdown) {retval = -ENODEV; goto done;}  
  if(unlikely( atomic_read(&cam_data->running_buffers) < SUBS_SUBMITTED)) {;
    sentech_submit_kurbs(cam_data, GFP_KERNEL);
  }
  if ((found_buf = find_oldest_buffer(cam_data, 1)) >= 0){
    if(debug!=0) printk("poll: Found data %d\n", found_buf);;
    retval |= POLLIN | POLLRDNORM;
  }else
    poll_wait(file, &(cam_data->wq), wait);
  if ((found_buf = find_oldest_buffer(cam_data, 1)) >= 0){
    if(debug!=0) printk("poll: Found data %d\n", found_buf);
    retval |= POLLIN | POLLRDNORM;
  }
  if (!atomic_read(&cam_data->running_buffers)
      )
    retval |= POLLERR;
 done:;
  lprintk(1,"Exiting %s with %d\n",__FUNCTION__,retval);
  return retval;
}
static unsigned int
sentech_poll_direct(struct file *file, struct poll_table_struct *wait){
  struct sentech_fh *fh = file->private_data;
  struct sentech_camera *cam_data = get_cam_data(fh);
  struct video_device *vfd = video_devdata(file);
  struct sentech_usb_buffer *sub;
  int retval = 0;
  lprintk(1,"In %s\n",__FUNCTION__);
  if(!vfd) {
    printk("sentech: No video device!\n");
    retval |= POLLERR;
    goto done;
  }
  if(cam_data == NULL) {
    printk("sentech: Video device is absent!\n");
    retval |= POLLERR;
    goto done;
  }
  if(!cam_data->connected) {
    printk("sentech poll: Video device is disconnecting!\n");
    retval |= POLLERR | POLLHUP; 
    goto done;
  }
  if(!cam_data->run) {
    printk("sentech: Video device is stopped!\n");
    retval |= POLLERR;
    goto done;
  }
  if(cam_data->shutdown) {retval = -ENODEV; goto done;}  
  if(unlikely( atomic_read(&cam_data->running_buffers) < SUBS_SUBMITTED)) {;
    sentech_submit_kurbs(cam_data, GFP_KERNEL);
  }
  poll_wait(file, &(cam_data->wq), wait);
  if ((sub=get_has_data(cam_data, 0))){
    if(debug!=0) printk("poll: Found data %d\n", sub->id);
    retval |= POLLIN | POLLRDNORM;
  }
  if (!atomic_read(&cam_data->running_buffers)
      )
    retval |= POLLERR;
 done:
  lprintk(1,"Exiting %s with %d\n",__FUNCTION__,retval);
  return retval;
}
static unsigned int
sentech_poll(struct file *file, struct poll_table_struct *wait) {
  int result;;
  if (param_have_buffers)
    result = sentech_poll_direct(file, wait);
  else
    result = sentech_poll_indirect(file, wait);;
  return result;
 }
void sentech_mmap_buffer_open_indirect (struct vm_area_struct *vma){
  int crt_buf;
  struct sentech_camera** _cam_data;
  struct sentech_camera* cam_data;;
  lprintk(2,"sentech: In %s\n",__FUNCTION__);
  _cam_data = (struct sentech_camera**)vma->vm_private_data;
  if(!_cam_data) {
    printk("<1> Nu sub for pgoff %ld in mmap!\n",
	   vma->vm_pgoff);
    return;
  };
  cam_data = *_cam_data;;
  crt_buf = _cam_data - cam_data->sib.vmalloc_buffers_context;;
  cam_data ->vmas ++;;
  cam_data->sib.vmalloc_mmaped[crt_buf] ++;;;
  lprintk(2,"Exiting %s\n",__FUNCTION__);
}
void sentech_mmap_buffer_open_direct (struct vm_area_struct *vma){
  struct sentech_usb_buffer *sub;
  struct sentech_camera* cam_data;;
  lprintk(2,"sentech: In %s\n",__FUNCTION__);
  sub=(struct sentech_usb_buffer*)vma->vm_private_data;
  if(!sub) {
    printk("<1> Nu sub for pgoff %ld in mmap!\n",
	   vma->vm_pgoff);
    return;
  }
  cam_data = sub->cam_data;
  cam_data ->vmas ++;
  sub->mmaped ++;;;
  lprintk(2,"Exiting %s\n",__FUNCTION__);
}
void sentech_mmap_buffer_open (struct vm_area_struct *vma){
  if (param_have_buffers)  sentech_mmap_buffer_open_direct (vma);
  else sentech_mmap_buffer_open_indirect (vma);
}
void sentech_mmap_buffer_close_indirect (struct vm_area_struct *vma){
  int crt_buf;
  struct sentech_camera **_cam_data;
  struct sentech_camera *cam_data;
  struct page* page; 
  unsigned int k;;
  lprintk(2,"sentech: In %s\n",__FUNCTION__);
  _cam_data = (struct sentech_camera**)vma->vm_private_data;
  if(!_cam_data) {
    printk("<1> Nu sub for pgoff %ld in mmap!\n",
	   vma->vm_pgoff);
    return;
  }
  cam_data = *_cam_data;
  crt_buf = _cam_data - cam_data->sib.vmalloc_buffers_context;
  cam_data->vmas --;
  cam_data->sib.vmalloc_mmaped[crt_buf] --;  
  if (!cam_data->sib.vmalloc_mmaped[crt_buf]) {
    if (!cam_data->sib.vmalloc_vbuffers[crt_buf] || cam_data->sib.vmalloc_vbuffers_size[crt_buf]==0){;
     } else {;
      for (k=0; k<cam_data->sib.vmalloc_vbuffers_size[crt_buf]; k += PAGE_SIZE) {
	page = vmalloc_to_page(cam_data->sib.vmalloc_vbuffers[crt_buf] + k);
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,14,0)
	put_page_testzero (page); 
#else
	if (page_count(page) > 1) put_page_testzero (page);
#endif
      }
      if (!param_keep_vbuffer) {
	if ((down_interruptible(&cam_data->format_lock))) {  
	  cam_data->sib.vmalloc_buffers_user[crt_buf] = 0;
	  return;
	}
	vfree(cam_data->sib.vmalloc_vbuffers[crt_buf]);;
	cam_data->sib.vmalloc_vbuffers[crt_buf] = 0;
	cam_data->sib.vmalloc_vbuffers_size[crt_buf] = 0;
	up(&cam_data->format_lock);
      }
    }
    cam_data->sib.vmalloc_buffers_user[crt_buf] = 0;
  };;
  lprintk(2,"sentech: Exiting %s\n",__FUNCTION__);
}
void sentech_mmap_buffer_close_direct (struct vm_area_struct *vma){
  struct sentech_usb_buffer *sub;
  struct sentech_camera *cam_data;
  struct page* page; 
  unsigned int k;;
  lprintk(2,"sentech: In %s\n",__FUNCTION__);
  sub=(struct sentech_usb_buffer*)vma->vm_private_data;
  if(!sub) {
    printk("<1> Nu sub for pgoff %ld in mmap!\n",
	   vma->vm_pgoff);
    return;
  }
  cam_data = sub->cam_data;
  cam_data->vmas --;
  sub->mmaped --;  
  if(!sub->mmaped) {
    if(!sub->vbuffer || sub->vbuffer_size==0){;
    }else{;
      for(k=0; k<sub->vbuffer_size; k += PAGE_SIZE) {
	page = vmalloc_to_page(sub->vbuffer + k);
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,14,0)
	put_page_testzero (page); 
#else
	if (page_count(page) > 1) put_page_testzero (page);
#endif
      }
      if (!param_keep_vbuffer) {
	vfree(sub->vbuffer);;
	sub->vbuffer = 0;
      }
    }
    sub->user_owned = 0;
  };;
  lprintk(2,"sentech: Exiting %s\n",__FUNCTION__);
}
void sentech_mmap_buffer_close (struct vm_area_struct *vma){
  if (param_have_buffers) sentech_mmap_buffer_close_direct (vma);
  else  sentech_mmap_buffer_close_indirect (vma);
}
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,24)
struct page* sentech_mmap_buffer_nopage (struct vm_area_struct *vma,
					 unsigned long address, int *type){
  unsigned long offset;
  unsigned long pgoff;
  struct sentech_usb_fragment* suf = NULL;
  struct sentech_usb_buffer* sub = 
    (struct sentech_usb_buffer*) vma->vm_private_data;
  struct page *page = NOPAGE_SIGBUS;
  void *pageptr = NULL;;
  lprintk(2,"sentech: In %s\n",__FUNCTION__);
  offset = (address - vma->vm_start);
;
  if(!sub->vbuffer) {
    if (offset >= sub->size) goto out;
    pgoff = (offset>>PAGE_SHIFT);
    list_for_each_entry(suf, &sub->suf_buffers, buffers) {
      if(pgoff > (suf->fragment_size>>PAGE_SHIFT)){
	pgoff -=   (suf->fragment_size>>PAGE_SHIFT);
	continue;
      }
      pageptr = suf->framebuffer_image + (pgoff<<PAGE_SHIFT);
      break;
    }
    if(!pageptr) goto out;
    if(virt_addr_valid(pageptr)) {
      page = virt_to_page(pageptr);
    }else{
      page = vmalloc_to_page(pageptr);
    }
  }else{
    if(offset>=sub->vbuffer_size){;
      goto out;
    }
    pageptr = sub->vbuffer + offset;;
    page = vmalloc_to_page(pageptr);
  };
  if (sub->vbuffer) {
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,14,0)
    get_page(page);
#else
    if (page_count(page) <= 0) init_page_count(page);
    else get_page(page);
#endif
  }
  if (type)
    *type = VM_FAULT_MINOR;
 out:;
  lprintk(1,"sentech: Exiting %s\n",__FUNCTION__);
  return page;
}
#endif
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,22)
int sentech_mmap_buffer_fault_indirect (struct vm_area_struct *vma,
				struct vm_fault *vmf){
  int flags=0;
  unsigned long off=0;
  struct sentech_camera** _cam_data;
  struct sentech_camera* cam_data;
  int crt_buf;
  struct page *page;
  void *pageptr = NULL; 
  int real_fragment_size;
;
 lprintk(1,"sentech: In %s\n",__FUNCTION__);
  _cam_data = (struct sentech_camera**) vma->vm_private_data;
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,24)
page = NULL;
#else
page = NOPAGE_SIGBUS;
#endif
  if(!_cam_data) {;
    goto out;
  }
  cam_data = *_cam_data;
  crt_buf =  _cam_data - cam_data->sib.vmalloc_buffers_context;
;
  off = (vmf->pgoff<<PAGE_SHIFT) - (crt_buf<<SBUF_IDX_SHIFT);
  if(!cam_data->sib.vmalloc_vbuffers[crt_buf]) {
    if(off>=cam_data->sib.vmalloc_buffers_size){;
      off = 0;
    }
    if(off>=cam_data->sib.vmalloc_buffers_size){;
      goto out;
    }
;
      real_fragment_size = cam_data->sib.vmalloc_buffers_size;
      if(off >= real_fragment_size){
	goto out;
      }
      pageptr = cam_data->sib.vmalloc_buffers[crt_buf] + off;;
    if(!pageptr) {;
      goto out;
    }
    if(virt_addr_valid(pageptr)) {
      page = virt_to_page(pageptr);
    }else{
      page = vmalloc_to_page(pageptr);
    }
  } else {
    if(off>=cam_data->sib.vmalloc_vbuffers_size[crt_buf]){;
      goto out;
    }
    pageptr = cam_data->sib.vmalloc_vbuffers[crt_buf] + off;;
    page = vmalloc_to_page(pageptr);
  };
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,14,0)
  get_page(page);
#else
  if (page_count(page) <= 0) init_page_count(page);
  else get_page(page);
#endif
 out:
  vmf->page = page;
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,7,0)
  vmf->flags |= VM_CAN_NONLINEAR;
#else
  vma->vm_flags |= VM_IO;
#endif
  if(vmf->page == NOPAGE_SIGBUS){
    flags |= VM_FAULT_SIGBUS;;
  };
  lprintk(1,"Exiting %s return %d\n",__FUNCTION__,flags);
  return flags;
}
int sentech_mmap_buffer_fault_direct (struct vm_area_struct *vma,
				struct vm_fault *vmf){
  int flags=0;
  unsigned long off=0;
  struct sentech_usb_fragment* suf = NULL;
  struct sentech_usb_buffer* sub = 
    (struct sentech_usb_buffer*) vma->vm_private_data;
  struct page *page;
  void *pageptr = NULL; 
  int real_fragment_size;
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,24)
page = NULL;
#else
page = NOPAGE_SIGBUS;
#endif
;
 lprintk(1,"sentech: In %s\n",__FUNCTION__);;
  if(!sub) {;
    goto out;
  }
  off = (vmf->pgoff<<PAGE_SHIFT) - (sub->id<<SBUF_IDX_SHIFT);
  if(!sub->vbuffer) {
    if(off>=sub->size){;
      off = 0;
    }
    if(off>=sub->size){;
      goto out;
    }
    list_for_each_entry(suf, &sub->suf_buffers, buffers) {;
      real_fragment_size = suf->fragment_size;
      if(suf->first_suf){
	real_fragment_size += sizeof(struct Sentech_Header);;
      }else{;
      }
      if(off >= real_fragment_size){
	off -=  (real_fragment_size);
	continue;
      }
      pageptr = suf->framebuffer_image + off;;
      break;
    }
    if(!pageptr) {;
      goto out;
    }
    if(virt_addr_valid(pageptr)) {
      page = virt_to_page(pageptr);
    }else{
      page = vmalloc_to_page(pageptr);
    }
  } else {
    if(off>=sub->vbuffer_size){;
      goto out;
    }
    pageptr = sub->vbuffer + off;;
    page = vmalloc_to_page(pageptr);
  };
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,14,0)
  get_page(page);
#else
  if (page_count(page) <= 0) init_page_count(page);
  else get_page(page);
#endif
 out:
  vmf->page = page;
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,7,0)
  vmf->flags |= VM_CAN_NONLINEAR;
#else
  vma->vm_flags |= VM_IO;
#endif
  if(vmf->page == NOPAGE_SIGBUS){
    flags |= VM_FAULT_SIGBUS;;
  };
  lprintk(1,"Exiting %s return %d\n",__FUNCTION__,flags);
  return flags;
}
int sentech_mmap_buffer_fault (struct vm_area_struct *vma,
				struct vm_fault *vmf){
  if (param_have_buffers) return sentech_mmap_buffer_fault_direct (vma, vmf);
  return sentech_mmap_buffer_fault_indirect (vma, vmf);
}
#endif
static struct vm_operations_struct sentech_vm_ops = {
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,22)
  .fault = sentech_mmap_buffer_fault,
#endif
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,24)
  .nopage = sentech_mmap_buffer_nopage,
#endif
  .open = sentech_mmap_buffer_open,
  .close = sentech_mmap_buffer_close,
};
static int 
sentech_mmap_mapper_indirect(struct sentech_camera *cam_data,
		    struct vm_area_struct *vma) {
  unsigned int vma_size=0;
  unsigned int size = 0, k=0;
  __u32 index = (vma->vm_pgoff<<PAGE_SHIFT)>>SBUF_IDX_SHIFT;
  int crt_buf;
  struct page* page; 
;
  lprintk(2,"In %s\n",__FUNCTION__);;
  vma_size = (unsigned long)vma->vm_end-(unsigned long)vma->vm_start;;
  vma->vm_ops = &sentech_vm_ops;
  crt_buf = index;
  vma->vm_private_data = &(cam_data->sib.vmalloc_buffers_context[crt_buf]);
  cam_data->sib.vmalloc_mapped_size[crt_buf] = vma_size;
  if (
      (
       (! cam_data->enable_native) ||
       (! param_bytesperline &&
	(cam_data->pix_format.width != cam_data->max_width))
       ||
       (cam_data->hflip)
       ||
       (
	((cam_data->pix_format.pixelformat != V4L2_PIX_FMT_RGB24)
	 || !(cam_data->mask_Bpp & STC_THREE_BYTES_PER_PIXEL_RGB8)) &&
	((cam_data->pix_format.pixelformat != V4L2_PIX_FMT_BGR24)
	 || !(cam_data->mask_Bpp & STC_THREE_BYTES_PER_PIXEL_RGB8)) &&
	(cam_data->pix_format.pixelformat != V4L2_PIX_FMT_SRGGB8) &&
	(cam_data->pix_format.pixelformat != V4L2_PIX_FMT_SGRBG8) &&
	(cam_data->pix_format.pixelformat != V4L2_PIX_FMT_SRGGB10) &&
	(cam_data->pix_format.pixelformat != V4L2_PIX_FMT_SRGGB12) &&
	(cam_data->pix_format.pixelformat != V4L2_PIX_FMT_SRGGB16) &&
	(cam_data->pix_format.pixelformat != V4L2_PIX_FMT_GREY)
	)
       ) 
      )
    {
      size = cam_data->pix_format.sizeimage;;
      if (size & ((1<<PAGE_SHIFT)-1)) {
	size += 1<<PAGE_SHIFT;
	size &= ~((1<<PAGE_SHIFT)-1);
      };
      if (param_keep_vbuffer) {
	if (cam_data->sib.vmalloc_vbuffers[crt_buf] && (cam_data->sib.vmalloc_vbuffers_size[crt_buf] != size+MARGIN)) {
	  if ((down_interruptible(&cam_data->format_lock))) {
	    return -ERESTARTSYS;
	  }
	  vfree(cam_data->sib.vmalloc_vbuffers[crt_buf]);;
	  cam_data->sib.vmalloc_vbuffers[crt_buf] = 0;
	  cam_data->sib.vmalloc_vbuffers_size[crt_buf] = 0;
	  up(&cam_data->format_lock);
	}
	if (! cam_data->sib.vmalloc_vbuffers[crt_buf]) {
	  cam_data->sib.vmalloc_vbuffers[crt_buf] = vmalloc( size + MARGIN);;
	  if (cam_data->sib.vmalloc_vbuffers[crt_buf] == NULL) {;
	    cam_data->sib.vmalloc_vbuffers_size[crt_buf] = 0;
	    return -EINVAL;
	  }
	}
	for (k = 0; k < size + MARGIN; k += PAGE_SIZE) {
	  page = vmalloc_to_page(cam_data->sib.vmalloc_vbuffers[crt_buf] + k);
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,14,0)
	  get_page (page); 
#else
	  if (page_count(page) <= 0) init_page_count(page);
	  else get_page(page);
#endif
	}    
	cam_data->sib.vmalloc_vbuffers_size[crt_buf] = size+MARGIN;;
      } else {
	if (! cam_data->sib.vmalloc_vbuffers[crt_buf]) {
	  cam_data->sib.vmalloc_vbuffers[crt_buf] = vmalloc( size + MARGIN);
	  if (cam_data->sib.vmalloc_vbuffers[crt_buf] == NULL) {;
	    cam_data->sib.vmalloc_vbuffers_size[crt_buf] = 0;
	    return -EINVAL;
	  }
	  for (k = 0; k < size + MARGIN; k += PAGE_SIZE) {
	    page = vmalloc_to_page(cam_data->sib.vmalloc_vbuffers[crt_buf] + k);
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,14,0)
	    get_page (page); 
#else
	    if (page_count(page) <= 0) init_page_count(page);
	    else get_page(page);
#endif
	  }    
	  cam_data->sib.vmalloc_vbuffers_size[crt_buf] = size+MARGIN;;
	} else {
	  printk("%s sentech camera user attempts 2nd mapping of index %u\n", __FUNCTION__, index);
	}
      }
      cam_data->native = 0;      
    }
  else
    {;;
      if (param_keep_vbuffer) {
	if (cam_data->sib.vmalloc_vbuffers[crt_buf] && (cam_data->sib.vmalloc_vbuffers_size[crt_buf] > 0)) {
	  if ((down_interruptible(&cam_data->format_lock))) {
	    return -ERESTARTSYS;
	  }
	  vfree(cam_data->sib.vmalloc_vbuffers[crt_buf]);;
	  cam_data->sib.vmalloc_vbuffers[crt_buf] = 0;
	  cam_data->sib.vmalloc_vbuffers_size[crt_buf] = 0;
	  up(&cam_data->format_lock);
	}
      }
      cam_data->native = 1;;
      lprintk(2,"%s Exiting\n",__FUNCTION__);
      if (vma_size  >  sentech_pagerounded(cam_data->max_frame_size)+MARGIN) 
	{
	  printk("VMA %lx larger than scanimage_size %x(max frame size %x)+Margin(%lx).\n",
		 vma->vm_end - vma->vm_start,
		 cam_data->scanimage_size,
		 cam_data->max_frame_size,
		 MARGIN);
	  return -EINVAL;
	}
    }
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,7,0)
  vma->vm_flags |= VM_RESERVED | VM_IO;
#else
  vma->vm_flags |= VM_PFNMAP | VM_IO;
#endif
  sentech_mmap_buffer_open(vma);;
  if (debug != 0) printk("sentech: %s Exiting\n",__FUNCTION__);
  return 0;
} 
static int 
sentech_mmap_mapper_direct(struct sentech_camera *cam_data,
		    struct vm_area_struct *vma) {
  unsigned int vma_size=0;
  unsigned int size = 0, k=0;
  __u32 index = (vma->vm_pgoff<<PAGE_SHIFT)>>SBUF_IDX_SHIFT;
  struct page* page; 
  struct sentech_usb_buffer *sub = NULL;;
  lprintk(2,"In %s\n",__FUNCTION__);;
  vma_size = (unsigned long)vma->vm_end-(unsigned long)vma->vm_start;;
  vma->vm_ops = &sentech_vm_ops;
  sub = get_sub(cam_data, index);
  vma->vm_private_data = sub;
  sub->mapped_size = vma_size;
  if (
     (
     (!cam_data->enable_native) ||
     (!param_bytesperline &&
      (cam_data->pix_format.width != cam_data->max_width))
     ||
     (cam_data->hflip)
     ||
     (
      ((cam_data->pix_format.pixelformat != V4L2_PIX_FMT_RGB24)
       || !(cam_data->mask_Bpp & STC_THREE_BYTES_PER_PIXEL_RGB8)) &&
      ((cam_data->pix_format.pixelformat != V4L2_PIX_FMT_BGR24)
       || !(cam_data->mask_Bpp & STC_THREE_BYTES_PER_PIXEL_RGB8)) &&
      (cam_data->pix_format.pixelformat != V4L2_PIX_FMT_SRGGB8) &&
      (cam_data->pix_format.pixelformat != V4L2_PIX_FMT_SGRBG8) &&
      (cam_data->pix_format.pixelformat != V4L2_PIX_FMT_SRGGB10) &&
      (cam_data->pix_format.pixelformat != V4L2_PIX_FMT_SRGGB12) &&
      (cam_data->pix_format.pixelformat != V4L2_PIX_FMT_SRGGB16) &&
      (cam_data->pix_format.pixelformat != V4L2_PIX_FMT_GREY)
      )
     ) 
     )
    {
      size = cam_data->pix_format.sizeimage;;
      if (size & ((1<<PAGE_SHIFT)-1)) {
	size += 1<<PAGE_SHIFT;
	size &= ~((1<<PAGE_SHIFT)-1);
      };
      if (param_keep_vbuffer) {
	if (sub->vbuffer && (sub->vbuffer_size != size+MARGIN)) {
	  vfree(sub->vbuffer);;
	  sub->vbuffer = 0;
	}
	if (! sub->vbuffer) {
	  sub->vbuffer = vmalloc( size + MARGIN);;
	  if (sub->vbuffer == NULL) {;;
	    return -EINVAL;
	  }
	};
	for (k=0; k<size+MARGIN; k += PAGE_SIZE) {
	  page = vmalloc_to_page(sub->vbuffer + k);
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,14,0)
	  get_page (page); 
#else
	  if (page_count(page) <= 0) init_page_count(page);
	  else get_page(page);
#endif
	}    
	sub->vbuffer_size = size+MARGIN;;
      } else {
	if (! sub->vbuffer) {
	  sub->vbuffer = vmalloc( size + MARGIN);
	  if (sub->vbuffer == NULL) {;
	    return -EINVAL;
	  }
;
	  for(k=0; k<size+MARGIN; k += PAGE_SIZE) {
	    page = vmalloc_to_page(sub->vbuffer + k);
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,14,0)
	    get_page (page); 
#else
	    if (page_count(page) <= 0) init_page_count(page);
	    else get_page(page);
#endif
	  }
	  sub->vbuffer_size = size+MARGIN;;
	} else {
	  printk("%s sentech camera user attempts 2nd mapping of index %u\n", __FUNCTION__, index);
	}
      }
      cam_data->native = 0;      
    } else {;
    if (param_keep_vbuffer) {
      if(sub->vbuffer && (sub->vbuffer_size > 0)) {
	vfree(sub->vbuffer);;
	sub->vbuffer = 0;
      }
    }
    cam_data->native = 1;;
    lprintk(2,"%s Exiting\n",__FUNCTION__);
    if(vma_size  >  sentech_pagerounded(cam_data->max_frame_size)+MARGIN) 
      {
	printk("VMA %lx larger than scanimage_size %x(max frame size %x)+Margin(%lx).\n",
	       vma->vm_end - vma->vm_start,
	       cam_data->scanimage_size,
	       cam_data->max_frame_size,
	       MARGIN);
	return -EINVAL;
      }
  }
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,7,0)
  vma->vm_flags |= VM_RESERVED | VM_IO;
#else
  vma->vm_flags |= VM_PFNMAP | VM_IO;
#endif
  sentech_mmap_buffer_open(vma);
  if(debug!=0)printk("sentech: %s Exiting\n",__FUNCTION__);
  return 0;
} 
static int 
sentech_mmap_mapper(struct sentech_camera *cam_data,
		    struct vm_area_struct *vma) {
  if (param_have_buffers) return sentech_mmap_mapper_direct(cam_data, vma);
  return sentech_mmap_mapper_indirect(cam_data, vma);
 }
static int
sentech_mmap(struct file *file, struct vm_area_struct * vma){
  struct sentech_camera *cam_data = get_cam_data(file->private_data);
  int retval = 0;
;
  lprintk(1,"In %s\n",__FUNCTION__);;
  if((! cam_data)||(! cam_data->connected)){
    retval = -ENODEV;
   goto mmap_exit_unlock;
  };;;
  retval=sentech_mmap_mapper(cam_data, vma);
;
 mmap_exit_unlock:
;
  lprintk(1,"Exiting %s\n",__FUNCTION__);
  return retval;
}
#ifdef HAVE_UNLOCKED_IOCTL
 inline int sentech_acquire_camera_ID(struct sentech_camera* cam_data, struct camera_id* ID){
  int res=0, retval =0;
  if(cam_data->usb3) {
    res=usb_control_msg(cam_data->udev, 
			cam_data->pipe_ctl_rcv, 
			Sentech_EEPROM_Request_Aptima, 
			USB_DIR_IN | USB_TYPE_VENDOR |USB_RECIP_DEVICE, 
			Sentech_EEPROM_PRODUCT_INFO_Aptima,  
			Sentech_EEPROM_100KHZ_Aptima, 
			ID, 
			sizeof(struct camera_id), CTRL_TIMEOUT);
  }else{
    res=usb_control_msg(cam_data->udev,
			cam_data->pipe_ctl_rcv,
			Sentech_EEPROM_Request, 
			USB_DIR_IN | USB_TYPE_VENDOR |USB_RECIP_DEVICE, 
			Sentech_EEPROM_PRODUCT_INFO, 
			Sentech_EEPROM_100KHZ, 
			ID,
			sizeof(struct camera_id), CTRL_TIMEOUT);
  }
  if(res!=sizeof(struct camera_id)){
    dev_err(&cam_data->interface->dev, "SentechUSB: resultID=%d\n",res);
    retval = res;
  }else{;
  }
  return retval;
}
 inline int sentech_set_camera_ID(struct sentech_camera* cam_data, struct camera_id* ID){
   int res=0, retval =0, msgl=0;
   __u8* payload;
   msgl = sizeof(struct camera_id)-STC_PRODUCT_INFO_LEN;;
  res = msgl;
  payload = (__u8*) ID;
  if(cam_data->usb3) {
    res=usb_control_msg(cam_data->udev,
			cam_data->pipe_ctl_snd,
			Sentech_EEPROM_Request_Aptima,
			USB_DIR_OUT | USB_TYPE_VENDOR |USB_RECIP_DEVICE, 
			Sentech_EEPROM_ID_ADDRESS_Aptima, 
			Sentech_EEPROM_100KHZ_Aptima, 
			payload+STC_PRODUCT_INFO_LEN,
			msgl, CTRL_TIMEOUT);
  }else{
    res=usb_control_msg(cam_data->udev,
			cam_data->pipe_ctl_snd,
			Sentech_EEPROM_Request,
			USB_DIR_OUT | USB_TYPE_VENDOR |USB_RECIP_DEVICE, 
			Sentech_EEPROM_ID_ADDRESS,
			Sentech_EEPROM_100KHZ,
			payload+STC_PRODUCT_INFO_LEN,
			msgl, CTRL_TIMEOUT);
  }
  if (res != msgl) {
    dev_err(&cam_data->interface->dev, "SentechUSB: resultID=%d\n", res);
    retval = res;
  } else {;
  }
  return retval;
}
 static void sentech_complete_save_settings(struct urb * urb ARG_COMPLETE) {
  int __attribute__((unused))  k;
  struct ctrl_fpga2_context* ctx = (struct ctrl_fpga2_context*)urb->context;;
  if (urb->status < 0);
  usb_free_urb(urb);
  complete(&ctx->done);
  ctx->end = 1;;
 }
 static int read_camera_settings(struct sentech_camera* cam_data);
 static int sentech_save_settings(struct sentech_camera* cam_data, struct sentech_settings* save_settings,
				  int memflags) {
  int retval;
  struct ctrl_fpga2_context* ctx;
  __u16 wVal = 0;
  __u16 wIndex = 0;
  __u16 wLen = 2; 
  __u8 _requesttype =
    USB_TYPE_VENDOR |
    USB_RECIP_DEVICE;
  void *data = cam_data->save_settings_cmd;;
  switch (save_settings->object) {
  case SENTECH_SETTINGS_SAVING_OBJ_PIX_BC:
  case SENTECH_SETTINGS_SAVING_OBJ_CAMERA:
    break;
  case 0: 
    if (save_settings->operation == SENTECH_SETTINGS_SAVING_OPP_RESET)
      break;
  default: return -EINVAL;
  }
  switch (save_settings->operation) {
  case SENTECH_SETTINGS_SAVING_OPP_READ:
    break;
  case SENTECH_SETTINGS_SAVING_OPP_WRITE:
    break;
  case SENTECH_SETTINGS_SAVING_OPP_RESET:
    if (save_settings->object == 0) break;
  default:
    return -EINVAL;
  }
  cam_data->save_settings_cmd[1] = (save_settings->operation << 4) | ((save_settings->object >> 8) & 0x0f);
  cam_data->save_settings_cmd[0] = (save_settings->object & 0x00ff);
  ctx = kmalloc(sizeof(struct ctrl_fpga2_context), memflags);
  if (! ctx) {
    retval = -ENOMEM;;
    return retval;
  }
  ctx->ID = cID++;
  ctx->end = 0;
  ctx->cam_data = cam_data;
  ctx->data = 0;
  ctx->p_data = 0; 
  init_completion(&ctx->done);
  list_add_tail(&ctx->queue, &cam_data->ctrl_queue);
  retval =
    usb_control_msg_async(cam_data->udev, cam_data->pipe_ctl_rcv,
			  0x81, 
			  _requesttype, 
			  wVal, 
			  wIndex, 
			  data, 
			  wLen, 
			  sentech_complete_save_settings, 
			  memflags,
			  &ctx->pb, ctx);;
  if (! retval) {;
    switch (save_settings->operation) {
    case SENTECH_SETTINGS_SAVING_OPP_WRITE:
      break;
    case SENTECH_SETTINGS_SAVING_OPP_RESET:
    case SENTECH_SETTINGS_SAVING_OPP_READ:;
      wait_for_completion_timeout(&ctx->done,HZ);
      read_camera_settings(cam_data);
      break;
    default:
	break;
    }
  }
;
  return retval;
 }
 static long sentech_unlocked_ioctl (struct file *filp, unsigned int cmd,
			      unsigned long arg) {
   struct sentech_camera* cam_data = get_cam_data(filp->private_data);
   long err, retval=-ENOTTY;
   __s16 * code;
   __s32 * release;
   struct camera_id* id;
   struct sentech_fpga2* fpga2cmd;
   struct sentech_settings* save_settings;
   unsigned long flags;
   int sync; 
   switch (cmd) {
   case SENTECH_IOC_VERSION:
     code = (__s16*)arg;
     *code = (SENTECH_MAJOR_VERSION<<8)+SENTECH_MINOR_VERSION;
     retval=0;
     break;
   case SENTECH_IOC_GET_SCAN_HEIGHT:
     code = (__s16*)arg;
     *code = cam_data->scan_height;
     retval=0;
     break;
   case SENTECH_IOC_GET_SCAN_WIDTH:
     code = (__s16*)arg;
     *code = cam_data->scan_width;
     retval=0;
     break;
   case SENTECH_IOC_SAVE_SETTINGS:
     if (! cam_data->saving_camera_settings_fpga_can) {
       retval = -EINVAL;
       break;
     }
     save_settings = (struct sentech_settings*) arg;
     retval = sentech_save_settings(cam_data, save_settings, GFP_KERNEL);
     break;
   case SENTECH_IOC_RELEASE:
     release = (__s32*)arg;
     *release = (SENTECH_MAJOR_VERSION<<8)+SENTECH_MINOR_VERSION+(SENTECH_RELEASE<<16);
     retval=0;
     break;
   case SENTECH_IOC_G_ID:
     id = (struct camera_id*)arg;
     err = !access_ok(VERIFY_WRITE, (void *)arg, _IOC_SIZE(cmd));
     if ((id == NULL) || err) {
       retval = -EFAULT;
       break;
     }
     if (sentech_acquire_camera_ID(cam_data, &(cam_data->ID_buf))) {
       printk(KERN_INFO "sentech: error get ID=%d len=%hhd: (%hhd) \"%s\"\n",
             cam_data->ID_buf.camera_ID, cam_data->ID_buf.length, cam_data->ID_buf.type, cam_data->ID_buf.name);
     } else {
       err = copy_to_user (id, &(cam_data->ID_buf), sizeof(struct camera_id));
       retval = err;
     }
     break;
   case SENTECH_IOC_S_ID:
     id = (struct camera_id*)arg;
     err = !access_ok(VERIFY_READ, (void *)arg, _IOC_SIZE(cmd));
     if((id==NULL) || err) {
       retval=-EFAULT;
       break;
     }
     err=copy_from_user (&(cam_data->ID_buf), id, sizeof(struct camera_id));
     if(err
	|| (cam_data->ID_buf.type & 0x7F)
	|| (cam_data->ID_buf.length > sizeof(cam_data->ID_buf.name))
	|| (!cam_data->ID_buf.type && 
	    (
	     cam_data->ID_buf.name[cam_data->ID_buf.length]
	     ||
	     (cam_data->ID_buf.length != strlen(cam_data->ID_buf.name)+1)
	     )
	    )
	) {
       printk(KERN_ALERT "Using SENTECH_S_ID memcopy error=%ld\n",err);
       retval=-EFAULT;
       break;
     }
     printk(KERN_INFO "sentech: trying to set (#%zu) ID=%d len=%hhd: (%hhx) \"%s\"\n",
	    sizeof(struct camera_id), cam_data->ID_buf.camera_ID, cam_data->ID_buf.length, cam_data->ID_buf.type, cam_data->ID_buf.name);
     if ((retval=sentech_set_camera_ID(cam_data, &cam_data->ID_buf))) {
       printk(KERN_INFO "sentech: error set ID=%d len=%hhd: (%hhd) \"%s\"\n",
	      cam_data->ID_buf.camera_ID, cam_data->ID_buf.length, cam_data->ID_buf.type, cam_data->ID_buf.name);
     }else retval=0;
     break;
   case SENTECH_IOC_S_CATEGORY:
     fpga2cmd = (struct sentech_fpga2*)arg;
     err = !access_ok(VERIFY_READ, (void *)arg, _IOC_SIZE(cmd));
     if ((fpga2cmd == NULL) || err || !cam_data->u3v) {
       retval = -EFAULT;
       break;
     }
     write_lock_irqsave(&cam_data->fpga2_lock, flags);
     err = copy_from_user (&(cam_data->fpga2_buf), fpga2cmd, sizeof(struct sentech_fpga2));
     if (err || (! cam_data->fpga2_buf.direct && ! cam_data->fpga2_offsets_known_flag)) {
       write_unlock_irqrestore(&cam_data->fpga2_lock, flags);
       retval = -ENODEV;
       break;
     }
     retval = sentech_fpga2_set(cam_data, GFP_KERNEL,
		       cam_data->fpga2_buf.category_offset,
		       cam_data->fpga2_buf.inside_category_offset,
		       &(cam_data->fpga2_buf.data), 4);
     write_unlock_irqrestore(&cam_data->fpga2_lock, flags);
     if (retval) {
       break;
     }
     break;
   case SENTECH_IOC_G_CATEGORY:
     fpga2cmd = (struct sentech_fpga2*)arg;
     err = ! access_ok(VERIFY_READ, (void *)arg, _IOC_SIZE(cmd));
     if ((fpga2cmd == NULL) || err || ! cam_data->u3v) {
       retval = -EFAULT;
       break;
     }
     err = ! access_ok(VERIFY_WRITE, (void *) arg, _IOC_SIZE(cmd));
     if (err) {
       retval = -EFAULT;
       break;
     }
     write_lock_irqsave(&cam_data->fpga2_lock, flags);
     err = copy_from_user (&(cam_data->fpga2_buf), fpga2cmd, sizeof(struct sentech_fpga2));
     if (err) {
       write_unlock_irqrestore(&cam_data->fpga2_lock, flags);
       retval = -EFAULT;
       break;
     }
     if (! cam_data->fpga2_buf.direct && ! cam_data->fpga2_offsets_known_flag) {
       write_unlock_irqrestore(&cam_data->fpga2_lock, flags);
       retval = -ENODEV;
       break;
     }
     retval = sentech_fpga2_get_sync(cam_data, GFP_KERNEL,
		       cam_data->fpga2_buf.category_offset,
		       cam_data->fpga2_buf.inside_category_offset,
		       &(cam_data->fpga2_buf.data), 4);
     if (retval) {
       write_unlock_irqrestore(&cam_data->fpga2_lock, flags);
       break;
     }
     err = copy_to_user (fpga2cmd, &(cam_data->fpga2_buf), sizeof(struct sentech_fpga2));
     write_unlock_irqrestore(&cam_data->fpga2_lock, flags);
     if (err) {
       retval = -EFAULT;
       break;
     }
     break;
   case SENTECH_IOC_LOAD_OFFSET:
     fpga2cmd = (struct sentech_fpga2*)arg;
     err = ! access_ok(VERIFY_READ, (void *)arg, _IOC_SIZE(cmd));
     if ((fpga2cmd == NULL) || err || ! cam_data->u3v) {
       retval = -EFAULT;
       break;
     }
     err = ! access_ok(VERIFY_WRITE, (void *) arg, _IOC_SIZE(cmd));
     if (err) {
       retval = -EFAULT;
       break;
     }
     write_lock_irqsave(&cam_data->fpga2_lock, flags);
     err = copy_from_user (&(cam_data->fpga2_buf), fpga2cmd, sizeof(struct sentech_fpga2));
     if (err) {
       write_unlock_irqrestore(&cam_data->fpga2_lock, flags);
       retval = -EFAULT;
       break;
     }
     if (! cam_data->u3v) {
       break;
     }
     sync = cam_data->fpga2_buf.inside_category_offset;;
     switch (cam_data->fpga2_buf.category_offset) {
     case CATEGORY_OFFSET_ANALOG_CTL:
       retval = sentech_fpga2_read_offset_address(cam_data, GFP_KERNEL, CATEGORY_OFFSET_ANALOG_CTL, &cam_data->fpga2_offset_analog_control, sync, sizeof(cam_data->fpga2_offset_analog_control));
       cam_data->fpga2_buf.inside_category_offset = cam_data->fpga2_offset_analog_control;
       break;
     case CATEGORY_OFFSET_VENDOR_UNIQ_CTRL:
       retval = sentech_fpga2_read_offset_address(cam_data, GFP_KERNEL, CATEGORY_OFFSET_VENDOR_UNIQ_CTRL, &cam_data->fpga2_offset_vendor_unique_control, sync, sizeof(cam_data->fpga2_offset_vendor_unique_control));
       cam_data->fpga2_buf.inside_category_offset = cam_data->fpga2_offset_vendor_unique_control;
       break;
     case CATEGORY_OFFSET_USB_ORIG_SAVE:
       retval = sentech_fpga2_read_offset_address(cam_data, GFP_KERNEL, CATEGORY_OFFSET_USB_ORIG_SAVE, &cam_data->fpga2_offset_USB_camera_original_function__save, sync, sizeof(cam_data->fpga2_offset_USB_camera_original_function__save));
       cam_data->fpga2_buf.inside_category_offset = cam_data->fpga2_offset_USB_camera_original_function__save;
       break;
     case CATEGORY_OFFSET_CAMERA_STATUS:
       retval = sentech_fpga2_read_offset_address(cam_data, GFP_KERNEL, CATEGORY_OFFSET_CAMERA_STATUS, &cam_data->fpga2_offset_USB_camera_status, sync, sizeof(cam_data->fpga2_offset_USB_camera_status));
       cam_data->fpga2_buf.inside_category_offset = cam_data->fpga2_offset_USB_camera_status;
       break;
     case CATEGORY_OFFSET_CAMERA_FEATURE:
       retval = sentech_fpga2_read_offset_address(cam_data, GFP_KERNEL, CATEGORY_OFFSET_CAMERA_FEATURE, &cam_data->fpga2_offset_USB_camera_feature, sync, sizeof(cam_data->fpga2_offset_USB_camera_feature));
       cam_data->fpga2_buf.inside_category_offset = cam_data->fpga2_offset_USB_camera_feature;
       break;
     case CATEGORY_OFFSET_HDR_CONTROL:
       retval = sentech_fpga2_read_offset_address(cam_data, GFP_KERNEL, CATEGORY_OFFSET_HDR_CONTROL, &cam_data->fpga2_offset_HDR_control, sync, sizeof(cam_data->fpga2_offset_HDR_control));
       cam_data->fpga2_buf.inside_category_offset = cam_data->fpga2_offset_HDR_control;
     }
     if (retval) {
       err = copy_to_user (fpga2cmd, &(cam_data->fpga2_buf), sizeof(struct sentech_fpga2));
       write_unlock_irqrestore(&cam_data->fpga2_lock, flags);
       break;
     }
     err = copy_to_user (fpga2cmd, &(cam_data->fpga2_buf), sizeof(struct sentech_fpga2));
     write_unlock_irqrestore(&cam_data->fpga2_lock, flags);
     if (err) {
       retval = -EFAULT;
       break;
     }
     break;
   default:
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,29)
     return sentech_ioctl(filp->f_dentry->d_inode, filp, cmd, arg);
#else
     return sentech_ioctl(filp, cmd, arg);
#endif
   }
   return retval;
 }
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,29)
static struct file_operations sentech_fops = {
	.owner		= THIS_MODULE,
	.llseek         = no_llseek,
	.open           = sentech_open,
	.release        = sentech_release,
	.read           = sentech_read,
	.poll           = sentech_poll,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,11)
	.compat_ioctl	= v4l_compat_ioctl32,
#endif
	.ioctl          = sentech_ioctl, 
	.mmap		= sentech_mmap,
 };
#else
static struct v4l2_file_operations sentech_fops = {
	.owner		= THIS_MODULE,
	.open           = sentech_open,
	.release        = sentech_release,
	.read           = sentech_read,
	.poll           = sentech_poll,
	.ioctl          = sentech_ioctl, 
#ifdef HAVE_UNLOCKED_IOCTL
	.unlocked_ioctl          = sentech_unlocked_ioctl,
#endif
	.mmap		= sentech_mmap,
};
#endif
static
struct v4l2_ioctl_ops sentech_ioctl_ops = {
  .vidioc_querycap = sentech_vidioc_querycap,
  .vidioc_g_priority = sentech_vidioc_g_priority,
  .vidioc_s_priority = sentech_vidioc_s_priority,
  .vidioc_enum_fmt_vid_cap = sentech_vidioc_enum_fmt_cap,
  .vidioc_g_fmt_vid_cap = sentech_vidioc_g_fmt_cap,
  .vidioc_s_fmt_vid_cap = sentech_vidioc_s_fmt_cap,
  .vidioc_try_fmt_vid_cap = sentech_vidioc_try_fmt_cap,
  .vidioc_reqbufs = sentech_vidioc_reqbufs,
  .vidioc_querybuf = sentech_vidioc_querybuf,
  .vidioc_qbuf = sentech_vidioc_qbuf,
  .vidioc_dqbuf = sentech_vidioc_dqbuf,
#ifdef HAVE_V4L1
#endif
  .vidioc_streamon = sentech_vidioc_streamon,
  .vidioc_streamoff = sentech_vidioc_streamoff,
  .vidioc_enum_input = sentech_vidioc_enum_input,
  .vidioc_g_input = sentech_vidioc_g_input,
  .vidioc_s_input = sentech_vidioc_s_input,
  .vidioc_queryctrl = sentech_vidioc_queryctrl,
  .vidioc_g_ctrl = sentech_vidioc_g_ctrl,
  .vidioc_s_ctrl = sentech_vidioc_s_ctrl,
  .vidioc_querymenu = sentech_vidioc_querymenu, 
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,27)
  .vidioc_log_status = sentech_vidioc_log_status,
#endif  
  .vidioc_cropcap = sentech_vidioc_cropcap,
  .vidioc_g_crop = sentech_vidioc_g_crop,
  .vidioc_s_crop = sentech_vidioc_s_crop,
  .vidioc_g_parm = sentech_vidioc_g_parm,
  .vidioc_s_parm = sentech_vidioc_s_parm,  
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,29)
  .vidioc_enum_framesizes = sentech_vidioc_enum_framesizes,
  .vidioc_enum_frameintervals = sentech_vidioc_enum_frameintervals,
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,18)
  .vidioc_querycap = sentech_vidioc_querycap,
  .vidioc_g_priority = sentech_vidioc_g_priority,
  .vidioc_s_priority = sentech_vidioc_s_priority,
  .vidioc_enum_fmt_cap = sentech_vidioc_enum_fmt_cap,
  .vidioc_g_fmt_cap = sentech_vidioc_g_fmt_cap,
  .vidioc_s_fmt_cap = sentech_vidioc_s_fmt_cap,
  .vidioc_try_fmt_cap = sentech_vidioc_try_fmt_cap,
  .vidioc_reqbufs = sentech_vidioc_reqbufs,
  .vidioc_querybuf = sentech_vidioc_querybuf,
  .vidioc_qbuf = sentech_vidioc_qbuf,
  .vidioc_dqbuf = sentech_vidioc_dqbuf,
#ifdef HAVE_V4L1
#endif
  .vidioc_streamon = sentech_vidioc_streamon,
  .vidioc_streamoff = sentech_vidioc_streamoff,
  .vidioc_enum_input = sentech_vidioc_enum_input,
  .vidioc_g_input = sentech_vidioc_g_input,
  .vidioc_s_input = sentech_vidioc_s_input,
  .vidioc_queryctrl = sentech_vidioc_queryctrl,
  .vidioc_g_ctrl = sentech_vidioc_g_ctrl,
  .vidioc_s_ctrl = sentech_vidioc_s_ctrl,
  .vidioc_querymenu = sentech_vidioc_querymenu, 
  .vidioc_cropcap = sentech_vidioc_cropcap,
  .vidioc_g_crop = sentech_vidioc_g_crop,
  .vidioc_s_crop = sentech_vidioc_s_crop,
  .vidioc_g_parm = sentech_vidioc_g_parm,
  .vidioc_s_parm = sentech_vidioc_s_parm,  
#endif
};
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,29)
static int sentech_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long ul_arg){
#else
static long sentech_ioctl(struct file *file, unsigned int cmd, unsigned long ul_arg){
#endif
  struct sentech_camera* cam_data = get_cam_data(file->private_data);
  void *arg = (void *)ul_arg;
  struct video_device *vfd = video_devdata(file);
  int retval=0;
;
  if(!vfd) {
    printk("sentech: No video device!\n");
    retval = -ENODEV;
    goto done;
  }
  if (! cam_data->connected && 
     (cmd != VIDIOC_STREAMOFF)
     ) {
    if (debug && printk_ratelimit()) printk(KERN_WARNING "sentech ioctl: Disconnecting!\n");
    retval = -ENODEV;
    goto done;
  }
#if  LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18)
  if (debug > 20 ){
  } else vfd->debug = 0;
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,29)
  retval = video_ioctl2(file, cmd, ul_arg);
#else
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18)
  retval = video_ioctl2(inode, file, cmd, ul_arg);
#else
  retval = video_ioctl2(inode, file, cmd, ul_arg, &sentech_ioctl_ops);
#endif 
#endif 
  if(debug>=4)printk("ioctl2 returns %d\n",retval);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,15)
  if(debug &&((!((cmd==VIDIOC_QBUF)||(cmd==VIDIOC_DQBUF)))||(debug>5)))
    v4l_printk_ioctl_arg("sentech(w) ", cmd, arg);
#endif
  if(debug &&((!((cmd==VIDIOC_QBUF)||(cmd==VIDIOC_DQBUF)))||(debug>5)))
    dprintk (3, "ioctl result %d %p\n", retval, arg);
 done:;
  return retval;
 }
static 
struct video_device sentech_template = {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,27)
  .ioctl_ops    = &sentech_ioctl_ops,
  .vfl_type		= VFL_TYPE_GRABBER,
#endif
  .name		= "Sentech_USB_Camera",
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
  .hardware	= 0,
#endif
  .fops           = &sentech_fops,
  .minor		= -1,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18)
  .debug          = 0, 
#endif
  .release	= video_device_release,
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,27)
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18)
  .type2	= VID_TYPE_CAPTURE|VID_TYPE_SUBCAPTURE|VID_TYPE_FRAMERAM|VID_TYPE_OVERLAY,
  .type		= VFL_TYPE_GRABBER,
  .vidioc_querycap = sentech_vidioc_querycap,
  .vidioc_g_priority = sentech_vidioc_g_priority,
  .vidioc_s_priority = sentech_vidioc_s_priority,
  .vidioc_enum_fmt_cap = sentech_vidioc_enum_fmt_cap,
  .vidioc_g_fmt_cap = sentech_vidioc_g_fmt_cap,
  .vidioc_s_fmt_cap = sentech_vidioc_s_fmt_cap,
  .vidioc_try_fmt_cap = sentech_vidioc_try_fmt_cap,
  .vidioc_reqbufs = sentech_vidioc_reqbufs,
  .vidioc_querybuf = sentech_vidioc_querybuf,
  .vidioc_qbuf = sentech_vidioc_qbuf,
  .vidioc_dqbuf = sentech_vidioc_dqbuf,
#ifdef HAVE_V4L1
#endif
  .vidioc_streamon = sentech_vidioc_streamon,
  .vidioc_streamoff = sentech_vidioc_streamoff,
  .vidioc_enum_input = sentech_vidioc_enum_input,
  .vidioc_g_input = sentech_vidioc_g_input,
  .vidioc_s_input = sentech_vidioc_s_input,
  .vidioc_queryctrl = sentech_vidioc_queryctrl,
  .vidioc_g_ctrl = sentech_vidioc_g_ctrl,
  .vidioc_s_ctrl = sentech_vidioc_s_ctrl,
  .vidioc_querymenu = sentech_vidioc_querymenu, 
  .vidioc_log_status = sentech_vidioc_log_status,
  .vidioc_cropcap = sentech_vidioc_cropcap,
  .vidioc_g_crop = sentech_vidioc_g_crop,
  .vidioc_s_crop = sentech_vidioc_s_crop,
  .vidioc_g_parm = sentech_vidioc_g_parm,
  .vidioc_s_parm = sentech_vidioc_s_parm,  
#endif
#endif
};
static 
int sentech_alloc_v4l2(struct sentech_camera* cam_data){;
  cam_data->v4l2dev = video_device_alloc();
  if(! cam_data->v4l2dev){
    dev_err(&cam_data->interface->dev, "SentechUSB: v4l2: Out of Memory\n");
    return -ENOMEM;
  };
  (*cam_data->v4l2dev) = sentech_template;
  if(1 ){
    strlcpy(cam_data->v4l2dev->name, cam_data->cam_name,31);
    cam_data->v4l2dev->name[31] = 0;
  }
  cam_data->v4l2dev->minor = param_minor; 
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,24)
  if (cam_data->monochrome) cam_data->v4l2dev->type2|=VID_TYPE_MONOCHROME;
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,11,0)
  cam_data->v4l2dev->v4l2_dev = &cam_data->v4l2_dev;
  cam_data->v4l2dev->dev_parent = &cam_data->udev->dev;
  dev_set_drvdata(cam_data->dev, cam_data);
#if defined(CONFIG_MEDIA_CONTROLLER)
  cam_data->v4l2_dev.mdev = &cam_data->media_dev;
  cam_data->media_dev.dev = cam_data->dev;
  snprintf(cam_data->media_dev.model, STC_CTRL_NAME_LEN, "%s", cam_data->cam_name);
  {
    int err;
    err = media_device_register(&cam_data->media_dev);
    if (!err)
      dprintk (0, "Fail to register media: %s", cam_data->cam_name);
  }
#endif
#ifdef PARAM_FIRST_NB
  cam_data->v4l2_dev.name = "Sentech_USB";
#else
  v4l2_device_set_name(&cam_data->v4l2_dev, "Sentech_USB", &first_nb);
#endif
#else
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,27)
  cam_data->v4l2dev->parent = &cam_data->udev->dev;
#else
  cam_data->v4l2dev->dev = &cam_data->udev->dev;
#endif
#endif
  video_set_drvdata(cam_data->v4l2dev,cam_data);;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,11,0)
  if(v4l2_device_register(cam_data->dev, &cam_data->v4l2_dev)){
      dev_err(&cam_data->interface->dev, "SentechUSB: v4l2: Registration error\n");
      v4l2_device_disconnect(&cam_data->v4l2_dev);
      v4l2_device_unregister(&cam_data->v4l2_dev);
      video_device_release(cam_data->v4l2dev);
      dev_err(&cam_data->interface->dev, "SentechUSB: v4l2: Registration canceled\n");
      goto error_register_V4L2;
  }
#endif
  if(video_register_device(cam_data->v4l2dev,
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,27)
			   cam_data->v4l2dev->type,
#else
			   cam_data->v4l2dev->vfl_type,
#endif
			   cam_data->v4l2dev->minor))
    {
      dev_err(&cam_data->interface->dev, "SentechUSB: v4l2: Registration error\n");
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,11,0)
      v4l2_device_disconnect(&cam_data->v4l2_dev);
      v4l2_device_unregister(&cam_data->v4l2_dev);
#endif
      video_device_release(cam_data->v4l2dev);
      dev_err(&cam_data->interface->dev, "SentechUSB: v4l2: Registration canceled\n");
      goto error_register_V4L2;
    };
  return 0;
 error_register_V4L2:;
  return -ENODEV;
}
inline void update_size(struct sentech_camera* cam_data){
  if((!cam_data->usb3) && ((__le32_to_cpu(cam_data->payload11.OR_Picture_size_modes_le) & STC_QSXGA) == 0 ) ) {
    cam_data->picture_size=le32_to_cpu(cam_data->payload12.Picture_size_mode_le);
  }
  cam_data->width=le32_to_cpu(cam_data->payload12.Horizontal_size_mode_le);
  cam_data->height=le32_to_cpu(cam_data->payload12.Vertical_size_mode_le);
  cam_data->crt_frame_size=cam_data->width*cam_data->height*
      cam_data->max_Bpp;
  if(cam_data->usb3) return;
    if(dynamic_scale){
      unsigned int i;
      __s32 max_exp, max_exp_us;
      max_exp = get_max_exposure(cam_data);
      max_exp_us = exposure_2_useconds(cam_data, max_exp);
      for (i = 0; i < ARRAY_SIZE(sentech_qctrl); i++) {
	if ((V4L2_CID_EXPOSURE == sentech_qctrl[i].id) ||
	    (V4L2_CID_CONTRAST == sentech_qctrl[i].id)) {
	  cam_data->qctrl[i].maximum = max_exp;
	}
	if (V4L2_CID_STC_EXPOSURE_US == sentech_qctrl[i].id) {
	  cam_data->qctrl[i].maximum = max_exp_us;
	  cam_data->qctrl[i].default_value = cam_data->qctrl[i].maximum;
	}
#ifdef L_BOUNDS_ANY
	sentech_exposure_bounds(cam_data, i, max_exp);
#endif
      }
    }
}
inline int sentech_reset_camera_data(struct sentech_camera* cam_data){
  int res=0, retval =0;
  res=usb_control_msg(cam_data->udev,
		      cam_data->pipe_ctl_snd,
		      0x20,
		      USB_DIR_OUT | USB_TYPE_VENDOR |USB_RECIP_DEVICE, 
		      0,0,
		      0,
		      0x0, CTRL_TIMEOUT);
  if(res!=0x0){
    dev_err(&cam_data->interface->dev, "SentechUSB: result11=%d\n",res);
    retval = res;
  }else{
  }
  return retval;
}
inline int sentech_acquire_camera_data(struct sentech_camera* cam_data){
  int res=0, retval =0;
  res=usb_control_msg(cam_data->udev,
		      cam_data->pipe_ctl_rcv,
		      0x11,
		      USB_DIR_IN | USB_TYPE_VENDOR |USB_RECIP_DEVICE, 
		      0,0,
		      &cam_data->payload11,
		      0x40, CTRL_TIMEOUT);
  if(res!=0x40){
    dev_err(&cam_data->interface->dev, "SentechUSB: result11=%d\n",res);
    retval = res;
  }else{
  }
  return retval;
}
inline int sentech_get_bytes_per_pixel(struct sentech_camera* cam_data){
  int res=0, retval =0;
  res=usb_control_msg(cam_data->udev,
		      cam_data->pipe_ctl_rcv,
		      0x13,
		      USB_DIR_IN | USB_TYPE_VENDOR |USB_RECIP_DEVICE, 
		      0,0,
		      &cam_data->state13,
		      0x40, CTRL_TIMEOUT);
  if(res!=0x06){
    dev_err(&cam_data->interface->dev, "SentechUSB: result15=%d\n",res);
    retval = res;
  }else{
    cam_data->payload13 = cam_data->state13;
  }
  return retval;
}
inline int sentech_get_picture_size(struct sentech_camera* cam_data){
  int res=0, retval = 0;;
  res=usb_control_msg(cam_data->udev,
		  cam_data->pipe_ctl_rcv,
		  0x12,
		  USB_DIR_IN | USB_TYPE_VENDOR |USB_RECIP_DEVICE, 
		  0,0,
		  &cam_data->payload12,
		  0x40, CTRL_TIMEOUT);
  cam_data->state12=cam_data->payload12;
  if(res == 0x40){
    update_size(cam_data);
    if(cam_data->usb3) return retval; 
    if(dynamic_scale){
      unsigned int i;
      __s32 max_exp, max_exp_us;
      max_exp = get_max_exposure(cam_data);
      max_exp_us = exposure_2_useconds(cam_data, max_exp);
      for (i = 0; i < ARRAY_SIZE(sentech_qctrl); i++) {
	if ((V4L2_CID_EXPOSURE == sentech_qctrl[i].id) ||
	    (V4L2_CID_CONTRAST == sentech_qctrl[i].id)) {
	  cam_data->qctrl[i].maximum = max_exp;
	}
	if (V4L2_CID_STC_EXPOSURE_US == sentech_qctrl[i].id) {
	  cam_data->qctrl[i].maximum = max_exp_us;
	  cam_data->qctrl[i].default_value = cam_data->qctrl[i].maximum;
	}
#ifdef L_BOUNDS_ANY
	sentech_exposure_bounds(cam_data, i, max_exp);
#endif
      }
    }else{
    }
  }else retval = res;
  return retval;
}
inline int sentech_reset_picture_size(struct sentech_camera* cam_data){
  int res=0;
  cam_data->payload12.Scan_mode_le = __le16_to_cpu(STC_FULL_SCREEN);
  res=usb_control_msg(cam_data->udev,
		  cam_data->pipe_ctl_snd,
		  0x12,
		  USB_DIR_OUT | USB_TYPE_VENDOR |USB_RECIP_DEVICE, 
		  0,0,
		  &cam_data->payload12,
		  0x40, CTRL_TIMEOUT);  
  return !res || sentech_get_picture_size(cam_data);
}
inline 
int sentech_set_picture_size(struct sentech_camera* cam_data){
  int res=0, retval = 0;
  res=usb_control_msg(cam_data->udev,
		  cam_data->pipe_ctl_snd,
		  0x12,
		  USB_DIR_OUT | USB_TYPE_VENDOR |USB_RECIP_DEVICE, 
		  0,0,
		  &cam_data->payload12,
		  0x40, CTRL_TIMEOUT);
  if(res == 0x40) update_size(cam_data);
  else retval = res;
  return retval;
}
inline 
int sentech_get_shutter_gain(struct sentech_camera* cam_data){
  int res=0;
  res=usb_control_msg(cam_data->udev,
		  cam_data->pipe_ctl_rcv,
		  0x21,
		  USB_DIR_IN | USB_TYPE_VENDOR |USB_RECIP_DEVICE, 
		  0,0,
		  &cam_data->payload21,
		  0x40, CTRL_TIMEOUT) == 0x40;
  cam_data->state21 = cam_data->payload21;
  return !res;
}
int sentech_update_shutter_gain(struct sentech_camera* cam_data){
  int res=0;
  res=usb_control_msg(cam_data->udev,
		  cam_data->pipe_ctl_rcv,
		  0x21,
		  USB_DIR_IN | USB_TYPE_VENDOR |USB_RECIP_DEVICE, 
		  0,0,
		  &cam_data->state21,
		  0x40, CTRL_TIMEOUT) == 0x40;
  return !res;
}
inline 
int sentech_set_shutter_gain(struct sentech_camera* cam_data){
  int res=0;
  res=usb_control_msg(cam_data->udev,
		  cam_data->pipe_ctl_snd,
		  0x21,
		  USB_DIR_OUT | USB_TYPE_VENDOR |USB_RECIP_DEVICE, 
		  0,0,
		  &cam_data->payload21,
		  0x40, CTRL_TIMEOUT) == 0x40;
  return !res;
}
inline int sentech_get_clock(struct sentech_camera* cam_data){
  int res=0;
  res=usb_control_msg(cam_data->udev,
		  cam_data->pipe_ctl_rcv,
		  0x22,
		  USB_DIR_IN | USB_TYPE_VENDOR |USB_RECIP_DEVICE, 
		  0,0,
		  &cam_data->payload22,
		  0x0a, CTRL_TIMEOUT) == 0x0a;
  cam_data->state22 = cam_data->payload22;
  return !res;
}
inline int sentech_reset_clock(struct sentech_camera* cam_data){
  int res=0;
  cam_data->payload22.Clock_mode_le = __le16_to_cpu(0);
  res=usb_control_msg(cam_data->udev,
		  cam_data->pipe_ctl_snd,
		  0x22,
		  USB_DIR_OUT | USB_TYPE_VENDOR |USB_RECIP_DEVICE, 
		  0,0,
		  &cam_data->payload22,
		  0x0a, CTRL_TIMEOUT) == 0x0a;
  return !res || sentech_get_clock(cam_data);
}
inline int 
sentech_get_white_balance(struct sentech_camera* cam_data){
  int res=0;
  res=usb_control_msg(cam_data->udev,
		  cam_data->pipe_ctl_rcv,
		  0x23,
		  USB_DIR_IN | USB_TYPE_VENDOR | USB_RECIP_DEVICE, 
		  0,0,
		  &cam_data->state23,
		  0x40, CTRL_TIMEOUT) == 0x40;
  return !res;
}
inline int 
sentech_set_white_balance(struct sentech_camera* cam_data){
  int res=0;
  res=usb_control_msg(cam_data->udev,
		  cam_data->pipe_ctl_snd,
		  0x23,
		  USB_DIR_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE, 
		  0,0,
		  &cam_data->payload23,
		  0x40, CTRL_TIMEOUT) == 0x40;
  return !res;
}
inline int 
sentech_set_FPGA_SOFT_TRIGGER(struct sentech_camera* cam_data){
  int res = 1;
  __s32 value;;
#define SET_FPGA_STATUS(PORT,FIELD) \
  res&=usb_control_msg(cam_data->udev,  \
		      cam_data->pipe_ctl_snd, \
		      0x80,                   \
		      USB_DIR_OUT | USB_TYPE_VENDOR |USB_RECIP_DEVICE, \
		      STC_FPGA_AUTOREFRESH_WVAL | PORT,0,                           \
		       &cam_data->payload80.FIELD,       \
		      sizeof(cam_data->payload80. FIELD), CTRL_TIMEOUT) == \
    sizeof(cam_data->payload80.FIELD);                                    \
  cam_data->state80.FIELD = cam_data->payload80.FIELD;  
  value = STC_TRIGGER_MODE_TRIGGER;
  cam_data->payload80.Trigger_Mode &= (~STC_FPGA_TRIGGER_MODE_MASK);
  cam_data->payload80.Trigger_Mode |= (value<<STC_FPGA_TRIGGER_MODE_SHIFT);
  value =  STC_TRIGGER_SOURCE_SOFTWARE;
  cam_data->payload80.Trigger_Mode &= (~STC_FPGA_TRIGGER_SRC_MASK);
  cam_data->payload80.Trigger_Mode |= (value<<STC_FPGA_TRIGGER_SRC_SHIFT);
  SET_FPGA_STATUS(0x30, Trigger_Mode);
  return res;
}
inline int 
sentech_get_FPGA(struct sentech_camera* cam_data){
  int res=1;;
#define REFRESH_FPGA_STATUS(PORT,FIELD) \
  res&=usb_control_msg(cam_data->udev,  \
		      cam_data->pipe_ctl_rcv, \
		      0x80,                   \
		      USB_DIR_IN | USB_TYPE_VENDOR |USB_RECIP_DEVICE, \
		      STC_FPGA_READ | PORT,0,                           \
		       &cam_data->state80.FIELD,       \
		      sizeof(cam_data->state80. FIELD), CTRL_TIMEOUT) == \
    sizeof(cam_data->state80.FIELD);                                    \
  cam_data->payload80.FIELD = cam_data->state80.FIELD;  
  REFRESH_FPGA_STATUS(0x00, IO_Directions);
  REFRESH_FPGA_STATUS(0x01, IO_Reversing);
  REFRESH_FPGA_STATUS(0x02, IO_Status);
  REFRESH_FPGA_STATUS(0x04, IO0_change);
  REFRESH_FPGA_STATUS(0x05, IO1_change);
  REFRESH_FPGA_STATUS(0x06, IO2_change);
  REFRESH_FPGA_STATUS(0x07, IO3_change);
  REFRESH_FPGA_STATUS(0x0F, LED);
  REFRESH_FPGA_STATUS(0x14, Digital_gain);
  REFRESH_FPGA_STATUS(0x30, Trigger_Mode);
  REFRESH_FPGA_STATUS(0x31, Trigger_Mask);
  REFRESH_FPGA_STATUS(0x32, EXP_delay_TRG);
  REFRESH_FPGA_STATUS(0x33, EXP_delay_TRG_high);
  REFRESH_FPGA_STATUS(0x34, Strobe_start_delay_TRG);
  REFRESH_FPGA_STATUS(0x35, Strobe_start_delay_TRG_high);
  REFRESH_FPGA_STATUS(0x36, Strobe_end_delay_TRG);
  REFRESH_FPGA_STATUS(0x37, Strobe_end_delay_TRG_high);
  REFRESH_FPGA_STATUS(0x38, TRG_delay_TRG);
  REFRESH_FPGA_STATUS(0x39, TRG_delay_TRG_high);
  REFRESH_FPGA_STATUS(0x3A, TRG_width_TRG);
  REFRESH_FPGA_STATUS(0x3B, TRG_width_TRG_high);
  REFRESH_FPGA_STATUS(0x3C, Readout_delay);
  REFRESH_FPGA_STATUS(0x3F, Image_Flip);
  REFRESH_FPGA_STATUS(0x40, White_Balance_R_gain);
  REFRESH_FPGA_STATUS(0x41, White_Balance_GR_gain);
  REFRESH_FPGA_STATUS(0x42, White_Balance_GB_gain);
  REFRESH_FPGA_STATUS(0x43, White_Balance_B_gain);
  REFRESH_FPGA_STATUS(0x44, White_Balance_mode);
  REFRESH_FPGA_STATUS(0x45, Max_exposure_Low_le);
  REFRESH_FPGA_STATUS(0x46, Max_exposure_Upp);
  REFRESH_FPGA_STATUS(0x47, Min_exposure_Low_le);
  REFRESH_FPGA_STATUS(0x48, Min_exposure_Upp);
  REFRESH_FPGA_STATUS(0x49, AGC_mode);
  REFRESH_FPGA_STATUS(0x4A, ACL_Target_brightness);
  REFRESH_FPGA_STATUS(0x4B, Maximum_gain_AGC);
  REFRESH_FPGA_STATUS(0x4C, Minimum_gain_AGC);
  REFRESH_FPGA_STATUS(0x4D, Gamma);
  REFRESH_FPGA_STATUS(0x4E, Pixel_blemish_correction);
  REFRESH_FPGA_STATUS(0x55, output_frame_rate_decimal);
  REFRESH_FPGA_STATUS(0x56, output_frame_rate);
  REFRESH_FPGA_STATUS(0x60, Altera_version);
  REFRESH_FPGA_STATUS(0x64, Variable_partial_ln_nb);
  REFRESH_FPGA_STATUS(0x65, USB3Vision_mode);
  REFRESH_FPGA_STATUS(0x66, SW_status);
  REFRESH_FPGA_STATUS(0x68, capabilities);
  REFRESH_FPGA_STATUS(0x69, capabilities_3);
  REFRESH_FPGA_STATUS(0x6B, capabilities_4);
  REFRESH_FPGA_STATUS(0x6C, USB3Vision_support);
  return !res;
}
 int set_SentechMode(struct sentech_camera* cam_data, int on){
  int res = 1;
  if(on){
    if((!((__le16_to_cpu(cam_data->state80.USB3Vision_support))&0x00001))
       ||(!((__le16_to_cpu(cam_data->state80.USB3Vision_mode))&0x00001)))
      return res;
    cam_data->payload80.USB3Vision_mode = __cpu_to_le16((__le16_to_cpu(cam_data->state80.USB3Vision_mode)) & ~0x00001);
  }else{
    if((!((__le16_to_cpu(cam_data->state80.USB3Vision_support))&0x00001))
       ||(((__le16_to_cpu(cam_data->state80.USB3Vision_mode))&0x00001)))
      return res;
    cam_data->payload80.USB3Vision_mode = __cpu_to_le16((__le16_to_cpu(cam_data->state80.USB3Vision_mode)) | 0x00001);
  }
  res&=usb_control_msg(cam_data->udev,
		       cam_data->pipe_ctl_snd,
		       0x80,
		       USB_DIR_OUT | USB_TYPE_VENDOR |USB_RECIP_DEVICE, 
		       STC_FPGA_AUTOREFRESH_WVAL|0x65,0,
		       &cam_data->payload80.USB3Vision_mode,
		       sizeof(cam_data->state80.USB3Vision_mode), CTRL_TIMEOUT) ==
    sizeof(cam_data->state80.USB3Vision_mode);
  res&=usb_control_msg(cam_data->udev,
		       cam_data->pipe_ctl_rcv,
		       0x80,
		       USB_DIR_IN | USB_TYPE_VENDOR |USB_RECIP_DEVICE, 
		       STC_FPGA_READ|0x65,0,
		       &cam_data->state80.USB3Vision_mode,
		       sizeof(cam_data->state80.USB3Vision_mode), CTRL_TIMEOUT) ==
    sizeof(cam_data->state80.USB3Vision_mode);
  cam_data->payload80.USB3Vision_mode = cam_data->state80.USB3Vision_mode;
  return res;
}
inline 
int sentech_set_clock(struct sentech_camera* cam_data){
  int res=0;
  res=usb_control_msg(cam_data->udev,
		  cam_data->pipe_ctl_snd,
		  0x22,
		  USB_DIR_OUT | USB_TYPE_VENDOR |USB_RECIP_DEVICE, 
		  0,0,
		  &cam_data->payload22,
		  0x40, CTRL_TIMEOUT) == 0x0a;
  return !res;
}
inline int sentech_detect_compute_frame_size(struct sentech_camera* cam_data){
  sentech_get_picture_size(cam_data);
  return cam_data->crt_frame_size;
}
inline 
int sentech_frame_size(struct sentech_camera* cam_data, int picture_size){
  int result = 0;
  switch(picture_size) {
  case STC_VGA:;
    result = 
      cam_data->crt_frame_size=
      (cam_data->width = 640)*
      (cam_data->height   = 480)*
      cam_data->max_Bpp;
    break;
  case STC_XGA:;
    result=
      cam_data->crt_frame_size=
      (cam_data->width = 0x400)*
      (cam_data->height = 0x300)*
      cam_data->max_Bpp;
    break;
  case STC_SXGA:;
    result =
      cam_data->crt_frame_size=
      (cam_data->width = SXGA_W)*
      (cam_data->height = 1024)*
      cam_data->max_Bpp;
    break;
  case STC_SXGA133:;
    result =
      cam_data->crt_frame_size=
      (cam_data->width = SXGA_133_W)*
      (cam_data->height = SXGA_133_H)*
      cam_data->max_Bpp;
    break;
  case STC_UXGA:;
    result =
      cam_data->crt_frame_size=
      (cam_data->width = 1600)*
      (cam_data->height = 1200)*
      cam_data->max_Bpp;
    break;
  case STC_QXGA:;
    result =
      cam_data->crt_frame_size=
      (cam_data->width = 2048)*
      (cam_data->height = 1536)*
      cam_data->max_Bpp;
    break;
  case STC_QSXGA:;
    result =
      cam_data->crt_frame_size=
      (cam_data->width = 2592)*
      (cam_data->height = 1944)*
      cam_data->max_Bpp;
    break;
  case STC_MBE132U3V_XGA:;
    result =
      cam_data->crt_frame_size=
      (cam_data->width = 1280)*
      (cam_data->height = 1024)*
      cam_data->max_Bpp;
    break;
  case STC_MBCM401U3V_XGA:;
    result =
      cam_data->crt_frame_size=
      (cam_data->width = 2048)*
      (cam_data->height = 2048)*
      cam_data->max_Bpp;
    break;
  case STC_MBCM200U3V_XGA:;
    result =
      cam_data->crt_frame_size=
      (cam_data->width = 2048)*
      (cam_data->height = 1088)*
      cam_data->max_Bpp;
    break;
  default:;
    if (cam_data->crt_frame_size) {
      result = cam_data->crt_frame_size;;
    } else {
      result = sentech_detect_compute_frame_size(cam_data);;
    }
  }
  return result;
}
static struct usb_device_id id_table [] = {
  { USB_DEVICE(SENTECH_VENDOR_ID, STC_C33USB) },
  { USB_DEVICE(SENTECH_VENDOR_ID, STC_C83USB) },
  { USB_DEVICE(SENTECH_VENDOR_ID, STC_B33USB) },
  { USB_DEVICE(SENTECH_VENDOR_ID, STC_B83USB) },
  { USB_DEVICE(SENTECH_VENDOR_ID, STC_C152USB) },
  { USB_DEVICE(SENTECH_VENDOR_ID, STC_C202USB) },
  { USB_DEVICE(SENTECH_VENDOR_ID, STC_B152USB) },
  { USB_DEVICE(SENTECH_VENDOR_ID, STC_B202USB) },
  { USB_DEVICE(SENTECH_VENDOR_ID, STC_C33USB_T) },
  { USB_DEVICE(SENTECH_VENDOR_ID, STC_C83USB_T) },
  { USB_DEVICE(SENTECH_VENDOR_ID, STC_B33USB_T) },
  { USB_DEVICE(SENTECH_VENDOR_ID, STC_B83USB_T) },
  { USB_DEVICE(SENTECH_VENDOR_ID, STC_C152USB_T) },
  { USB_DEVICE(SENTECH_VENDOR_ID, STC_C202USB_T) },
  { USB_DEVICE(SENTECH_VENDOR_ID, STC_B152USB_T) },
  { USB_DEVICE(SENTECH_VENDOR_ID, STC_B202USB_T) },
  { USB_DEVICE(SENTECH_VENDOR_ID, STC_B133USB_T) },
  { USB_DEVICE(SENTECH_VENDOR_ID, STC_C133USB_T) },
  { USB_DEVICE(SENTECH_VENDOR_ID, STC_MB33USB) },
  { USB_DEVICE(SENTECH_VENDOR_ID, STC_MC33USB) },
  { USB_DEVICE(SENTECH_VENDOR_ID, STC_MB83USB) },
  { USB_DEVICE(SENTECH_VENDOR_ID, STC_MC83USB) },
  { USB_DEVICE(SENTECH_VENDOR_ID, STC_MB133USB) },
  { USB_DEVICE(SENTECH_VENDOR_ID, STC_MC133USB) },
  { USB_DEVICE(SENTECH_VENDOR_ID, STC_MB152USB) },
  { USB_DEVICE(SENTECH_VENDOR_ID, STC_MC152USB) },
  { USB_DEVICE(SENTECH_VENDOR_ID, STC_MB202USB) },
  { USB_DEVICE(SENTECH_VENDOR_ID, STC_MC202USB) },
  { USB_DEVICE(SENTECH_VENDOR_ID, STC_MBA5MUSB3) },
  { USB_DEVICE(SENTECH_VENDOR_ID, STC_MCA5MUSB3) },
  { USB_DEVICE(SENTECH_VENDOR_ID, STC_MBE132U3V) },
  { USB_DEVICE(SENTECH_VENDOR_ID, STC_MCE132U3V) },
  { USB_DEVICE(SENTECH_VENDOR_ID, STC_MBCM401U3V) },
  { USB_DEVICE(SENTECH_VENDOR_ID, STC_MCCM401U3V) },
  { USB_DEVICE(SENTECH_VENDOR_ID, STC_MBCM200U3V) },
  { USB_DEVICE(SENTECH_VENDOR_ID, STC_MCCM200U3V) },
  { },
};
MODULE_DEVICE_TABLE (usb, id_table);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,35)
#include "core/hcd.h"
#endif
static
  void setMaxBpp(struct sentech_camera* cam_data) {
  int i;
  cam_data->max_Bpp = 1;
  cam_data->Bpp = 1;
  cam_data->mask_Bpp = STC_ONE_BYTE_PER_PIXEL;
  cam_data->bpp = 8;
  if(cam_data->supports_bpp) {
    cam_data->max_Bpp = getMaxBpp(__le32_to_cpu(cam_data->payload11.OR_Bytes_per_pixel_le));
    cam_data->mask_Bpp = __le32_to_cpu(cam_data->payload11.OR_Bytes_per_pixel_le);
    cam_data->bpp = getMax_bpp(__le32_to_cpu(cam_data->payload11.OR_Bytes_per_pixel_le));;
    for(i = 0; i < ARRAY_SIZE(sentech_qctrl); i++) {
      switch (sentech_qctrl[i].id) {
      case (V4L2_CID_STC_BYTES):
	cam_data->qctrl[i].maximum = cam_data->max_Bpp;
	cam_data->qctl_regs[i] = cam_data->Bpp;
	break;
      }
    }
    sentech_setBytes(cam_data, STC_ONE_BYTE_PER_PIXEL);
  }
 }
static 
void setCrtBpp(struct sentech_camera* cam_data) {
  if(cam_data->supports_bpp) {
    cam_data->Bpp = getMaxBpp(__le32_to_cpu(cam_data->state13.Bytes_per_pixel_le));
  }  
 }
 static int ensure_positive_error(int err){
   return (err<0)?-err:err;
 }
 static void fill_controls_from_device(struct sentech_camera* cam_data) {
   unsigned int i;
   __s32 max_exp = 0, max_exp_us = 0;
   cam_data->h_flip_fpga_can = (cam_data->state80.capabilities_3 & 0x0040)?1:0;
   cam_data->v_flip_fpga_can = (cam_data->state80.capabilities_3 & 0x0080)?1:0;
   cam_data->auto_wb_fpga_can = (cam_data->state80.capabilities_3 & 0x0100)?1:0;
   cam_data->auto_gain_fpga_can = (cam_data->state80.capabilities_3 & 0x0200)?1:0;
   cam_data->auto_shutters_fpga_can = (cam_data->state80.capabilities_3 & 0x0400)?1:0;
   cam_data->pixel_blemish_correction_fpga_can = (cam_data->state80.capabilities_4 & 0x00800)?1:0;
   cam_data->saving_camera_settings_fpga_can = (cam_data->state80.capabilities_4 & 0x00400)?1:0;
   if (cam_data->auto_shutters_fpga_can) {
     cam_data->auto_shutters = (cam_data->state80.AGC_mode & 0x0001)?1:0;
   }
   if (cam_data->auto_gain_fpga_can) {
     cam_data->auto_gain = (cam_data->state80.AGC_mode & 0x0002)?1:0;
   }
   cam_data->color_mode=__le16_to_cpu(cam_data->payload11.Color_layout_le);
   if (dynamic_scale) {
     max_exp = get_max_exposure(cam_data); 
     max_exp_us = exposure_2_useconds(cam_data, max_exp);
   }   
   for (i = 0; i < ARRAY_SIZE(sentech_qctrl); i++) {
     if (V4L2_CID_STC_FRAME_START == sentech_qctrl[i].id) {
       cam_data->qctrl[i].maximum = __le32_to_cpu(cam_data->state12.Vertical_size_mode_le);
     }
     if (V4L2_CID_STC_FRAME_HEIGHT == sentech_qctrl[i].id) {
       cam_data->qctrl[i].maximum = __le32_to_cpu(cam_data->state12.Vertical_size_mode_le);
     }
     if (cam_data->usb3) {
       if (V4L2_CID_AUTOGAIN == sentech_qctrl[i].id) {
	 if (cam_data->auto_gain_fpga_can)
	   cam_data->qctrl[i].default_value = cam_data->auto_gain;
       }
       if (V4L2_CID_GAIN == sentech_qctrl[i].id) {
	 cam_data->qctrl[i].maximum = __le16_to_cpu(cam_data->state12.maximum_gain_le);
	 cam_data->qctrl[i].default_value = __le16_to_cpu(cam_data->state12.initial_gain_le);
       }
       if (V4L2_CID_STC_AUTOSHUTTERS == sentech_qctrl[i].id) {
	 if (cam_data->auto_shutters_fpga_can) {
	   cam_data->qctrl[i].maximum = 1;
	   cam_data->qctrl[i].default_value = 1;
	 } else {
	   cam_data->qctrl[i].maximum = 0;
	   cam_data->qctrl[i].flags = V4L2_CTRL_FLAG_DISABLED;
	   cam_data->qctrl[i].default_value = 0;
	 }
       }
#ifdef  SENTECH_AUTO_AWB_CIDS
       if (V4L2_CID_STC_AUTOGAIN == sentech_qctrl[i].id) {
	 if (cam_data->auto_gain_fpga_can)
	   cam_data->qctrl[i].default_value = 1;
       }
#endif
     } else {
       if (V4L2_CID_STC_AUTOSHUTTERS == sentech_qctrl[i].id) {
	 if (0&&cam_data->auto_shutters_fpga_can){
	   cam_data->qctrl[i].maximum = 1;
	   cam_data->qctrl[i].default_value = 1;
	 } else {
	   cam_data->qctrl[i].maximum = 0;
	   cam_data->qctrl[i].flags = V4L2_CTRL_FLAG_DISABLED;
	   cam_data->qctrl[i].default_value = 0;
	 }
       }
     }
     if (1 || ! cam_data->usb3 || ! cam_data->auto_gain_fpga_can) {
       if (V4L2_CID_STC_GAIN_MIN == sentech_qctrl[i].id) {
	 if (! cam_data->usb3 || ! cam_data->auto_gain_fpga_can)
	   cam_data->qctrl[i].maximum = __le16_to_cpu(cam_data->state12.initial_gain_le);
	 else cam_data->qctrl[i].maximum =  __le16_to_cpu(cam_data->state12.maximum_gain_le);
       }
       if (V4L2_CID_STC_GAIN_MAX == sentech_qctrl[i].id) {
	 if (! cam_data->usb3 || ! cam_data->auto_gain_fpga_can)
	   cam_data->qctrl[i].minimum = __le16_to_cpu(cam_data->state12.initial_gain_le);
	 else  cam_data->qctrl[i].minimum = 0;
	 cam_data->qctrl[i].maximum = __le16_to_cpu(cam_data->state12.maximum_gain_le);
	 cam_data->qctrl[i].default_value = __le16_to_cpu(cam_data->state12.maximum_gain_le);
       }
     }
     if (V4L2_CID_GAIN == sentech_qctrl[i].id) {
       cam_data->qctrl[i].maximum = __le16_to_cpu(cam_data->state12.maximum_gain_le);
       cam_data->qctrl[i].default_value = __le16_to_cpu(cam_data->state12.initial_gain_le);
     }
     if (dynamic_scale) {
       if ((V4L2_CID_EXPOSURE == sentech_qctrl[i].id) ||
	   ( V4L2_CID_CONTRAST == sentech_qctrl[i].id)) {
	 cam_data->qctrl[i].maximum = max_exp;
       }
       if (V4L2_CID_STC_EXPOSURE_US == sentech_qctrl[i].id) {
	 cam_data->qctrl[i].maximum = max_exp_us;
	 cam_data->qctrl[i].default_value = cam_data->qctrl[i].maximum;
       }
#ifdef L_BOUNDS_ANY
       sentech_exposure_bounds(cam_data, i, max_exp); 
#endif
     }
    switch (sentech_qctrl[i].id) {
    case (V4L2_CID_STC_TRIGGER_MODE):
      cam_data->qctl_regs[i] = 
      (cam_data->state80.Trigger_Mode & STC_FPGA_TRIGGER_MODE_MASK)>>
      STC_FPGA_TRIGGER_MODE_SHIFT;
      break;
    case (V4L2_CID_STC_TRIGGER_SOURCE):
      cam_data->qctl_regs[i] = 
      (cam_data->state80.Trigger_Mode & STC_FPGA_TRIGGER_SRC_MASK)>>
      STC_FPGA_TRIGGER_SRC_SHIFT;
      break;
    case (V4L2_CID_STC_FRAME_START):
      cam_data->qctl_regs[i] = 0;
      break;
    case (V4L2_CID_STC_FRAME_HEIGHT):
      cam_data->qctl_regs[i] = 0;
      break;
#ifdef  SENTECH_AUTO_AWB_CIDS
    case (V4L2_CID_STC_AUTOGAIN):
#endif
    case (V4L2_CID_AUTOGAIN):
      if(cam_data->auto_gain_fpga_can) cam_data->qctl_regs[i] = cam_data->auto_gain;
      break;
    case (V4L2_CID_GAIN):
      if(cam_data->auto_gain_fpga_can) cam_data->qctl_regs[i] = __le16_to_cpu(cam_data->state21.Gain_le);
      break;
    case (V4L2_CID_STC_AUTOSHUTTERS):
      if(cam_data->auto_shutters_fpga_can) cam_data->qctl_regs[i] = cam_data->auto_shutters;
      break;
    case (V4L2_CID_STC_EXPOSURE_US):
      if(cam_data->auto_shutters_fpga_can){
	cam_data->qctl_regs[i] = get_crt_exposure(cam_data);
	cam_data->qctrl[i].maximum =
	  exposure_2_useconds(cam_data,
			      (USB3_EXP_UP_MASK_4BITS << USB3_EXP_UP_SHIFT) +
			      (USB3_EXP_LOW_MASK_16BITS << USB3_EXP_LOW_SHIFT));
      }
      break;
    case (V4L2_CID_EXPOSURE):
    case (V4L2_CID_CONTRAST):
      if(cam_data->auto_shutters_fpga_can){
	cam_data->qctl_regs[i] = get_crt_exposure(cam_data);
	cam_data->qctrl[i].maximum =
	  (USB3_EXP_UP_MASK_4BITS << USB3_EXP_UP_SHIFT) +
	  (USB3_EXP_LOW_MASK_16BITS << USB3_EXP_LOW_SHIFT);
      }
      break;
    case (V4L2_CID_STC_EXPOSURE_MIN):
      if(cam_data->auto_shutters_fpga_can){
	cam_data->qctrl[i].maximum = 
	  exposure_2_useconds(cam_data,
			      (USB3_EXP_UP_MASK_4BITS << USB3_EXP_UP_SHIFT) +
			      (USB3_EXP_LOW_MASK_16BITS << USB3_EXP_LOW_SHIFT));
	cam_data->qctrl[i].step = 1 << USB3_EXP_LOW_SHIFT;
 	cam_data->qctrl[i].default_value = USB3_MIN_EXP_DEFAULT << USB3_EXP_LOW_SHIFT;
      }
      break;
    case (V4L2_CID_STC_EXPOSURE_MAX):
      if(cam_data->auto_shutters_fpga_can){
	cam_data->qctrl[i].maximum = 
	  exposure_2_useconds(cam_data,
			      (USB3_EXP_UP_MASK_4BITS << USB3_EXP_UP_SHIFT) +
			      (USB3_EXP_LOW_MASK_16BITS << USB3_EXP_LOW_SHIFT));
 	cam_data->qctrl[i].step = 1 << USB3_EXP_LOW_SHIFT;
 	cam_data->qctrl[i].default_value = USB3_MAX_EXP_DEFAULT << USB3_EXP_LOW_SHIFT;
     }
      break;
    case (V4L2_CID_STC_HORIZONTAL_SIZE_CAPTURE):
      if(cam_data->usb3 && sentech_support_AOI(cam_data)){
	cam_data->qctl_regs[i] = cam_data->scan_width;
	cam_data->qctrl[i].maximum = __le32_to_cpu(cam_data->payload11.Maximum_picture_width_le);
	cam_data->qctrl[i].flags = 0;
      }
      break;
    case (V4L2_CID_STC_VERTICAL_SIZE_CAPTURE):
      if(cam_data->usb3 && sentech_support_AOI(cam_data)){
	cam_data->qctl_regs[i] = cam_data->scan_height;
	cam_data->qctrl[i].maximum = __le32_to_cpu(cam_data->payload11.Maximum_picture_height_le);
	cam_data->qctrl[i].flags = 0;
      }
      break;
    case (V4L2_CID_STC_HORIZONTAL_OFFSET_CAPTURE):
      if(cam_data->usb3 && sentech_support_AOI(cam_data)){
	cam_data->qctl_regs[i] = cam_data->scan_horizontal_offset;
	cam_data->qctrl[i].maximum = __le32_to_cpu(cam_data->payload11.Maximum_picture_width_le);
	cam_data->qctrl[i].flags = 0;
      }
      break;
    case (V4L2_CID_STC_VERTICAL_OFFSET_CAPTURE):
      if(cam_data->usb3 && sentech_support_AOI(cam_data)){
	cam_data->qctl_regs[i] = cam_data->scan_vertical_offset;
	cam_data->qctrl[i].maximum = __le32_to_cpu(cam_data->payload11.Maximum_picture_height_le);
	cam_data->qctrl[i].flags = 0;
      }
      break;
    default:
      cam_data->qctl_regs[i] = cam_data->qctrl[i].default_value;
    }
    switch (sentech_qctrl[i].id) {
    case (V4L2_CID_STC_RED): 
    case V4L2_CID_RED_BALANCE:
      {cam_data->qctl_regs[i] = cam_data->red; break;}
    case (V4L2_CID_STC_GREEN): 
      {cam_data->qctl_regs[i] = cam_data->green; break;}
    case (V4L2_CID_STC_BLUE): 
    case V4L2_CID_BLUE_BALANCE:
      {cam_data->qctl_regs[i] = cam_data->blue; break;}
    }
   }
 }
 static int read_camera_settings(struct sentech_camera* cam_data) {
   struct usb_interface * interface = cam_data->interface;;
   if (sentech_acquire_camera_data(cam_data)) {
     dev_info(&interface->dev, "Sentech USB Camera: error crtl 11\n");
     return ENODEV;
   }
;
   if(sentech_get_picture_size(cam_data))
     dev_info(&interface->dev, "Sentech USB Camera: error crtl 12\n");
   sentech_reset_picture_size(cam_data);;
   if (sentech_get_bytes_per_pixel(cam_data))
     dev_info (&interface->dev, "Sentech USB Camera: error crtl 13\n");;
   if (sentech_get_shutter_gain(cam_data))
     dev_info(&interface->dev, "Sentech USB Camera: error crtl 21\n");;
   if (sentech_get_clock(cam_data))
     dev_info(&interface->dev, "Sentech USB Camera: error crtl 22\n");
   sentech_reset_clock(cam_data);; 
   if(sentech_get_white_balance(cam_data)) {
     dev_info(&interface->dev, "Sentech USB Camera: error crtl 23\n");
     return ENODEV;
   }; 
   if (sentech_get_FPGA(cam_data))
     dev_info(&interface->dev, "Sentech USB Camera: simple camera model\n");; 
;
   fill_controls_from_device(cam_data);;
   return 0;
 }
static 
  int SentechUSB_probe(struct usb_interface *interface, const struct usb_device_id * UNUSED(id)) {
  unsigned int i;
  int retval=0;
  int _60fps=1,max_fps=600;
  int  picture_size=0, monochrome=0;
  struct usb_device *udev;
  struct sentech_camera *cam_data = NULL;
  char* cam_name="Unknown";
  int minor = interface->minor;
  u16 model;
  __u8 iSerial;
  int old_style = 0;
  if(debug!=0)printk("sentech: In %s\n",__FUNCTION__);
  udev = interface_to_usbdev(interface);
  iSerial = udev->descriptor.iSerialNumber;
  if (iSerial == 0);
  else {;
  }
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,7)
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,0,0)
#endif
    switch (udev->speed) {
    case USB_SPEED_UNKNOWN:;
      break;
    case USB_SPEED_LOW:
    case USB_SPEED_FULL:;
      break;
    case USB_SPEED_HIGH:;
      break;
    case USB_SPEED_WIRELESS:;
      break;
    case USB_SPEED_SUPER:;
      break;
    default:;
    }
  model = __le16_to_cpu(udev->descriptor.idProduct);
  cam_data = kmalloc(sizeof(struct sentech_camera), GFP_KERNEL);
  if (cam_data == NULL) {
    dev_err(&interface->dev, "SentechUSB: Out of memory\n");
    retval = ENOMEM;
    goto probe_error_alloc1;
  };
  memset (cam_data, 0x00, sizeof (*cam_data));
  cam_data->has_data = 1;
  INIT_LIST_HEAD(&cam_data->ctrl_queue);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,9)
  kref_init(&cam_data->kref);
#else
  kref_init(&cam_data->kref, cam_data_delete);
#endif
  init_MUTEX(&cam_data->lock);
  sema_init(&cam_data->format_lock, 1);
  cam_data->model = model;
  cam_data->udev = usb_get_dev(udev);
  cam_data->bus = cam_data->udev->bus;
  cam_data->interface = interface;
  cam_data->dev = &interface->dev;
  cam_data->pipe_image   = usb_rcvbulkpipe(udev,IMAGE_BULK_ENDPOINT);
  cam_data->pipe_ctl_rcv = usb_rcvctrlpipe(udev,IMAGE_CTL_ENDPOINT);
  cam_data->pipe_ctl_snd = usb_sndctrlpipe(udev,IMAGE_CTL_ENDPOINT);
  cam_data->output.top = cam_data->output.left = 1;
  switch(model) {
  case STC_MBCM401U3V:
  case STC_MCCM401U3V:
  case STC_MBCM200U3V:
  case STC_MCCM200U3V:
    cam_data->u3v = 1;
  case STC_MBA5MUSB3:
  case STC_MCA5MUSB3:
  case STC_MBE132U3V:
  case STC_MCE132U3V:
    cam_data->usb3 = 1; 
  case STC_MB33USB:
  case STC_MC33USB:
  case STC_MB83USB:
  case STC_MC83USB:
  case STC_MB133USB:
  case STC_MC133USB:
  case STC_MB152USB:
  case STC_MC152USB:
  case STC_MB202USB:
  case STC_MC202USB:
    if (param_supports_bpp)
      cam_data->supports_bpp = 1;
  case STC_C33USB:
  case STC_C83USB:
  case STC_C33USB_T:
  case STC_C83USB_T:
  case STC_B33USB:
  case STC_B83USB:
  case STC_B33USB_T:
  case STC_B83USB_T:
    old_style = 1;
  }
  memcpy (cam_data->qctrl, sentech_qctrl, sizeof(sentech_qctrl));;
;
  if (! old_style) {
    if (sentech_reset_camera_data(cam_data)){
      dev_info(&interface->dev, "Sentech USB Camera: need manual reset\n");
    }
  }
  if (sentech_acquire_camera_data(cam_data)) {
    dev_info(&interface->dev, "Sentech USB Camera: error crtl 11\n");
    return ENODEV;
  }
;
  setMaxBpp(cam_data);
  if(sentech_get_picture_size(cam_data))
    dev_info(&interface->dev, "Sentech USB Camera: error crtl 12\n");
    sentech_reset_picture_size(cam_data);
;
  if (sentech_get_bytes_per_pixel(cam_data))
    dev_info (&interface->dev, "Sentech USB Camera: error crtl 13\n");;
  setCrtBpp(cam_data);
  if (sentech_get_shutter_gain(cam_data))
    dev_info(&interface->dev, "Sentech USB Camera: error crtl 21\n");
;
  if (sentech_get_clock(cam_data))
    dev_info(&interface->dev, "Sentech USB Camera: error crtl 22\n");
    sentech_reset_clock(cam_data);
; 
  if(sentech_get_white_balance(cam_data)) {
    dev_info(&interface->dev, "Sentech USB Camera: error crtl 23\n");
    return ENODEV;
  }
; 
    if (sentech_get_FPGA(cam_data))
      dev_info(&interface->dev, "Sentech USB Camera: simple camera model\n");
; 
  if ( 
      param_init_softtrigger)
    sentech_set_FPGA_SOFT_TRIGGER(cam_data);
  if((((__le16_to_cpu(cam_data->state80.USB3Vision_support))&0x00001))) {
    if(interface->altsetting!=NULL){
      if(interface->altsetting->extralen == 0){
	printk("sentech: secondary interface detection in %s\n",__FUNCTION__);
	return ENODEV;
      }else{
      }
    }
  }
  if(param_detect_multiple_prrobe_u3v) {
    if((((__le16_to_cpu(cam_data->state80.USB3Vision_support))&0x00001))
       &&(!((__le16_to_cpu(cam_data->state80.USB3Vision_mode))&0x00001)))
      return ENODEV;  
    set_SentechMode(cam_data, 1);
  }
  rwlock_init(&cam_data->fpga2_lock);
  if (cam_data->u3v) {
    int sync = 0;;
    sentech_fpga2_read_offset_address(cam_data, GFP_KERNEL, CATEGORY_OFFSET_ANALOG_CTL, &cam_data->fpga2_offset_analog_control, sync, sizeof(cam_data->fpga2_offset_analog_control));
    sentech_fpga2_read_offset_address(cam_data, GFP_KERNEL, CATEGORY_OFFSET_VENDOR_UNIQ_CTRL, &cam_data->fpga2_offset_vendor_unique_control, sync, sizeof(cam_data->fpga2_offset_vendor_unique_control));
    sentech_fpga2_read_offset_address(cam_data, GFP_KERNEL, CATEGORY_OFFSET_USB_ORIG_SAVE, &cam_data->fpga2_offset_USB_camera_original_function__save, sync, sizeof(cam_data->fpga2_offset_USB_camera_original_function__save));
    sentech_fpga2_read_offset_address(cam_data, GFP_KERNEL, CATEGORY_OFFSET_CAMERA_STATUS, &cam_data->fpga2_offset_USB_camera_status, sync, sizeof(cam_data->fpga2_offset_USB_camera_status));
    sentech_fpga2_read_offset_address(cam_data, GFP_KERNEL, CATEGORY_OFFSET_CAMERA_FEATURE, &cam_data->fpga2_offset_USB_camera_feature, sync, sizeof(cam_data->fpga2_offset_USB_camera_feature));
    sentech_fpga2_read_offset_address(cam_data, GFP_KERNEL, CATEGORY_OFFSET_HDR_CONTROL, &cam_data->fpga2_offset_HDR_control, sync, sizeof(cam_data->fpga2_offset_HDR_control));
  }
  cam_data->h_flip_fpga_can = (cam_data->state80.capabilities_3 & 0x0040)?1:0;
  cam_data->v_flip_fpga_can = (cam_data->state80.capabilities_3 & 0x0080)?1:0;
  cam_data->auto_wb_fpga_can = (cam_data->state80.capabilities_3 & 0x0100)?1:0;
  cam_data->auto_gain_fpga_can = (cam_data->state80.capabilities_3 & 0x0200)?1:0;
  cam_data->auto_shutters_fpga_can = (cam_data->state80.capabilities_3 & 0x0400)?1:0;
  cam_data->pixel_blemish_correction_fpga_can = (cam_data->state80.capabilities_4 & 0x00800)?1:0;
  cam_data->saving_camera_settings_fpga_can = (cam_data->state80.capabilities_4 & 0x00400)?1:0;
  if (cam_data->auto_shutters_fpga_can) {
    cam_data->auto_shutters = (cam_data->state80.AGC_mode & 0x0001)?1:0;
  }
  if (cam_data->auto_gain_fpga_can) {
    cam_data->auto_gain = (cam_data->state80.AGC_mode & 0x0002)?1:0;
  }
  cam_data->color_mode=__le16_to_cpu(cam_data->payload11.Color_layout_le);
  if(__le32_to_cpu(cam_data->payload11.OR_Picture_size_modes_le)&STC_QSXGA)
    picture_size = STC_QSXGA;
  else
    if(__le32_to_cpu(cam_data->payload11.OR_Picture_size_modes_le)&STC_UXGA)
      picture_size = STC_UXGA;
    else
      if(__le32_to_cpu(cam_data->payload11.OR_Picture_size_modes_le)&STC_SXGA)
	picture_size = STC_SXGA;
      else
	if(__le32_to_cpu(cam_data->payload11.OR_Picture_size_modes_le)&STC_SXGA133)
	  picture_size = STC_SXGA133;
	else
	  if(__le32_to_cpu(cam_data->payload11.OR_Picture_size_modes_le)&STC_XGA)
	    picture_size = STC_XGA;
	  else
	    if(__le32_to_cpu(cam_data->payload11.OR_Picture_size_modes_le)&STC_VGA)
	      picture_size = STC_VGA;
  switch(model) {
  case STC_C33USB:
    cam_name="Sentech  C33 USB Camera 100:000"; monochrome = 0; picture_size=STC_VGA; max_fps=600; break;
  case STC_C83USB:
    cam_name="Sentech  C83 USB Camera 100:000"; monochrome = 0; picture_size=STC_XGA; _60fps=0; max_fps=300; break;
  case STC_B33USB:
    cam_name="Sentech  B33 USB Camera 100:000"; monochrome = 1; picture_size=STC_VGA; max_fps=600; break;
  case STC_B83USB:
    cam_name="Sentech  B83 USB Camera 1000:000"; monochrome = 1; picture_size=STC_XGA; _60fps=0; max_fps=300; break;
  case STC_C33USB_T:
    cam_name="Sentech TC33 USB Camera 100:000"; monochrome = 0; picture_size=STC_VGA; max_fps=600; break;
  case STC_MB33USB:
    cam_name="Sentech MB33 USB Camera 100:000"; monochrome = 1; picture_size=STC_VGA; max_fps=600; break;
  case STC_MC33USB:
    cam_name="Sentech MC33 USB Camera 100:000"; monochrome = 0; picture_size=STC_VGA; max_fps=600; break;
  case STC_MB83USB:
    cam_name="Sentech MB83 USB Camera 100:000"; monochrome = 1; picture_size=STC_XGA; max_fps=300; _60fps=0; break;
  case STC_MC83USB:
    cam_name="Sentech MC83 USB Camera 100:000"; monochrome = 0; picture_size=STC_XGA; max_fps=300; _60fps=0;
    break;
  case STC_MB133USB:
    cam_name="Sentech MB133 USB Camera 100:000"; monochrome = 1; picture_size=STC_SXGA133; max_fps=220; _60fps=0; break;
  case STC_MC133USB:
    cam_name="Sentech MC133 USB Camera 100:000"; monochrome = 0; picture_size=STC_SXGA133; max_fps=220; _60fps=0; break;
  case STC_MB152USB:
    cam_name="Sentech MB152 USB Camera 100:000"; monochrome = 1; picture_size=STC_SXGA; max_fps=190; _60fps=0; break;
  case STC_MC152USB:
    cam_name="Sentech MC152 USB Camera 100:000"; monochrome = 0; picture_size=STC_SXGA; max_fps=190; _60fps=0; break;
  case STC_MB202USB:
    cam_name="Sentech MB202 USB Camera 100:000"; monochrome = 1; picture_size=STC_UXGA; max_fps=75; _60fps=0; break;
  case STC_MC202USB:
    cam_name="Sentech MC202 USB Camera 100:000"; monochrome = 0; picture_size=STC_UXGA; max_fps=75; _60fps=0; break;
  case STC_C83USB_T:
    cam_name="Sentech TC83 USB Camera 100:000"; monochrome = 0; picture_size=STC_XGA; _60fps=0; max_fps=300; break;
  case STC_C152USB_T:
    cam_name="Sentech TC152USB Camera 100:000"; monochrome = 0; picture_size=STC_SXGA; _60fps=0; max_fps=190; break;
  case STC_C202USB_T:
    cam_name="Sentech TC202USB Camera 100:000"; monochrome = 0; picture_size=STC_UXGA; _60fps=0; max_fps=75; break;
  case STC_C133USB_T:
    cam_name="Sentech TC133USB Camera 100:000"; monochrome = 0; picture_size=STC_SXGA133; _60fps=0; max_fps=220; break;
  case STC_B33USB_T:
    cam_name="Sentech TB33 USB Camera 100:000"; monochrome = 1; picture_size=STC_VGA; max_fps=600; break;
  case STC_B83USB_T:
    cam_name="Sentech TB83 USB Camera 100:000"; monochrome = 1; picture_size=STC_XGA; _60fps=0; max_fps=300; break;
  case STC_B152USB_T:
    cam_name="Sentech TB152USB Camera 100:000"; monochrome = 1; picture_size=STC_SXGA; _60fps=0; max_fps=190; break;
  case STC_B202USB_T:
    cam_name="Sentech TB202USB Camera 100:000"; monochrome = 1; picture_size=STC_UXGA; _60fps=0; max_fps=75; break;
  case STC_B133USB_T:
    cam_name="Sentech TB133USB Camera 100:000"; monochrome = 1; picture_size=STC_SXGA133; _60fps=0; max_fps=220; break;
  case STC_MCA5MUSB3:
    cam_name="Sentech MCA5MUSB3 Camera 100:000"; monochrome = 0; picture_size=STC_QSXGA; _60fps=0; max_fps=220; break;
  case STC_MBA5MUSB3:
    cam_name="Sentech MBA5MUSB3 Camera 100:000"; monochrome = 1; picture_size=STC_QSXGA; _60fps=0; max_fps=220; break;
  case STC_MCE132U3V:
    cam_name="Sentech MCE132U3V Camera 100:000"; monochrome = 0; picture_size=STC_MBE132U3V_XGA; _60fps=0; max_fps=220; break;
  case STC_MBE132U3V:
    cam_name="Sentech MBE132U3V Camera 100:000"; monochrome = 1; picture_size=STC_MBE132U3V_XGA; _60fps=0; max_fps=220; break;
  case STC_MCCM401U3V:
    cam_name="Sentech MCCM401U3V Camera 100:000"; monochrome = 0; picture_size=STC_MBCM401U3V_XGA; _60fps=0; max_fps=220; break;
  case STC_MBCM401U3V:
    cam_name="Sentech MBCM401U3V Camera 100:000"; monochrome = 1; picture_size=STC_MBCM401U3V_XGA; _60fps=0; max_fps=86; break;
  case STC_MCCM200U3V:
    cam_name="Sentech MCCM200U3V Camera 100:000"; monochrome = 0; picture_size=STC_MBCM200U3V_XGA; _60fps=0; max_fps=150; break;
  case STC_MBCM200U3V:
    cam_name="Sentech MBCM200U3V Camera 100:000"; monochrome = 1; picture_size=STC_MBCM200U3V_XGA; _60fps=0; max_fps=150; break;
  default:
    cam_name="Sentech Generic  Camera 100:000";
    monochrome = (cam_data->color_mode==STC_BW);
    _60fps=0; max_fps=300; 
    if(0 && ! picture_size) { 
      dev_err(&interface->dev, "SentechUSB: Device %x not supported\n", model);
      retval = ENODEV;
      goto probe_error_identification;
    }
  }
;
  if ((!cam_data->usb3)&&(cam_data->picture_size != picture_size)) {
    printk(KERN_ALERT "Camera not fully supported (tell msilaghi@fit.edu): Image_size %xh believed %xh\n", cam_data->picture_size, picture_size);
  }
  if ((__le32_to_cpu(cam_data->payload11.OR_Picture_size_modes_le) != picture_size)
      && ((__le32_to_cpu(cam_data->payload11.OR_Picture_size_modes_le) & STC_QSXGA) == 0 )) {
    printk(KERN_ALERT "Camera not yet supported (tell msilaghi@fit.edu): Image_size %xh believed %xh\n", __le32_to_cpu(cam_data->payload11.OR_Picture_size_modes_le), picture_size);
    if (__le32_to_cpu(cam_data->payload11.OR_Picture_size_modes_le) > 0)
      picture_size = __le32_to_cpu(cam_data->payload11.OR_Picture_size_modes_le);
  }
  strncpy(cam_data->cam_name, cam_name, CAM_NAME_LENGTH);
#ifdef PARAM_FIRST_NB
  cam_data->cam_number = first_nb;
  first_nb++;
  if(first_nb>999) first_nb=0;
#else
  cam_data->cam_number = atomic_inc_return(&first_nb) - 1;
  if(cam_data->cam_number>999) cam_data->cam_number=999;
#endif
  snprintf(cam_data->cam_name+24, 8, "%3d:%3d", minor, cam_data->cam_number);
  dev_info(&interface->dev, "Detected: Sentech Camera \"%s\" with minor=%d\n", cam_data->cam_name, minor);
  list_add_tail(&cam_data->sentech_devlist, &sentech_camlist);
  mb();
  for(;cam_data->model_idx < ARRAY_SIZE(sentech_fps);
      cam_data->model_idx++) {
    if(model==sentech_fps[cam_data->model_idx].model)
      break;
  }
  if(cam_data->model_idx >= ARRAY_SIZE(sentech_fps)) {;
    if(monochrome) cam_data->model_idx = ARRAY_SIZE(sentech_fps)-2; 
    else cam_data->model_idx = ARRAY_SIZE(sentech_fps)-1; 
  }
  cam_data->monochrome = monochrome;
  cam_data->try_fragments = param_try_fragments;
  cam_data->connected = 1;
  cam_data->enable_native = DEFAULT_NATIVE_MODES;
  cam_data->exclusive = 0;
#ifdef L_BOUNDS
  cam_data->exposure_max = 1;
  cam_data->exposure_min = 0;
  cam_data->gain_max = 255;
  cam_data->gain_min = 0;
#endif
  atomic_set(&cam_data->running_buffers, 0);
  cam_data->crt_frame_size = sentech_frame_size(cam_data,picture_size);
  cam_data->max_frame_size = cam_data->crt_frame_size;
  if(param_max_frame_size)
    cam_data->max_frame_size = param_max_frame_size;
  cam_data->crop.width = cam_data->max_width = cam_data->width;
  cam_data->crop.height = cam_data->max_height  = cam_data->height;;
  if (cam_data->usb3) {
    cam_data->scan_width =
      __le32_to_cpu(cam_data->state12.Horizontal_size_mode_le) & ~3;
    cam_data->scan_horizontal_offset =
      __le32_to_cpu(cam_data->state12.Horizontal_offset_le) & ~1;;
    cam_data->scan_height =
      __le32_to_cpu(cam_data->state12.Vertical_size_mode_le) & ~3;
    cam_data->scan_vertical_offset =
      __le32_to_cpu(cam_data->state12.Vertical_offset_le) & ~1;;
    cam_data->scanimage_size =  cam_data->Bpp*cam_data->scan_height*cam_data->scan_width;;
    cam_data->scan_size =  cam_data->scanimage_size+sizeof(struct Sentech_Header);;
    cam_data->defrect.width = cam_data->max_width;
    cam_data->defrect.height = cam_data->max_height;
    cam_data->bounds.width = cam_data->scan_width;
    cam_data->bounds.height = cam_data->scan_height;
    cam_data->output.width = cam_data->width;
    cam_data->output.height = cam_data->height;;;;
  } else {
    cam_data->scan_width = cam_data->max_width;
    if (__le32_to_cpu(cam_data->payload12.Horizontal_size_mode_le)
	!= cam_data->width) {
      printk(KERN_ALERT "Camera not yet supported (tell msilaghi@fit.edu): Image_width %x believed %x\n", __le32_to_cpu(cam_data->payload12.Horizontal_size_mode_le), cam_data->width);
      cam_data->width = __le32_to_cpu(cam_data->payload12.Horizontal_size_mode_le);
    }
    if (__le32_to_cpu(cam_data->payload12.Vertical_size_mode_le)
	!= cam_data->height) {
      printk(KERN_ALERT "Camera not yet supported (tell msilaghi@fit.edu): Image_height %x believed %x\n", __le32_to_cpu(cam_data->payload12.Vertical_size_mode_le), cam_data->height);
      cam_data->height = __le32_to_cpu(cam_data->payload12.Vertical_size_mode_le);
    }
    {
      int supposed_size=cam_data->width*cam_data->height*cam_data->max_Bpp;
      if(cam_data->crt_frame_size != supposed_size){
	printk(KERN_ALERT "Camera not yet supported (tell msilaghi@fit.edu): Image_bytes %x believed %x\n", supposed_size, cam_data->crt_frame_size);
	cam_data->crt_frame_size = supposed_size;
      }
    }
    {
      switch(__le16_to_cpu(cam_data->state12.Scan_mode_le)){
      case STC_FULL_SCREEN:cam_data->scan_height = cam_data->max_height;
	break;
      case STC_HALF_SCREEN:cam_data->scan_height = 
	  sentech_fps[cam_data->model_idx].fsize[1].height;;
	break;
      case STC_QUART_SCREEN:cam_data->scan_height = 
	  sentech_fps[cam_data->model_idx].fsize[2].height;
	break;
      default: 
	cam_data->scan_height =  __le32_to_cpu(cam_data->state12.Vertical_size_mode_le);
      }
    };
    cam_data->scanimage_size =  cam_data->Bpp*cam_data->scan_height*cam_data->scan_width;;
    cam_data->scan_size =  cam_data->scanimage_size+sizeof(struct Sentech_Header);;
    cam_data->defrect.width = cam_data->max_width;
    cam_data->defrect.height = cam_data->max_height;
    cam_data->bounds.width = cam_data->scan_width;
    cam_data->bounds.height = cam_data->scan_height;
    cam_data->output.width = cam_data->width;
    cam_data->output.height = cam_data->height;;;;
  }
#ifdef SENTECH_AWB
  if (!cam_data->usb3) {
    cam_data->awb_buffer = vmalloc(cam_data->max_frame_size);;
    if (!cam_data->awb_buffer) {;
      retval = ENOMEM;
      goto probe_exit_enqued;
    }
    if (!cam_data->awb_object) {
      int awb_object_size;
      awb_object_size = Sentech_GetAWBObjectSize();
      cam_data->awb_object = kmalloc(awb_object_size, GFP_KERNEL);;    
    };
    if (!cam_data->awb_object) {;
      retval = ENOMEM;
      goto probe_exit_enqued;
    };
    Sentech_InitAWBObject(cam_data->awb_object,
			cam_data->awb_buffer,
			&cam_data->state12.line_numbers_le,
			&cam_data->state12.one_line_clock_numbers_le,
			&cam_data->payload21.Shutter_line_numbers_le,
			&cam_data->payload21.Shutter_clock_numbers_le,
			&cam_data->payload21.Gain_le,
			&cam_data->max_width,
			&cam_data->scan_height,
			&cam_data->modif_flag21,
			&cam_data->red,
			&cam_data->green,
			&cam_data->blue,
			&cam_data->state23.R_Gain_le,
			__le16_to_cpu(cam_data->state12.initial_gain_le),
			__le16_to_cpu(cam_data->state12.initial_digital_gain_le),
			cam_data->trigger_version 
#ifdef L_BOUNDS
			,&cam_data->gain_min
			,&cam_data->gain_max
			,&cam_data->exposure_min
			,&cam_data->exposure_max
#endif
			);
    if (!monochrome) {
      cam_data->auto_white_balance = 1;
      Sentech_SetAutoAWB(cam_data->awb_object, STCAM_WB_FULLAUTO);
    }
    cam_data->auto_gain = 1;
    Sentech_SetAutoGain(cam_data->awb_object,
			(STCAM_AGC_ON|STCAM_AUTO_SHUTTER_ON)
			);
    Sentech_SetAutoExposure(cam_data->awb_object,
			    (STCAM_AGC_ON|STCAM_AUTO_SHUTTER_ON)
			    );
    {
      BYTE target[]= {STC_DEFAULT_EXPOSURE_TARGET,5,5};
      Sentech_CLuminance_SetTarget(cam_data->awb_object,
				   target);
    }
  }
#endif
  cam_data->exposure_target = STC_DEFAULT_EXPOSURE_TARGET;
  cam_data->freq_auto = param_freq_auto?param_freq_auto:3;
  cam_data->red = 
    cam_data->blue =
    cam_data->green = (1<<COLOR_SHIFT);
  setSaturations(cam_data->saturation = SATDEF);
  setSaturations_16(cam_data->saturation = SATDEF);
  cam_data->gamma = default_gamma; 
  cam_data->brightness = default_brightness; 
  cam_data->gain = default_gain;;
  cam_data->pix_format.priv = 0;
  cam_data->pix_format.field = V4L2_FIELD_NONE;
  cam_data->pix_format.height = cam_data->max_height;
  cam_data->pix_format.bytesperline = cam_data->pix_format.width = cam_data->max_width;
  cam_data->pix_format.sizeimage = cam_data->scanimage_size;
  switch (sentech_fps[cam_data->model_idx].native_fourcc) {
  case V4L2_PIX_FMT_GREY:
    if (grey) {;
      cam_data->pix_format.pixelformat = V4L2_PIX_FMT_GREY;
      cam_data->fmt = &format_grey;
    } else if(y16) {;
      cam_data->pix_format.pixelformat = V4L2_PIX_FMT_Y16;
      cam_data->fmt = &format_y16;
    } else {;
      cam_data->pix_format.pixelformat = V4L2_PIX_FMT_YUYV;
      cam_data->fmt = &format_yuyv; break;
    }
    break;
  case V4L2_PIX_FMT_SRGGB8:
    if(rggb) {;
      cam_data->pix_format.pixelformat = V4L2_PIX_FMT_SRGGB8;
      cam_data->fmt = &format_rggb; break;
    } else {;
      cam_data->pix_format.pixelformat = V4L2_PIX_FMT_YUYV;
      cam_data->fmt = &format_yuyv; break;
    }
  case V4L2_PIX_FMT_SGRBG8:
    if(rggb) {;
      cam_data->pix_format.pixelformat = V4L2_PIX_FMT_SGRBG8;
      cam_data->fmt = &format_grbg; break;
    } else {;
      cam_data->pix_format.pixelformat = V4L2_PIX_FMT_YUYV;
      cam_data->fmt = &format_yuyv; break;
    }
  default:
;
    cam_data->pix_format.pixelformat = V4L2_PIX_FMT_YUYV;
    cam_data->fmt = &format_yuyv;;
  }
  cam_data->continuous_capture = param_continuous_capture;
  cam_data->preallocate = param_preallocate;
  cam_data->requested_buffers = param_buffers;;
  cam_data->max_fps = max_fps;
  for (i = 0; i < ARRAY_SIZE(sentech_qctrl); i++) {
    if (V4L2_CID_STC_FRAME_START == sentech_qctrl[i].id) {
      cam_data->qctrl[i].maximum = __le32_to_cpu(cam_data->state12.Vertical_size_mode_le);
    }
    if (V4L2_CID_STC_FRAME_HEIGHT == sentech_qctrl[i].id) {
      cam_data->qctrl[i].maximum = __le32_to_cpu(cam_data->state12.Vertical_size_mode_le);
    }
    if(cam_data->usb3) {
      if (V4L2_CID_AUTOGAIN == sentech_qctrl[i].id) {
	if(cam_data->auto_gain_fpga_can)
	  cam_data->qctrl[i].default_value = cam_data->auto_gain;
      }
      if (V4L2_CID_GAIN == sentech_qctrl[i].id) {
	  cam_data->qctrl[i].maximum = __le16_to_cpu(cam_data->state12.maximum_gain_le);
	  cam_data->qctrl[i].default_value = __le16_to_cpu(cam_data->state12.initial_gain_le);
      }
      if (V4L2_CID_STC_AUTOSHUTTERS == sentech_qctrl[i].id) {
	if(cam_data->auto_shutters_fpga_can){
	  cam_data->qctrl[i].maximum = 1;
	  cam_data->qctrl[i].default_value = 1;
	}else{
	  cam_data->qctrl[i].maximum = 0;
	  cam_data->qctrl[i].flags = V4L2_CTRL_FLAG_DISABLED;
	  cam_data->qctrl[i].default_value = 0;
	}
      }
#ifdef  SENTECH_AUTO_AWB_CIDS
      if (V4L2_CID_STC_AUTOGAIN == sentech_qctrl[i].id) {
	if(cam_data->auto_gain_fpga_can)
	  cam_data->qctrl[i].default_value = 1;
      }
#endif
    }else{
      if (V4L2_CID_STC_AUTOSHUTTERS == sentech_qctrl[i].id) {
	if(0&&cam_data->auto_shutters_fpga_can){
	  cam_data->qctrl[i].maximum = 1;
	  cam_data->qctrl[i].default_value = 1;
	}else{
	  cam_data->qctrl[i].maximum = 0;
	  cam_data->qctrl[i].flags = V4L2_CTRL_FLAG_DISABLED;
	  cam_data->qctrl[i].default_value = 0;
	}
      }
    }
  }
  for (i = 0; i < ARRAY_SIZE(sentech_qctrl); i++) {
      if(1 || !cam_data->usb3 || !cam_data->auto_gain_fpga_can) {
	if (V4L2_CID_STC_GAIN_MIN == sentech_qctrl[i].id) {
	  if(!cam_data->usb3 || !cam_data->auto_gain_fpga_can)
	    cam_data->qctrl[i].maximum = __le16_to_cpu(cam_data->state12.initial_gain_le);
	  else cam_data->qctrl[i].maximum =  __le16_to_cpu(cam_data->state12.maximum_gain_le);
	}
	if (V4L2_CID_STC_GAIN_MAX == sentech_qctrl[i].id) {
	  if(!cam_data->usb3 || !cam_data->auto_gain_fpga_can)
	    cam_data->qctrl[i].minimum = __le16_to_cpu(cam_data->state12.initial_gain_le);
	  else  cam_data->qctrl[i].minimum = 0;
	  cam_data->qctrl[i].maximum = __le16_to_cpu(cam_data->state12.maximum_gain_le);
	  cam_data->qctrl[i].default_value = __le16_to_cpu(cam_data->state12.maximum_gain_le);
	}
      }
      if (V4L2_CID_GAIN == sentech_qctrl[i].id) {
	cam_data->qctrl[i].maximum = __le16_to_cpu(cam_data->state12.maximum_gain_le);
	cam_data->qctrl[i].default_value = __le16_to_cpu(cam_data->state12.initial_gain_le);
      }
    }
  if (dynamic_scale) {
    unsigned int i;
    __s32 max_exp, max_exp_us;
    max_exp = get_max_exposure(cam_data);
    max_exp_us = exposure_2_useconds(cam_data, max_exp);
    for (i = 0; i < ARRAY_SIZE(sentech_qctrl); i++) {
      if ((V4L2_CID_EXPOSURE == sentech_qctrl[i].id) ||
	  ( V4L2_CID_CONTRAST == sentech_qctrl[i].id)) {
	cam_data->qctrl[i].maximum = max_exp;
      }
      if (V4L2_CID_STC_EXPOSURE_US == sentech_qctrl[i].id) {
	cam_data->qctrl[i].maximum = max_exp_us;
	cam_data->qctrl[i].default_value = cam_data->qctrl[i].maximum;
      }
#ifdef L_BOUNDS_ANY
      sentech_exposure_bounds(cam_data, i, max_exp);
#endif
    }
  }
  if(!_60fps)
    for (i = 0; i < ARRAY_SIZE(sentech_qctrl); i++) 
      if (V4L2_CID_STC_FPS == sentech_qctrl[i].id)
	switch(cam_data->max_fps){
	case 300:
	  strncpy((char*)cam_data->qctrl[i].name,
		  "Frames per Second 7.5 15 30",
		  sizeof(sentech_qctrl[i].name));
	  break;
	case 190:
	  strncpy((char*)cam_data->qctrl[i].name,
		  "Frames per Second 4.25 9.5 19",
		  sizeof(sentech_qctrl[i].name));
	  break;
	case 150:
	  strncpy((char*)cam_data->qctrl[i].name,
		  "Frames per Second 3.75 7.5 15",
		  sizeof(sentech_qctrl[i].name));
	  break;
	case 75:
	  strncpy((char*)cam_data->qctrl[i].name,
		  "Frames per Second 1.875 3.75 7.5",
		  sizeof(sentech_qctrl[i].name));
	case 220:
	  strncpy((char*)cam_data->qctrl[i].name,
		  "Frames per Second 5.5 11 22",
		  sizeof(sentech_qctrl[i].name));
	}
  for(i = 0; i < ARRAY_SIZE(sentech_qctrl); i++) {
    switch (sentech_qctrl[i].id) {
    case (V4L2_CID_STC_TRIGGER_MODE):
      cam_data->qctl_regs[i] = 
      (cam_data->payload80.Trigger_Mode & STC_FPGA_TRIGGER_MODE_MASK)>>
      STC_FPGA_TRIGGER_MODE_SHIFT;
      break;
    case (V4L2_CID_STC_TRIGGER_SOURCE):
      cam_data->qctl_regs[i] = 
      (cam_data->payload80.Trigger_Mode & STC_FPGA_TRIGGER_SRC_MASK)>>
      STC_FPGA_TRIGGER_SRC_SHIFT;
      break;
    case (V4L2_CID_STC_FRAME_START):
      cam_data->qctl_regs[i] = 0;
      break;
    case (V4L2_CID_STC_FRAME_HEIGHT):
      cam_data->qctl_regs[i] = 0;
      break;
#ifdef  SENTECH_AUTO_AWB_CIDS
    case (V4L2_CID_STC_AUTOGAIN):
#endif
    case (V4L2_CID_AUTOGAIN):
      if(cam_data->auto_gain_fpga_can) cam_data->qctl_regs[i] = cam_data->auto_gain;
      break;
    case (V4L2_CID_GAIN):
      if(cam_data->auto_gain_fpga_can) cam_data->qctl_regs[i] = __le16_to_cpu(cam_data->payload21.Gain_le);
      break;
    case (V4L2_CID_STC_AUTOSHUTTERS):
      if(cam_data->auto_shutters_fpga_can) cam_data->qctl_regs[i] = cam_data->auto_shutters;
      break;
    case (V4L2_CID_STC_EXPOSURE_US):
      if(cam_data->auto_shutters_fpga_can){
	cam_data->qctl_regs[i] = get_crt_exposure(cam_data);
	cam_data->qctrl[i].maximum =
	  exposure_2_useconds(cam_data,
			      (USB3_EXP_UP_MASK_4BITS << USB3_EXP_UP_SHIFT) +
			      (USB3_EXP_LOW_MASK_16BITS << USB3_EXP_LOW_SHIFT));
      }
      break;
    case (V4L2_CID_EXPOSURE):
    case (V4L2_CID_CONTRAST):
      if(cam_data->auto_shutters_fpga_can){
	cam_data->qctl_regs[i] = get_crt_exposure(cam_data);
	cam_data->qctrl[i].maximum =
	  (USB3_EXP_UP_MASK_4BITS << USB3_EXP_UP_SHIFT) +
	  (USB3_EXP_LOW_MASK_16BITS << USB3_EXP_LOW_SHIFT);
      }
      break;
    case (V4L2_CID_STC_EXPOSURE_MIN):
      if(cam_data->auto_shutters_fpga_can){
	cam_data->qctrl[i].maximum = 
	  exposure_2_useconds(cam_data,
			      (USB3_EXP_UP_MASK_4BITS << USB3_EXP_UP_SHIFT) +
			      (USB3_EXP_LOW_MASK_16BITS << USB3_EXP_LOW_SHIFT));
	cam_data->qctrl[i].step = 1 << USB3_EXP_LOW_SHIFT;
 	cam_data->qctrl[i].default_value = USB3_MIN_EXP_DEFAULT << USB3_EXP_LOW_SHIFT;
      }
      break;
    case (V4L2_CID_STC_EXPOSURE_MAX):
      if(cam_data->auto_shutters_fpga_can){
	cam_data->qctrl[i].maximum = 
	  exposure_2_useconds(cam_data,
			      (USB3_EXP_UP_MASK_4BITS << USB3_EXP_UP_SHIFT) +
			      (USB3_EXP_LOW_MASK_16BITS << USB3_EXP_LOW_SHIFT));
 	cam_data->qctrl[i].step = 1 << USB3_EXP_LOW_SHIFT;
 	cam_data->qctrl[i].default_value = USB3_MAX_EXP_DEFAULT << USB3_EXP_LOW_SHIFT;
     }
      break;
    case (V4L2_CID_STC_HORIZONTAL_SIZE_CAPTURE):
      if(cam_data->usb3 && sentech_support_AOI(cam_data)){
	cam_data->qctl_regs[i] = cam_data->scan_width;
	cam_data->qctrl[i].maximum = __le32_to_cpu(cam_data->payload11.Maximum_picture_width_le);
	cam_data->qctrl[i].flags = 0;
      }
      break;
    case (V4L2_CID_STC_VERTICAL_SIZE_CAPTURE):
      if(cam_data->usb3 && sentech_support_AOI(cam_data)){
	cam_data->qctl_regs[i] = cam_data->scan_height;
	cam_data->qctrl[i].maximum = __le32_to_cpu(cam_data->payload11.Maximum_picture_height_le);
	cam_data->qctrl[i].flags = 0;
      }
      break;
    case (V4L2_CID_STC_HORIZONTAL_OFFSET_CAPTURE):
      if(cam_data->usb3 && sentech_support_AOI(cam_data)){
	cam_data->qctl_regs[i] = cam_data->scan_horizontal_offset;
	cam_data->qctrl[i].maximum = __le32_to_cpu(cam_data->payload11.Maximum_picture_width_le);
	cam_data->qctrl[i].flags = 0;
      }
      break;
    case (V4L2_CID_STC_VERTICAL_OFFSET_CAPTURE):
      if(cam_data->usb3 && sentech_support_AOI(cam_data)){
	cam_data->qctl_regs[i] = cam_data->scan_vertical_offset;
	cam_data->qctrl[i].maximum = __le32_to_cpu(cam_data->payload11.Maximum_picture_height_le);
	cam_data->qctrl[i].flags = 0;
      }
      break;
    default:
      cam_data->qctl_regs[i] = cam_data->qctrl[i].default_value;
    }
  }
  cam_data->psat_skip = param_saturation_skip;
  cam_data->psat_auto = param_auto_saturation;
  cam_data->psat_step = param_saturation_step;
  cam_data->psat_min = param_saturation_min;
  cam_data->psat_max = param_saturation_max;
  cam_data->psat_acc = param_sat_acc;
  cam_data->psat_lo = param_saturation_lower_limit;
  cam_data->psat_up = param_saturation_upper_limit;
  cam_data->psat_exp = param_saturation_exp_inc;
  cam_data->psat_gain = param_saturation_gain_inc;
  cam_data->psat_amortiz = param_sat_amortization;
  cam_data->psat_amortiz_length = param_sat_amortization_length;
  INIT_LIST_HEAD(&cam_data->buffers);
  init_waitqueue_head(&cam_data->autoq);
  init_waitqueue_head(&cam_data->wq);
  rwlock_init(&cam_data->buffers_lock);
  rwlock_init(&cam_data->ctrl_lock);
  if (param_have_buffers) {;
    if(param_preallocate) {;
      if((retval = ensure_positive_error(sentech_alloc_subs(cam_data)))){ 
	printk(KERN_ALERT "sentech: Failed to allocate buffers!\n"); 
	goto probe_error_alloc_buffers;
      }
    }
  } else {;
    if ((retval = sentech_alloc_non_dma_buffers(cam_data))) {;
      goto probe_error_alloc_buffers;
    }
  }
  usb_set_intfdata (interface, cam_data);
  if(cam_data->payload11.Color_layout_le == STC_BW)
    if(!cam_data->monochrome)
      dev_info(&interface->dev, "Sentech USB Camera monochrome\n");
  cam_data->red = __le16_to_cpu(cam_data->state23.R_Gain_le) >> (7-COLOR_SHIFT);
  cam_data->blue = __le16_to_cpu(cam_data->state23.B_Gain_le) >> (7-COLOR_SHIFT);
  cam_data->green =
    (__le16_to_cpu(cam_data->state23.Gr_Gain_le)+
     __le16_to_cpu(cam_data->state23.Gr_Gain_le)) >> (1+7-COLOR_SHIFT);
  for(i = 0; i < ARRAY_SIZE(sentech_qctrl); i++) {
    switch (sentech_qctrl[i].id) {
    case (V4L2_CID_STC_RED): 
    case V4L2_CID_RED_BALANCE:
      {cam_data->qctl_regs[i] = cam_data->red; break;}
    case (V4L2_CID_STC_GREEN): 
      {cam_data->qctl_regs[i] = cam_data->green; break;}
    case (V4L2_CID_STC_BLUE): 
    case V4L2_CID_BLUE_BALANCE:
      {cam_data->qctl_regs[i] = cam_data->blue; break;}
    }
  }
  PROBE(5410);
;
  if ((retval = ensure_positive_error(sentech_alloc_v4l2(cam_data)))) {
    dev_err(&interface->dev, "SentechUSB: No V4l2 initialization\n");
    goto probe_error_v4l2;  
  }
  if(device_create_file(&interface->dev, &dev_attr_first_frame));
  if(device_create_file(&interface->dev, &dev_attr_last_frame));
  if(device_create_file(&interface->dev, &dev_attr_lost_frames));
#if  LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,12)
  if(device_create_file(&interface->dev, &dev_attr_serial));
  if(device_create_file(&interface->dev, &dev_attr_product));
  if(device_create_file(&interface->dev, &dev_attr_manufacturer));
#endif
  dprintksubs(cam_data);
  if(!cam_data->usb3) start_qthread(cam_data);
;
  retval = 0;
  cam_data->probe_done = 1;
  goto probe_exit_unlock;
  video_unregister_device(cam_data->v4l2dev);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,11,0)
  v4l2_device_disconnect(&cam_data->v4l2_dev);
  v4l2_device_unregister(&cam_data->v4l2_dev);   
#if defined(CONFIG_MEDIA_CONTROLLER)
  media_device_unregister(&cam_data->media_dev);
#endif
#endif
 probe_error_v4l2:
  if (!param_have_buffers) {;
    sentech_free_non_dma_buffers(cam_data);
  }
 probe_error_alloc_buffers:
#ifdef SENTECH_AWB
 probe_exit_enqued:
#endif
  if(cam_data->sentech_devlist.next &&
     cam_data->sentech_devlist.next != &cam_data->sentech_devlist) {
    list_del(&cam_data->sentech_devlist);
    cam_data->sentech_devlist.prev = cam_data->sentech_devlist.next = 0;
  }
 probe_error_identification:
  if (cam_data)
    kref_put(&cam_data->kref, cam_data_delete);
 probe_error_alloc1:
  printk(KERN_ALERT "sentech: Error %d, in %s\n", retval, __FUNCTION__);
 probe_exit_unlock:
  if(debug!=0)printk("sentech: Exiting %s\n",__FUNCTION__);
  return retval;
}
static void SentechUSB_disconnect(struct usb_interface *interface){
  struct sentech_camera *cam_data;
  int minor = interface->minor;
  unsigned long flags;
  int loops;
  struct ctrl_context* ctx;
  int cam_number = -1;
  if(debug!=0)printk("sentech: In %s\n",__FUNCTION__);
  lock_kernel();
  cam_data = usb_get_intfdata (interface);
  if(!cam_data){
    dev_info(&interface->dev, "Sentech USB Camera disconnected\n");
    goto disconnect_exit_unlock;
  }
  cam_data->run = 0;
  cam_data->connected = 0;
  stop_qthread(cam_data);
  usb_set_intfdata(interface, NULL);
  if(cam_data->v4l2dev) {
    video_unregister_device(cam_data->v4l2dev);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,11,0)
    v4l2_device_disconnect(&cam_data->v4l2_dev);
    v4l2_device_unregister(&cam_data->v4l2_dev);
#if defined(CONFIG_MEDIA_CONTROLLER)
  media_device_unregister(&cam_data->media_dev);
#endif
#endif
  }
  else
    dev_info(&interface->dev, "No Video Device with this camera\n");
  device_remove_file(&interface->dev, &dev_attr_first_frame);
  device_remove_file(&interface->dev, &dev_attr_last_frame);
  device_remove_file(&interface->dev, &dev_attr_lost_frames);
#if  LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,12)
  device_remove_file(&interface->dev, &dev_attr_serial);
  device_remove_file(&interface->dev, &dev_attr_product);
  device_remove_file(&interface->dev, &dev_attr_manufacturer);
#endif
 disconnect_exit_unlock:
  unlock_kernel();
  if(cam_data) {
    dprintksubs(cam_data);
    usb_put_dev(cam_data->udev); 
    cam_number = cam_data->cam_number;
;
    while(!list_empty(&cam_data->ctrl_queue)){
      write_lock_irqsave(&cam_data->ctrl_lock, flags);
      ctx = list_entry(cam_data->ctrl_queue.next,
		       struct ctrl_context, queue);
      list_del (&ctx->queue);
      write_unlock_irqrestore(&cam_data->ctrl_lock, flags);;
      wait_for_completion_timeout(&ctx->done,HZ);;
      kfree(ctx);
    }
    loops=-1;
    while (cam_data->kthread_running){;
      loops--;
      if(loops == 0) {;
	break;
      }
    }
    dev_info(&interface->dev, "Sentech USB Camera \"%s\" #%d:%d now disconnecting\n", cam_data->cam_name, minor, cam_number);
    kref_put(&cam_data->kref, cam_data_delete);
    dev_info(&interface->dev, "Sentech USB Camera now disconnected\n");
  }
  if(debug!=0)printk("sentech: Exiting %s\n",__FUNCTION__);
}
static 
int SentechUSB_suspend (struct usb_interface *interface, pm_message_t message){
  struct sentech_camera *cam_data;
  int retval = 0;
  if(debug!=0)printk("sentech: In %s\n",__FUNCTION__);
  cam_data = usb_get_intfdata (interface);
  cam_data->suspended_running = cam_data->run;
  cam_data->run = 0;
  if(debug!=0)printk("sentech: Exiting %s\n",__FUNCTION__);
  return retval;
}
static 
int SentechUSB_resume (struct usb_interface *interface){
  struct sentech_camera *cam_data;
  int retval = 0;
  if(debug!=0)printk("sentech: In %s\n",__FUNCTION__);
  cam_data = usb_get_intfdata (interface);
  cam_data->run = cam_data->suspended_running;
  atomic_set(&cam_data->running_buffers, 0);
  if(cam_data->run){
    if(cam_data->shutdown) {retval = -ENODEV; goto resume_done;}
    cam_data->status = 0;
    cam_data->suspended_running = 0;
    sentech_submit_kurbs(cam_data, GFP_KERNEL);
  }  
 resume_done:
  if(debug!=0)printk("sentech: Exiting %s\n",__FUNCTION__);
  return retval;
}
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,23)
static 
int SentechUSB_reset_resume (struct usb_interface *interface){
  struct sentech_camera *cam_data;
  unsigned long flags;
  int retval = 0;
  if(debug!=0)printk("sentech: In %s\n",__FUNCTION__);
  cam_data = usb_get_intfdata (interface);
  if(cam_data->scan_height != cam_data->max_height)
    setFlag(&cam_data->modif_flag12);
  if(cam_data->supports_bpp)
    setFlag(&cam_data->modif_flag13); 
  setFlag(&cam_data->modif_flag21); 
  setFlag(&cam_data->modif_flag22);
  setFlag(&cam_data->modif_flag_80_00);
  setFlag(&cam_data->modif_flag_80_04);
  setFlag(&cam_data->modif_flag_80_30);
  setFlag(&cam_data->modif_flag_80_0F);
  setFlag(&cam_data->modif_flag_80_14);
  write_lock_irqsave(&cam_data->ctrl_lock, flags);
  set_controls(cam_data, GFP_KERNEL);
  write_unlock_irqrestore(&cam_data->ctrl_lock, flags);
  mb();
  cam_data->run = cam_data->suspended_running;
  atomic_set(&cam_data->running_buffers, 0);
  if(cam_data->run){
    if(cam_data->shutdown) {retval = -ENODEV; goto reset_resume_done;}
    cam_data->status = 0;
    cam_data->suspended_running = 0;
    sentech_submit_kurbs(cam_data, GFP_KERNEL);
  }
 reset_resume_done:
  if(debug!=0)printk("sentech: Exiting %s\n",__FUNCTION__);
  return retval;
}
#endif
static struct usb_driver SentechUSB_driver = {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,15)
  	.owner =	THIS_MODULE,
#endif
	.name =		"SentechTUSB",
	.probe =	SentechUSB_probe,
	.disconnect =	SentechUSB_disconnect,
	.id_table =	id_table,
	.suspend = SentechUSB_suspend,
	.resume = SentechUSB_resume,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,23)
	.reset_resume = SentechUSB_reset_resume,
#endif
};
#ifdef __LP64__
#define IFOR "lx"
#else
#define IFOR "x"
#endif
 static void __init zero_gamma16(void) {
#define INITDATA_DISCARDED
#ifdef INITDATA_DISCARDED
#endif
  int G=0;
  if(!param_gamma16) return;
  for(; G<ARRAY_SIZE(_sentech_gamma16); G++){
    sentech_gamma16[G] = 0;
  }
 }
static void __init ini_gamma16(void) {
#define INITDATA_DISCARDED
#ifdef INITDATA_DISCARDED
  int G=0;
  if(!param_gamma16) return;
  for(; G<ARRAY_SIZE(_sentech_gamma16); G++){
    sentech_gamma16[G] = kmalloc(SIZE_GAMMAS16*sizeof(__u16), GFP_KERNEL);
    memcpy(sentech_gamma16[G], _sentech_gamma16[G], SIZE_GAMMAS16*sizeof(__u16));
  }
#else
  int G=0;
  if(!param_gamma16) return;
  for(; G<ARRAY_SIZE(_sentech_gamma16); G++)
    sentech_gamma16[G] = _sentech_gamma16[G];
#endif
 }
 static void free_gamma16(void) {
#define INITDATA_DISCARDED
#ifdef INITDATA_DISCARDED
  int G=0;
  if (!param_gamma16) return;
  for(; G<ARRAY_SIZE(_sentech_gamma16); G++){
    if (sentech_gamma16[G]) {
      kfree(sentech_gamma16[G]);
    }
  }
  dprintk (0, "free_gamma16\n");
#endif
 }
static int __init dbg_headers(void) {
  if(sizeof(struct Sentech_Header) != 0x40){
    dprintk (1, "Header Size=%" IFOR ". BUG:msilaghi@fit.edu\n",
	    sizeof(struct Sentech_Header));
    goto dbg_probe_exit;
  }
  if(sizeof(struct Sentech_Camera_Data_Payload) != 0x40){
    dprintk (1, "Sentech_Camera_Data_Payload=%" IFOR ". BUG:msilaghi@fit.edu\n",
	    sizeof(struct Sentech_Camera_Data_Payload));
    goto dbg_probe_exit;
  }
  if(sizeof(struct Sentech_Picture_Size_Payload) != 0x40){
    dprintk (1, "Sentech_Picture_Size_Payload=%" IFOR ". BUG:msilaghi@fit.edu\n",
	    sizeof(struct Sentech_Picture_Size_Payload));
    goto dbg_probe_exit;
  }
  if(sizeof(struct Sentech_Bytes_Per_Pixel_Data_Payload) != 0x06){
    dprintk (1, "Sentech_Picture_Size_Payload=%" IFOR ". BUG:msilaghi@fit.edu\n",
	    sizeof(struct Sentech_Picture_Size_Payload));
    goto dbg_probe_exit;
  }
  if(sizeof(struct Sentech_Shutter_Gain_Payload) != 0x40){
    dprintk (1, "Sentech_Shutter_Gain_Payload=%" IFOR ". BUG:msilaghi@fit.edu\n",
	    sizeof(struct Sentech_Shutter_Gain_Payload));
    goto dbg_probe_exit;
  }
  if(sizeof(struct Sentech_Clock_Payload) != 0x0A){
    dprintk (1, "Sentech_Clock_Payload=%" IFOR ". BUG:msilaghi@fit.edu\n",
	    sizeof(struct Sentech_Clock_Payload));
    goto dbg_probe_exit;
  }
  if(sizeof(struct Sentech_White_Balance_Payload) != 0x40){
    dprintk (1, "Sentech_White_Balance_Payload=%" IFOR ". BUG:msilaghi@fit.edu\n",
	    sizeof(struct Sentech_White_Balance_Payload));
    goto dbg_probe_exit;
  }
  return 1;
 dbg_probe_exit:
  return 0;
}
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,9)
#define sen_err(format, arg...)                                     \
  printk(KERN_ERR KBUILD_MODNAME ": " format "\n", ##arg)
#else
#define sen_err(format, arg...)                                     \
  printk(KERN_ERR "Sentech : " format "\n", ##arg)
#endif
static int __init usb_SentechUSB_init(void){
  int i;
  int retval = 0;;
  if (param_buffers < 1) {
    sen_err("SentechUSB: requested param_buffers<1, fixed to 1\n");
    param_buffers=1;
  }
  SUBS_SUBMITTED = param_SUBS_SUBMITTED;
  if (SUBS_SUBMITTED < 1) {
    sen_err("SentechUSB: requested SUBS_SUBMITTED<1, fixed to 1\n");
    SUBS_SUBMITTED=1;
  }
  if ((param_buffers <= SUBS_SUBMITTED)){
    if (auto_SUBS_SUBMITTED) {
      if (SUBS_SUBMITTED > 1){
	SUBS_SUBMITTED = max(1,param_buffers-1);
	sen_err("SentechUSB: requested param_buffers<=SUBS_SUBMITTED, fixed SUBS_SUBMITTED to %d\n", SUBS_SUBMITTED);
      }
      if (param_buffers == SUBS_SUBMITTED) {
	param_use_submitted = 1;
	sen_err("SentechUSB: requested param_buffers==SUBS_SUBMITTED, fixed param_use_submitted to 1\n");
      }
    } else {
      sen_err("SentechUSB: requested param_buffers<=SUBS_SUBMITTED\n");
      if ( !param_use_submitted ) {
	sen_err("SentechUSB: requested param_buffers<=SUBS_SUBMITTED, please fix!\n");
	return -1;
      }
      if (param_buffers < SUBS_SUBMITTED) {
	SUBS_SUBMITTED = param_buffers;
	sen_err("SentechUSB: requested param_buffers<SUBS_SUBMITTED, fixed SUBS_SUBMITTED to %d\n", SUBS_SUBMITTED);
      }
    }
  }
  if (param_use_submitted) {
    if (!param_submit_user_buffers) {
	sen_err("SentechUSB: requested param_submit_user_buffers<param_use_submitted, fixed param_submit_user_buffers to 1\n");
      param_submit_user_buffers = 1;
    }
  };
  if (!dbg_headers()) {
    sen_err("\n<2> SentechUSB: usb_register failed. Compiler version inconsistency! Announce driver author: msilaghi@fit.edu!\n");
    return -1;
  };
  fillm();
;
  initSaturation();;
  if (param_supports_bpp) {;
    ini_gamma16();;
    fillm_16();;
    initSaturation_16();
  } else {
    zero_gamma16();
  }
  for (i = 0; i < ARRAY_SIZE(sentech_qctrl); i++) {
    switch (sentech_qctrl[i].id) {
    case (V4L2_CID_BRIGHTNESS):
      if (default_brightness < sentech_qctrl[i].minimum)
	default_brightness = sentech_qctrl[i].minimum;
      if (default_brightness > sentech_qctrl[i].maximum)
	default_brightness = sentech_qctrl[i].maximum;
      sentech_qctrl[i].default_value = default_brightness;
      break;
    case (V4L2_CID_GAIN):
      if (default_gain < sentech_qctrl[i].minimum)
	default_gain = sentech_qctrl[i].minimum;
      if (default_gain > sentech_qctrl[i].maximum)
	default_gain = sentech_qctrl[i].maximum;
      sentech_qctrl[i].default_value = default_gain;
      break;
    case (V4L2_CID_STC_TARGET):
      if (default_exposure_target < sentech_qctrl[i].minimum)
	default_exposure_target = sentech_qctrl[i].minimum;
      if (default_exposure_target > sentech_qctrl[i].maximum)
	default_exposure_target = sentech_qctrl[i].maximum;
      sentech_qctrl[i].default_value = default_exposure_target;
      break;
    case (V4L2_CID_CONTRAST):
    case (V4L2_CID_EXPOSURE):
      if (default_exposure < sentech_qctrl[i].minimum)
	default_exposure = sentech_qctrl[i].minimum;
      if (default_exposure > sentech_qctrl[i].maximum)
	default_exposure = sentech_qctrl[i].maximum;
      sentech_qctrl[i].default_value = default_exposure;
      break;
    case (V4L2_CID_GAMMA):
    case (V4L2_CID_STC_GAMMA):
      if (default_gamma < sentech_qctrl[i].minimum)
	default_gamma = sentech_qctrl[i].minimum;
      if (default_gamma > sentech_qctrl[i].maximum)
	default_gamma = sentech_qctrl[i].maximum;
      sentech_qctrl[i].default_value = default_gamma;
      break;
    }
  }
  retval = usb_register(&SentechUSB_driver);
  if (retval) {
    sen_err("\n<2> SentechUSB: usb_register failed. Error number %d\n", retval);
    if (param_supports_bpp) {
      free_gamma16();
    }
  }
  else;
  return retval;
}
static void free_global_subs(void){
  struct sentech_usb_buffer* sub=NULL;
  unsigned long flags;
;
  for(;;) {
    write_lock_irqsave(&subs_list_lock, flags);
    if(!list_empty(&sentech_sublist)) {
      sub =
	list_first_entry(&sentech_sublist,
			 struct sentech_usb_buffer,
			 sub_buffers);
      list_del(&sub->sub_buffers);
    }else sub = 0;
    write_unlock_irqrestore(&subs_list_lock, flags);
    if(sub) free_sub(sub, NULL);
    else break;
  };
  return;
}
static void __exit usb_SentechUSB_exit(void){;
  usb_deregister(&SentechUSB_driver);
  unloading_module = 1;
  free_global_subs();
  if (param_supports_bpp) {
    free_gamma16();
  };
}
#ifndef MOD_INIT
int init_module(void){
  return usb_SentechUSB_init();
}
void cleanup_module(void){
  usb_SentechUSB_exit();
}
#else
module_init (usb_SentechUSB_init);
module_exit (usb_SentechUSB_exit);
#endif
MODULE_AUTHOR(DRIVER_AUTHOR);
MODULE_DESCRIPTION(DRIVER_DESC);
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,7)
MODULE_LICENSE("GPL\0is not the licence, but Proprietary MCS/Sentech America");
#else
MODULE_LICENSE("GPL");
#endif
MODULE_SUPPORTED_DEVICE("SentechTUSB");
